/*	target_negotiate.h	*/

/* chap support - CHONG */
#include "chap.h"
#include "misc_func.h"

/* stores everything related to an iscsi connection */
struct iscsi_conn
{
        int                     active;

        /* prev: pointer to the previous session */
        struct iscsi_conn       *prev;

        /* next: pointer to the next session */
        struct iscsi_conn       *next;
          
        /* conn_id: the id for this connection */
        int                     conn_id;

        /* cid: connection id assigned by the Initiator */
        __u16                   cid;

        /* conn_socket: socket used by this connection */
        struct socket           *conn_socket;

        /* cmnd_list: list of commands received by this connection */
        struct iscsi_cmnd       *cmnd_list;

        /* session: what session does this connection belong to */
        struct iscsi_session    *session;

        /* dev: device on which this connection was received */
        struct iscsi_global     *dev;

        /* rx_thread: task struct for rx_thread */
        struct task_struct      *rx_thread;

        /* tx_thread: task struct for tx_thread */
        struct task_struct      *tx_thread;

        /* tx_sem: semaphore to control operation of tx_thread */
        struct semaphore        tx_sem;

        /* thr_kill_sem: semaphore to kill rx and tx threads */
        struct semaphore        thr_kill_sem;

        /* bad_hdr: complete header of bad PDU for reject */
        unsigned char           bad_hdr[ISCSI_HDR_LEN];

        /* hdr_crc: header digest */
        int                     hdr_crc; 

        /* data_crc: data digest */
        int                     data_crc;

        /* connection_flags: bits to control stuff during login and FFP */
        unsigned int            connection_flags;

        /* Added the r2t retransmit timer - SAI */
        struct timer_list       *r2t_timer;

        /* 
         * Connection WIDE COUNTERS: stat_sn
         */
        __u32                   stat_sn;
        __u32                   max_send_length;/*initiator's MaxRecvPDULength*/
        __u32                   max_recv_length;/*target's MaxRecvPDULength*/
};



/* stores everything related to an iscsi session */
struct iscsi_session
{
        /* prev: pointer to the previous session */
        struct iscsi_session    *prev;

        /* next: pointer to the next session */
        struct iscsi_session    *next;

        /* address: this session is connected to an IP address */
        struct sockaddr_in      *address;

        /* isid: id assigned to a session by the Initiator */
        __u8                    isid[6];

        /* tsih: id assigned to a session by the Target */
        __u16                   tsih;
        
        /* cmnd_id: these are the values assigned as target_tags */
        __u32                   cmnd_id;

        /* cmnd_list: list of commands received within session */
        struct iscsi_cmnd       *cmnd_list;
 
        /* conn_sem: semaphore to add/remove commands */
        struct semaphore        cmnd_sem;

        /* nconn: no of active connections */  
        int                     nconn;

        /* conn_list: list of connections within session */
        struct iscsi_conn       *conn_list;

        /* conn_sem: semaphore to add/remove connections */  
        struct semaphore        conn_sem;

        /* devdata: pointer to the device specific data */
        struct iscsi_global     *devdata;
        
 	/* nop_period: timeout period to send periodic NopIns to initiator */
        unsigned int            nop_period;

        /* SNACK flags for Error Recovery - SAI */
        unsigned char           targ_snack_flg;

		/* error recovery ver ref18_04*/
		struct task_struct      *retran_thread;
		struct semaphore        retran_sem;
		struct semaphore        thr_kill_sem;
		
	/* draft_number: draft number target will enforce */
	unsigned int		draft_number;

        /* For parameters */
        struct parameter_type           (*session_params)[MAX_CONFIG_PARAMS] ;
        struct session_operational_parameters   *oper_param;
        
        /*  do we need these?  */
        unsigned int            text_param_len;
        void                    *text_param;   

        /* version identifiers */
        __u8                    version_min;
        __u8                    version_max;
        __u8                    version_active;

        /* init_task_tag: session wide initiator assigned task tag */
        __u32                   init_task_tag;

        /* init_task_tag: session wide initiator assigned task tag */
        __u32                   target_xfer_tag;

        /* initiator name */
        __u8                    *init_name;
        __u8                    *targ_name;
        
        /*
         * SESSION WIDE COUNTERS: exp_cmd_sn and max_cmd_sn
         */
        __u32                   cmd_sn;
        __u32                   exp_cmd_sn;    
        __u32                   max_cmd_sn;
};


/*************************************************************************
 *This function sends a login response with proper status code when error
 * occurs in target during login phase
 *
 ************************************************************************/

int login_reject( struct socket *sock, int class, int detail,
    	unsigned int draft_number, struct generic_pdu *outputpdu, int flags )
{

  TRACE(TRACE_ENTER_LEAVE, "Enter login_reject\n");

  outputpdu->status_class = class;
  outputpdu->status_detail = detail;

 /* Draft 18, Section 4.3.1 Login Phase Start
  "-Login Response with Login Reject.  This is an immediate rejection
  from the target that causes the connection to terminate and the session
  to terminate if this is the first (or only) connection of a new
  session.  The T bit and the CSG and NSG fields are reserved."
 */

  /* Draft 9, Section 5.1 Login Phase Start
  "-Login Response with Login Reject.  This is an immediate rejection
  from the target that causes the session to terminate.  The T bit of
  the response MUST be set to 1 and the CSG and NSG are reserved."
  */

  outputpdu->flags &= ~CSG;   /* set the CSN to zero */
  outputpdu->flags &= ~NSG;   /* set the NSG to zero */
 
  if( draft_number <=DRAFT9 )
	outputpdu->flags |= T_BIT;
  else
	outputpdu->flags &= ~T_BIT;
  
  /* draft 9 P96, draft 8 p91: StatSN, ExpCmdSN, MaxCmdSN
     are valid only if Status-Class is 0 */

  outputpdu->cmd_sn = 0;		
  outputpdu->exp_stat_sn = 0;		
  outputpdu->max_cmd_sn = 0;		

  /* send out the outputpdu which never has any data attached to it */
  outputpdu->text_length = 0;
  if( iscsi_send_msg(sock, draft_number, outputpdu, flags) < 0 )
    {
      TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
      return -1;
    }

  TRACE(TRACE_ENTER_LEAVE, "Leave login_reject\n");

  return 0;
}


/*******************************************************************
 *This function checks every login request except the first login for
 * new connection or session
 *
 *******************************************************************/
int check_other_login( struct socket *sock , int correct_CSG,
			unsigned int draft_number,
			struct generic_pdu *inputpdu,
			struct generic_pdu *outputpdu,
			int flags )
{
       TRACE(TRACE_ENTER_LEAVE, "Enter check_other_login\n");
      
       /* check if we get a login command, send a login reject if a non-login 
	* pdu is received
        */ 
	if( (inputpdu->opcode & ISCSI_OPCODE ) != ISCSI_INIT_LOGIN_CMND ) 
	{
                int detail; 
                if( draft_number <= DRAFT9 )
                        detail = STAT_DETAIL_ERR;
                else
                        detail = STAT_DETAIL_INVALID_DURING_LOGIN;   

                TRACE_ERROR("Receive PDU has opcode %x, which is not a "
			"login request \n", (inputpdu->opcode & ISCSI_OPCODE));
		
		/* initiator error */
	        if ( login_reject(sock, STAT_CLASS_INITIATOR, detail,
					draft_number, outputpdu, flags) < 0 ) 
		{
			TRACE_ERROR( "Error in login_reject\n" );
		} 
		return -1;
	}

 	/*check if the login has I bit set for draft 8 and 9 */
	if( draft_number <= DRAFT9 || draft_number >= DRAFT18 )
	{
		if( !(inputpdu->opcode & I_BIT) )  
		{
			/* protocol error  */
	 		TRACE_ERROR( "The login request doesn't have I bit set! \n" );
                
	       	 	if ( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
					draft_number, outputpdu, flags) < 0 ) 
			{
				TRACE_ERROR( "Error in login_reject\n" );
			} 
			return -1;
		}
	}
	else
	{
		if( inputpdu->opcode & I_BIT )  
		{
			/* protocol error */
	 		TRACE_ERROR( "The login request has the I bit set, the "
				"I bit should be reserved for draft 11, 12 and 13! \n" );
		}
	}

	/* check if the login has proper version set */
	if( (inputpdu->version_max > ISCSI_MAX_VERSION) ||
		(inputpdu->version_max < inputpdu->version_active) ) 
	{
		/* protocol error  */
	 	TRACE_ERROR( "The target doesn't support the version "
				"specified by the login request\n" );
		if ( login_reject(sock, STAT_CLASS_INITIATOR,
					STAT_DETAIL_VERSION_NOT_SUPPORTED,
					draft_number, outputpdu, flags) < 0 )
		{
			TRACE_ERROR( "Error in login_reject\n" );
		}
		return -1;

	}

	if( ((inputpdu->flags & CSG) >> CSG_SHIFT) != correct_CSG )
	{
		/* protocol error */
		TRACE_ERROR( "Initiator changed state unexpectly\n");
		if ( login_reject(sock, STAT_CLASS_INITIATOR,  STAT_DETAIL_ERR,
					draft_number, outputpdu, flags) < 0 )
		{
			TRACE_ERROR( "Error in login_reject\n" );
		}		
               	return  -1;
	}

       TRACE(TRACE_ENTER_LEAVE, "Leave check_other_login\n");

	return 0;

}

/*****************************************************************************
 *  This function checks if the first login has Initiator name, target name
 *  and sessiontype(for draft 8,9)
 *  Send login reject if any parameter is missed.
 *
 ****************************************************************************/
int check_flags( struct iscsi_conn *conn, unsigned long long login_flags,
                struct generic_pdu *outputpdu)
{

   int flag = 1;

   TRACE(TRACE_ENTER_LEAVE, "Enter check_flags\n");

   if( !(login_flags & INITIATORNAME_FLAG) )
   {
      TRACE_ERROR("Initiator name not given in initial login\n");
      flag = -1;
   }
   if( flag > 0 )
   {
      if( !(login_flags & TARGETNAME_FLAG) && !(login_flags & DISCOVERY_FLAG) )
        {
          TRACE_ERROR("Target name not given in initial login to NORMAL session\n");
          flag = -1;
        }
   }
   if( flag > 0 )
   {
      if( !(login_flags & SESSIONTYPE_FLAG) &&
                                        conn->session->draft_number <= DRAFT8 )
        {
          TRACE_ERROR( "Sessiontype key not given in the initial login in "
				"draft %d\n", conn->session->draft_number);
          flag = -1;
        }
   }

   /* send a login reject if an error was seen */
   if( flag < 0 )
   {
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR,
                                STAT_DETAIL_MISSING_PARAMETER,
                                conn->session->draft_number, outputpdu,
                                conn->connection_flags) < 0 )
        {
           TRACE_ERROR( "Error in login_reject\n" );
        }
      return -1;
   }
    
  TRACE(TRACE_ENTER_LEAVE, "Leave check_flags\n");

   return flag;
}

/****************************************************************
 * This function checks the first login in a new connection
 *
 ****************************************************************/
             
int check_first_login( struct iscsi_conn *conn,
                       struct generic_pdu *inputpdu,
                       struct generic_pdu *outputpdu)
{
   int retval = 1;
              
   TRACE(TRACE_ENTER_LEAVE, "Enter check_first_login\n");

   /* check MaxConnections */
   if(conn->session->nconn > conn->session->oper_param->MaxConnections)
   {
      TRACE_ERROR("check_first_login: current nconn %d > MaxConnections %d, "
                   "terminate the connection\n", conn->session->nconn,
                   conn->session->oper_param->MaxConnections);
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR, 
		STAT_DETAIL_TOO_MANY_CONNECTIONS, conn->session->draft_number,
		 outputpdu, conn->connection_flags) < 0 )
      {      
        TRACE_ERROR( "Error in login_reject\n" );
      }
      retval = -1;
      goto out;
   }

   /* check X bit for draft8  */
   if ( (inputpdu->opcode & OLD_X_BIT) == OLD_X_BIT 
	&& conn->session->draft_number == DRAFT8 )
   {
      TRACE_ERROR("check_first_login: reinstate a failed connection or session"
                      " is not implemented yet, terminate the connection\n");
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR,
		 STAT_DETAIL_ERR,conn->session->draft_number, outputpdu, 
		conn->connection_flags) < 0 )
      {
        TRACE_ERROR( "Error in login_reject\n" );
      }
      retval = -1;
      goto out;
   }

   /* check I bit for draft 8&9 and 18 and after*/
   if( conn->session->draft_number <= DRAFT9 || conn->session->draft_number >= DRAFT18)
   {   
      if ( (inputpdu->opcode & I_BIT) != I_BIT ) 
      {
         TRACE_ERROR("check_first_login: I bit is not 1 for draft %d, terminate the"
                      " connection\n", conn->session->draft_number);
         if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
                        conn->session->draft_number, outputpdu, 
		        conn->connection_flags) < 0 )
         {
            TRACE_ERROR( "Error in login_reject\n" );
         }
      
         retval = -1;
         goto out;
      }
       
   }
   else
   {
      if ( (inputpdu->opcode & I_BIT) == I_BIT )
      {
         TRACE_ERROR("handle_login: I bit is not 0, should be reserved for"
                     " draft 11, 12 and 13\n");
      }
   }         

   /* check CSG */
   if ( (inputpdu->opcode & CSG) >= CSG2 )
   {
      TRACE_ERROR("check_first_login: stage error , terminate the connection\n");
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
                        conn->session->draft_number, outputpdu, 
			conn->connection_flags) < 0 )
      {
         TRACE_ERROR( "Error in login_reject\n" );
      }
      retval = -1;
      goto out;
   }

   /* check versions */
   if( inputpdu->version_active != conn->session->version_min )
   {   
      TRACE_ERROR("check_first_login: unsupported version, terminate the connection\n"); 
       
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR, STAT_DETAIL_VERSION_NOT_SUPPORTED,
                        conn->session->draft_number, outputpdu, 
			conn->connection_flags) < 0 )
      {
            TRACE_ERROR( "Error in login_reject\n" );
      }
      retval = -1;
      goto out;
   }
      
   conn->session->version_active = conn->session->version_max;
       
   /* check cmd_sn */
   if (inputpdu->cmd_sn > conn->session->max_cmd_sn ||
                 inputpdu->cmd_sn < conn->session->exp_cmd_sn)
   {
      TRACE_ERROR("check_first_login: cmd_sn error, Terminate the session\n");
      if ( login_reject(conn->conn_socket, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
                        conn->session->draft_number, outputpdu, 
			conn->connection_flags) < 0 )
      {
         TRACE_ERROR( "Error in login_reject\n" );
      }
                   
      retval = -1;
      goto out;
   }
       
out:
       
  TRACE(TRACE_ENTER_LEAVE, "Leave check_first_login\n");
  return retval;
}


int target_check_login(struct iscsi_conn *conn,
		struct generic_pdu * inputpdu,
		struct generic_pdu * outputpdu,
		unsigned int when_called,
		int noperational,
		unsigned long long * login_flags,
		__u16 tsih_value,
		struct unknown_key ** unknown_key_list)
{
	struct parameter_type * p_param_tbl = *conn->session->session_params;
	int *max_send_length = &conn->max_send_length;
	struct socket * sock = conn->conn_socket;
	int retval = 0;
	int correct_CSG = (inputpdu->flags & CSG) >> CSG_SHIFT;
	int add_length;
	int what_to_process;

	TRACE(TRACE_DEBUG, "Entering target check login\n");
	if (((inputpdu->flags & NSG) == NSG2) ||
		(!(inputpdu->flags & T_BIT) && (inputpdu->flags & NSG)) ||
		((inputpdu->flags & T_BIT) && 
			((inputpdu->flags & NSG) <= ((inputpdu->flags & CSG) >> CSG_SHIFT))))
	{
		TRACE_ERROR("invalid NSG %d\n", inputpdu->flags & NSG);
		retval = -1;
	}	
	if ((retval < 0) && (inputpdu->flags & T_BIT))
	{
		if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
			conn->session->draft_number, outputpdu,
			conn->connection_flags) < 0)
			TRACE_ERROR("error in login_reject\n");
		goto out;
	}
	outputpdu->flags &= ~CSG;
	outputpdu->flags |= (inputpdu->flags & CSG);
	if (inputpdu->flags & T_BIT)
	{
			outputpdu->flags &= ~NSG;
			outputpdu->flags |= (inputpdu->flags & NSG);
	}
	outputpdu->flags &= ~T_BIT;
	outputpdu->flags |= ( inputpdu->flags & T_BIT );
	if (!(*login_flags & FIRST_FLAG))
	{
		if (check_other_login(sock, correct_CSG, conn->session->draft_number,
			inputpdu, outputpdu, conn->connection_flags) < 0)
			goto out;
	}
	else
	{
		if (check_first_login(conn, inputpdu, outputpdu) < 0)
			goto out;
	}
      	if( (outputpdu->flags & CSG) == 0 )
	{/* now in security parameter negotiation stage */
		what_to_process = SECURITY_PARAM | INFORMATIONAL_PARAM;
		if( noperational > 0 )
		{/* we want to offer operational keys, force next stage to 1 */
			inputpdu->flags &= ~NSG;
			inputpdu->flags |= NSG1;
		}
	}
	else
	{/* now in operational parameter negotiation stage */
		what_to_process = OPERATIONAL_PARAM | INFORMATIONAL_PARAM;
	}

	add_length = scan_input_and_process(sock, p_param_tbl,
			what_to_process,
			TARGETNAME_FLAG | INITIATORNAME_FLAG | SESSIONTYPE_FLAG,
			TARGET,
			max_send_length,
			conn->session->draft_number,
			when_called,
			inputpdu,
			outputpdu,
			conn->connection_flags,
			login_flags,
			unknown_key_list);
	if (add_length < 0)
	{
		if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
			conn->session->draft_number, outputpdu,
			conn->connection_flags) < 0)
		{
			TRACE_ERROR("error in login reject\n");
		}
		return add_length;
	}
	if ((*login_flags) & FIRST_FLAG)
	{
		if (check_flags(conn, *login_flags, outputpdu) < 0)
		{
			return -1;
		}
	}

	outputpdu->text_length = add_length;
	add_length = scan_table_and_process(sock, p_param_tbl,
			what_to_process,
			TARGET,
			conn->session->draft_number,
			inputpdu, outputpdu,
			conn->connection_flags,
			login_flags);
	if (add_length < 0)
	{
		if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
			conn->session->draft_number, outputpdu,
			conn->connection_flags) < 0)
		{
			TRACE_ERROR("error in login reject\n");
		}
		return add_length;
	}
	outputpdu->text_length += add_length;

	outputpdu->status_class = 0;
	outputpdu->status_detail = 0;

	outputpdu->cmd_sn = cpu_to_be32(conn->stat_sn+1);
	outputpdu->exp_stat_sn = cpu_to_be32(conn->session->exp_cmd_sn); 
	outputpdu->max_cmd_sn = cpu_to_be32(conn->session->max_cmd_sn);

	/* For a new session TSIH MUST be set by the target in the final response and MUST be 0 otherwise */
	if( ( (inputpdu->tsih) == 0) && (!(inputpdu->flags & T_BIT)
		|| (inputpdu->flags&CSG) != CSG1 || (inputpdu->flags & NSG) != NSG3
		|| !(outputpdu->flags & T_BIT) || (outputpdu->flags&CSG) != CSG1
		|| (outputpdu->flags & NSG)!=NSG3 ) )
	{
		outputpdu->tsih = 0;
	}
	else
	{
		outputpdu->tsih = tsih_value;
	}
  
out:

	TRACE(TRACE_DEBUG, "Leave target_check_login\n");
	return retval;
}

int target_security_negotiate(struct iscsi_conn *conn,
		struct generic_pdu * inputpdu,
		struct generic_pdu * outputpdu,
		unsigned int when_called,
		unsigned long long * login_flags,
		int noperational,
		struct chap_parameter_type chap_param,
		struct unknown_key ** unknown_key_list)
{
	struct parameter_type * p_param_tbl = *conn->session->session_params;
	struct socket * sock = conn->conn_socket;
	int retval = 0;
	int padding = 0;
	struct parameter_type * auth_p;
	char * dummy_string;
	struct unknown_key	*key, *key_next, **key_prev;

	int neg_flags = 0;
	char * chap_r = NULL;
	char * chap_c = NULL;
	int chap_a = 0;
	int chap_i;
	char * chap_n;
	int tmp_len;
	int add_len;
	int tmp_val;
	__u16 tsih_value = outputpdu->tsih;
	
	ALLOCATE_MAX_TEXT_LEN(dummy_string);

	TRACE(TRACE_DEBUG, "Entering target security negotiate\n");
	if ((auth_p = find_flag_parameter(AUTHMETHOD_FLAG, p_param_tbl)) == NULL)
	{
		TRACE_ERROR("AuthMethod parameter not found\n");
		retval = -1;
		goto out;
	}
	if (target_check_login(conn,
			inputpdu,
			outputpdu,
			when_called,
			noperational,
			login_flags,
			tsih_value,
			unknown_key_list) < 0)
	{
		TRACE_ERROR("check login failed\n");
		retval = -1;
		goto out;
	}
	
	for( key = *unknown_key_list;  key != NULL;  key = key->next )
	{
		if (is_securitykey(key->keyname))
		{
			TRACE_ERROR("security key %s received is not allowed now\n", key->keyname);
			retval = -1;
			if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
				conn->session->draft_number, outputpdu,
				conn->connection_flags) < 0)
			{
				TRACE_ERROR("error in login reject\n");
			}
			goto out;
		}
	}

	if (iscsi_send_msg(sock, conn->session->draft_number, outputpdu, conn->connection_flags) < 0)
	{
		TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
		retval = -1;
		goto out;
	}
	else
	{
	  conn->stat_sn++;
	}

	outputpdu->text_length = 0;
	*login_flags &= ~FIRST_FLAG;

	if (!strcmp(auth_p->str_value, CHAP))	
	{
		if (iscsi_recv_msg(sock, ISCSI_HDR_LEN, (char *) inputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
			retval = -1;
			goto out;
		}

		inputpdu->text_length = be32_to_cpu(inputpdu->length);
		padding = -inputpdu->text_length & 3;
		if (inputpdu->text_length < MAX_TEXT_LEN)
		{
			if (iscsi_recv_msg(sock, inputpdu->text_length + padding,
				inputpdu->text, conn->connection_flags) < 0)
			{
				TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
				retval = -1;
				goto out;
			}
		}
		else
		{
			TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
				inputpdu->text_length, MAX_TEXT_LEN);
			retval = -1;
			goto out; 
		}
	
		if (target_check_login(conn,
				inputpdu,
				outputpdu,
				when_called,
				noperational,
				login_flags,
				tsih_value,
				unknown_key_list) < 0)
		{
			TRACE_ERROR("check login failed\n");
			retval = -1;
			goto out;
		}

		for( key = *(key_prev = unknown_key_list);  key != NULL;  key=key_next )
		{
			key_next = key->next;
			if (is_securitykey(key->keyname))
			{
				if (!strcmp(key->keyname, "CHAP_A"))
				{
					printk("got key chap_a %s\n", key->keyvalue);
					if( (chap_a=CHAP_SelectAlgorithm(key->keyvalue)) <= 0 )
						{
						TRACE_ERROR("unable to select algorithm\n");
						retval = -1;
						if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
							conn->session->draft_number, outputpdu,
							conn->connection_flags) < 0)
							{
							TRACE_ERROR("error in login reject\n");
							}
						goto out;
						}
					printk("algorithm selected is %d\n", chap_a);
					CHAP_SetAlgorithm(chap_a, chap_param.local_ctx);
					CHAP_SetAlgorithm(chap_a, chap_param.peer_ctx);
					neg_flags |= 0x01;
				}
				else
				{
					TRACE_ERROR("invalid security key %s\n", key->keyname);
					retval = -1;
					if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
						conn->session->draft_number, outputpdu,
						conn->connection_flags) < 0)
					{
						TRACE_ERROR("error in login reject\n");
					}
					goto out;
				}
				/* now delete this key from the list */
				my_kfree((void **)&key->keyvalue, "Key Value");
				my_kfree((void **)&key->keyname, "Key Name");
				my_kfree((void **)&key, "Unknown key");
				*key_prev = key_next;
			}
			else
			{
				key_prev = &key->next;
			}
		}

		if (neg_flags != 0x01)
		{
			TRACE_ERROR("security key chap_a not found\n");
			retval = -1;
			if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
				conn->session->draft_number, outputpdu,
				conn->connection_flags) < 0)
			{
				TRACE_ERROR("error in login reject\n");
			}
			goto out;
		}
		/*	generate an id number and a challenge string to send to him */
		chap_i = CHAP_GetIdentifier(chap_param.local_ctx);
		chap_c = CHAP_GetChallenge(chap_param.local_ctx);
		/* RDR start */
		if( chap_c == NULL )
		{
			TRACE_ERROR("CHAP exchange failed\n");
			goto out;
		}
		/* RDR end */
		tmp_len = sprintf(dummy_string, "CHAP_A=%d",chap_a);
		add_len = tmp_len + 1;
		printk("attach key chap_a %d\n",chap_a);
		tmp_len = sprintf(dummy_string + add_len, "CHAP_I=%d", chap_i);
		add_len += tmp_len + 1;
		printk("attach key chap_i %d\n", chap_i);
		tmp_len = sprintf(dummy_string + add_len, "CHAP_C=%s", chap_c);
		add_len += tmp_len + 1;
		outputpdu->text_length = add_len;
		memcpy(outputpdu->text, dummy_string, add_len);

		/*	finished with printable string version of our challenge to him */
		my_kfree((void **)&chap_c, "CHAP Challenge");
		
		if (iscsi_send_msg(sock, conn->session->draft_number, outputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
			retval = -1;
			goto out;
		}
		else
		{
	  		conn->stat_sn++;
		}	

		outputpdu->text_length = 0;

		if (iscsi_recv_msg(sock, ISCSI_HDR_LEN, (char *) inputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
			retval = -1;
			goto out;
		}

		inputpdu->text_length = be32_to_cpu(inputpdu->length);
		padding = -inputpdu->text_length & 3;
		if (inputpdu->text_length < MAX_TEXT_LEN)
		{
			if (iscsi_recv_msg(sock, inputpdu->text_length + padding,
				inputpdu->text, conn->connection_flags) < 0)
			{
				TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
				retval = -1;
				goto out;
			}
		}
		else
		{
			TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
				inputpdu->text_length, MAX_TEXT_LEN);
			retval = -1;
			goto out; 
		}

		if (target_check_login(conn,
				inputpdu,
				outputpdu,
				when_called,
				noperational,
				login_flags,
				tsih_value,
				unknown_key_list) < 0)
		{
			TRACE_ERROR("check login failed\n");
			retval = -1;
			goto out;
		}

		for( key = *(key_prev = unknown_key_list);  key != NULL;  key=key_next )
		{
			key_next = key->next;
			if (is_securitykey(key->keyname))
			{
				if (!strcmp(key->keyname, "CHAP_N"))
				{
					if (strcmp(key->keyvalue, chap_param.local_ctx->name))
					{
						TRACE_ERROR("security authentication failed\n");
						retval = -1;
						if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_NOT_AUTH,
							conn->session->draft_number, outputpdu,
							conn->connection_flags) < 0)
						{
							TRACE_ERROR("error in login reject\n");
						}
						goto out;
					}
					neg_flags |= 0x02;
					printk("got key chap_n %s\n", key->keyvalue);
				}
				else if (!strcmp(key->keyname, "CHAP_R"))
				{
					if ((chap_r = my_kmalloc(strlen(key->keyvalue)+1, "CHAP Response")) == NULL)
					{
						retval = -1;
						goto out;
					}
					strcpy(chap_r, key->keyvalue);
					neg_flags |= 0x04;
					printk("got key chap_r %s\n", key->keyvalue);
				}
				else if (!strcmp(key->keyname, "CHAP_I"))
				{
					tmp_val = simple_strtoul(key->keyvalue, NULL, 0);
					neg_flags |= 0x08;
					if ((tmp_val > 255) || (tmp_val < 0))
					{
						TRACE_ERROR("invalid chap_i %d\n", tmp_val);
						retval = -1;
						if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
							conn->session->draft_number, outputpdu,
							conn->connection_flags) < 0)
						{
							TRACE_ERROR("error in login reject\n");
						}
						goto out;
					}
					chap_i = tmp_val;
					printk("got key chap_i %d\n", tmp_val);
				}
				else if (!strcmp(key->keyname, "CHAP_C"))
				{
					if ((chap_c = my_kmalloc(strlen(key->keyvalue)+1, "CHAP Challenge")) == NULL)
					{
						retval = -1;
						goto out;
					}
					strcpy(chap_c, key->keyvalue);
					neg_flags |= 0x10;
					printk("got key chap_c %s\n", key->keyvalue);
				}
				else
				{
					TRACE_ERROR("invalid security key %s\n", key->keyname);
					retval = -1;
					if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
						conn->session->draft_number, outputpdu,
						conn->connection_flags) < 0)
					{
						TRACE_ERROR("error in login reject\n");
					}
					goto out;
				}
				/* now delete this key from the list */
				my_kfree((void **)&key->keyvalue, "Key Value");
				my_kfree((void **)&key->keyname, "Key Name");
				my_kfree((void **)&key, "Unknown key");
				*key_prev = key_next;
			}
			else
			{
				key_prev = &key->next;
			}
		}

		if ((neg_flags != 0x07) && (neg_flags != 0x1F))
		{
			TRACE_ERROR("not enough security keys received\n");
			retval = -1;
			if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
				conn->session->draft_number, outputpdu,
				conn->connection_flags) < 0)
			{
				TRACE_ERROR("error in login reject\n");
			}
			goto out;
		}
		else
		{
			if (CHAP_CheckResponse(chap_r, chap_param.local_ctx) <= 0)
			{
				TRACE_ERROR("security authentication failed\n");
				retval = -1;
				if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_NOT_AUTH,
					conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0)
				{
					TRACE_ERROR("error in login reject\n");
				}
				goto out;
			}
			/*	finished with his response to our challenge */
			my_kfree((void **)&chap_r, "CHAP Response");
		}
		if (neg_flags == 0x1F)
		{
			if ((chap_r = CHAP_GetResponse(chap_i, chap_c, chap_param.peer_ctx)) == NULL)
			{
				TRACE_ERROR("CHAP_R to Initiator cannot be generated\n");
				retval = -1;
				if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
					conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0)
				{
					TRACE_ERROR("error in login reject\n");
				}
				goto out;
			}
			if ((chap_n = CHAP_GetName(chap_param.peer_ctx)) == NULL )
			{
				TRACE_ERROR("CHAP_N to Initiator not configured\n");
				retval = -1;
				if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
					conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0)
				{
					TRACE_ERROR("error in login reject\n");
				}
				goto out;
			}
			tmp_len = sprintf(dummy_string, "CHAP_N=%s",chap_n);
			add_len = tmp_len + 1;
			printk("attach key chap_n %s\n",chap_n);
			/*	finished with our name to him */
			my_kfree((void **)&chap_n, "CHAP Get Name");

			tmp_len = sprintf(dummy_string + add_len, "CHAP_R=%s", chap_r);
			add_len += tmp_len + 1;
			printk("attach key chap_r %s\n", chap_r);
			outputpdu->text_length = add_len;
			memcpy(outputpdu->text, dummy_string, add_len);
		}

		if (iscsi_send_msg(sock, conn->session->draft_number, outputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
			retval = -1;
			goto out;
		}
		else
		{
	  		conn->stat_sn++;
		}

		outputpdu->text_length = 0;
	}

	while (!(outputpdu->flags & T_BIT))
	{
		if (iscsi_recv_msg(sock, ISCSI_HDR_LEN, (char *) inputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
			retval = -1;
			goto out;
		}

		inputpdu->text_length = be32_to_cpu(inputpdu->length);
		padding = -inputpdu->text_length & 3;
		if (inputpdu->text_length < MAX_TEXT_LEN)
		{
			if (iscsi_recv_msg(sock, inputpdu->text_length + padding,
				inputpdu->text, conn->connection_flags) < 0)
			{
				TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
				retval = -1;
				goto out;
			}
		}
		else
		{
			TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
				inputpdu->text_length, MAX_TEXT_LEN);
			retval = -1;
			goto out; 
		}

		if (target_check_login(conn,
				inputpdu,
				outputpdu,
				when_called,
				noperational,
				login_flags,
				tsih_value,
				unknown_key_list) < 0)
		{	
			TRACE_ERROR("check login failed\n");
			retval = -1;
			goto out;
		}

		for( key = *unknown_key_list;  key != NULL;  key = key->next )
		{
			if (is_securitykey(key->keyname))
			{
				TRACE_ERROR("security key %s received is not allowed now\n", key->keyname);
				retval = -1;
				if (login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
					conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0)
				{
					TRACE_ERROR("error in login reject\n");
				}
				goto out;
			}
		}

		if (iscsi_send_msg(sock, conn->session->draft_number, outputpdu, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
			retval = -1;
			goto out;
		}
		else
		{
	  		conn->stat_sn++;
		}

		outputpdu->text_length = 0;
	}

	if (iscsi_recv_msg(sock, ISCSI_HDR_LEN, (char *) inputpdu, conn->connection_flags) < 0)
	{
		TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
		retval = -1;
		goto out;
	}

	inputpdu->text_length = be32_to_cpu(inputpdu->length);
	padding = -inputpdu->text_length & 3;
	if (inputpdu->text_length < MAX_TEXT_LEN)
	{
		if (iscsi_recv_msg(sock, inputpdu->text_length + padding,
			inputpdu->text, conn->connection_flags) < 0)
		{
			TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
			retval = -1;
			goto out;
		}
	}
	else
	{
		TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
			inputpdu->text_length, MAX_TEXT_LEN);
		retval = -1;
		goto out; 
	}

	/* end up security phase */


out:
	my_kfree((void **)&chap_r, "CHAP Response");
	my_kfree((void **)&chap_c, "CHAP Challenge");

	FREE_STRING(dummy_string);
	TRACE(TRACE_DEBUG, "Leaving target security negotiate\n");
	return retval;
}


/************************************************************************/
/* This function performs parameter negotiation on target side         */
/* arguments: socket id, table of session parameters                   */
/*		length to be read from the socket in case of first     */
/*		login command sent from the initiator to the target.   */    
/************************************************************************/

static int target_parameter_negotiate(struct iscsi_conn *conn, 
			struct generic_pdu *inputpdu,
			struct generic_pdu *outputpdu,
			unsigned int when_called,
			struct chap_parameter_type chap_param, /* chap support - CHONG */
			struct unknown_key **unknown_key_list,
			int target_num	
			)
{
  struct parameter_type *p_param_tbl;
  struct socket *sock = conn->conn_socket;  
  int *max_send_length = &conn->max_send_length; 
  int retval=0, count;
  int padding, nsecurity, ninformational, noperational;
  int add_length;
  int correct_CSG;
  unsigned long long login_flags = FIRST_FLAG;
  unsigned int	what_to_process;
  __u16 tsih_value;
  
  TRACE(TRACE_ENTER_LEAVE, "Entering target_parameter_negotiate\n");
  
  p_param_tbl = *conn->session->session_params;

  /* find out number of each type of key we want to offer */
  scan_table_and_count(p_param_tbl, conn->session->draft_number,
  			 &nsecurity, &ninformational, &noperational);

  /* Include the padding bytes too  */
  padding = (-inputpdu->text_length) & 3;
  
  /* Get the input data - received parameter request */
  if( inputpdu->text_length > 0 )
    {
      if(iscsi_recv_msg(sock, inputpdu->text_length+padding, inputpdu->text,
						conn->connection_flags) < 0)
	{
	  TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
	  return -1;
	}
    }

  /* target always starts in the same state as initiator */
  correct_CSG = (inputpdu->flags & CSG) >> CSG_SHIFT;

  if( nsecurity > 0  &&  (inputpdu->flags & CSG) != 0 )
    {
      TRACE_ERROR("Initiator starting in operational negotiation stage "
      		  "but target wants to negotiate security keys\n");
      if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
	  				conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0 )
	{
	  TRACE_ERROR("Error in login_reject\n");
	}
      return -1;
    }


  /* store the session tsih of outputpdu */
   tsih_value = outputpdu->tsih;

  /* chap support - CHONG */
  if ((inputpdu->flags & CSG) == 0) 
  {
    if ((retval = target_security_negotiate(conn,
			    	inputpdu,
			    	outputpdu,
				when_called,
				&login_flags,
				noperational,
			 	chap_param,
				unknown_key_list)) < 0)
      goto out; 

    correct_CSG = (inputpdu->flags & CSG) >> CSG_SHIFT;
  }

  /* use count to control the loop times for login negotiation */
  count = 0;
  do
    {
      /* check if the input CSG is valid */
      if( (inputpdu->flags & CSG) >= CSG2 )
        {
	  TRACE_ERROR("Invalid CSG %d should be 0 or 1\n",
	  				(inputpdu->flags & CSG) >> CSG_SHIFT);
	  retval = -2;
      	}

      /* check if the input NSG is valid:
	 NSG can never be 2,
	 NSG must be 0 if the T bit is 0,
	 NSG must be greater than CSG if T bit is 1.
      */
      if( (inputpdu->flags & NSG) == NSG2  ||
      	  (!(inputpdu->flags & T_BIT)  &&  (inputpdu->flags & NSG))  ||
      	  ((inputpdu->flags & T_BIT)  &&
		  (inputpdu->flags & NSG) <= ((inputpdu->flags & CSG)>>CSG_SHIFT)) )
        {
	  TRACE_ERROR("Invalid NSG %d \n", inputpdu->flags & NSG);
	  retval = -2;
      	}

      if( retval == -2 &&  (inputpdu->flags & T_BIT)) /* if T bit is set and CSG/NSG is wrong */
	{
            if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
	  				conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0 )
	     {
	       TRACE_ERROR("Error in login_reject\n");
	     }
	     goto out;
	}
      
      outputpdu->flags &= ~CSG; /* output.CSG = inputpdu->CSG */
      outputpdu->flags |= (inputpdu->flags & CSG);

      outputpdu->flags &= ~T_BIT;  /* output.T = inputpdu->T */
      outputpdu->flags |= ( inputpdu->flags & T_BIT );
      
      if ( !(login_flags & FIRST_FLAG) )
	{
	  /* only check the non-first login */
	  if (check_other_login(sock, correct_CSG, conn->session->draft_number,
			  inputpdu, outputpdu, conn->connection_flags) < 0 )
	     goto out;
	} 
      else
       { /*check the first login */
	  if( check_first_login(conn, inputpdu, outputpdu)<0 )
	     goto out;
       }      
      
      if( (outputpdu->flags & CSG) == 0 )
	{/* now in security parameter negotiation stage */
	  what_to_process = SECURITY_PARAM | INFORMATIONAL_PARAM;
	  if( noperational > 0 )
	    {/* we want to offer operational keys, force next stage to 1 */
	      inputpdu->flags &= ~NSG;
	      inputpdu->flags |= NSG1;
	    }
	}
      else
        {/* now in operational parameter negotiation stage */
          what_to_process = OPERATIONAL_PARAM | INFORMATIONAL_PARAM;
	}

      count++;  /* to avoid infinite loops */
      if( count >= LOOP_TIMES )
      {
        TRACE_ERROR("Infinite loop in parameter negotiations\n");
        if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
                           conn->session->draft_number, outputpdu,
                           conn->connection_flags) < 0 )
        {
        	TRACE_ERROR("Error in login_reject\n");
        }
		retval = -1;
        goto out;
      }


      /* process each key attached to input */
      add_length = scan_input_and_process( sock, p_param_tbl, 
		   what_to_process,
		   TARGETNAME_FLAG | INITIATORNAME_FLAG| SESSIONTYPE_FLAG,
		   TARGET,
		   max_send_length,
		   conn->session->draft_number,
		   when_called,
		   inputpdu,
		   outputpdu,
		   conn->connection_flags,
		   &login_flags,
		   unknown_key_list );
	  
      if( add_length < 0 )
    	{/* error */
	  if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
	  				conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0 )
	    {
	      TRACE_ERROR("Error in login_reject\n");
	    }
	  return add_length;
    	}
  
      /* check if the initial login includes initiator name, target name,
        sessiontype( draft 8,9) */

      if( login_flags & FIRST_FLAG )
      {
        if(  check_flags( conn, login_flags, outputpdu ) < 0  )
         {
            return -1;
         }              
      } 

      outputpdu->text_length = add_length;
	
      /* scan through our table and process */
      add_length = scan_table_and_process(  sock,
					p_param_tbl,
					what_to_process,
					TARGET,
					conn->session->draft_number,
					inputpdu, outputpdu,
					conn->connection_flags,
					&login_flags);

      if( add_length < 0 )
    	{/* error */
	  if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
	  				conn->session->draft_number, outputpdu,
					conn->connection_flags) < 0 )
	    {
	      TRACE_ERROR("Error in login_reject\n");
	    }
	  return add_length;
    	}

      outputpdu->text_length += add_length;
	
      /*  after finishing attaching all keys to output */
      if( outputpdu->flags & T_BIT )	
        { /* check whether all parameters that were sent in this stage
             have received their responses */
          if( check_neg_responses(p_param_tbl, 0) < 0 )
            {
              if( count < LOOP_TIMES )
                {/* reset the T bit to be 0 for more negotiations */
                  outputpdu->flags &= (~T_BIT);
                  outputpdu->flags &= (~NSG);
                }
              else /* It seems the initiator doesn't reply the key */
                { /* error */
                  TRACE_ERROR("Target didn't receive all the responses\n");
                  if( login_reject(sock, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR,
                                    conn->session->draft_number, outputpdu,
                                    conn->connection_flags) < 0 )
                  {
                    TRACE_ERROR("Error in login_reject\n");
                  }
                  return -1;
                }
            }
          else
            { /* output.NSG = input.NSG, we will change to the next state now */
                outputpdu->flags &= (~NSG);
                outputpdu->flags |= inputpdu->flags & NSG;
            }
        } 
      else
        {
          /* not changing the state */
           outputpdu->flags &= (~NSG);
        }
           	
      outputpdu->status_class = 0;
      outputpdu->status_detail = 0;

      outputpdu->cmd_sn = cpu_to_be32(conn->stat_sn+1);
      outputpdu->exp_stat_sn = cpu_to_be32(conn->session->exp_cmd_sn); 
      outputpdu->max_cmd_sn = cpu_to_be32(conn->session->max_cmd_sn);

      /* For a new session TSIH MUST be set by the target in the final
	 response and MUST be 0 otherwise */
      if( ( (inputpdu->tsih) == 0) && (!(inputpdu->flags & T_BIT)  
	|| (inputpdu->flags&CSG) != CSG1 || (inputpdu->flags & NSG) != NSG3 
	|| !(outputpdu->flags & T_BIT) || (outputpdu->flags&CSG) != CSG1 
	|| (outputpdu->flags & NSG)!=NSG3 ) )
        {
                outputpdu->tsih = 0;
        }
      else
        {
                outputpdu->tsih = tsih_value;
        }

      /* send out the output pdu */
      if( iscsi_send_msg(sock, conn->session->draft_number, outputpdu,
						conn->connection_flags) < 0 )
	{
	  TRACE(TRACE_DEBUG, "iscsi_send_msg failed\n");
	  retval = -1;
	  goto out;
	}
      else
	{
	  conn->stat_sn++;
	}
	
      /* after sending the output pdu, reset its length to 0 */
      outputpdu->text_length = 0;
	
      login_flags &= ~FIRST_FLAG;	
	
      if( outputpdu->flags & T_BIT )	
	{ /* both sides change state now after target sends a reply */

	  /*  for draft 8 or before, enable digests at end of security stage.
	      Draft 8, section 5.2: "All PDUs sent after the security
	      negotiation stage MUST be built using the agreed security."

	      for draft 9 or after, enable digests at end of login phase.
	      Draft 9, Appendix D 10: "When Initiator and Target agree on a
	      digest, this digest MUST be used for every PDU in Full
	      Feature Phase."
	  */
	  if( (conn->session->draft_number <= DRAFT8  &&
				      (outputpdu->flags & CSG) == 0)  ||
	      (conn->session->draft_number >= DRAFT9  &&
				      (outputpdu->flags & NSG) == NSG3) )
	    {/*  enable digests now */
	    set_digestflags(p_param_tbl, &conn->connection_flags);
	    }

	  if( (outputpdu->flags & NSG) == NSG3 )
	    {/* go to full feature phase, end of login phase */
	      goto out;
	    }
	  
	  /* change to a new (operational) state */
	  outputpdu->flags &= (~CSG);
	  outputpdu->flags |= ((inputpdu->flags & NSG) << CSG_SHIFT);
	}

      /* keep the last outputpdu.CSG, since that is the state we are in now */
      correct_CSG = (outputpdu->flags & CSG) >> CSG_SHIFT;
	
      /*  wait for input from initiator  */
      if( iscsi_recv_msg(sock, ISCSI_HDR_LEN, (char *)inputpdu,
     						 conn->connection_flags) < 0 )
	{
	  TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
	  return -1;
	}

      TRACE(TRACE_ISCSI, "Got Login Command, ITT %u\n",
                                        ntohl(inputpdu->init_task_tag));
     
      if( TRACE_TEST(TRACE_ISCSI_FULL) )
        print_init_login_cmnd((struct iscsi_init_login_cmnd *)inputpdu);
      
      /*  Also get the text associated with it  */
      inputpdu->text_length = be32_to_cpu(inputpdu->length);
      
      /* Include the padding bytes too */
      padding = -inputpdu->text_length & 3;
      
      if( inputpdu->text_length <= MAX_TEXT_LEN )
	{
	  if( iscsi_recv_msg(sock, inputpdu->text_length + padding,
	 			 inputpdu->text, conn->connection_flags) < 0 )
	    {
		TRACE(TRACE_DEBUG, "iscsi_recv_msg failed\n");
		retval = -1;
		goto out;
	    }
	}
      else
	{
	  TRACE_ERROR("DSL %u greater than default MaxRecvDataSegmentLength %d\n",
					inputpdu->text_length, MAX_TEXT_LEN);
	  retval = -1;
	  goto out;
	}

    }while(1);
  
  out:

  TRACE(TRACE_ENTER_LEAVE, "Leave target_parameter_negotiate\n");
  
  return retval;
    
}


int parameter_negotiate( struct iscsi_conn *conn,
			 struct iscsi_init_login_cmnd *loginpdu,
			 unsigned int when_called, 
			 struct chap_parameter_type chap_param,
			 int target_num)
{
  int				retval = 0;
  struct generic_pdu		*inputpdu, *outputpdu;
  struct unknown_key		*uptr, *unknown_key_list = NULL;


  TRACE(TRACE_ENTER_LEAVE, "Enter parameter_negotiate\n");

  if( (inputpdu=my_kmalloc(sizeof(struct generic_pdu), "Input pdu")) == NULL )
  	{
	retval = -1;
	goto out;
	}

  if( (outputpdu=my_kmalloc(sizeof(struct generic_pdu), "Output pdu")) == NULL )
  	{
	my_kfree((void **)&inputpdu, "Input pdu");
	retval = -1;
	goto out;
	}

  /* Get memory for text portion of inputpdu and outputpdu */
  if( (inputpdu->text = my_kmalloc(MAX_TEXT_LEN, "Input text")) == NULL )
  	{
	my_kfree((void **)&outputpdu, "Output pdu");
	my_kfree((void **)&inputpdu, "Input pdu");
	retval = -1;
	goto out;
	}

  if( (outputpdu->text = my_kmalloc(MAX_TEXT_LEN, "Output text")) == NULL )
  	{
	my_kfree((void **)&inputpdu->text, "Input text");
	my_kfree((void **)&outputpdu, "Output pdu");
	my_kfree((void **)&inputpdu, "Input pdu");
	retval = -1;
	goto out;
	}

  /* fill inputpdu with copy of login pdu header received from initiator */
  memcpy(inputpdu, loginpdu, ISCSI_HDR_LEN);
  inputpdu->text_length = loginpdu->length;	/* still have to read this! */

  /* reset outputpdu to all 0 so reserved fields will be 0 */
  memset(outputpdu, 0, ISCSI_HDR_LEN);
  outputpdu->text_length = 0;

  /* set the initiator task tag */
  outputpdu->init_task_tag = cpu_to_be32(loginpdu->init_task_tag);
      
  /*set the ISID and TSIH for this session */
  memcpy(outputpdu->isid, conn->session->isid, 6);
  outputpdu->tsih = cpu_to_be16(conn->session->tsih);

  /* Set opcode */
  outputpdu->opcode = ISCSI_TARG_LOGIN_RSP;

  /* Set the first bit in byte 0 in all responses for draft 8 and before */
  /* Set the I bit in all responses for draft 9 and before */
  if( conn->session->draft_number <= DRAFT8 )
      {/* draft 8 and before */
      outputpdu->opcode |= OLD_X_BIT|I_BIT;
      }
  else if( conn->session->draft_number <= DRAFT9 )
      {/* draft 9 */
      outputpdu->opcode |= I_BIT;
      }

  /* set the version_max and version_active for the session  */
  outputpdu->version_max = conn->session->version_max;
  outputpdu->version_active = conn->session->version_min;

  retval = target_parameter_negotiate(conn, 
				      inputpdu,
				      outputpdu,
				      when_called,
					  chap_param,
				      &unknown_key_list,
				      target_num
				  );

  /*     Draft 13, Section 4.2 Text Mode Negotiation
         "Some parameters may be subject to integrity rules (e.g., parameter-x
         must not exceed parameter-y or parameter-y not 1 implies parameter-v
         to by Yes).  Whenever required integrity rules are specified with the
         keys.  Checking for compliance with the integrity rule MUST NOT be
         performed before all the negotiation parameters are available (the
         existent and newly negotiated).  An iSCSI target MUST perform integ-
         rity checking before committing new values for parameters.  An
         initiator MAY perform integrity checking."
   */
   check_integrity_rules( *conn->session->session_params, conn->session->draft_number);
 
  /*  if the target sent a MaxRecvDataSegmentLength (draft 13)
      value to the target, we have to now start using it in FFP 
  */ 
  set_connection_recv_length( *conn->session->session_params, &conn->max_recv_length, conn->session->draft_number);

  /* free up memory allocated for any unknown keys we received */
  while( unknown_key_list != NULL )
    {
    uptr = unknown_key_list->next;
	my_kfree((void **)&unknown_key_list->keyvalue, "Key Value");
	my_kfree((void **)&unknown_key_list->keyname, "Key Name");
	my_kfree((void **)&unknown_key_list, "Unknown key");
    unknown_key_list = uptr;
    }
  
  /* Free all allocated memory */
  my_kfree((void **)&outputpdu->text, "Output text");
  my_kfree((void **)&inputpdu->text, "Input text");
  my_kfree((void **)&outputpdu, "Output pdu");
  my_kfree((void **)&inputpdu, "Input pdu");
  
out:

  TRACE(TRACE_ENTER_LEAVE, "Leave parameter_negotiate, retval %d\n", retval);

  return retval;
}


/* set back the leading only keys if these keys were set to be
 * " KEY_TO_BE_NEGOTIATED" in the leading connection negotation.*/
void reset_parameter_table(struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS])
{
	struct parameter_type *p = NULL;
	int i;
	TRACE(TRACE_ENTER_LEAVE, "Enter reset_parameter_table\n");
	for( i = 0, p= p_param_tbl; i < MAX_CONFIG_PARAMS; i++, p++  )
	{
		if( IS_LEADING_ONLY(p->type)  
			&& IS_KEY_TO_BE_NEGOTIATED(p->neg_info) )
			p->neg_info &=~KEY_TO_BE_NEGOTIATED;
	}
	
	TRACE(TRACE_ENTER_LEAVE, "Leave reset_parameter_table\n");
}
