/***************************************************************************
*	ref18/target/iscsi_target.c
*
*
* 	This is the iscsi target front-end which interfaces with the STML. 
*	The front-end has been written to the specifications of Draft 13 of the 
*	iSCSI spec.
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
***************************************************************************/

#include "iscsi_target.h"
#include "target_error_rec.h"
/* chap support - CHONG */
#include "../iscsi-unh-common/chap.c"
#include "../iscsi-unh-common/misc_func.c"


/* 
 * initialize stuff in devdata.
 */
int init_target (struct iscsi_global *devdata)
{
	devdata->conn_id = 0;
	devdata->ntsih = 0;
	devdata->nsessions = 0;
	devdata->session_list = NULL;
	devdata->nop_period = 0;
	devdata->force = 0;
	/* Initialize SNACK flags - SAI */
	devdata->targ_snack_flg = 0;
	devdata->draft_number = DEFAULT_DRAFT;
	devdata->param_tbl = kmalloc (MAX_CONFIG_PARAMS * sizeof (struct 
		parameter_type), GFP_KERNEL);
	/* chap support - CHONG */
	devdata->chap_parameter.chap_flags = 0;
	devdata->chap_parameter.local_ctx = CHAP_InitializeContext();
	devdata->chap_parameter.peer_ctx = CHAP_InitializeContext();
	
	if (!(devdata->param_tbl))
	{
		TRACE_ERROR("init_target: Unable to allocate memory to param_tbl\n");
		return -1;
	}

	/* Copy the default parameters */
	param_tbl_cpy (*devdata->param_tbl, config_params);

	return 0;
}


/*
 * iscsi_detect: this function sets up the server so that it can accept 
 * connections.
 */
int iscsi_detect (Scsi_Target_Template *tmpt)
{
	struct socket	*sock;
	struct sockaddr_in	sock_data;
	int			backlog = 5;

	TRACE(TRACE_DEBUG, "Entering iSCSI detect\n");
	
	devdata = (struct iscsi_global*) kmalloc (sizeof (struct iscsi_global), 
		GFP_KERNEL);

	if (!devdata) 
	{
		TRACE_ERROR("iscsi_detect: Unable to allocate memory to devdata\n");
		return -1;
	}

	/* Initialize devdata */
	if (init_target (devdata) < 0)
	{
		TRACE_ERROR("iscsi_detect: Initialize devdata failed\n");
		return -1;
	}
	
	init_MUTEX		(&devdata->session_sem);
	init_MUTEX_LOCKED 	(&devdata->server_sem);

	TRACE(TRACE_DEBUG, "iscsi_detect: iSCSI initialization completed\n");
	
	/* create a socket */
	if (sock_create (PF_INET, SOCK_STREAM, 0, &sock) < 0) 
	{
		TRACE_ERROR("iscsi_detect: Could not create socket\n");
		return -2;
	}

	TRACE(TRACE_NET, "iscsi_detect: Socket created as %p\n", sock);

	/* bind */
	sock_data.sin_family		= AF_INET;
	sock_data.sin_port		= htons (ISCSI_PORT);
	sock_data.sin_addr.s_addr	= htonl (INADDR_ANY);

	if (sock->ops->bind (sock, (struct sockaddr*) &sock_data, sizeof (struct 
		sockaddr))) 
	{
		TRACE_ERROR("iscsi_detect: Could not bind to socket\n");
		return -3;
	}

	TRACE(TRACE_NET, "iscsi_detect: Socket bound to %p\n", sock);
	
	/* listen */
	if ((sock->ops->listen (sock, backlog))) 
	{
		TRACE_ERROR("iscsi_detect: Problems with listen\n");
		return -4;
	}

	TRACE(TRACE_NET, "iscsi_detect: Now listening at %p\n", sock);
	
	devdata->server_socket = sock;

	/* create a server_thread that can accept connections */
	kernel_thread ((int (*)(void *))iscsi_server_thread, (void*)devdata, 0);

	TRACE(TRACE_NET, "iscsi_detect: Server_thread spawned at %p\n", sock);
	
	/* register the front end now */
	devdata->device = register_target_front_end (tmpt);

	if (!devdata->device) 
	{
		TRACE_ERROR("iscsi_detect: Device registration failed\n");
		return -5;
	}

	devdata->device->dev_specific = (void*) devdata;

	TRACE(TRACE_DEBUG, "iscsi_detect: Registration complete\n");
	
	return 1; /* one device detected */
}


/*
 * iscsi_release: function to release the iSCSI device as required by
 * the Mid-Level
 */
int iscsi_release (Scsi_Target_Device *device)
{
	if ((struct iscsi_global*) device->dev_specific != devdata) 
	{
		TRACE_ERROR("iscsi_release: This is not an iscsi device\n");
		return -1;
	}

	while (devdata->session_list) 
	{
		TRACE(TRACE_ISCSI, "iscsi%d: Releasing session %.2x%.2x\n", (int) 
			devdata->device->id, devdata->session_list->isid[0], 
			devdata->session_list->tsih);

		if (iscsi_release_session(devdata->session_list) < 0)
		{
			TRACE_ERROR("iscsi_release: Error releasing session\n");
		}
	}

	if (devdata->server_thr)
	{
		send_sig (SIGKILL, devdata->server_thr, 1);
		down_interruptible (&devdata->server_sem);
	}

	if (devdata->server_socket)
	{
		sock_release (devdata->server_socket);
	}

	if (devdata->param_tbl)
	{
		param_tbl_uncpy (*devdata->param_tbl);
		kfree (devdata->param_tbl);
	}
	
	/* chap support - CHONG */
	CHAP_FinalizeContext(devdata->chap_parameter.local_ctx);
	CHAP_FinalizeContext(devdata->chap_parameter.peer_ctx);

	kfree (devdata);
	devdata = NULL;

	/* check for any unfreed memory */
	my_kempty();

	TRACE(TRACE_DEBUG, "iscsi_release: Exiting iscsi_release\n");
	
	return 0;
}


/*
 * iscsi_server_thread: This thread is responsible for accepting
 * connections
 */
void iscsi_server_thread (void *param)
{
	struct iscsi_conn	*new_conn;
	struct iscsi_global 	*dev;
	struct socket	*newsock, *sock;

	TRACE(TRACE_DEBUG, "Starting the iSCSI server\n");

	dev = (struct iscsi_global*) param;

	if (!dev)
	{
		TRACE_ERROR("iscsi_server_thread: NULL parameter\n");
		return;
	}

	lock_kernel();

	siginitsetinv (&current->blocked, ISCSI_SHUTDOWN_SIGS);
	daemonize();
	sprintf (current->comm, "iscsi_thr");
	dev->server_thr = current;

	unlock_kernel();

	sock = dev->server_socket;

	/* now create a loop for accepting connections */
	while (1) 
	{
		/* accept connections */
		newsock = sock_alloc();

		if (!newsock) 
		{
			TRACE_ERROR("iscsi_server_thread: Could not allocate a socket\n");
			goto iscsi_server_thread_out;
		}

		newsock->type	= sock->type;
		newsock->ops	= sock->ops;

		TRACE(TRACE_NET,"before accept: sock %p newsock %p\n",sock,newsock);

		if (sock->ops->accept (sock, newsock, 0) < 0)
		{
			TRACE_ERROR("iscsi_server_thread: Trouble at accept\n");
			sock_release(newsock); //Edward
			goto iscsi_server_thread_out;
		}

		TRACE(TRACE_NET,"after accept: sock %p newsock %p\n",sock,newsock);

		/* clear bad sessions and connections here */
		clean_bad_session (devdata->session_list);

		/* fill in connection details */
		new_conn = (struct iscsi_conn*) kmalloc (sizeof (struct iscsi_conn), 
			GFP_KERNEL);

		if (!new_conn) 
		{
			TRACE_ERROR("iscsi_server_thread: Could not create new connection\n");
			goto iscsi_server_thread_out;
		}

		TRACE(TRACE_NET,"connection: new_conn %p newsock %p\n",new_conn,newsock);

		memset(new_conn, 0, sizeof(struct iscsi_conn));

		new_conn->active	= 1;
		new_conn->conn_id	= ++dev->conn_id;
		new_conn->conn_socket	= newsock;
		new_conn->dev		= dev;
		new_conn->max_send_length = 8192;
		new_conn->max_recv_length = 8192;

		init_MUTEX_LOCKED (&new_conn->tx_sem);
		init_MUTEX_LOCKED (&new_conn->thr_kill_sem);

		/* 
		 * NOTE THAT WE CANNOT DO MUCH OF ANYTHING WITH THIS
		 * CONNECTION AS WE DO NOT KNOW WHAT SESSION IT BELONGS
		 * TO. THIS WILL HAVE TO BE A PART OF THE RX_THREAD
		 */

		/* start the rx thread */
		kernel_thread ((int(*)(void*)) iscsi_rx_thread, (void*) new_conn, 0);

		/* start the tx thread */
		kernel_thread ((int(*)(void*)) iscsi_tx_thread, (void*) new_conn, 0);
	}

iscsi_server_thread_out:
	dev->server_thr = NULL;

	if (devdata->session_list)
		iscsi_release_session(devdata->session_list);

	up (&dev->server_sem);

	return;
}


void clean_bad_session (struct iscsi_session *session)
{
	struct iscsi_session *next_session = NULL; // Edward

	while (session)
	{
		int active = 1;
		struct iscsi_conn *conn = session->conn_list;

		while (conn)
		{
			if (conn->active == 1)
			{
				conn = conn->next;
			}
			else
			{
				active = 0;
				break;
			}
		}

		if (active == 1)
		{
			session = session->next;
			continue;
		}

		TRACE(TRACE_ISCSI, "iscsi%d: Releasing session %.2x%.2x\n", (int) 
			devdata->device->id, session->isid[0], session->tsih);

		next_session = session->next;

		if (iscsi_release_session (session) < 0) 	
		{
			TRACE_ERROR("clean_bad_session: Error releasing session\n");
		}

		session = next_session;
	}
}


/*
 * iscsi_release_session: This function is responsible for closing out
 * a session. I do not know the level of granularity that is expected
 * by iscsi. Does a connection have special release mechanisms ?
 * INPUT: session to be released
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
static inline int iscsi_release_session (struct iscsi_session *session)
{
	struct iscsi_cmnd *cmnd;
	
	if (!session) 
	{
		TRACE_ERROR("iscsi_release_session: Cannot release a NULL session\n");
		return -1;
	}

	/* free commands */
	while (session->cmnd_list) 
	{
		cmnd = session->cmnd_list;
		session->cmnd_list = cmnd->next;

		if (cmnd->cmnd != NULL)
		{
			if(scsi_release (cmnd->cmnd) < 0)
			{
				TRACE_ERROR("iscsi_release_session: Trouble releasing command\n");
			}
		}

		kfree (cmnd);
		cmnd = NULL;
	}

	/* free connections */
	while (session->conn_list) 
	{
		TRACE(TRACE_ISCSI, "iscsi%d: releasing connection %d\n", (int) 
			session->devdata->device->id, session->conn_list->conn_id);

		if (iscsi_release_connection (session->conn_list) < 0)
		{
			TRACE_ERROR("iscsi_release_session: Trouble releasing connection\n");
		}
	}

	/* dequeue session */
	if (!(session->prev) && !(session->next) && 
		(session->devdata->session_list != session))
	{
		goto out;
	}
	
	down_interruptible(&session->devdata->session_sem);

	if (session->prev)
		session->prev->next = session->next;
	else	
		session->devdata->session_list = session->next;
	
	if (session->next)
		session->next->prev = session->prev;

	up(&session->devdata->session_sem);

	/* error recovery ver new 18_04 - SAI */
	if (session->retran_thread)
	{
		send_sig (SIGKILL, session->retran_thread, 1);
		down_interruptible(&session->thr_kill_sem);
	}
	
out:
	/* free session */
	if (session->address)
		kfree (session->address);

	if (session->session_params)
		kfree (session->session_params);

	if (session->oper_param)
		kfree (session->oper_param);

	kfree (session);
	session = NULL;

	return 0;
}


/*
 * iscsi_release_connection: releases all the stuff related to a conn.
 * Right now, it does not do anything for releasing commands specific
 * to a connection since commands are more specific to a session as
 * against a connection. It also dequeues the connection from the list
 */
static inline int iscsi_release_connection (struct iscsi_conn *conn)
{
	if (!conn) 
	{
		TRACE_ERROR("iscsi_release_connection: Cannot release NULL connection");
		return -1;
	}

	/* Kill the rx and tx thread */
	TRACE(TRACE_ISCSI, "Releasing iscsi_rx_%d\n", conn->conn_id);

	if (conn->rx_thread)
	{
		send_sig (SIGKILL, conn->rx_thread, 1);
		down_interruptible (&conn->thr_kill_sem);
	}
		
	TRACE(TRACE_ISCSI, "Releasing iscsi_tx_%d\n", conn->conn_id);
	
	if (conn->tx_thread)
	{
		send_sig (SIGKILL, conn->tx_thread, 1);
		down_interruptible (&conn->thr_kill_sem);
	}
		
	TRACE(TRACE_ISCSI, "Releasing R2T timer %p \n", conn->r2t_timer);

    /* Delete r2t timer - SAI */
    if (conn->r2t_timer != NULL)
    {
        TRACE(TRACE_DEBUG, "Deleting r2t timer %p\n", conn->r2t_timer);
        del_timer(conn->r2t_timer);
        TRACE(TRACE_DEBUG, "Deleted r2t timer\n");
        kfree(conn->r2t_timer);
    }

	TRACE(TRACE_ISCSI, "Releasing conn socket %p \n", conn->conn_socket);

	/* Release socket */
	if (conn->conn_socket != NULL) 
	{
		TRACE(TRACE_NET, "iscsi_release_connection: releasing socket %p "
			"for connection %d\n", conn->conn_socket, conn->conn_id);

		sock_release (conn->conn_socket);
	}

	TRACE(TRACE_ISCSI, "Dequeue connection conn->prev %p conn->next %p \n", conn->prev, conn->next);

	/* Dequeue connection */
	down_interruptible(&conn->session->conn_sem);

	if (conn->prev)
		conn->prev->next = conn->next;
	else
		conn->session->conn_list= conn->next;
	
	if (conn->next)
		conn->next->prev = conn->prev;

	up(&conn->session->conn_sem);

	/* Free connection */
	kfree (conn);
	conn = NULL;

	return 0;
}


/* 
 * iscsi_rx_thread: This thread is responsible for receiving iSCSI PDUs
 * and messages from the Initiator and then figure out what to do with
 * them
 */
void iscsi_rx_thread (void *param)
{
	struct iscsi_conn	*conn = (struct iscsi_conn*) param;
	struct msghdr	msg;
	struct iovec		iov;
	struct targ_error_rec	err_rec;	
	__u8			buffer[ISCSI_HDR_LEN];
	__u32		digest;
	__u32		hdr_crc = 0;

	if (!conn)
	{
		TRACE_ERROR("iscsi_rx_thread: Invalid parameter\n");
		return;
	}

	lock_kernel();

	siginitsetinv (&current->blocked, ISCSI_SHUTDOWN_SIGS);
	daemonize();
	sprintf (current->comm, "iscsi_rx_%d", conn->conn_id);
	conn->rx_thread = current;

	unlock_kernel();

	TRACE(TRACE_ISCSI, "Starting iscsi_rx_%d\n", conn->conn_id);

	/* receive loop */
	while (1) 
	{
		memset (buffer, 0x0, ISCSI_HDR_LEN);
		iov.iov_base	= buffer;
		iov.iov_len	= ISCSI_HDR_LEN;
		msg.msg_iov	= &iov;
		msg.msg_iovlen	= 1;
		
		/* receive iSCSI header */
		if (rx_data (conn, &msg, ISCSI_HDR_LEN) <= 0)
		{
			TRACE_ERROR("iscsi_rx_thread: Trouble receiving data\n");
			sock_release (conn->conn_socket);
            conn->conn_socket = NULL;       /* sock_release does not re-initialize socket - SAI */
			goto iscsi_rx_thread_out;
		}

		if (TRACE_TEST(TRACE_ISCSI_FULL))
		{
			int i=0;
			for (i=0; i<ISCSI_HDR_LEN; i++)
			{
				conn->bad_hdr[i] = buffer[i];
				printk("%02x ", buffer[i]);
				if (i==24) printk("\n");
			}
			printk("\n");
		}

		// if get full feature phase pdu before login phase
		if ((buffer[0] & ISCSI_OPCODE) != ISCSI_INIT_LOGIN_CMND && 
			!(conn->session))
		{
			TRACE_ERROR("iscsi_rx_thread: Expect login pdu\n");

			if (conn->dev->draft_number <= DRAFT9)
				iscsi_tx_rjt (conn, REASON_FULL_BEFORE_LOGIN);
			else if (conn->dev->draft_number <= DRAFT12)
				iscsi_tx_login_rsp (conn, STAT_CLASS_INITIATOR, 
					STAT_DETAIL_INVALID_DURING_LOGIN);

			sock_release (conn->conn_socket);
			goto iscsi_rx_thread_out;
		}

		// for header digest in full feature phase
		if ((buffer[0] & ISCSI_OPCODE) != ISCSI_INIT_LOGIN_CMND && 
			conn->hdr_crc == 1)
		{
			TRACE(TRACE_DEBUG, "iscsi_rx_thread: Header digest check\n");
			do_crc (buffer, ISCSI_HDR_LEN, 0x3, &hdr_crc);

			iov.iov_base	= &digest;
			iov.iov_len	= CRC_LEN;
			msg.msg_iov	= &iov;
			msg.msg_iovlen	= 1;

			if (rx_data (conn, &msg, CRC_LEN) <= 0)
			{
				TRACE_ERROR("iscsi_rx_thread: Trouble receiving data\n");
				sock_release (conn->conn_socket);
				goto iscsi_rx_thread_out;
			}

			if (hdr_crc != digest)
			{
				TRACE_ERROR("iscsi_rx_thread: Header crc 0x%08x computed 0x%08x\n", 
					ntohl(digest), ntohl(hdr_crc));
				/* Error Recovery Code Added by SAI */
				err_rec.curr_conn = conn;
				err_rec.pdu_hdr  = (struct generic_pdu *)buffer;
				err_rec.pdu_hdr->length = be32_to_cpu(err_rec.pdu_hdr->length);
				err_rec.err_type = HEADER_DIGERR;
				targ_do_error_recovery(&err_rec);
				continue;
			}
			else
			{
				TRACE(TRACE_ISCSI, "Got header crc 0x%08x\n", ntohl(digest));
			}
		}

		switch (buffer[0] & ISCSI_OPCODE) 
		{
			case ISCSI_INIT_LOGIN_CMND:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got login request\n", 
					conn->conn_id);

				if (handle_login (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_login\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			case ISCSI_INIT_TEXT_CMND:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got text request\n", 
					conn->conn_id);

				if (handle_discovery (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_discovery\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			case ISCSI_INIT_SCSI_CMND:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got SCSI command\n", 
					conn->conn_id);	

				if (handle_cmnd (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_cmnd\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}
			
			case ISCSI_INIT_SCSI_DATA_OUT:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got SCSI data-out\n", 
					conn->conn_id);

				if (handle_data (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_data\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			case ISCSI_INIT_TASK_MGMT_CMND:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got SCSI task mgt. command\n", 
					conn->conn_id);

				if (handle_task_mgt_command(conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_task_mgt_command\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			case ISCSI_INIT_LOGOUT_CMND:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got logout request\n", 
					conn->conn_id);

				if (handle_logout (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_logout\n");
					goto iscsi_rx_thread_out;
				}

				goto iscsi_rx_thread_out;
			}

			case ISCSI_INIT_NOP_OUT:
			{
				TRACE(TRACE_ISCSI, "iscsi_rx_%d: Got NOP_OUT\n", conn->conn_id);

				if (handle_nopout (conn, buffer) < 0)
				{
					TRACE_ERROR("iscsi_rx_thread: Trouble in handle_nopout\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			/* SNACK Handling by Target - SAI */
			case ISCSI_INIT_SNACK:
			{
				TRACE(TRACE_DEBUG, "iscsi_rx_%d: SCSI SNACK Request command received\n", conn->conn_id);

				if (handle_snack(conn, buffer) < 0 )
				{
					TRACE_ERROR("iscsi_rx_thread: handle_snack returned an error\n");
					goto iscsi_rx_thread_out;
				}

				break;
			}

			case ISCSI_TARG_NOP_IN:
			case ISCSI_TARG_SCSI_RSP:
			case ISCSI_TARG_TASK_MGMT_RSP:
			case ISCSI_TARG_LOGIN_RSP:
			case ISCSI_TARG_TEXT_RSP:
			case ISCSI_TARG_SCSI_DATA_IN:
			case ISCSI_TARG_LOGOUT_RSP:
			case ISCSI_TARG_R2T:
			case ISCSI_TARG_ASYNC_MSG:
			case ISCSI_TARG_RJT:
			{
				TRACE_ERROR("iscsi_rx_thread: Got target opcode %.2x\n", 
					buffer[0] & ISCSI_OPCODE);

				/* Send a Reject PDU and close connection */
				iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
				goto iscsi_rx_thread_out;
			}

			default:
			{	
				TRACE_ERROR("iscsi_rx_thread: Got unknown opcode %.2x\n", 
					buffer[0] & ISCSI_OPCODE);

				/* Send a Reject PDU and close connection */
				iscsi_tx_rjt (conn, REASON_COMMAND_NOT_SUPPORTED);
				goto iscsi_rx_thread_out;
			}
		}
	}

iscsi_rx_thread_out:
	conn->active = 0;
	conn->rx_thread = NULL;
	up (&conn->thr_kill_sem);

	return;
}


/*
 * iscsi_tx_thread: This thread is responsible for the transmission of
 * responses and/or status information to the initiator
 */
void iscsi_tx_thread (void *param)
{
	struct iscsi_conn	*conn = (struct iscsi_conn*) param;
	struct iscsi_cmnd	*cmnd;

	if (!conn) 
	{
		TRACE_ERROR("iscsi_tx_thread: Invalid parameter\n");
		return;
	}
	
	lock_kernel();

	siginitsetinv (&current->blocked, ISCSI_SHUTDOWN_SIGS);
	daemonize();
	sprintf (current->comm, "iscsi_tx_%d", conn->conn_id);
	conn->tx_thread = current;

	unlock_kernel();

	TRACE(TRACE_ISCSI, "Starting iscsi_tx_%d\n", conn->conn_id);

	while (1)
	{
		down_interruptible (&conn->tx_sem);
		
		/* is signal received */
		if (signal_pending (current))
		{
			goto iscsi_tx_thread_out;
		}
		
		/* is a command ready for me */
		for (cmnd = conn->session->cmnd_list; cmnd != NULL; )
		{
			if (cmnd->conn == conn)
			{
				switch (cmnd->state) 
				{
					case ISCSI_DISCOVERY:
					{
						if (handle_discovery_rsp (cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in handle_discovery_rsp\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_LOGOUT:
					{
						if (handle_logout_rsp (cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in handle_logout_rsp\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_PING:
					{
						if (handle_nopin (cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in handle_nopin\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_DONE:
					{
						if (handle_iscsi_done (cmnd, 0, 0) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in handle_iscsi_done\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_MGT_FN_DONE:
					{
						if (handle_iscsi_mgt_fn_done(cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in handle_iscsi_mgt_fn_done\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_BUFFER_RDY:
					{
						if (iscsi_tx_r2t (cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in iscsi_tx_r2t\n");
							goto iscsi_tx_thread_out;
						}
						cmnd = cmnd->next;
						break;
					}

					case ISCSI_DEQUEUE:
					{
						if (iscsi_dequeue (&cmnd) < 0)
						{
							TRACE_ERROR("iscsi_tx_thread: Trouble in iscsi_dequeue\n");
							goto iscsi_tx_thread_out;
						}
						break;
					}

					case ISCSI_CMND_RECEIVED:
					case ISCSI_NEW_CMND:
					case ISCSI_SENT:
					case ISCSI_R2T_SENT:
					case ISCSI_IMMEDIATE_DATA_IN:
					case ISCSI_UNSOLICITED_DATA_IN:
					case ISCSI_DATA_IN:
					{
						/* Not much to do */
						cmnd = cmnd->next;
						break;
					}

					default: /* NEVER HERE */
					{
						TRACE_ERROR("iscsi_tx_thread: Unknown command state\n");
						goto iscsi_tx_thread_out;
					}
				}
			}
			else
			{
					cmnd = cmnd->next;
			}
		}
	}

iscsi_tx_thread_out:
	conn->active = 0;
	conn->tx_thread = NULL;
	up (&conn->thr_kill_sem);

	return;
}


/*
 * rx_data: will receive a fixed amount of data.
 * INPUT: struct iscsi_conn (what connection), struct msghdr, total
 * amount of data in bytes
 * OUTPUT: int total bytes read if everything is okay
 * 	       < 0 if trouble
 */
static int rx_data (struct iscsi_conn *conn, struct msghdr *msg, int data)
{
	int 		rx_loop, total_rx = 0;
	__u32	checksum=0, data_crc=0;
	int		niov=1, i=0;
	struct iovec	*iov, *iov_old;
	mm_segment_t oldfs;
	
	if (!conn->conn_socket)
		return -1;

	niov = msg->msg_iovlen;
	iov = msg->msg_iov;

	iov_old = (struct iovec*) kmalloc (niov * sizeof(struct iovec), GFP_KERNEL);

	for( i=0; i<niov; i++)
	{
		iov_old[i].iov_base = iov[i].iov_base;
		iov_old[i].iov_len = iov[i].iov_len;
	}

	while (total_rx < data) 
	{
		oldfs = get_fs(); 
		set_fs(get_ds());
		rx_loop = sock_recvmsg (conn->conn_socket,msg,(data - total_rx),0);
		set_fs(oldfs);

		if (rx_loop <= 0) 
		{
			TRACE_ERROR("rx_data: rx_loop %d total_rx %d\n", rx_loop, total_rx);
			return rx_loop;
		}

		total_rx += rx_loop;

		TRACE(TRACE_DEBUG, "rx_data: rx_loop %d total_rx %d\n", rx_loop, total_rx);
	}

	if (niov > 1 && conn->data_crc == 1)
	{
		do_crc (iov_old[0].iov_base, iov_old[0].iov_len, 0x1, &data_crc);

		for (i=1; i<niov-1; i++)
			do_crc (iov_old[i].iov_base, iov_old[i].iov_len, 0x0, &data_crc);

		checksum = *(__u32 *)(iov_old[niov-1].iov_base);

		if (ntohl(checksum) != data_crc)
		{
			TRACE_ERROR("rx_data: Data crc 0x%08x computed 0x%08x\n", 
				ntohl(checksum), data_crc);
			/* Data Digest Error Recovery - SAI */
			return PAYLOAD_DIGERR;
		}
		else
		{
			TRACE(TRACE_ISCSI, "Got data crc 0x%08x\n", ntohl(checksum));
		}
		
		kfree(iov_old);
	}

	return total_rx;
}


/*
 * tx_data: will transmit a fixed amount of data.
 * INPUT: struct iscsi_conn (what connection), struct msg hdr, total
 * amount of data in bytes,
 * OUTPUT: int total bytes read if everything is okay
 * 	       < 0 if trouble
 */
static int tx_data (struct iscsi_conn *conn, struct msghdr *msg, int data)
{

# ifdef DEBUG_DATA
	int i, j;
	struct iovec *debug_iov;
	unsigned char *to_print;
# endif

	int 		total_tx = 0, tx_loop;
	__u32	hdr_crc=0, data_crc=0, checksum=0;
	int		niov=1, data_len=0, k=0;
	struct iovec	*iov;
	mm_segment_t oldfs;

	if (!msg) 
	{
		TRACE_ERROR("tx_data: NULL msghdr struct\n");
		return -1;
	}

	if (!conn->conn_socket)
	{
		TRACE_ERROR("tx_data: NULL conn_socket\n");
		return -1;
	}

# ifdef DEBUG_DATA
	TRACE(TRACE_DEBUG, "msghdr: %p msg_iovlen: %d\n", msg, msg->msg_iovlen);

	debug_iov = msg->msg_iov;

	for (i = 0; i < msg->msg_iovlen; i++) 
	{
		to_print = (unsigned char*) debug_iov->iov_base;

		for (j = 0; ((j < debug_iov->iov_len)&&(j < 64)); j++) 
		{
			TRACE(TRACE_DEBUG, "%.2x ", to_print[j]);

			if ( ((j+1) % 16) == 0 ) 
				TRACE(TRACE_DEBUG,"\n");
			else if ( ((j+1) % 4) == 0 ) 
				TRACE(TRACE_DEBUG,"    ");
		}

		TRACE(TRACE_DEBUG, "\n");

		debug_iov++;
	}
# endif

	/* send data digest */
	niov = msg->msg_iovlen;
	iov = msg->msg_iov;

	if (conn->session && conn->hdr_crc == 1)
	{
		do_crc (iov[0].iov_base, ISCSI_HDR_LEN, 0x3, &hdr_crc);

		iov[1].iov_base = &hdr_crc;
		iov[1].iov_len  = CRC_LEN;
		TRACE(TRACE_ISCSI, "Send header crc 0x%08x\n", hdr_crc);
	}

	if (conn->session && conn->data_crc == 1 && niov > conn->hdr_crc+2)
	{
		k = conn->hdr_crc + 1;

		do_crc (iov[k].iov_base, iov[k].iov_len, 1, &data_crc);
		data_len = iov[k].iov_len;

		for (k++; k<niov-1; k++)
		{
			do_crc (iov[k].iov_base, iov[k].iov_len, 0, &data_crc);
			data_len += iov[k].iov_len;
		}

		checksum = htonl(data_crc);

		iov[niov-1].iov_base = &checksum;
		iov[niov-1].iov_len = CRC_LEN;

		TRACE(TRACE_ISCSI, "Send data_len %d data crc 0x%08x\n",
		data_len, data_crc);
	}

	while (total_tx < data) 
	{
		oldfs = get_fs(); set_fs(get_ds());

		tx_loop = sock_sendmsg (conn->conn_socket,msg,(data-total_tx));

		set_fs(oldfs);

		if (tx_loop <= 0) 
		{
			TRACE_ERROR("tx_data: tx_loop %d total_tx %d\n",tx_loop,total_tx);
			return tx_loop;
		}

		total_tx += tx_loop;

		TRACE(TRACE_DEBUG,"tx_data: tx_loop %d total_tx %d\n",tx_loop,total_tx);
	}
	
	return total_tx;
}


/**************************************************************************/
/*                          handle income pdu                             */
/**************************************************************************/

/*
 * handle_login: This function is responsible for interpreting the login
 * message and then making sure the necessary details get filled in, a
 * new session is opened if required by the login and the response is returned
 *
 * INPUT: Details of the connection, buffer containing the login
 * OUTPUT: 0 if everything is okay, < 0 if there is a problem
 * Note a 0 value returned may still mean that a connection is rejected
 */
static inline int handle_login (struct iscsi_conn *conn, unsigned char *buffer)
{
	struct iscsi_init_login_cmnd *pdu = (struct iscsi_init_login_cmnd*) buffer;
	struct iscsi_session		*session = NULL, *temp = NULL;
	struct iscsi_conn 		*temp_conn = NULL;
	int				err;
	unsigned int                 when_called;
	struct chap_parameter_type chap_param; /*chap support - CHONG */

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_login_cmnd(pdu);
	
	pdu->length		= be32_to_cpu (pdu->length);
	pdu->tsih		= be16_to_cpu (pdu->tsih);
	pdu->init_task_tag	= be32_to_cpu (pdu->init_task_tag);
	pdu->cid		= be16_to_cpu (pdu->cid);
	pdu->cmd_sn		= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 	= be32_to_cpu (pdu->exp_stat_sn);

	/* a new session */
	if (pdu->tsih == 0x0000)
	{
		session = (struct iscsi_session*) kmalloc (sizeof (struct 
			iscsi_session), GFP_KERNEL);

		if (!session) 
		{
			TRACE_ERROR("handle_login: Unable to malloc for session\n");
			sock_release (conn->conn_socket);
			return -1;
		}

		memset(session, 0, sizeof(struct iscsi_session));

		conn->cid			= pdu->cid;
		conn->session		= session;
		conn->stat_sn 		= pdu->exp_stat_sn;
		session->nconn		= 1;
		session->conn_list	= conn;
		session->devdata		= conn->dev;

		if (conn->dev->draft_number <= DRAFT8)
		{  /* draft 8 and before */
			session->version_max	= 2;
			session->version_min	= 2;
		}
		else if (conn->dev->draft_number <= DRAFT9)
		{  /* draft 9 */
			session->version_max	= 3;
			session->version_min	= 3;
		}
		else
		{  /* draft 11 and after */
			session->version_max	= 0;
			session->version_min	= 0;
		}

		session->cmd_sn 		= pdu->cmd_sn;
		session->exp_cmd_sn 	= pdu->cmd_sn;
		session->max_cmd_sn 	= pdu->cmd_sn + QUEUE_DEPTH_ALLOWED - 1;

		memcpy(session->isid, pdu->isid, 6);

		/* add session to the session list */
		down_interruptible (&session->devdata->session_sem);

		if (!(session->session_params = kmalloc (MAX_CONFIG_PARAMS * 
			sizeof(struct parameter_type), GFP_KERNEL)))
		{
			TRACE_ERROR("handle_login: Unable to malloc for session_params\n");
			goto err_out;
		}

		if (!(session->oper_param = kmalloc (MAX_CONFIG_PARAMS * 
			sizeof(struct session_operational_parameters), GFP_KERNEL)))
		{
			TRACE_ERROR("handle_login: Unable to malloc for oper_param\n");
			goto err_out;
		}
	
		/* copy the parameters information from the global structure */
		conn->connection_flags	= conn->dev->force;
		session->nop_period	= conn->dev->nop_period;
		session->draft_number 	= conn->dev->draft_number;

        /* Store SNACK flags as part of Session - SAI */
        session->targ_snack_flg = devdata->targ_snack_flg;

		param_tbl_cpy (*session->session_params, *conn->dev->param_tbl);
		set_session_parameters (session->oper_param, *session->session_params, 
			session->draft_number);

		up (&session->devdata->session_sem);
		
		init_MUTEX (&session->conn_sem);
		init_MUTEX (&session->cmnd_sem);
		/* initialize retran_thread new 18_04 SAI */
		session->retran_thread = NULL;
		

		if (!(session->address = (struct sockaddr_in*) kmalloc (sizeof 
			(struct sockaddr_in), GFP_KERNEL)))
		{
			TRACE_ERROR("handle_login: Unable to malloc for sockaddr_in\n");
			goto err_out;
		}

		if (conn->conn_socket->ops->getname (conn->conn_socket, (struct 
			sockaddr*) session->address, &err, 1) < 0) 	
		{
			TRACE_ERROR("handle_login: Trouble at getname\n");
			goto err_out;
		}

		/* check isid */
		for (temp = conn->dev->session_list; temp != NULL; temp = temp->next)
		{
			//check isid for draft 9 and after or 8 and before
			if (((session->draft_number >= DRAFT9 && memcmp(temp->isid, 
				pdu->isid, 6) == 0) || (session->draft_number <= DRAFT8 && 
				temp->isid[4] == pdu->isid[4] && temp->isid[5] == 
				pdu->isid[5])) && (temp->address->sin_addr.s_addr == 
				session->address->sin_addr.s_addr))
			{
				TRACE_ERROR("handle_login: Exist a session with the same ISID "
					"from the same IP address\n");
				iscsi_tx_login_rsp (conn, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR);
				goto err_out;
			}
		}

		TRACE(TRACE_ISCSI, "handle_login: A new session started for 0x%08x\n", 
			session->address->sin_addr.s_addr);

		/* add session to the session list */
		down_interruptible (&session->devdata->session_sem);

		session->tsih = ++session->devdata->ntsih;
		session->devdata->nsessions++;
		session->next = session->devdata->session_list;

		if (session->devdata->session_list)
			session->devdata->session_list->prev = session;

		session->devdata->session_list = session;

		up (&session->devdata->session_sem);

		/* set flags to allow keys for new session (leading) login */
		when_called = LEADING_ONLY|INITIAL_ONLY|ALL;
	}
	else
	{
		/* check through the session list */
		for (session = conn->dev->session_list; session != NULL; 
			session = session->next)
		{
			if (session->tsih == pdu->tsih)
				break;
		}

		if (!session) 
		{
			if (session->draft_number >= DRAFT12)
			{
				TRACE_ERROR("handle_login: No session with a TSIH %.4x "
					"terminate this connection ...\n", pdu->tsih);
			}
			else
			{
				TRACE_ERROR("handle_login: No session with a TSID %.4x "
					"terminate this connection ...\n", pdu->tsih);
			}
			sock_release (conn->conn_socket);
			return -1;
		}

		conn->cid = pdu->cid;
		conn->session = session;
		conn->stat_sn = pdu->exp_stat_sn;
		conn->connection_flags  = conn->dev->force;

		/* set back the leading only keys if these keys were set to be
		 * " KEY_TO_BE_NEGOTIATED" in the leading connection negotation.*/
		param_tbl_uncpy(*session->session_params);
		reset_parameter_table(*conn->dev->param_tbl);
		param_tbl_cpy (*session->session_params, *conn->dev->param_tbl);
		
		/* check isid */
		// check draft 9 and after or 8 and before
		if ((session->draft_number >= DRAFT9 && memcmp(pdu->isid, 
			session->isid, 6) != 0) || (session->draft_number <= DRAFT8 && 
			(pdu->isid[4] != session->isid[4] || pdu->isid[5] != 
			session->isid[5])))
		{
			TRACE_ERROR("handle_login: The session has a different ISID " 
				"terminate the session\n");
			iscsi_tx_login_rsp (conn, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR);
			sock_release (conn->conn_socket);
			return -1;
		}

		/* check cid */
		for (temp_conn = conn->session->conn_list; temp_conn!=NULL; 
			temp_conn = temp_conn->next)
		{
			if (temp_conn->cid == conn->cid) 
			{
				TRACE_ERROR("handle_login: Exist a connection with the same cid" 
					" Terminate the connection\n");
				iscsi_tx_login_rsp (conn, STAT_CLASS_INITIATOR, STAT_DETAIL_ERR);
				return -1;
			}
		}

		TRACE(TRACE_ISCSI, "handle_login: An old session found for 0x%08x\n", 
			session->address->sin_addr.s_addr);

		/* insert connection into connection list */
		down (&session->conn_sem);

		conn->next = session->conn_list;

		if (session->conn_list)
			session->conn_list->prev = conn;

		session->conn_list = conn;
		session->nconn++;

		up (&session->conn_sem);

		/* set flags to allow keys for new connection (only) login */
		when_called = INITIAL_ONLY|ALL;
	}
		/* chap support - CHONG */
	chap_param.chap_flags = 0;
	chap_param.local_ctx = CHAP_CloneContext(conn->dev->chap_parameter.local_ctx);
	chap_param.peer_ctx = CHAP_CloneContext(conn->dev->chap_parameter.peer_ctx);

	if (parameter_negotiate(conn, pdu, when_called, chap_param, target_num) < 0)
	{
		TRACE_ERROR("handle_login: Parameter negotiation failed\n");
		/* chap support - CHONG */
		CHAP_FinalizeContext(chap_param.local_ctx);
		CHAP_FinalizeContext(chap_param.peer_ctx);
		return -1;
	}
	else
		conn->stat_sn++;

	/* chap support - CHONG */
	CHAP_FinalizeContext(chap_param.local_ctx);
	CHAP_FinalizeContext(chap_param.peer_ctx);

	if (conn->stat_sn < 0)
	{
		TRACE_ERROR("handle_login: Parameter negotiation with error stat_sn\n");
		return -1;
	}

	/* set the operational parameters to the negotiated value */
	if (pdu->tsih == 0x0000)
	{
		set_session_parameters (session->oper_param, *session->session_params, 
			session->draft_number);

		TRACE(TRACE_ISCSI, "Creating R2T timer for connection %p in session %p \n", conn, session);

        /* Initialize R2T timer - ref18_09c SAI */
        if (session->oper_param->ErrorRecoveryLevel > 0 && devdata->r2t_period > 0)
        {
            if ((conn->r2t_timer = (struct timer_list *)kmalloc(sizeof(struct timer_list), GFP_KERNEL)) == NULL)
            {
                goto err_out;
            }
            init_timer(conn->r2t_timer);
			/* error recovery ver ref18_04 */
			/* create a retransmit_thread for handling error recovery */
			init_MUTEX_LOCKED(&session->retran_sem);
			init_MUTEX(&session->thr_kill_sem);

			kernel_thread ((int (*)(void *))iscsi_retran_thread, (void*)session, 0);
            restart_r2t_timer(conn, devdata->r2t_period*HZ);
			TRACE(TRACE_ISCSI, "R2T timer %p created and started for connection %p in session %p\n", 
								conn->r2t_timer, conn, session);
        }
        else
        {
			session->retran_thread = NULL;
            conn->r2t_timer = NULL;
        }

	}
	else
	{
		TRACE(TRACE_ISCSI, "Creating R2T timer for connection %p in session %p \n", conn, session);

        /* Initialize R2T timer - ref18_09c SAI */
        if (session->oper_param->ErrorRecoveryLevel > 0 && devdata->r2t_period > 0)
        {
            if ((conn->r2t_timer = (struct timer_list *)kmalloc(sizeof(struct timer_list), GFP_KERNEL)) == NULL)
            {
                goto err_out;
            }
            init_timer(conn->r2t_timer);
            /*restart_r2t_timer(conn, devdata->r2t_period*HZ);*/

			TRACE(TRACE_ISCSI, "R2T timer %p created and started for connection %p in session %p\n", 
								conn->r2t_timer, conn, session);
        }
        else
        {
            conn->r2t_timer = NULL;
        }
	}

	conn->hdr_crc = (conn->connection_flags & USE_HEADERDIGEST) ? 1:0;
	conn->data_crc = (conn->connection_flags & USE_DATADIGEST) ? 1:0;

	return 0;

err_out:
	if(conn->conn_socket)
		sock_release (conn->conn_socket);

	/* free session */
	if (session->address)
		kfree (session->address);

	if (session->session_params)
	{
		param_tbl_uncpy(*session->session_params);
		kfree (session->session_params);
	}

	if (session->oper_param)
		kfree (session->oper_param);

	kfree (session);
	session = NULL;

	return -1;
}


/*
 * handle_discovery: this function is responsible for processing a TEXT command
 * in discovery session, currently only support SendTarget command.
 * INPUT: connection and the buffer containing the command structure
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int handle_discovery (struct iscsi_conn* conn, unsigned char* buffer)
{
	struct iscsi_init_text_cmnd *pdu = (struct iscsi_init_text_cmnd*) buffer;
	struct iscsi_cmnd		*cmnd, *temp=NULL;
	struct iovec			*iov;
	struct msghdr		msg;
	int				size, niov=1, padding, err, data_crc_len=0;
	char 			*ptr=NULL, *ptr2;
	__u32			digest, pad_bytes;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	print_init_text_cmnd(pdu);

	pdu->length			= be32_to_cpu (pdu->length);
	pdu->init_task_tag		= be32_to_cpu (pdu->init_task_tag);
	pdu->target_xfer_tag 	= be32_to_cpu (pdu->target_xfer_tag);
	pdu->cmd_sn			= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 		= be32_to_cpu (pdu->exp_stat_sn);

	cmnd = (struct iscsi_cmnd*) kmalloc(sizeof(struct iscsi_cmnd),GFP_KERNEL);

	if (!cmnd)
	{
		TRACE_ERROR("handle_discovery: Could not allocate space for cmnd\n");
		return -1;
	}

	memset(cmnd, 0, sizeof(struct iscsi_cmnd));

	cmnd->init_task_tag		= pdu->init_task_tag;
	cmnd->conn			= conn;
	cmnd->session		= conn->session;
	cmnd->cmd_sn 		= pdu->cmd_sn;
	cmnd->first_burst_len 	= pdu->length;

	// check cmd_sn
	if (cmnd->cmd_sn > conn->session->max_cmd_sn)
	{
		TRACE_ERROR("handle_discovery: cmd_sn is bigger than max_cmd_sn" 
			" Terminate the session\n");
		iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
		return -1;
	}

	while (1)
	{
		if (cmnd->cmd_sn < cmnd->session->exp_cmd_sn)
		{
			TRACE_ERROR("handle_discovery: cmd_sn is less than exp_cmd_sn"
				" Terminate the session\n");
			iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
			return -1;
		}
		else if (cmnd->cmd_sn > cmnd->session->exp_cmd_sn) 
		{
			TRACE(TRACE_DEBUG, "handle_discovery: Wait for sync\n");
		}
		else 
		{
			if ((pdu->opcode & I_BIT) != I_BIT)
				cmnd->session->exp_cmd_sn++;
			break;
		}
	}

	down (&conn->session->cmnd_sem);

	cmnd->state			= ISCSI_DISCOVERY;
	cmnd->target_xfer_tag	= pdu->target_xfer_tag;
	cmnd->stat_sn		= pdu->exp_stat_sn;

	/* receive this data */
	padding = -(pdu->length)&3;

	if (padding)
		niov++;

	if (cmnd->conn->data_crc == 1)
		niov++;
	
	iov = (struct iovec*) kmalloc (niov * sizeof(struct iovec), GFP_KERNEL);

	if (!iov)
	{
		TRACE_ERROR("handle_immediate_data: Could not malloc iov\n");
		return -1;
	}

	buffer = (char *) kmalloc (pdu->length * sizeof(char), GFP_KERNEL);
	iov[0].iov_base = buffer;
	iov[0].iov_len = pdu->length;

	if (padding)
	{
		iov[niov-1 - cmnd->conn->data_crc].iov_base = &pad_bytes;
		iov[niov-1 - cmnd->conn->data_crc].iov_len = padding;
	}

	if (cmnd->conn->data_crc == 1)
	{
		data_crc_len = CRC_LEN;
		iov[niov-1].iov_base = &digest;
		iov[niov-1].iov_len = CRC_LEN;
	}

	fill_msg_hdr (&msg);
	msg.msg_iov		= iov;
	msg.msg_iovlen	= niov;

	err = rx_data (conn, &msg, pdu->length + padding + data_crc_len);

	if (err != (pdu->length + padding + data_crc_len)) 
	{
		TRACE_ERROR("handle_discovery: Trouble in rx_data\n");
		return -1;
	}

	size = pdu->length;

	if (conn->session->oper_param->SessionType == 1)
	{  /* in discovery session, print out all keys sent by initiator */
		if (buffer != NULL && size > 0)
		{  /* print out the info sent by initiator */
			for (ptr = buffer;  size > 0;  ptr = ptr2)
			{
				for (ptr2 = ptr;  *ptr2 != '\0'  &&  size > 0; ptr2++, size--)
				{  /* skip over key itself */
				}

				if (*ptr2 != '\0')
				{
					*ptr2 = '\0';
					TRACE_ERROR("Unterminated key %s\n", ptr);
				}

				printk("iscsi discovery: %s\n", ptr);

				for ( ; *ptr2 == '\0'  &&  size > 0; ptr2++, size-- )
				{/* skip over nulls between keys */
				}
			}
		}
	}

	// check cmnd in the queue
	for (temp = cmnd->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			pdu->exp_stat_sn) && (temp->state == ISCSI_SENT))
		{
			TRACE(TRACE_DEBUG, "handle_discovery: command statsn %d,"
				"received exp_stat_sn %d, command state %d\n", 
				temp->stat_sn, pdu->exp_stat_sn, temp->state); 
			temp->state = ISCSI_DEQUEUE;
		}

		if (temp->next == NULL)
			break;
	}

	/* Add command to the queue */
	if (temp)
		temp->next = cmnd;
	else
		cmnd->session->cmnd_list = cmnd;

	up (&conn->session->cmnd_sem);
	up (&cmnd->conn->tx_sem);

	return 0;
}


static inline int handle_logout (struct iscsi_conn* conn, unsigned char* buffer)
{
	struct iscsi_init_logout_cmnd *pdu = (struct iscsi_init_logout_cmnd*)buffer;
	struct iscsi_cmnd *cmnd, *temp = NULL;
	
	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_logout_cmnd(pdu);

	pdu->length		= be32_to_cpu (pdu->length);
	pdu->init_task_tag	= be32_to_cpu (pdu->init_task_tag);
	pdu->cid		= be16_to_cpu (pdu->cid);
	pdu->cmd_sn		= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 	= be32_to_cpu (pdu->exp_stat_sn);

	cmnd = (struct iscsi_cmnd*) kmalloc (sizeof(struct iscsi_cmnd),GFP_KERNEL);

	if (!cmnd)
	{
		TRACE_ERROR("handle_logout: Could not allocate space for cmnd\n");
		return -1;
	}

	memset(cmnd, 0, sizeof(struct iscsi_cmnd));

	cmnd->init_task_tag	= pdu->init_task_tag;
	cmnd->conn		= conn;
	cmnd->session		= conn->session;
	cmnd->cmd_sn 		= pdu->cmd_sn;

	// check cmd_sn
	if (cmnd->cmd_sn > conn->session->max_cmd_sn)
	{
		TRACE_ERROR("handle_logout: cmd_sn is bigger than max_cmd_sn, "
			"Terminate the session\n");
		iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
		return -1;
	}

	while (1)
	{
		if (cmnd->cmd_sn < cmnd->session->exp_cmd_sn)
		{
			TRACE_ERROR("handle_logout: cmd_sn is less than exp_cmd_sn," 
				" Terminate the session\n");
			iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
			return -1;
		}
		else if (cmnd->cmd_sn > cmnd->session->exp_cmd_sn) 
		{
			TRACE_ERROR("handle_logout: Wait for sync\n");
		}
		else 
		{
			if ((pdu->opcode & I_BIT) != I_BIT)
				cmnd->session->exp_cmd_sn++;
			break;
		}
	}

	down (&conn->session->cmnd_sem);

	cmnd->state		= ISCSI_LOGOUT;
	cmnd->stat_sn		= pdu->exp_stat_sn;

	// check cmnd in the queue
	for (temp = cmnd->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			pdu->exp_stat_sn) && (temp->state == ISCSI_SENT))
		{
			TRACE(TRACE_DEBUG, "handle_logout: command statsn %d,"
				"received exp_stat_sn %d, command state %d\n", 
				temp->stat_sn, pdu->exp_stat_sn, temp->state); 
			temp->state = ISCSI_DEQUEUE;
		}

		if (temp->next == NULL)
			break;
	}

	/* Add command to the queue */
	if (temp)
		temp->next = cmnd;
	else
		cmnd->session->cmnd_list = cmnd;

	up (&conn->session->cmnd_sem);
	up (&cmnd->conn->tx_sem);

	return 0;
}


static inline int handle_nopout (struct iscsi_conn* conn, unsigned char* buffer)
{
	struct iscsi_init_nopout 	*pdu = (struct iscsi_init_nopout*) buffer;
	struct iscsi_cmnd		*cmnd, *temp=NULL;
	struct iovec			*iov;
	struct msghdr			msg;
	int				niov=1, padding, err, data_crc_len=0;
	__u32				digest, pad_bytes;

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_nopout(pdu);

	pdu->length			= be32_to_cpu (pdu->length);
	pdu->init_task_tag		= be32_to_cpu (pdu->init_task_tag);
	pdu->target_xfer_tag		= be32_to_cpu (pdu->target_xfer_tag);
	pdu->cmd_sn			= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 		= be32_to_cpu (pdu->exp_stat_sn);

	cmnd = (struct iscsi_cmnd*) kmalloc (sizeof(struct iscsi_cmnd),GFP_KERNEL);

	if (!cmnd)
	{
		TRACE_ERROR("handle_nopout: Could not allocate space for cmnd\n");
		return -1;
	}

	memset(cmnd, 0, sizeof(struct iscsi_cmnd));

	cmnd->data_length		= pdu->length;
	cmnd->init_task_tag		= pdu->init_task_tag;
	cmnd->target_xfer_tag		= pdu->target_xfer_tag;
	cmnd->conn			= conn;
	cmnd->session			= conn->session;
	cmnd->cmd_sn 			= pdu->cmd_sn;

	// check cmd_sn
	if (cmnd->cmd_sn > conn->session->max_cmd_sn)
	{
		TRACE_ERROR("handle_nopout: cmd_sn is bigger than max_cmd_sn,"
			" Terminate the session\n");
		iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
		return -1;
	}

	while (1)
	{
		if (cmnd->cmd_sn < cmnd->session->exp_cmd_sn)
		{
			TRACE_ERROR("handle_nopout: cmd_sn is less than exp_cmd_sn," 
				" Terminate the session\n");
			iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
			return -1;
		}
		else if (cmnd->cmd_sn > cmnd->session->exp_cmd_sn) 
		{
			TRACE_ERROR("handle_nopout: Wait for sync\n");
		}
		else 
		{
			if ((pdu->opcode & I_BIT) != I_BIT)
				cmnd->session->exp_cmd_sn++;
			break;
		}
	}

	down (&conn->session->cmnd_sem);
	
	cmnd->state		= ISCSI_PING;
	cmnd->stat_sn		= pdu->exp_stat_sn;

	/* receive data */
	if (pdu->length > 0)
	{
		padding = -(pdu->length)&3;

		if (padding)
			niov++;
	
		if (cmnd->conn->data_crc == 1)
			niov++;
	
		iov = (struct iovec*) kmalloc (niov*sizeof(struct iovec),GFP_KERNEL);
	
		if (!iov)
		{
			TRACE_ERROR("handle_nopout: Could not malloc iov\n");
			return -1;
		}

		cmnd->ping_data = (char *) kmalloc (pdu->length * sizeof(char), GFP_KERNEL);
		iov[0].iov_base = cmnd->ping_data;
		iov[0].iov_len = pdu->length;

		if (padding)
		{
			iov[niov-1 - cmnd->conn->data_crc].iov_base = &pad_bytes;
			iov[niov-1 - cmnd->conn->data_crc].iov_len = padding;
		}

		if (cmnd->conn->data_crc == 1)
		{
			data_crc_len = CRC_LEN;
			iov[niov-1].iov_base = &digest;
			iov[niov-1].iov_len = CRC_LEN;
		}

		fill_msg_hdr (&msg);
		msg.msg_iov	= iov;
		msg.msg_iovlen = niov;

		err = rx_data (conn, &msg, pdu->length + padding + data_crc_len);

		if (err != (pdu->length + padding + data_crc_len)) 
		{
			TRACE_ERROR("handle_nopout: Trouble in rx_data\n");
			return -1;
		}
	}

	// check cmnd in the queue
	for (temp = cmnd->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			pdu->exp_stat_sn) && (temp->state == ISCSI_SENT))
		{
			TRACE(TRACE_DEBUG, "handle_nopout: command statsn %d,"
				"received exp_stat_sn %d, command state %d\n", 
				temp->stat_sn, pdu->exp_stat_sn, temp->state); 
			temp->state = ISCSI_DEQUEUE;
		}

		if (temp->next == NULL)
			break;
	}

	/* Add command to the queue */
	if (temp)
		temp->next = cmnd;
	else
		cmnd->session->cmnd_list = cmnd;

	up (&conn->session->cmnd_sem);
	up (&cmnd->conn->tx_sem);

	return 0;
}


/*
 * handle_cmnd: this function is responsible for processing a SCSI
 * command received from an initiator, making the mid-level aware of the
 * command and then queueing up the command in the command queue for the
 * session
 * INPUT: connection and the buffer containing the command structure
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int handle_cmnd (struct iscsi_conn* conn, unsigned char* buffer)
{
	struct iscsi_init_scsi_cmnd *pdu = (struct iscsi_init_scsi_cmnd*) buffer;
	struct iscsi_cmnd *cmnd, *temp=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter handle_cmnd\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_cmnd(pdu);
	
	pdu->length		= be32_to_cpu (pdu->length);
	// pdu->lun		= be64_to_cpu (pdu->lun);
	pdu->init_task_tag	= be32_to_cpu (pdu->init_task_tag);
	pdu->xfer_len 		= be32_to_cpu (pdu->xfer_len);
	pdu->cmd_sn		= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 	= be32_to_cpu (pdu->exp_stat_sn);

	/* check data length, immediate data should not bigger than 
		MaxRecvPDULength or FirstBurstSize */
	if ((pdu->length > conn->max_recv_length 
			&& conn->session->draft_number >= DRAFT9) 
		|| (pdu->length > conn->session->oper_param->DataPDULength 
			&& conn->session->draft_number <= DRAFT8) 
		|| (pdu->length > conn->session->oper_param->FirstBurstLength
			&& conn->session->draft_number >= DRAFT18)
		|| (pdu->length > conn->session->oper_param->FirstBurstSize
			&& conn->session->draft_number <= DRAFT13))
	{
		printk("DSL: %d", pdu->length);
		printk("max_recv_length: %d", conn->max_recv_length);
		printk("DataPDULength: %d", conn->session->oper_param->DataPDULength);
		TRACE_ERROR( "handle_cmnd: data length is too big\n" );
		return -1;
	}

	cmnd = (struct iscsi_cmnd*) kmalloc (sizeof(struct iscsi_cmnd),GFP_KERNEL);

	if (!cmnd)
	{
		TRACE_ERROR("handle_cmnd: Could not allocate space for cmnd\n");
		return -1;
	}

	memset(cmnd, 0, sizeof(struct iscsi_cmnd));

	cmnd->init_task_tag		= pdu->init_task_tag;
	cmnd->data_length 		= pdu->xfer_len;
	cmnd->lun			= pdu->lun;
	cmnd->cdb[0]			= pdu->cdb[0];
	cmnd->cdb[4]			= pdu->cdb[4];
	cmnd->data_dir			= pdu->flags & (READ_TYPE_CMND|WRITE_TYPE_CMND);
	cmnd->conn		        = conn;
	cmnd->session			= conn->session;
	cmnd->cmd_sn 			= pdu->cmd_sn;
	cmnd->first_burst_len		= pdu->length;

    /* Retransmit Flag for Error Recovery - SAI */

    /* error recovery ver ref18_04 - SAI */

	init_MUTEX_LOCKED (&cmnd->unsolicited_data_sem);

	// check cmd_sn --- v8 ---
	if (cmnd->cmd_sn > conn->session->max_cmd_sn)
	{
		TRACE_ERROR("handle_cmnd: cmd_sn is bigger than max_cmd_sn,"
			" Terminate the session\n");
		iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
		return -1;
	}

	while (1)
	{
		if (cmnd->cmd_sn < cmnd->session->exp_cmd_sn)
		{
            /* The initiator could have acted pro-actively in error recovery
               level > 0. If so discard the command draft 17 sec 5.2.1
               "usage of retry" - SAI */
            if (conn->session->oper_param->ErrorRecoveryLevel > 0)
            {
                TRACE(TRACE_ISCSI, "handle_cmnd: Discarding already allegiant command\n");

                cmnd->immediate_data_present = (pdu->length ? 1:0);

                if (cmnd->immediate_data_present)
                {
                    targ_drop_pdu_data (conn, pdu->length);
                }
                return 0;
            }

			TRACE_ERROR("handle_cmnd: cmd_sn is less than exp_cmd_sn," 
				" Terminate the session\n");
			iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
			return -1;
		}
		else if (cmnd->cmd_sn > cmnd->session->exp_cmd_sn) 
		{
			TRACE_ERROR("handle_cmnd: Wait for sync\n");
		}
		else
		{
			if ((pdu->opcode & I_BIT) != I_BIT)
				cmnd->session->exp_cmd_sn++;
			break;
		}
	}

	down (&conn->session->cmnd_sem);

	cmnd->cmnd = rx_cmnd(conn->dev->device,conn->session->oper_param->TargetName,
			pdu->lun,pdu->cdb,ISCSI_CDB_LEN,pdu->xfer_len, pdu->flags);

	if (!(cmnd->cmnd))
	{
		TRACE_ERROR("handle_cmnd: rx_cmnd returned a NULL\n");
		return -1;
	}

	cmnd->state		= ISCSI_NEW_CMND;
	cmnd->target_xfer_tag	= ++cmnd->session->cmnd_id;
	cmnd->stat_sn		= pdu->exp_stat_sn;

	// check cmnd in the queue
	for (temp = cmnd->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			pdu->exp_stat_sn) && (temp->state == ISCSI_SENT))
		{
			TRACE(TRACE_DEBUG, "handle_cmnd: command statsn %d,"
				"received exp_stat_sn %d, command state %d\n", 
				temp->stat_sn, pdu->exp_stat_sn, temp->state); 
			temp->state = ISCSI_DEQUEUE;
		}

		if (temp->next == NULL)
			break;
	}

	/* Add command to the queue */
	if (temp)
		temp->next = cmnd;
	else	
		cmnd->session->cmnd_list = cmnd;

	up (&conn->session->cmnd_sem);

	cmnd->immediate_data_present = (pdu->length ? 1:0);
	cmnd->unsolicited_data_present = !(F_BIT == (pdu->flags & F_BIT));

	TRACE(TRACE_DEBUG,"unsolicited_data_present %d flags %.2x\n", 
		cmnd->unsolicited_data_present, pdu->flags);

	if (cmnd->immediate_data_present)
	{	
		/* immediate data follows */
		TRACE(TRACE_DEBUG, "Blocked on unsolicited_data_sem\n");
		down(&cmnd->unsolicited_data_sem);
		TRACE(TRACE_DEBUG, "Unblocked on unsolicited_data_sem\n");

		handle_immediate_data (cmnd, pdu->length);

		if (cmnd->data_done >= cmnd->data_length)
		{
			TRACE(TRACE_DEBUG, "%d received for cmnd %p\n", pdu->length, cmnd);

			cmnd->state = ISCSI_DATA_IN;

			down (&cmnd->session->cmnd_sem);
 
			if (scsi_rx_data (cmnd->cmnd) < 0)
			{
				TRACE_ERROR("handle_cmnd: scsi_rx_data returned an error\n");
				return -1;
			}

			up (&cmnd->session->cmnd_sem);
			up (&cmnd->conn->tx_sem);
		}

		if(!(cmnd->unsolicited_data_present) && (pdu->length<cmnd->data_length))
			up (&cmnd->conn->tx_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave handle_cmnd\n");

	return 0;
}


static inline int handle_task_mgt_command(struct iscsi_conn* conn, 
                                          unsigned char* buffer)
{
	struct iscsi_init_task_mgt_command *pdu = (struct 
		iscsi_init_task_mgt_command *) buffer;
	struct iscsi_cmnd *new_command = NULL, *ref_command = NULL, *temp = NULL;
	int i;   

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_task_mgt_command(pdu);

	for (i=0; i<ISCSI_HDR_LEN; i++)
		conn->bad_hdr[i] = buffer[i];

	pdu->lun 		= be64_to_cpu (pdu->lun);
	pdu->init_task_tag 	= be32_to_cpu (pdu->init_task_tag);
	pdu->ref_task_tag	= be32_to_cpu (pdu->ref_task_tag);    
	pdu->cmd_sn 		= be32_to_cpu (pdu->cmd_sn);
	pdu->exp_stat_sn 	= be32_to_cpu (pdu->exp_stat_sn);
	pdu->ref_cmd_sn	= be32_to_cpu (pdu->ref_cmd_sn);
	
	if (pdu->function & F_BIT)
	{
		TRACE_ERROR("handle_task_mgt_command: F_BIT is 1\n");
	}

	if ((pdu->function != ABORT_TASK) && (pdu->ref_task_tag != 0xFFFFFFFF))
	{
		TRACE_ERROR("handle_task_mgt_command: referenced task tag field "
			"is not 0xFFFFFFFF\n");
	}

	if (pdu->function == ABORT_TASK)
	{
		ref_command = search_tags (conn, pdu->ref_task_tag, 0xFFFFFFFF);

		//check ref_cmd_sn in the future
		if(ref_command==NULL)
			TRACE(TRACE_DEBUG, "No command with init tag %d, so dropping cmdsn %d\n", 
				pdu->ref_task_tag, pdu->cmd_sn);

		return 0;
	}

	new_command = (struct iscsi_cmnd*) kmalloc (sizeof (struct iscsi_cmnd), 
		GFP_KERNEL);

	if (!new_command)
	{
		TRACE_ERROR("handle_task_mgt_command: Could not allocate space for cmnd\n");
		return -1;
	}

	new_command->next   			= NULL;
	new_command->state 			= ISCSI_CMND_RECEIVED;
	new_command->session    		= conn->session;
	new_command->conn   			= conn;
	new_command->init_task_tag   		= pdu->init_task_tag;
	new_command->ref_task_tag 		= pdu->ref_task_tag;
	new_command->ref_task_lun 		= pdu->lun;
	new_command->ref_cmd_sn		= pdu->ref_cmd_sn;
	new_command->cmd_sn 			= pdu->cmd_sn;
	new_command->stat_sn			= pdu->exp_stat_sn;
	new_command->scatter_list_offset 	= 0;
	new_command->scatter_list_count 	= 0;
	new_command->data_done  		= 0;
	new_command->cmnd 			= NULL;

	// check cmd_sn
	if (new_command->cmd_sn > conn->session->max_cmd_sn)
	{
		TRACE_ERROR("handle_task_mgt_command: cmd_sn is bigger than max_cmd_sn,"
			" Terminate the session\n");
		iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
		return -1;
	}

	while (1)
	{
		if (new_command->cmd_sn < new_command->session->exp_cmd_sn)
		{
			TRACE_ERROR("handle_task_mgt_command: cmd_sn is less than exp_cmd_sn," 
				" Terminate the session\n");
			iscsi_tx_rjt (conn, REASON_PROTOCOL_ERR);
			return -1;
		}
		else if (new_command->cmd_sn > new_command->session->exp_cmd_sn) 
		{
			TRACE_ERROR("handle_task_mgt_command: Wait for sync\n");
			return 0;   /* Added to avoid an end-less loop during command drop-outs - SAI */

		}
		else 
		{
			if ((pdu->opcode & I_BIT) != I_BIT)
				new_command->session->exp_cmd_sn++;
			break;
		}
	}

	down (&conn->session->cmnd_sem);

	new_command->message = rx_task_mgmt_fn(conn->dev->device, pdu->function, 
		(ref_command ? ref_command->cmnd:NULL));

	if (!new_command->message)
	{
		TRACE_ERROR("handle_task_mgt_command: rx_task_mgt_fn returned an error\n");
		return -1;
	}
   
	new_command->state  		= ISCSI_NEW_CMND;
	new_command->target_xfer_tag	= ++new_command->session->cmnd_id;
	new_command->stat_sn 		= pdu->exp_stat_sn;

	/* Add command to the queue */
	for (temp = new_command->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			pdu->exp_stat_sn) && (temp->state == ISCSI_SENT))
			temp->state = ISCSI_DEQUEUE;

		if (temp->next == NULL)
			break;
	}

	if (temp)
		temp->next = new_command;
	else   
		new_command->session->cmnd_list = new_command;

	up (&conn->session->cmnd_sem);

	return 0;
}


/*
 * handle_data: This function is responsible for receiving data that
 * corresponds to a command. It receives the data in the buffers that
 * are allocated by the STML. It then has to change the state of the
 * command and inform the STML about the data received
 * INPUT: connection and buffer containing the ISCSI_DATA (write) PDU
 * OUTPUT: 0, if everything is okay, < 0 otherwise
 */
static inline int handle_data (struct iscsi_conn *conn, unsigned char *buffer)
{
	struct iscsi_init_scsi_data_out *hdr=(struct iscsi_init_scsi_data_out*)buffer;
	struct iscsi_cmnd	*cmd, *temp;
	struct iovec		*iov=NULL;
	struct msghdr		msg;
	struct targ_error_rec	err_rec;
	struct scatterlist	*st_list;
	int			offset, i, niov, padding, err, data_crc_len=0, retval=0;
	__u32			data_sn, digest, pad_bytes;

	TRACE(TRACE_ENTER_LEAVE, "Entered handle_data\n");

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_init_scsi_data_out(hdr);	

	for(i=0; i<ISCSI_HDR_LEN; i++)
		conn->bad_hdr[i]	= buffer[i];

	hdr->length		= be32_to_cpu (hdr->length);
	hdr->lun		= be64_to_cpu (hdr->lun);
	hdr->init_task_tag	= be32_to_cpu (hdr->init_task_tag);
	hdr->target_xfer_tag	= be32_to_cpu (hdr->target_xfer_tag);
	hdr->exp_stat_sn	= be32_to_cpu (hdr->exp_stat_sn);
	hdr->offset		= be32_to_cpu (hdr->offset);
	data_sn			= be32_to_cpu (hdr->data_sn);

	/* check data length, should not bigger than MaxRecvPDULength */
	if ((hdr->length > conn->max_recv_length && 
		conn->session->draft_number >= DRAFT9) || (hdr->length > 
		conn->session->oper_param->DataPDULength && 
		conn->session->draft_number <= DRAFT8))
	{
		TRACE_ERROR( "handle_data: data length is too big\n" );
		return -1;
	}

	/* search command queue to match command */
	TRACE(TRACE_DEBUG, "Before search_tags\n");
	cmd = search_tags (conn, hdr->init_task_tag, hdr->target_xfer_tag);
	TRACE(TRACE_DEBUG, "After search_tags\n");

	if (!cmd) 
	{
		TRACE_ERROR("handle_data: could not find a match for command "
			"init_task_tag: %.8x target_xfer_tag %.8x\n", 
			hdr->init_task_tag, hdr->target_xfer_tag);
		return -1;
	}

	if (data_sn > cmd->data_sn || data_sn < cmd->data_sn)
	{
		/* Check if the offset is what we expected - SAI */
		if (cmd->data_done < hdr->offset)
		{
			/* Sequence Error Recovery - SAI */

			TRACE_ERROR("handle_data: got data_sn %d, should be %d\n", 
				data_sn, cmd->data_sn );

			err_rec.curr_conn = conn;
			err_rec.pdu_hdr = (struct generic_pdu *)hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = SEQUENCE_ERR;
			targ_do_error_recovery(&err_rec);
			goto end_handle_data;
		}
	}

	/* error recovery ver ref18_04 - SAI */
	if (cmd->data_done > hdr->offset)
	{
		TRACE(TRACE_ISCSI, "Probably a duplicate packet - drop it\n");
		targ_drop_pdu_data(conn, hdr->length);	
		goto end_handle_data;
	}
	else if (cmd->data_done < hdr->offset)
	{
		/* offset error - expected packets might have been dropped in
							the line*/
		TRACE_ERROR("handle_data: got data_sn %d, should be %d\n", 
			data_sn, cmd->data_sn );

		err_rec.curr_conn = conn;
		err_rec.pdu_hdr = (struct generic_pdu *)hdr;
		err_rec.cmd = cmd;
		err_rec.err_type = SEQUENCE_ERR;
		targ_do_error_recovery(&err_rec);
		goto end_handle_data;
	}

	cmd->data_sn = data_sn + 1;	

	/* check FirstBurstSize and MaxBurstSize */
	if (cmd->unsolicited_data_present)
	{
		cmd->first_burst_len += hdr->length;

		if ((cmd->first_burst_len > conn->session->oper_param->FirstBurstSize
				&& conn->session->draft_number <= DRAFT13)
			|| (cmd->first_burst_len > conn->session->oper_param->FirstBurstLength
				&& conn->session->draft_number >= DRAFT18))
		{
			TRACE_ERROR( "handle_data: data length is too big\n" );
			return -1;
		}

		if (F_BIT == (hdr->flags & F_BIT))
		{
			cmd->unsolicited_data_present = 0;
			cmd->data_sn = 0;
		}
	}
	else
	{
		cmd->next_burst_len += hdr->length;

		if ((cmd->next_burst_len > conn->session->oper_param->MaxBurstSize
				&& conn->session->draft_number <= DRAFT13)
			|| (cmd->next_burst_len > conn->session->oper_param->MaxBurstLength
				&& conn->session->draft_number >= DRAFT18))
		{
			TRACE_ERROR("handle_data: data length is too big\n");
			return -1;
		}

		if (F_BIT == (hdr->flags & F_BIT))
		{
			cmd->next_burst_len = 0;
			cmd->data_sn = 0;
			cmd->outstanding_r2t--;
			up (&cmd->conn->tx_sem);
		}
	}

	/* make sure you use the status information */
	down (&cmd->session->cmnd_sem);

	for (temp = cmd->session->cmnd_list; temp != NULL; temp = temp->next)
	{
		if (temp->stat_sn && temp->conn == conn && (temp->stat_sn < 
			hdr->exp_stat_sn) && (temp->state == ISCSI_SENT))
		temp->state = ISCSI_DEQUEUE;
	}

	up (&cmd->session->cmnd_sem);

	TRACE(TRACE_DEBUG, "handle_data: After Dequeueing check\n");

	if(!(cmd->state & ISCSI_BUFFER_RDY)) 
	{
		TRACE(TRACE_DEBUG, "handle_data: Blocked on unsolicited_data_sem\n");
		down(&cmd->unsolicited_data_sem);
		TRACE(TRACE_DEBUG, "handle_data: Unblocked on unsolicited_data_sem\n");
	}

	/* receive this data */
	st_list = (struct scatterlist*) cmd->cmnd->req->sr_buffer;
	offset = cmd->scatter_list_offset;
	st_list += cmd->scatter_list_count;

	TRACE(TRACE_DEBUG, "scatter list offset %d, count %d\n", 
		cmd->scatter_list_offset, cmd->scatter_list_count); 

	niov = find_iovec_needed (hdr->length, st_list, offset);

	if (niov < 0) 
	{
		TRACE_ERROR("handle_data: find_iovec_needed returned an error\n");
		return -1;
	}

	padding = -(hdr->length)&3;

	if (padding)
		niov++;

	if (conn->data_crc == 1)
		niov++;

	iov = (struct iovec*) kmalloc (niov * sizeof(struct iovec), GFP_KERNEL);

	if (!iov) 
	{
		TRACE_ERROR("handle_data: Could not malloc iov\n");
		return -1;
	}

	TRACE(TRACE_DEBUG, "handle_data: the no. of iovec needed %d\n", niov);

	/* Store scatter list count for recovery purposes - SAI */
	retval = fill_iovec (iov, niov, st_list, &offset, 
		hdr->length); 
	cmd->scatter_list_count += retval;
	cmd->scatter_list_offset = offset;
	cmd->data_done += hdr->length;

	if (padding)
	{
		iov[niov-1 - conn->data_crc].iov_base = &pad_bytes;
		iov[niov-1 - conn->data_crc].iov_len = padding;
	}

	if (cmd->conn->data_crc == 1)
	{
		data_crc_len = CRC_LEN;
		iov[niov-1].iov_base = &digest;
		iov[niov-1].iov_len = CRC_LEN;
	}

	fill_msg_hdr (&msg);
	msg.msg_iov	= iov;
	msg.msg_iovlen = niov;

	err = rx_data (conn, &msg, hdr->length + padding + data_crc_len);

	if (err != (hdr->length + padding + data_crc_len)) 
	{
		/* Payload Digest Error Recovery - SAI */
		if (err == PAYLOAD_DIGERR)
		{
			cmd->scatter_list_offset -= offset;
			cmd->scatter_list_count -= retval;
			err_rec.curr_conn = conn;
			err_rec.pdu_hdr = (struct generic_pdu *)hdr;
			err_rec.cmd = cmd;
			err_rec.err_type = PAYLOAD_DIGERR;
			targ_do_error_recovery(&err_rec);
			goto end_handle_data;
		}

		TRACE_ERROR("handle_data: not receive the complete data\n");
		return -1;
	}

	TRACE(TRACE_DEBUG, "handle_data: receiving data for cmd_sn %.8x "
		"init_task_tag %.8x target_xfer_tag %.8x\n", 
		cmd->cmd_sn, cmd->init_task_tag, cmd->target_xfer_tag);

	/* error recovery ver ref18_04 - SAI */
	search_data_q(cmd);

	if (cmd->data_done >= cmd->data_length) 
	{
		cmd->state = ISCSI_DATA_IN;

		down (&cmd->session->cmnd_sem);

		if (scsi_rx_data (cmd->cmnd) < 0) 
		{
			TRACE_ERROR("handle_data: scsi_rx_data returned an error\n");
			return -1;
		}

		up (&cmd->session->cmnd_sem);
		up (&conn->tx_sem);
	}

	if ((hdr->flags & F_BIT) && (cmd->unsolicited_data_present) && 
		(cmd->data_done < cmd->data_length)) 
	{
		up (&conn->tx_sem);
	}

/* Added for Error Recovery - SAI */
end_handle_data:

	TRACE(TRACE_ENTER_LEAVE, "Leave handle_data\n");

	kfree (iov);

	return 0;
}

/* Added by SAI for Handling SNACK requests from Initiator as part of Error Recovery */

static inline int handle_snack(struct iscsi_conn *conn, unsigned char *buffer)
{
        int             i=0, runlen=0, begrun=0;
        struct iscsi_init_snack         *pdu = (struct iscsi_init_snack *) buffer;
        struct iscsi_cmnd               *cmd=NULL;
        struct iscsi_session            *session;

        TRACE(TRACE_ENTER_LEAVE, "Enter handle_snack \n");

        for(i=0; i<ISCSI_HDR_LEN; i++)
                conn->bad_hdr[i]        = buffer[i];

        print_init_snack(pdu);

        pdu->length             = be32_to_cpu (pdu->length);
        pdu->lun                = be64_to_cpu (pdu->lun);
        pdu->init_task_tag      = be32_to_cpu (pdu->init_task_tag);
        pdu->target_xfer_tag    = be32_to_cpu (pdu->target_xfer_tag);
        pdu->exp_stat_sn        = be32_to_cpu (pdu->exp_stat_sn);
        pdu->begrun             = be32_to_cpu (pdu->begrun);
        pdu->runlen             = be32_to_cpu (pdu->runlen);

        session = conn->session;
		runlen = pdu->runlen;
		begrun = pdu->begrun;

		if (session->oper_param->ErrorRecoveryLevel < 1)
		{
			TRACE_ERROR("Received SNACK request in Error Recovery Level < 1\n");
			return 0;
		}

        if ( (pdu->flags & SNACK_TYPE) == DATA_R2T_SNACK)
        {
                if (!(session->targ_snack_flg & DATA_SNACK_ENABLE))
                {
                        TRACE_ERROR("Target Data SNACK is not Enabled - Ignoring Snack Request\n");
                        iscsi_tx_rjt(conn, REASON_DATA_SNACK);
                        return 0;
                }
        }

        if ( (pdu->flags & SNACK_TYPE) == STATUS_SNACK)
        {
                if (!(session->targ_snack_flg & STATUS_SNACK_ENABLE))
                {
                        TRACE_ERROR("Target Status SNACK is not Enabled - Ignoring Snack Request\n");
                        return 0;
                }
        }
        if (session->targ_snack_flg & DATA_SNACK_REJECT )
        {
                if ( (pdu->flags & SNACK_TYPE) == DATA_R2T_SNACK)
                {
                        TRACE(TRACE_ISCSI, "handle_snack : Sending Data-SNACK Reject for SNACK Request\n");
                        iscsi_tx_rjt(conn, REASON_DATA_SNACK);
                        return 0;
                }
        }

        for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next)
        {
                if (cmd->init_task_tag == pdu->init_task_tag && (pdu->flags & SNACK_TYPE) == DATA_R2T_SNACK )
                {
                        if (cmd->state == ISCSI_SENT)
                        {
                                if (cmd->data_sn >= pdu->begrun)
                                {
                                        /* Found the corresponding Data PDU - re-transmit it */

                                        TRACE(TRACE_ISCSI, "handle_snack : Re-Transmitting Data on SNACK Request\n");
                                        cmd->retransmit_flg = 1;
                                        cmd->scatter_list_count = 0;
                                        if (!pdu->runlen)
                                            handle_iscsi_done(cmd, pdu->begrun, cmd->data_sn);
                                        else
                                            handle_iscsi_done(cmd, pdu->begrun, pdu->runlen);

                                }
                        }
                        else if (cmd->state == ISCSI_R2T_SENT)
                        {
                                if (cmd->r2t_sn >= pdu->begrun)
                                {
                                        /* Found the corresponding R2T PDU - re-transmit it */
                                        TRACE(TRACE_ISCSI, "handle_snack : Re-Transmitting R2T on SNACK Request\n");

                                        if (!pdu->runlen)
										{
											while (begrun <= cmd->r2t_sn)
											{
												retran_r2t(cmd, begrun);
												begrun++;
											}
										}
                                        else
                                            retran_r2t(cmd, pdu->begrun);
                                }
                        }
                }
                else if ( (pdu->flags & SNACK_TYPE) == STATUS_SNACK )
                {
                        if (cmd->state == ISCSI_SENT)
                        {
                                if ( !(session->targ_snack_flg & DATA_SNACK_REJECT ) )
                                {
                                        if (cmd->stat_sn >= pdu->begrun)
                                        {
                                                /* Found the corresponding Status PDU - re-transmit it */
                                                TRACE(TRACE_ISCSI, "handle_snack : Re-Transmitting iscsi Response on SNACK Request\n");

                                                cmd->conn->stat_sn --;
                                                send_iscsi_response(cmd);
                                                runlen --;

                                                if (runlen <= 0)
                                                        break;
                                        }
                                }
                                else
                                {
										/* new 18_04 SAI */
										iscsi_tx_rjt(conn, REASON_DATA_SNACK);

                                        /***TRACE(TRACE_ISCSI, "handle_snack : Ignoring STATUS Snack request - status already sent\n"); ***/
                                }
                        }
                }
        }

        /* wake up the transmit thread */
        if (cmd)
                up (&cmd->conn->tx_sem);

        TRACE(TRACE_ENTER_LEAVE, "Leave handle_snack \n");

        return 0;
}


/*
 * handle_immediate_data: this function is responsible for processing a
 * 							
 *
 */
static inline int handle_immediate_data(struct iscsi_cmnd *cmd, int length)
{
	struct iovec		*iov;
	struct msghdr		msg;
	struct scatterlist 	*st_list;
	int			niov=1, padding, err;
	int			offset=0, data_crc_len=0;
	__u8			*pad_bytes[4];
	__u32			digest;

	TRACE(TRACE_DEBUG, "Entering handle_immediate data\n");
	
	/* receive this data */
	st_list = (struct scatterlist*) cmd->cmnd->req->sr_buffer;
	niov = find_iovec_needed (length, st_list, offset);

	if (niov < 0)
	{
		TRACE_ERROR("handle_immediate_data: Trouble in find_iovec_needed\n");
		return -1;
	}
	
	padding = -(length)&3;

	if (padding)
		niov++;

	if (cmd->conn->data_crc == 1)
		niov++;
	
	iov = (struct iovec*) kmalloc (niov * sizeof(struct iovec), GFP_KERNEL);

	if (!iov)
	{
		TRACE_ERROR("handle_immediate_data: Could not malloc iov\n");
		return -1;
	}

	cmd->scatter_list_count = fill_iovec (iov,niov,st_list,&offset,length); 

	if (padding)
	{
		iov[niov-1 - cmd->conn->data_crc].iov_base = pad_bytes;
		iov[niov-1 - cmd->conn->data_crc].iov_len = padding;
	}

	if (cmd->conn->data_crc == 1)
	{
		data_crc_len = CRC_LEN;
		iov[niov-1].iov_base = &digest;
		iov[niov-1].iov_len = CRC_LEN;
	}

	fill_msg_hdr (&msg);
	msg.msg_iov = iov; 
	msg.msg_iovlen = niov;

	err = rx_data (cmd->conn, &msg, length + padding + data_crc_len);

	if (err != (length + padding + data_crc_len))
	{
		TRACE_ERROR("handle_immediate_data: Trouble in rx_data\n");
		return -1;
	}

	cmd->scatter_list_offset = offset;
	cmd->data_done += length; 
	
	TRACE(TRACE_DEBUG, "Leaving handle_immediate_data\n");

	return err;
}


/*
 * iscsi_xmit_response: responsible for transmitting data and scsi
 * response
 */
int iscsi_xmit_response (Target_Scsi_Cmnd *cmnd)
{
	struct iscsi_cmnd *cmd;

	cmd = search_iscsi_cmnd (cmnd);

	if (!cmd) 
	{
		TRACE_ERROR("iscsi_xmit_response: Could not find cmnd %d\n", cmnd->id);
		return -1;
	}

	cmd->state = ISCSI_DONE;

	/* wake up the transmit thread */
	up (&cmd->conn->tx_sem);
	
	return 0;
}


/*
 * iscsi_rdy_to_xfer: function changes the state of the command and
 * so that the transmit thread can send a R2T to the Initiator
 */
int iscsi_rdy_to_xfer (Target_Scsi_Cmnd *cmnd)
{
	struct iscsi_cmnd *cmd;

	cmd = search_iscsi_cmnd (cmnd);
	
	if (!cmd) 
	{
		TRACE_ERROR("iscsi_rdy_to_xfer: Could not find cmnd %d\n", cmnd->id);
		return -1;
	}

	if (cmd->data_length == 0)
		cmd->state = ISCSI_DONE;
	else
		cmd->state = ISCSI_BUFFER_RDY;

	if (cmd->immediate_data_present || cmd->unsolicited_data_present)
	{
		TRACE(TRACE_DEBUG, "iscsi_rdy_to_xfer: unblocking unsolicited_data_sem\n");
		up (&cmd->unsolicited_data_sem);
	}
	else
		up (&cmd->conn->tx_sem);

	return 0;
}


/*
 * iscsi_task_mgt_fn_done:
 *
 *
 */
void iscsi_task_mgt_fn_done (Target_Scsi_Message *msg)
{
	struct iscsi_cmnd *related_command;

	related_command = search_task_mgt_command(msg);

	if (!related_command)
	{
		TRACE_ERROR("iscsi_task_mgt_fn_done: Could not find msg %d\n", msg->message);
		return;
	}

	related_command->state = ISCSI_MGT_FN_DONE;

	/* wake up the transmit thread */
	up (&related_command->conn->tx_sem);

	return; 
}


/*
 * fill_msg_hdr: fills up the msg_hdr to the default values - returns
 * 0 if everything is okay, < 0 if there is trouble. Input value is the
 * msg_hdr struct that needs to be filled
 */
static int fill_msg_hdr (struct msghdr *msg)
{
	if (!msg) 
	{
		TRACE_ERROR("fill_msg_hdr: NULL struct given\n");
		return -1;
	}

	msg->msg_controllen	= 0;
	msg->msg_control	= NULL;
	msg->msg_flags	= 0;
	msg->msg_name		= NULL;
	msg->msg_namelen	= 0;

	return 0;
}


/*
 * search_iscsi_cmnd: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given command.
 * INPUT: Target_Scsi_Cmnd
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static inline struct iscsi_cmnd* search_iscsi_cmnd (Target_Scsi_Cmnd *cmnd)
{
	struct iscsi_cmnd	*cmd = NULL;
	struct iscsi_session	*session;
	struct iscsi_global	*host;
	
	if (!cmnd) 
	{
		TRACE_ERROR("search_iscsi_cmnd: cannot search for a NULL command\n");
		return NULL;
	}

	host	= (struct iscsi_global*) cmnd->device->dev_specific;
	session = host->session_list;

	for ( ; session != NULL; session = session->next)
	{
		down (&session->cmnd_sem);

		for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next)
		{
			if (cmd->cmnd == cmnd) 
			{
				up (&session->cmnd_sem);
				goto search_iscsi_cmnd_out;
			}
		}

		up (&session->cmnd_sem);
	}
	
search_iscsi_cmnd_out:
	return cmd;
}


/*
 * search_iscsi_message: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given message.
 * INPUT: Target_Scsi_Message
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static inline struct iscsi_cmnd *search_task_mgt_command 
	(Target_Scsi_Message *message)
{
	struct iscsi_cmnd *related_task_mgt_command = NULL;
	struct iscsi_session *related_session = NULL;
	struct iscsi_global *host = NULL;

	if (!message)
	{
		TRACE_ERROR("search_iscsi_message: cannot search for a NULL command\n");
		return NULL;
	}

	host = (struct iscsi_global*) message->device->dev_specific;
	related_session = host->session_list;

	for ( ; related_session != NULL; related_session = related_session->next)
	{
		down (&related_session->cmnd_sem);

		for (related_task_mgt_command = related_session->cmnd_list; 
			related_task_mgt_command!= NULL; 
			related_task_mgt_command = related_task_mgt_command->next)
		{
			if (related_task_mgt_command->message == message)
			{
				up (&related_session->cmnd_sem);
				goto out;
			}
		}

		up (&related_session->cmnd_sem);
	}

out:
	return related_task_mgt_command;
}


/*
 * search_tags: This goes through individual sessions and then
 * through the command queues for that session to look for a match to
 * a given set of tags.
 * INPUT: connection, initiator and target tag
 * OUTPUT: struct iscsi_cmnd if it finds one, NULL otherwise
 */
static inline struct iscsi_cmnd* search_tags (struct iscsi_conn *conn,
	__u32 init_task_tag, __u32 target_xfer_tag)
{
	struct iscsi_cmnd *cmd = NULL;
	struct iscsi_session *session;
/* struct iscsi_global *host; */

	if (!conn) 
	{
		TRACE_ERROR("search_tags: cannot search for a NULL command\n");
		return NULL;
	}

/*
	host = (struct iscsi_global*) conn->dev;
	session = host->session_list;
*/
	/* get session from conn structure directly to support multiple sessions */
	session = conn->session;
	
	down (&session->cmnd_sem);

	for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next)
	{
		TRACE(TRACE_DEBUG, "scsi cmnd %p, init_task_tag %d target_xfer_tag %d"
			" data_done %d xfer length %d data dir %d\n", 
			cmd->cmnd, cmd->init_task_tag, cmd->target_xfer_tag, 
			cmd->data_done, cmd->data_length, cmd->data_dir);			
	}

	for (cmd = session->cmnd_list; cmd != NULL; cmd = cmd->next)
	{
		if ((cmd->init_task_tag == init_task_tag) && ((cmd->target_xfer_tag 
			== target_xfer_tag) || (target_xfer_tag == 0xFFFFFFFF)))
		{
			up (&session->cmnd_sem);

			TRACE(TRACE_DEBUG, "Searched the following command\n");
			TRACE(TRACE_DEBUG, "scsi cmnd %p, init_task_tag %d target_xfer_tag %d"
				" data_done %d xfer length %d data dir %d\n", cmd->cmnd, 
				cmd->init_task_tag, cmd->target_xfer_tag, cmd->data_done,
				cmd->data_length, cmd->data_dir);			

			goto search_tags_out;
		}
	}

	up (&session->cmnd_sem);

search_tags_out:
	return cmd;
}


/**************************************************************************/
/*                            handle response                             */
/**************************************************************************/

static inline int handle_discovery_rsp (struct iscsi_cmnd *cmnd )
{
	__u8                		iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_text_rsp   	*hdr;
	struct iovec        		*iov;
	struct msghdr       		msg;
	char				*buffer, *ptr, *pad_bytes=NULL;
	int err, padding, niov=1, hdr_crc_len=0, data_crc_len=0, size=0, i;

	/* must allocate a multiple of 4 bytes in case of padding */
	if (!( buffer = (char *) kmalloc (64 * target_num * sizeof(char), GFP_KERNEL)))
	{
		TRACE_ERROR("handle_discovery_rsp: Unable to allocate space\n");
		return -1;
	}

	/* send the '\0' at the end of the string too */
	ptr = buffer;
	for( i=0; i<target_num; i++ )
	{
		int n=0; char temp[64];

		n = 1 + sprintf(temp, "TargetName=%s%d", TARGETNAME_HEADER, i);
		strcpy( ptr, temp );
		ptr += n;
		size += n;
	}

	if (cmnd->conn->hdr_crc == 1)
		niov++;

	niov++; // plus one for text

	/* ADD PADDING over here */
	padding = -(size)&3;

	if( padding )
		niov++; // one for padding

	if( cmnd->conn->data_crc == 1 )
		niov++; // one for data digest
	
	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_text_rsp *) iscsi_hdr;

	hdr->opcode 		= ISCSI_TARG_TEXT_RSP;

	if( cmnd->conn->session->draft_number <= DRAFT8 )
		hdr->opcode 	|= OLD_X_BIT;

	if( cmnd->conn->session->draft_number <= DRAFT9 )
		hdr->opcode 	|= I_BIT;

	hdr->flags		|= F_BIT;
	hdr->length		= cpu_to_be32 (size);
	hdr->init_task_tag	= cpu_to_be32 (cmnd->init_task_tag);
	hdr->target_xfer_tag	= cpu_to_be32 (cmnd->target_xfer_tag);
	hdr->stat_sn		= cpu_to_be32 (cmnd->conn->stat_sn++);
	hdr->exp_cmd_sn	= cpu_to_be32 (cmnd->session->exp_cmd_sn);
	hdr->max_cmd_sn	= cpu_to_be32 (cmnd->session->max_cmd_sn);

	iov = (struct iovec*) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	iov[0].iov_base 	= iscsi_hdr;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	if (cmnd->conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		iov[1].iov_len = CRC_LEN;
		iov[2].iov_base = buffer;
		iov[2].iov_len = size;
	}
	else
	{
		iov[1].iov_base = buffer;
		iov[1].iov_len = size;
	}

	if (padding)
	{
		pad_bytes = (__u8*) kmalloc (padding*sizeof(__u8), GFP_KERNEL);
		memset (pad_bytes, 0x0, padding);
		iov[niov-1 - cmnd->conn->data_crc].iov_base = pad_bytes;
		iov[niov-1 - cmnd->conn->data_crc].iov_len  = padding;
		TRACE(TRACE_DEBUG, "padding attached: %d bytes\n", padding );
	}

	if (cmnd->conn->data_crc == 1)
	{
		data_crc_len = CRC_LEN;
		iov[niov-1].iov_len = CRC_LEN;
	}

	fill_msg_hdr (&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len + size + 
		padding + data_crc_len);

	if (err != ISCSI_HDR_LEN + hdr_crc_len + size + padding + data_crc_len)
	{
		TRACE_ERROR("handle_discovery_rsp: tx_data returned < ISCSI_HDR_LEN\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", iscsi_hdr[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: discovery_response command sent\n", 
		cmnd->conn->conn_id);

	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_text_rsp(hdr);

	kfree (iov);
	kfree (pad_bytes);
	kfree (buffer);
	cmnd->state = ISCSI_SENT;

	return 0;
}


static inline int handle_logout_rsp (struct iscsi_cmnd *cmnd )
{
	__u8                		iscsi_hdr[ISCSI_HDR_LEN];
	int                 		err, niov=1, hdr_crc_len=0;
	struct iscsi_targ_logout_rsp  *hdr;
	struct iovec        		*iov;
	struct msghdr       		msg;

	if (cmnd->conn->hdr_crc == 1)
		niov++;

	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_logout_rsp *) iscsi_hdr;

	hdr->opcode 		= ISCSI_TARG_LOGOUT_RSP;

	if( cmnd->conn->session->draft_number <= DRAFT8 )
		hdr->opcode 	|= OLD_X_BIT;

	if( cmnd->conn->session->draft_number <= DRAFT9 )
		hdr->opcode 	|= I_BIT;

	hdr->flags		|= F_BIT;

	hdr->init_task_tag	= cpu_to_be32 (cmnd->init_task_tag);
	hdr->stat_sn		= cpu_to_be32 (cmnd->conn->stat_sn++);
	hdr->exp_cmd_sn	= cpu_to_be32 (cmnd->session->exp_cmd_sn);
	hdr->max_cmd_sn	= cpu_to_be32 (cmnd->session->max_cmd_sn);

	iov = (struct iovec*) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	iov[0].iov_base 	= iscsi_hdr;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	if (cmnd->conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		iov[1].iov_len = CRC_LEN;
	}

	fill_msg_hdr (&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len);

	if (err != ISCSI_HDR_LEN + hdr_crc_len)
	{
		TRACE_ERROR("handle_logout_rsp: tx_data returned error\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", iscsi_hdr[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: logout_response command sent\n", 
		cmnd->conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_logout_rsp(hdr);

	kfree (iov);
	cmnd->state = ISCSI_SENT;

	return 0;
}


static inline int handle_nopin (struct iscsi_cmnd *cmnd )
{
	__u8                		iscsi_hdr[ISCSI_HDR_LEN];
	struct iscsi_targ_nopin   	*hdr;
	struct iovec        		*iov;
	struct msghdr       		msg;
	char				*pad_bytes=NULL;
	int err, niov=1, hdr_crc_len=0, padding=0, data_crc_len=0;

	if (cmnd->conn->hdr_crc == 1)
		niov++;

	if (cmnd->data_length > 0)
	{
		niov++; 	// plus one for text

		padding	= -(cmnd->data_length)&3;

		if( padding )
			niov++; // one for padding

		if( cmnd->conn->data_crc == 1 )
			niov++; // one for data digest
	}
	
	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_nopin *) iscsi_hdr;

	hdr->opcode 		= ISCSI_TARG_NOP_IN;

	if( cmnd->conn->session->draft_number <= DRAFT8 )
		hdr->opcode 	|= OLD_X_BIT;

	if( cmnd->conn->session->draft_number <= DRAFT9 )
		hdr->opcode 	|= I_BIT;

	hdr->flags		|= F_BIT;

	hdr->length		= cpu_to_be32 (cmnd->data_length);
	hdr->init_task_tag	= cpu_to_be32 (cmnd->init_task_tag);
	hdr->target_xfer_tag	= ALL_ONES;
	hdr->stat_sn		= cpu_to_be32 (cmnd->conn->stat_sn++);
	hdr->exp_cmd_sn	= cpu_to_be32 (cmnd->session->exp_cmd_sn);
	hdr->max_cmd_sn	= cpu_to_be32 (cmnd->session->max_cmd_sn);

	iov = (struct iovec*) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	iov[0].iov_base 	= iscsi_hdr;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	if (cmnd->conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		iov[1].iov_len = CRC_LEN;
	}

	if (cmnd->data_length > 0)
	{
		iov[cmnd->conn->hdr_crc + 1].iov_base = cmnd->ping_data;
		iov[cmnd->conn->hdr_crc + 1].iov_len = cmnd->data_length;

		if (padding)
		{
			pad_bytes = (__u8*) kmalloc (padding*sizeof(__u8), GFP_KERNEL);
			memset (pad_bytes, 0x0, padding);
			iov[niov-1 - cmnd->conn->data_crc].iov_base = pad_bytes;
			iov[niov-1 - cmnd->conn->data_crc].iov_len  = padding;
			TRACE(TRACE_DEBUG, "padding attached: %d bytes\n", padding );
		}

		if (cmnd->conn->data_crc == 1)
		{
			data_crc_len = CRC_LEN;
			iov[niov-1].iov_len = CRC_LEN;
		}
	}

	fill_msg_hdr (&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len + 
		cmnd->data_length + padding + data_crc_len);

	if (err != ISCSI_HDR_LEN + hdr_crc_len + cmnd->data_length + 
		padding + data_crc_len)
	{
		TRACE_ERROR("handle_nopin: tx_data returned error\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", iscsi_hdr[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: nopin command sent\n", 
		cmnd->conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_nopin (hdr);

	kfree (iov);

	if (cmnd->data_length > 0)
		kfree (cmnd->ping_data);

	cmnd->state = ISCSI_SENT;

	return 0;
}

/* Seperated Data-In from Status Response PDU sending for Status Error Recovery - SAI */

static inline int send_iscsi_response (struct iscsi_cmnd *cmnd)
{
	struct iscsi_targ_scsi_rsp     *rsp;
	struct iovec    *iov;
	struct msghdr   msg;
	Scsi_Request    *req;
	__u8  iscsi_hdr[ISCSI_HDR_LEN];
	int   hdr_crc_len=0, data_crc_len=0, niov=0, err;
	int   data_length_left=0;
	int   transfer = 0;
	__u8  sense_data[16];
	int   sense_data_length = 0;
	int   filemark = 0;
  int   send_sense = 0;
	
	/* cdeng 3/24/02 */
	int   residual_count = 0;	
	int   overflow = 0;
	int   underflow = 0;

	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	req = cmnd->cmnd->req;

	if ((req->sr_data_direction == SCSI_DATA_READ) 
		&& (((req->sr_result >> 16) & 0xff) == DID_OK))
	{
		data_length_left = req->sr_bufflen;

		/* code from ramesh for tape device only */
		TRACE(TRACE_DEBUG, "Sense: %2x %2x %2x %2x %2x %2x %2x %2x\n",
			req->sr_sense_buffer[0], req->sr_sense_buffer[1],
			req->sr_sense_buffer[2], req->sr_sense_buffer[3],
			req->sr_sense_buffer[4], req->sr_sense_buffer[5],
			req->sr_sense_buffer[6], req->sr_sense_buffer[7]);

		if ((req->sr_sense_buffer[0] & 0x70) == 0x70)				
		{
			if ( (req->sr_sense_buffer[2] & 0xe0) != 0 )
			{
				/* on EOM, EOF or ILI calculate residual count*/
				transfer = (req->sr_sense_buffer[3] << 24) |
					(req->sr_sense_buffer[4] << 16) |
					(req->sr_sense_buffer[5] << 8) |
					req->sr_sense_buffer[6];
				TRACE(TRACE_DEBUG, "information in sense data: %d\n", transfer);
			}

			if ( (int)req->sr_sense_buffer[2] != 0)
			{
				send_sense = 1;
			}
			
			if ((req->sr_sense_buffer[2] & 0x80) != 0)
			{	/* EOF */
				filemark = 1;
			}
			else if ((req->sr_sense_buffer[2] & 0x60) != 0)
			{	/* EOM, or ILI */
				if ((req->sr_sense_buffer[0] & 0x80) != 0)
				{
					if (req->sr_sense_buffer[2] & 0x20)
					{	/* ILI */
						data_length_left -= transfer;
					}
				}
			}
		}

		/* check overflow or underflow problem, cdeng 3/24/02 */
		if (filemark == 1 || 1 == send_sense)
			data_length_left = 0;
		else if (data_length_left > cmnd->data_length)
		{
			residual_count = data_length_left - cmnd->data_length;
			data_length_left = cmnd->data_length;
			overflow = 1;
		}
		else if (data_length_left < cmnd->data_length)
		{
			residual_count = cmnd->data_length - data_length_left;
			underflow = 1;
		}

	}

	TRACE(TRACE_DEBUG, "Before sending scsi response\n");

	if (cmnd->conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		niov = 2;
	}
	else
		niov = 1;

	if (filemark == 1 || underflow == 1 || send_sense == 1)
	{
		if (cmnd->conn->data_crc == 1)
		{
			data_crc_len = CRC_LEN;
			niov += 2;
		}
		else
			niov += 1;
	}

	/* let us now send status information */
	iov = (struct iovec*) kmalloc (niov * sizeof (struct iovec), GFP_KERNEL);
	if (!iov)
	{
		TRACE_ERROR("handle_iscsi_done: could not allocate iovec\n");
		return -1;
	}

	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	rsp = (struct iscsi_targ_scsi_rsp *) iscsi_hdr;

	rsp->opcode 		= ISCSI_TARG_SCSI_RSP;

	// only for version 8
	if( cmnd->conn->session->draft_number <= DRAFT8 )
		rsp->opcode 	|= OLD_X_BIT;

	// for draft 8 and 9
	if( cmnd->conn->session->draft_number <= DRAFT9 )
		rsp->opcode 	|= I_BIT;

	// for draft not 12
	if( cmnd->conn->session->draft_number != DRAFT12 )
		rsp->flags 	|= F_BIT;
	
	/* cdeng 3/24/02 */
	if (overflow)
	{
		rsp->flags		|= O_BIT;

		if( cmnd->conn->session->draft_number <= DRAFT9 )
			rsp->target_xfer_tag	= cpu_to_be32 ( residual_count );
		else
			rsp->resid 	= cpu_to_be32 ( residual_count );
	}
	else if( underflow )
	{
		rsp->flags		|= U_BIT;

		if( cmnd->conn->session->draft_number <= DRAFT9 )
			rsp->target_xfer_tag	= cpu_to_be32 ( residual_count );
		else
			rsp->resid = cpu_to_be32 ( residual_count );
	}

	rsp->init_task_tag = cpu_to_be32 (cmnd->init_task_tag);
	cmnd->stat_sn      = cmnd->conn->stat_sn;
	rsp->stat_sn       = cpu_to_be32 (cmnd->conn->stat_sn++);
	rsp->exp_cmd_sn    = cpu_to_be32 (cmnd->session->exp_cmd_sn);
	rsp->max_cmd_sn    = cpu_to_be32 (cmnd->session->max_cmd_sn);

	if ((req->sr_data_direction == SCSI_DATA_READ)
		&& (((req->sr_result >> 16) & 0xff) == DID_OK))
	{
		rsp->exp_data_sn = cpu_to_be32 (cmnd->data_sn);
	}

	if (filemark == 1)
	{
		TRACE(TRACE_DEBUG, "filemark is found\n");
		rsp->status = CONDITION_GOOD;
		sense_data_length = 16;
		memset (sense_data, 0x0, sense_data_length);
		sense_data[1] = 0x0e;
		sense_data[2] = 0xf0;
		sense_data[4] = 0x80;
		sense_data[9] = 0x07;
		sense_data[15] = 0x01;
	}
	else if(underflow == 1)
	{
		TRACE(TRACE_DEBUG, "underflow is found\n");
		rsp->status	= CONDITION_GOOD;
		sense_data_length = 16;
	  	memset (sense_data, 0x0, sense_data_length);
		sense_data[1] = 0x0e;
		sense_data[2] = 0xf0;
		sense_data[4] = 0x20;
		sense_data[5] = req->sr_sense_buffer[3];
		sense_data[6] = req->sr_sense_buffer[4];
		sense_data[7] = req->sr_sense_buffer[5];
		sense_data[8] = req->sr_sense_buffer[6];
		sense_data[9] = 0x07;
	}
	else if (send_sense == 1)
	{
		rsp->status = CONDITION_GOOD;
		sense_data_length = 16;
		memset (sense_data, 0x0, sense_data_length);
		sense_data[1] = 0x0e;
		sense_data[2] = req->sr_sense_buffer[0];
		sense_data[4] = req->sr_sense_buffer[2];
		sense_data[5] = req->sr_sense_buffer[3];
		sense_data[6] = req->sr_sense_buffer[4];
		sense_data[7] = req->sr_sense_buffer[5];
		sense_data[8] = req->sr_sense_buffer[6];
		sense_data[9] = 0x07;
	}
																												 
   	rsp->length = cpu_to_be32 (sense_data_length);
      
	iov[0].iov_base	= iscsi_hdr;
	iov[0].iov_len	= ISCSI_HDR_LEN;

	if (sense_data_length > 0)
	{
		if (cmnd->conn->hdr_crc == 1)
		{
			iov[2].iov_base = sense_data;
			iov[2].iov_len = sense_data_length;
		}
		else
		{
			iov[1].iov_base = sense_data;
			iov[1].iov_len = sense_data_length;
		}
	}
	      
	fill_msg_hdr (&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	TRACE(TRACE_DEBUG, "handle_iscsi_done: sending status for cmnd_rn %.8x "
		"init_task_tag %.8x target_xfer_tag %.8x\n", cmnd->cmd_sn, 
	        cmnd->init_task_tag, cmnd->target_xfer_tag);

	if (sense_data_length == 0)
		data_crc_len = 0;
  
	err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + sense_data_length + hdr_crc_len + data_crc_len);
  
	if (err != ISCSI_HDR_LEN + hdr_crc_len + sense_data_length + data_crc_len) 
	{
		TRACE(TRACE_DEBUG, "data need to be transfered: %d\n", ISCSI_HDR_LEN + hdr_crc_len 
			+ sense_data_length + data_crc_len);
		TRACE_ERROR("handle_iscsi_done: did not transmit all of status\n");
		return -1;
	}
	
	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", ((u_char *)iov[0].iov_base)[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: scsi_response command sent\n", 
		cmnd->conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_scsi_rsp(rsp);

	kfree (iov);
	cmnd->state = ISCSI_SENT;

	return (0);
}

/*
 * handle_iscsi_done: This is responsible for building the data (if any)
 * and response headers and transmitting them to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is done
 * executing
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int handle_iscsi_done (struct iscsi_cmnd *cmnd, int startsn, int endsn)
{
	struct iscsi_targ_scsi_data_in *hdr;
	struct scatterlist             *st_list;
	struct iovec    *iov;
	struct msghdr   msg;
	Scsi_Request    *req;
	__u8  iscsi_hdr[ISCSI_HDR_LEN], *pad_bytes = NULL;
	int   hdr_crc_len=0, data_crc_len=0, padding, niov=0, err=0;
	int   data_payload_length =0;
	int   data_length_left=0;
	int data_send_length=0, offset;
	int   transfer = 0;
	int   filemark = 0;
	int   send_sense = 0;
  
	/* cdeng 3/24/02 */
	int   residual_count = 0;	
	int   overflow = 0;
	int   underflow = 0;
	int   prevsn = cmnd->prev_data_sn;

	req = cmnd->cmnd->req;

	if ((req->sr_data_direction == SCSI_DATA_READ) 
		&& (((req->sr_result >> 16) & 0xff) == DID_OK))
	{
		TRACE(TRACE_DEBUG, "handle_iscsi_done: sending data for cmnd_rn "
			"%.8x init_task_tag %.8x target_xfer_tag %.8x\n", cmnd->cmd_sn, 
			cmnd->init_task_tag, cmnd->target_xfer_tag);

		data_length_left = req->sr_bufflen;

		/* for our initiator which send bad exp_data_len */
		if (cmnd->cdb[0] == INQUIRY)
			cmnd->data_length = (int) cmnd->cdb[4];

		/* code from ramesh for tape device only */
		TRACE(TRACE_DEBUG, "Sense: %2x %2x %2x %2x %2x %2x %2x %2x\n",
			req->sr_sense_buffer[0], req->sr_sense_buffer[1],
			req->sr_sense_buffer[2], req->sr_sense_buffer[3],
			req->sr_sense_buffer[4], req->sr_sense_buffer[5],
			req->sr_sense_buffer[6], req->sr_sense_buffer[7]);

		if ((req->sr_sense_buffer[0] & 0x70) == 0x70)				
		{
			if ( (req->sr_sense_buffer[2] & 0xe0) != 0 )
			{
				/* on EOM, EOF or ILI calculate residual count*/
				transfer = (req->sr_sense_buffer[3] << 24) |
					(req->sr_sense_buffer[4] << 16) |
					(req->sr_sense_buffer[5] << 8) |
					req->sr_sense_buffer[6];
				TRACE(TRACE_DEBUG, "information in sense data: %d\n", transfer);
			}
			
			if ( (int)req->sr_sense_buffer[2] != 0)
			{
				send_sense = 1;
			}
												 
			if ((req->sr_sense_buffer[2] & 0x80) != 0)
			{	/* EOF */
				filemark = 1;
			}
			else if ((req->sr_sense_buffer[2] & 0x60) != 0)
			{	/* EOM, or ILI */
				if ((req->sr_sense_buffer[0] & 0x80) != 0)
				{
					if (req->sr_sense_buffer[2] & 0x20)
					{	/* ILI */
						data_length_left -= transfer;
					}
				}
			}
		}

		/* check overflow or underflow problem, cdeng 3/24/02 */
		if (filemark == 1 || 1 == send_sense)
			data_length_left = 0;
		else if (data_length_left > cmnd->data_length)
		{
			residual_count = data_length_left - cmnd->data_length;
			data_length_left = cmnd->data_length;
			overflow = 1;
		}
		else if (data_length_left < cmnd->data_length)
		{
			residual_count = cmnd->data_length - data_length_left;
			underflow = 1;
		}

        /* Store the previous data sequence for error recovery - SAI */
        if (!cmnd->retransmit_flg)
            cmnd->prev_data_sn = cmnd->data_sn;

		while (data_length_left > 0)
		{
			/* cdeng, Nov. 20: handle MaxBurstLength in DataIn PDU */
			if ((data_length_left <= cmnd->conn->session->oper_param->MaxBurstLength
			&& cmnd->conn->session->draft_number >= DRAFT9) ||
			(data_length_left <= cmnd->conn->session->oper_param->MaxBurstSize
			&& cmnd->conn->session->draft_number <= DRAFT8))
			{
				data_send_length = data_length_left;
			}
			else if (cmnd->conn->session->draft_number >= DRAFT9)
			{
				data_send_length = cmnd->conn->session->oper_param->MaxBurstLength;
			}
			else
			{
				data_send_length = cmnd->conn->session->oper_param->MaxBurstSize;
			}
			
			offset = req->sr_bufflen - data_length_left;
			data_length_left -= data_send_length;
			
			while (data_send_length >0)
			{
				TRACE(TRACE_DEBUG,"data_length_left: %d, data_send_length: %d\n", data_length_left, data_send_length);	
				
			/* we have data to send */
			if (data_send_length > ISCSI_DATA_MAX) 
			{
				TRACE_ERROR("handle_iscsi_done: this needs data greater than "
					"ISCSI_DATA_MAX and not ready to deal with this\n");
				cmnd->state = ISCSI_DEQUEUE;
				return -1;
			}

			memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
			hdr = (struct iscsi_targ_scsi_data_in *) iscsi_hdr;

			hdr->opcode 	= ISCSI_TARG_SCSI_DATA_IN;

			// only for version 8
			if (cmnd->conn->session->draft_number <= DRAFT8)
				hdr->opcode 	|= OLD_X_BIT;

			if (cmnd->conn->session->draft_number <= DRAFT9)
				hdr->opcode 	|= I_BIT;

			if ((data_send_length <= cmnd->conn->max_send_length && 
				cmnd->conn->session->draft_number >=  DRAFT9) || 
				(data_send_length <= cmnd->conn->session->oper_param->DataPDULength 
				&& cmnd->conn->session->draft_number <= DRAFT8))
			{
				data_payload_length = data_send_length;
				hdr->flags 	|= F_BIT;
			}
			else if (cmnd->conn->session->draft_number >= DRAFT9)
			{
				data_payload_length = cmnd->conn->max_send_length;
			}
			else
			{
				data_payload_length=cmnd->conn->session->oper_param->DataPDULength;
			}

			/* ADD PADDING over here */
			padding = -(data_payload_length)&3;

			hdr->length = cpu_to_be32 (data_payload_length);
			hdr->init_task_tag = cpu_to_be32 (cmnd->init_task_tag);

			if (cmnd->conn->session->draft_number > DRAFT9)
				hdr->target_xfer_tag	= ALL_ONES;

			// if S flag set
			// hdr->stat_sn		= cpu_to_be32 (cmnd->conn->stat_sn++);
			hdr->exp_cmd_sn		= cpu_to_be32 (cmnd->session->exp_cmd_sn);
			hdr->max_cmd_sn		= cpu_to_be32 (cmnd->session->max_cmd_sn);
            /* re-transmit only the requested Data PDU - SAI */
            if (!cmnd->retransmit_flg)
                hdr->data_sn        = cpu_to_be32 (cmnd->data_sn++);
            else
                hdr->data_sn        = cpu_to_be32 (prevsn);
			hdr->offset		= cpu_to_be32 (offset);

			if (underflow && cmnd->cdb[4] == cmnd->data_length)
				hdr->offset = cpu_to_be32 (offset - residual_count);
			
			st_list=((struct scatterlist*)req->sr_buffer)+cmnd->scatter_list_count;

			TRACE(TRACE_DEBUG, "offset within sg entry %d offset within sg list"
				" %d\n",cmnd->scatter_list_offset, cmnd->scatter_list_count);

			niov = find_iovec_needed (data_payload_length, st_list, 
				cmnd->scatter_list_offset);

			if (niov < 0) 
			{
				TRACE_ERROR("handle_iscsi_done: Trouble in find_iovec_needed\n");
				return niov;
			}

			niov++; /* Add one for header */

			if (cmnd->conn->hdr_crc == 1)
			{
				hdr_crc_len = CRC_LEN;
				niov++; /* add one for hdr_crc */
			}

			if (padding)
				niov++; /* add one for padding */

			if (cmnd->conn->data_crc == 1)
				niov++; /* add one for data_crc */

			iov = (struct iovec*) kmalloc(niov*sizeof(struct iovec),GFP_KERNEL);
		
			if (!iov) 
			{
				TRACE(TRACE_DEBUG,"handle_iscsi_done: Could not allocate space for iov\n");
				return -1;
			}

			iov[0].iov_base = iscsi_hdr;
			iov[0].iov_len  = ISCSI_HDR_LEN;

			if (cmnd->conn->hdr_crc == 1)
			{
				cmnd->scatter_list_count += fill_iovec (&iov[2], niov, st_list, 
					&cmnd->scatter_list_offset, data_payload_length);
			}
			else
			{
				cmnd->scatter_list_count += fill_iovec (&iov[1], niov, st_list, 
					&cmnd->scatter_list_offset, data_payload_length);
			}
			
			if (padding) 
			{
				pad_bytes = (__u8*) kmalloc (padding*sizeof(__u8), GFP_KERNEL);
				memset (pad_bytes, 0x0, padding);
				iov[niov-1 - cmnd->conn->data_crc].iov_base = pad_bytes;
				iov[niov-1 - cmnd->conn->data_crc].iov_len  = padding;
				TRACE( TRACE_DEBUG, "padding attached: %d bytes\n", padding );
			}

			if (cmnd->conn->data_crc == 1)
			{
				data_crc_len = CRC_LEN;
				iov[niov-1].iov_len = CRC_LEN;
			}

			fill_msg_hdr (&msg);
			msg.msg_iov = iov;
			msg.msg_iovlen = niov;

            /* skip the data PDU re-transmissions that are not asked for - SAI */

            if (cmnd->retransmit_flg)
            {
                if (prevsn < startsn || prevsn > endsn)
                    goto skip_transmission;
            }

			err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len + 
				data_payload_length + padding + data_crc_len);

			if (err != ISCSI_HDR_LEN + hdr_crc_len + data_payload_length + 
				padding + data_crc_len) 
			{
				TRACE_ERROR("handle_iscsi_done: Did not transmit all bytes\n");
				return -1;
			}

			if (TRACE_TEST(TRACE_ISCSI_FULL))
			{
				int i=0;
				for (i=0; i<ISCSI_HDR_LEN; i++)
				{
					printk("%02x ", ((u_char *)iov[0].iov_base)[i]);
					if (i==24) printk("\n");
				}
				printk("\n");
			}

			TRACE(TRACE_ISCSI, "iscsi_tx_%d: data_in command sent\n", 
				cmnd->conn->conn_id);

			if (TRACE_TEST(TRACE_ISCSI_FULL))
				print_targ_scsi_data_in(hdr);


    /* Added for error recovery - SAI */
    skip_transmission:

			kfree (iov);

			if (padding)
				kfree (pad_bytes);

			data_send_length -= data_payload_length;
			offset += data_payload_length;
			
            /* Increment the retransmitted data sequence number - SAI */
            prevsn ++;

			TRACE(TRACE_DEBUG, "data sent %d data_left %d sg_list_offset %d "
				"sg_entry_offset %d\n", err, data_send_length, 
				cmnd->scatter_list_count, cmnd->scatter_list_offset);
			}
		}
	}
	
	/* Seperated Response for Error Recovery - SAI */
	if (cmnd->retransmit_flg)
	{
		cmnd->state = ISCSI_SENT;
		cmnd->retransmit_flg = 0;
	}
	else
	{
		send_iscsi_response(cmnd);
	}

	return (0);

}


/*
 * handle_iscsi_mgt_fn_done: This is responsible for building the data (if any)
 * and response headers and transmitting them to the correct initiator.
 * The command state is set to DEQUEUE to reflect that it is done
 * executing
 * INPUT: command that is done.
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int handle_iscsi_mgt_fn_done(struct iscsi_cmnd *cmnd)
{
	struct iscsi_targ_task_mgt_response rsp;
	struct iscsi_cmnd *temp, *prev = NULL, *aborted_command = NULL;
	struct iovec *iov;
	struct msghdr msg;
	int err=0, hdr_crc_len=0, niov=1;

	if (cmnd->conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		niov = 2;
	}

	iov = (struct iovec*) kmalloc (niov * sizeof (struct iovec), GFP_KERNEL);

	if (!iov)
	{
		TRACE_ERROR("handle_iscsi_mgt_fn_done: could not allocate iovec\n");
		return -1;
	}
	
	memset(&rsp, 0x0, ISCSI_HDR_LEN);

	rsp.opcode 		= ISCSI_TARG_TASK_MGMT_RSP;

	// only for version 8
	if( cmnd->conn->session->draft_number <= DRAFT8 )
		rsp.opcode 		|= F_BIT;

	if( cmnd->conn->session->draft_number <= DRAFT9 )
		rsp.opcode 		|= I_BIT;

	rsp.flags 		|= F_BIT;
	rsp.response 		= FUNCTION_COMPLETE;
	rsp.qualifier		= 0;
	rsp.init_task_tag 	= htonl(cmnd->init_task_tag);
	rsp.ref_task_tag 	= htonl(cmnd->ref_task_tag);
	rsp.stat_sn 		= htonl(cmnd->conn->stat_sn++);
	rsp.exp_cmd_sn 	= htonl(cmnd->session->exp_cmd_sn);
	rsp.max_cmd_sn 	= htonl(cmnd->session->max_cmd_sn);

	iov[0].iov_base 	= &rsp;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	fill_msg_hdr(&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	TRACE(TRACE_DEBUG, "handle_iscsi_mgt_fn_done: sending task_mgt_repsonse"
		" for cmnd_rn %.8x init_tag %.8x target_xfer_tag %.8x\n", 
		cmnd->cmd_sn, cmnd->init_task_tag, cmnd->target_xfer_tag);

	err = tx_data(cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len);

	if(err != ISCSI_HDR_LEN + hdr_crc_len)
	{
		TRACE_ERROR("handle_iscsi_mgt_fn_done: Trouble in tx_data\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", ((u_char *)iov[0].iov_base)[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_rx_%d: mgt_response command sent\n", 
		cmnd->conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_task_mgt_response(&rsp);

	down_interruptible (&cmnd->session->cmnd_sem);

	for (temp = cmnd->session->cmnd_list; temp != NULL; prev = temp, 
		temp = temp->next)
	{
		if (temp == cmnd)
		break;
	}

	if (!temp)
	{
		TRACE_ERROR("iscsi_iscsi_mgt_fn_done: match not found\n");
		err = -1;
		goto out;
	}

	if (!prev)
		cmnd->session->cmnd_list = cmnd->next;
	else
		prev->next = cmnd->next;

	kfree(temp);
	temp = NULL;

	/* clear the aborted command struct entry */
	aborted_command = search_tags(cmnd->conn, ntohl(rsp.ref_task_tag), 
		0xFFFFFFFF);

	for (temp = aborted_command->session->cmnd_list; temp != NULL; 
		prev = temp, temp = temp->next)
	{
		if (temp == aborted_command)
		break;
	}

	if (!temp)
	{
		TRACE_ERROR("handle_iscsi_mgt_fn_done: match not found\n");
		err = -1;
		goto out;
	}

	if (!prev)
		cmnd->session->cmnd_list = cmnd->next;
	else
		prev->next = cmnd->next;

	kfree(temp);
	temp = NULL;

out:
	up (&cmnd->session->cmnd_sem);

	return(err);
}


/*
 * iscsi_dequeue: this function dequeues the given command from the
 * session list and sets the entered pointer to the start of the queue
 * This is necessary to avoid any confusion about the command and its
 * value
 * INPUT: command which has to be dequeued
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int iscsi_dequeue (struct iscsi_cmnd **cmnd_pptr)
{
	int			err = 0;
	struct iscsi_cmnd 	*temp, *prev = NULL;
	
	TRACE(TRACE_DEBUG, "iscsi_dequeue: init_task_tag %.8x\n", 
		(*cmnd_pptr)->init_task_tag);

	down_interruptible (&(*cmnd_pptr)->session->cmnd_sem);

	for (temp = (*cmnd_pptr)->session->cmnd_list; temp != NULL; prev = temp, 
		temp = temp->next)
	{
		if (temp == (*cmnd_pptr))
			break;
	}

	if (!temp) 
	{
		TRACE_ERROR("iscsi_dequeue: match not found\n");
		err = -1;
		goto iscsi_dequeue_out;
	}

	if (!prev)
		(*cmnd_pptr)->session->cmnd_list = (*cmnd_pptr)->next;
	else	
		prev->next = (*cmnd_pptr)->next;

	(*cmnd_pptr) = (*cmnd_pptr)->session->cmnd_list;

	(*cmnd_pptr)->session->max_cmd_sn++;	

	if (temp->cmnd != NULL)
	{
		if (scsi_target_done (temp->cmnd) < 0) 
		{
			TRACE_ERROR("iscsi_dequeue: scsi_target_done returned an error\n");
			goto iscsi_dequeue_out;
		}
	}

	/* Free the R2T cookie if any - SAI */
	free_r2t_cookie(temp);

	kfree (temp);
	temp = NULL;
	
iscsi_dequeue_out:
	up (& (*cmnd_pptr)->session->cmnd_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_dequeue \n");

	return (err);
}


/*
 * iscsi_tx_reject: this function transmits a reject to the
 * Initiator
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int iscsi_tx_rjt (struct iscsi_conn *conn, __u8 reason)
{
	__u8                	iscsi_hdr[ISCSI_HDR_LEN], *data=NULL;
	int                 	err, i, niov = 2, hdr_crc_len = 0, data_crc_len=0;
	struct iscsi_targ_rjt *hdr;
	struct iovec        	*iov;
	struct msghdr       	msg;

	if (conn->session && conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		niov++;
	}

	iov = (struct iovec*) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_rjt *) iscsi_hdr;

	hdr->opcode 		= ISCSI_TARG_RJT;

	if( conn->dev->draft_number <= DRAFT8 )
		hdr->opcode 	|= F_BIT;

	if( conn->dev->draft_number <= DRAFT9 )
		hdr->opcode 	|= I_BIT;

	hdr->flags		|= F_BIT;
	hdr->reason 		= reason;
	hdr->length		= cpu_to_be32 (ISCSI_HDR_LEN);
	/* init task tag must be all ones - new 18_04 SAI */
	hdr->init_task_tag = ALL_ONES;

	if(conn)
		hdr->stat_sn	= cpu_to_be32 (conn->stat_sn++);

	if(conn->session)
	{
		hdr->exp_cmd_sn	= cpu_to_be32 (conn->session->exp_cmd_sn);
		hdr->max_cmd_sn	= cpu_to_be32 (conn->session->max_cmd_sn);
	}

	iov[0].iov_base 	= iscsi_hdr;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	data = (__u8*) kmalloc (ISCSI_HDR_LEN * sizeof(__u8), GFP_KERNEL);
	memset (data, 0x0, ISCSI_HDR_LEN);

	for(i=0; i<ISCSI_HDR_LEN; i++)
		data[i] = conn->bad_hdr[i];

	iov[niov-1].iov_base	= data;
	iov[niov-1].iov_len 	= ISCSI_HDR_LEN;

    /* Added the missing data_crc checks - SAI */
    if (conn->data_crc == 1)
    {
        data_crc_len = CRC_LEN;
        niov++;
        iov[niov-1].iov_len = CRC_LEN;
    }

	fill_msg_hdr (&msg);
	msg.msg_iov 		= iov;
	msg.msg_iovlen  	= niov; /* Removed Hardcoded value - SAI */

    /* Added data_crc field for calls to tx_data - SAI */
    err = tx_data (conn, &msg, ISCSI_HDR_LEN * 2 + hdr_crc_len + data_crc_len);

    /* Added data_crc field for checks after tx_data - SAI */
    if (err != ISCSI_HDR_LEN * 2 + hdr_crc_len + data_crc_len)
	{
		TRACE_ERROR("iscsi_tx_rjt: tx_data returned < ISCSI_HDR_LEN\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", iscsi_hdr[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: reject command sent\n", conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_rjt(hdr);

	return 0;
}


static inline int iscsi_tx_login_rsp (struct iscsi_conn *conn, 
	__u8 status_class, __u8 status_detail)
{
	__u8                		iscsi_hdr[ISCSI_HDR_LEN];
	int                 		err, niov=1, hdr_crc_len = 0;
	struct iscsi_targ_login_rsp   *hdr;
	struct iovec        		*iov;
	struct msghdr       		msg;

	if (conn->session && conn->hdr_crc == 1)
	{
		hdr_crc_len = CRC_LEN;
		niov++;
	}

	iov = (struct iovec*) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
	hdr = (struct iscsi_targ_login_rsp *) iscsi_hdr;

	hdr->opcode 		= ISCSI_TARG_LOGIN_RSP;

	// only for version 8
	if( conn->dev->draft_number <= DRAFT8 )
		hdr->opcode 	|= F_BIT;

	if( conn->dev->draft_number <= DRAFT9 )
		hdr->opcode 	|= I_BIT;

	hdr->status_class	= status_class;
	hdr->status_detail	= status_detail;

	if( conn->dev->draft_number <= DRAFT8 )
	{/* draft 8 and before */
		hdr->version_max	= 2;
		hdr->version_active	= 2;
	}
	else if( conn->dev->draft_number <= DRAFT9 )
	{/* draft 9 */
		hdr->version_max	= 3;
		hdr->version_active	= 3;
	}
	else
	{/* draft 11 and after */
		hdr->version_max	= 0;
		hdr->version_active	= 0;
	}
        
	if(conn->session)
	{
		// send correct isid
		if( conn->dev->draft_number >= DRAFT9 )
			memcpy(hdr->isid, conn->session->isid, 6);
		else if( conn->session )
		{
			hdr->isid[0] = 0;
			hdr->isid[1] = 0;
			hdr->isid[2] = 0;
			hdr->isid[3] = 0;
			hdr->isid[4] = conn->session->isid[4];
			hdr->isid[5] = conn->session->isid[5];
		}

		hdr->tsih		= cpu_to_be16 (conn->session->tsih);
		hdr->init_task_tag	= cpu_to_be32 (conn->session->init_task_tag);
		hdr->exp_cmd_sn	= cpu_to_be32 (conn->session->exp_cmd_sn);
	}

	iov[0].iov_base 	= iscsi_hdr;
	iov[0].iov_len 	= ISCSI_HDR_LEN;

	fill_msg_hdr (&msg);
	msg.msg_iov 		= iov;
	msg.msg_iovlen  	= niov;

	err = tx_data (conn, &msg, ISCSI_HDR_LEN + hdr_crc_len);

	if (err != ISCSI_HDR_LEN + hdr_crc_len)
	{
		TRACE_ERROR("iscsi_tx_login_rsp: tx_data returned < ISCSI_HDR_LEN\n");
		return -1;
	}

	if (TRACE_TEST(TRACE_ISCSI_FULL))
	{
		int i=0;
		for (i=0; i<ISCSI_HDR_LEN; i++)
		{
			printk("%02x ", iscsi_hdr[i]);
			if (i==24) printk("\n");
		}
		printk("\n");
	}

	TRACE(TRACE_ISCSI, "iscsi_tx_%d: login_response command sent\n", 
		conn->conn_id);

	if (TRACE_TEST(TRACE_ISCSI_FULL))
		print_targ_login_rsp(hdr);

	return 0;
}

/*
 * iscsi_tx_r2t: this function transmits a ready-to-transfer to the
 * Initiator so that it can transmit the data necessary for a WRITE
 * INPUT: command for which R2T needs to be transferred
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static inline int iscsi_tx_r2t (struct iscsi_cmnd *cmnd)
{
	__u8			iscsi_hdr[ISCSI_HDR_LEN];
	int			err, niov, hdr_crc_len=0;
	struct iscsi_targ_r2t	*hdr;
	struct iovec		*iov;
	struct msghdr		msg;
	__u32			data_length_left=0;
	int			max_burst_len=0;
	struct iscsi_cookie *cookie=NULL;
	
	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_tx_r2t\n");

	data_length_left = cmnd->data_length - cmnd->first_burst_len;
	
	while (data_length_left > 0)
	{
		niov = 1;
		
		if (cmnd->conn->hdr_crc == 1)
		{
			hdr_crc_len = CRC_LEN;
			niov++;
		}

		iov = (struct iovec*) kmalloc(niov*sizeof (struct iovec),GFP_KERNEL);

		/* 
		 * since the entire space will have been malloced by the STML
		 * we need to send just one R2T to the Initiator.
		 */
		memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
		hdr = (struct iscsi_targ_r2t *) iscsi_hdr;
	
		hdr->opcode		= ISCSI_TARG_R2T;

		// only for version 8
		if( cmnd->conn->session->draft_number <= DRAFT8 )
			hdr->opcode 	|= F_BIT;
		if( cmnd->conn->session->draft_number <= DRAFT9 )
			hdr->opcode	|= I_BIT;

		hdr->flags 		|= F_BIT;
		hdr->init_task_tag	= cpu_to_be32 (cmnd->init_task_tag);
		hdr->target_xfer_tag= cpu_to_be32 (cmnd->target_xfer_tag);
		hdr->stat_sn	= cpu_to_be32 (cmnd->conn->stat_sn);
		hdr->exp_cmd_sn	= cpu_to_be32 (cmnd->session->exp_cmd_sn);
		hdr->max_cmd_sn	= cpu_to_be32 (cmnd->session->max_cmd_sn);
        hdr->r2t_sn     = cpu_to_be32 (cmnd->r2t_sn++);
		hdr->offset 	= cpu_to_be32 (cmnd->data_length - data_length_left);
		
		if( cmnd->conn->session->draft_number <= DRAFT13 )
			max_burst_len = cmnd->conn->session->oper_param->MaxBurstSize;
		else if( cmnd->conn->session->draft_number >= DRAFT18 )
			max_burst_len = cmnd->conn->session->oper_param->MaxBurstLength;
			
		/* form an R2T cookie to be used for re-transmissions - SAI */
		if (cmnd->conn->session->oper_param->ErrorRecoveryLevel > 0)
		{
			cookie = create_r2t_cookie(cmnd);

			if (cookie)
			{
				cookie->seq    	= cmnd->r2t_sn-1;
				cookie->offset   = cmnd->data_length - data_length_left;
				cookie->xfer_len = (data_length_left<=max_burst_len)?data_length_left:max_burst_len;
			}
		}

		if( data_length_left <= max_burst_len )
		{
			hdr->xfer_len = cpu_to_be32 (data_length_left); 
			data_length_left = 0;
		}
		else
		{
			hdr->xfer_len = cpu_to_be32 (max_burst_len); 
			data_length_left -= max_burst_len;
		}

		iov[0].iov_base	= iscsi_hdr;
		iov[0].iov_len	= ISCSI_HDR_LEN;
	
		fill_msg_hdr (&msg);
		msg.msg_iov	= iov;
		msg.msg_iovlen	= niov;

		err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len);
		
		if (err != ISCSI_HDR_LEN + hdr_crc_len) 
		{
			TRACE_ERROR("iscsi_tx_r2t: tx_data returned < ISCSI_HDR_LEN\n");
			return -1;
		}

		if (TRACE_TEST(TRACE_ISCSI_FULL))
		{
			int i=0;
			for (i=0; i<ISCSI_HDR_LEN; i++)
			{
				printk("%02x ", iscsi_hdr[i]);
				if (i==24) printk("\n");
			}
			printk("\n");
		}
	
		TRACE(TRACE_ISCSI, "iscsi_tx_%d: r2t command sent\n", 
			cmnd->conn->conn_id);

		if (TRACE_TEST(TRACE_ISCSI_FULL))
			print_targ_r2t(hdr);

        /* Increment the outstanding R2T for every R2T sent - SAI */
        cmnd->outstanding_r2t++;

    	/* store the r2t time stamp - SAI */
		/* error recovery ver ref18_04 */
		cmnd->timestamp = jiffies;

		if (data_length_left>0 && cmnd->session->oper_param->MaxOutstandingR2T 
			<= cmnd->outstanding_r2t)
			down_interruptible (&cmnd->conn->tx_sem);
	}

	/* error recovery ver ref18_04 */
	cmnd->state = ISCSI_R2T_SENT;

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_tx_r2t\n");

	return (0);
}


/*
 * find_iovec_needed: this function returns the number of iovecs that
 * are needed to receive some data
 * INPUT: amount of data needed, pointer to scattergather list (from the
 * 	  point where the data is to start for this round, offset within
 * 	  the first scattergather element - in case it was not used up
 * 	  completely
 * OUTPUT: number of iovectors needed, else < 0 if there is trouble
 */
static int find_iovec_needed (unsigned int data_len, 
	struct scatterlist *st_list, unsigned int offset)
{
	int i = 0;

	while (data_len) 
	{
		if (data_len > (st_list->length-offset))
			data_len -= (st_list->length - offset);
		else
			data_len = 0;

		i++;
		offset = 0;
	}

	return (i);
}


/*
 * fill_iovec: This function fills in a given iovec structure so as to
 * reflect a given scattergather list upto the length that is required
 * No checks are made on the length on the iovec given except that it
 * needs to be greater than the size needed. Also, the scatterlist is
 * also not checked. The integer offset upon return reflects the posn of
 * the offset in the last scatterlist that was used for this function
 * INPUT: struct iovec, no. of iovecs, scatterlist, ptr to offset within
 * 	  the first element, and the total length required
 * OUTPUT: how many st_list elements are used, < 0 if there is trouble
 */
static int fill_iovec (struct iovec *iov, int niov, 
       struct scatterlist *st_list, int *offset, unsigned int data)
{
	int i = 0, count = 0;
	
	TRACE(TRACE_DEBUG, "offset: %d, data_len: %d\n", *offset, data);

	while ((data != 0) && (i < niov)) 
	{
		iov[i].iov_base = st_list[i].address + (*offset);

		if ((st_list[i].length - *offset) > data) 
		{
			iov[i].iov_len	= data;
			*offset		+= data;
		}
		else 
		{
			iov[i].iov_len	= st_list[i].length - *offset;
			*offset		= 0;
			count++;
		}

		data -= iov[i].iov_len;

		TRACE(TRACE_DEBUG, "iov: %p i: %d iov_base: %p iov_len: %d\n", iov+i, 
			i, iov[i].iov_base, iov[i].iov_len);

		i++;
	}

	return (count);
}


/*  called when a file in the /proc/scsi_target/iscsi/ directory is read or written */ 
int iscsi_proc_info(char *buffer, char **start, off_t offset, 
	int length, int inode, int inout)
{
	int result;
	int int_val;
	int param_neg_info;
	char *ptr_to_keytext;
	//struct parameter_type   *p;


	TRACE(TRACE_DEBUG, "Entering iscsi_proc_info\n");

	if (inout == 1)        
	{
		/* write to this file, buffer should be "iscsi_initiator ip 1234abcd 
			target 6 lun 0" */
		if( buffer == NULL  ||  length > PAGE_SIZE )
			result = -EINVAL;
		else
		{
			result = length;

			if( length == PAGE_SIZE )
				length--;

			buffer[length] = '\0';  /* make the buffer into a string */

			TRACE(TRACE_DEBUG, "string got: %s\n", buffer);

			if( length >= 20 && strncmp(buffer, "iscsi_target manage ", 20) == 0)
			{
				/* this is for configuring the login parameters */

				/* this buffer has expected insterface, scan UP/DOWN */
				param_neg_info = simple_strtoul( buffer+20, &ptr_to_keytext, 0);

				/* Move a position to ignore space */
				ptr_to_keytext++;

				/* Now configure the parameters */
				if( param_neg_info == FORCE )
				{
					/* special to force login subphases */
					if( strcmp(ptr_to_keytext, "s") == 0 )
						devdata->force |= USE_SECURITY_PHASE;
					else if( strcmp(ptr_to_keytext, "o") == 0 )
						devdata->force |= USE_OPERATIONAL_PHASE;
					else if( strcmp(ptr_to_keytext, "f") == 0 )
						devdata->force |= USE_FLAT_SPACE_LUN;
					else if( strcmp(ptr_to_keytext, "r") == 0 )
						devdata->force |= USE_FULL_REPLIES;
					else if( strncmp(ptr_to_keytext, "n=", 2) == 0 )
						devdata->nop_period = simple_strtoul(ptr_to_keytext+2, NULL, 0);
					else if( strncmp(ptr_to_keytext, "v=", 2) == 0 )
					{ /* set draft number to enforce */
						devdata->draft_number = simple_strtoul(ptr_to_keytext+2, NULL, 0);
						/* now adjust Burst parameter units to match version */
						adjust_version_diff(*devdata->param_tbl, devdata->draft_number);
					}
                    /* set the r2t period - SAI */
                    else if (strncmp(ptr_to_keytext, "r2tp=", 5) == 0)
                    {
                        devdata->r2t_period =simple_strtoul(ptr_to_keytext+5, NULL, 0);
                    }
					/* chap support - CHONG */
					else if (strcmp(ptr_to_keytext, "t") ==0)
					{
						TRACE(TRACE_ISCSI, "target confirmation enabled\n");
						devdata->chap_parameter.chap_flags |= USE_TARGET_CONFIRMATION;
					}
					else if (strcmp(ptr_to_keytext, "b") == 0)
					{
						TRACE(TRACE_ISCSI, "base64 number enabled\n");
						devdata->chap_parameter.chap_flags |= USE_BASE64;
						devdata->chap_parameter.local_ctx->number_format = BASE64_FORMAT;
						devdata->chap_parameter.peer_ctx->number_format = BASE64_FORMAT;
					}
					else if (strncmp(ptr_to_keytext, "px=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "peer secret set to %s\n", ptr_to_keytext + 3);
						CHAP_SetSecret(ptr_to_keytext + 3, devdata->chap_parameter.peer_ctx);
					}
					else if (strncmp(ptr_to_keytext, "pn=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "peer name set to %s\n", ptr_to_keytext + 3);
						CHAP_SetName(ptr_to_keytext + 3, devdata->chap_parameter.peer_ctx);
					}
					else if (strncmp(ptr_to_keytext, "lx=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "local secret set to %s\n", ptr_to_keytext + 3);
						CHAP_SetSecret(ptr_to_keytext + 3, devdata->chap_parameter.local_ctx);
					}
					else if (strncmp(ptr_to_keytext, "ln=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "local name set to %s\n", ptr_to_keytext + 3);
						CHAP_SetName(ptr_to_keytext + 3, devdata->chap_parameter.local_ctx);
					}
					else if (strncmp(ptr_to_keytext, "cl=", 3) == 0)
					{
						int_val = simple_strtoul(ptr_to_keytext + 3, NULL, 0);
						if ((int_val <= 0) || (int_val > MAX_CHALLENGE_LENGTH))
						{
							TRACE_ERROR("invalid length of challenge: %d\n", int_val);
						}	
						TRACE(TRACE_ISCSI, "challenge length set to %d\n", int_val);
						CHAP_SetChallengeLength(int_val, devdata->chap_parameter.local_ctx);
					}
					else
					{
						TRACE_ERROR("iscsi_proc_info: unknown force \"%s\"\n", ptr_to_keytext);
					}
				}
				else if( param_neg_info == RESTORE )
				{ /* restore all settings to default values */

					/* reset parameter list to copy of defaults */
					param_tbl_uncpy(*devdata->param_tbl);
					param_tbl_cpy(*devdata->param_tbl, config_params);
                                       
					/* reset the draft number to default */
					devdata->draft_number = DEFAULT_DRAFT;
        
					/* reset all other fields to 0 */
					devdata->nop_period = 0;
					devdata->force = 0;
				}
                else if (param_neg_info == SNACK)
                {
                    /* Error Recovery options to enable or disable SNACK - SAI */

                    if (strncmp(ptr_to_keytext, "d=", 2) == 0)
                    {
                        if ( (strcmp(ptr_to_keytext+2, "y") == 0)  || strcmp(ptr_to_keytext+2, "Y") == 0)
                        {
                            devdata->targ_snack_flg |= DATA_SNACK_ENABLE;
                        }
                    }
                    if (strncmp(ptr_to_keytext, "s=", 2) == 0)
                    {
                        if ( (strcmp(ptr_to_keytext+2, "y") == 0)  || strcmp(ptr_to_keytext+2, "Y") == 0)
                        {
                            devdata->targ_snack_flg |= STATUS_SNACK_ENABLE;
                        }
                    }
                    if (strncmp(ptr_to_keytext, "h=", 2) == 0)
                    {
                        if (simple_strtoul(ptr_to_keytext+2, NULL, 0) == 1)
                            devdata->targ_snack_flg |= DATA_SNACK_RETRANSMIT;
                        else if (simple_strtoul(ptr_to_keytext+2, NULL, 0) == 2)
                            devdata->targ_snack_flg |= DATA_SNACK_REJECT;

                        TRACE(TRACE_DEBUG, "SNACK Flags : datasnack %d statsnack %d DataSnackRexmit %d  DataSnackReject %d\n",
                                    (devdata->targ_snack_flg & DATA_SNACK_ENABLE),
                                    ( (devdata->targ_snack_flg & STATUS_SNACK_ENABLE) >> 1),
                                    ( (devdata->targ_snack_flg & DATA_SNACK_RETRANSMIT) >> 7),
                                    ( (devdata->targ_snack_flg & DATA_SNACK_REJECT) >> 6) );
                    }

                }
				else
				{
					configure_parameter(param_neg_info, ptr_to_keytext, 
						*devdata->param_tbl);
				}
			}
			else
				result = -EINVAL;
		}
	}
	else
	{
		result =  sprintf( buffer, "The iSCSI Target front end\n");
		result += sprintf( buffer+result, "Number of active sessions: %d\n", 
			devdata ? devdata->nsessions : 0 );		
	}

	TRACE( TRACE_DEBUG, "Leaving iscsi_proc_info\n");

	return result;
}


/*      Note the name "driver_template" is required for "scsi_module.c" to work! */
/*      Used by: "/usr/src/linux/drivers/scsi/scsi_module.c" */
/*      Wrap this in an ifdef because "scsi_module.c" says to! */
#ifdef MODULE

MODULE_AUTHOR("UNH IOL ISCSI");
MODULE_DESCRIPTION("UNH-IOL iSCSI reference target");
MODULE_SUPPORTED_DEVICE("sd");
#ifdef MODULE_LICENSE
        MODULE_LICENSE("GPL");
#endif

#endif

