#define	SNACK_TYPE				0x0f
#define DATA_R2T_SNACK			0x00
#define	STATUS_SNACK			0x01
#define	DATACK_SNACK			0x02

#define SESSION_RECOVERY		0
#define	DIGEST_RECOVERY			1
#define	CONNECTION_RECOVERY		2

#define DATA_SNACK_ENABLE		0x01
#define STATUS_SNACK_ENABLE		0x02
#define DATA_SNACK_RETRANSMIT		0x80
#define DATA_SNACK_REJECT		0x40

#define MAX_MALLOC_SIZE			1024

struct semaphore	targ_sess_recovery_sem;

struct targ_error_rec
	{
	struct iscsi_conn	*curr_conn;
	struct generic_pdu	*pdu_hdr;
	struct iscsi_cmnd	*cmd;
	__u8			err_type;
	};

/* This data structure is used for maintaining R2T cookies and also
   for queueing out-of-order dataouts received */
struct iscsi_cookie
	{
	__u32					seq;
	int						xfer_len;
	int						offset;
	int						list_offset;
	int						list_count;
	struct iscsi_cookie 	*next;
	};

void targ_session_recovery(struct iscsi_conn *current_connection);
int targ_drop_pdu_data(struct iscsi_conn *curr_conn, int size);
void targ_recovery_thread(void *arg);
int targ_digest_recovery(struct targ_error_rec *err_rec);
int send_recovery_r2t (struct iscsi_cmnd *cmnd, int data_offset, int data_length, struct iscsi_cookie*);
int queue_data(struct targ_error_rec *err_rec);

/************************************************************************
*  The Main Error Recovery Handler Called from iscsi_initiator when it  *
*  finds any error.                                                     *
************************************************************************/

void targ_do_error_recovery(struct targ_error_rec *err_rec)
	{
	
	int retval=1;
	struct iscsi_conn	*curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_do_error_recovery\n");

	if (err_rec == NULL)
		{
		TRACE_ERROR("Error Recovery structure is NULL\n");
		return;
		}

	if (err_rec->curr_conn == NULL)
		{
		TRACE_ERROR("Connection Pointer is NULL in Error Recovery structure\n");
		return;
		}

	curr_conn = err_rec->curr_conn;

	switch(curr_conn->session->oper_param->ErrorRecoveryLevel)
		{
		case SESSION_RECOVERY:
				init_MUTEX_LOCKED(&targ_sess_recovery_sem);
				targ_session_recovery(curr_conn);
				break;
		case DIGEST_RECOVERY:
				retval = targ_digest_recovery(err_rec);
				break;
		case CONNECTION_RECOVERY:
				TRACE_ERROR("FUNCTION Not Implemented yet\n");
				break;
		}

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_do_error_recovery\n");

	return;
	}

/************************************************************************
* This function handles the session recovery. It creates a recovery     *
* thread that will do session recovery.                                 *
************************************************************************/

void targ_session_recovery(struct iscsi_conn *current_connection)
{
	TRACE(TRACE_ISCSI, "Executing Target Session Recovery - cancelling Receive Thread\n");

	/* Create a recovery thread */
	kernel_thread((int (*)(void *))targ_recovery_thread, (void *)current_connection, 0);

	/* Wait till the recovery thread completes processing */
	down_interruptible(&targ_sess_recovery_sem);

	/* Release the attained recovery lock */
	up(&targ_sess_recovery_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_session_recovery\n");

	return;
}

/************************************************************************
* This thread handles the session recovery. Session Recovery will       *
* abort all pending tasks and closes all transport connection           *
* Section (6.12.4)                                                      *
************************************************************************/

void targ_recovery_thread(void *arg)
	{
	struct iscsi_conn	*current_connection;
	struct iscsi_session	*current_session;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_recovery_thread\n");

	current_connection = (struct iscsi_conn *)arg;
	current_session = current_connection->session;

	TRACE(TRACE_ISCSI, "target_recovery_thread started\n");

	iscsi_release_session(current_session);

	up(&targ_sess_recovery_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_recovery_thread\n");

	return;
	}

/************************************************************************
* This function handles the digest errors and sequence errors that may  *
* have caused because of previous digest errors.                        *
************************************************************************/

int targ_digest_recovery(struct targ_error_rec *err_rec)
	{
	int			retval=0, data_length=0;
	int			size=0, err_type=0, data_offset=0;
	unsigned int		opcode=0;
	struct iscsi_conn	*curr_conn;
	struct iscsi_init_scsi_data_out	*data_out_hdr;
	struct iscsi_cmnd	*cmd;

	curr_conn = err_rec->curr_conn;
	err_type = err_rec->err_type;
	cmd = err_rec->cmd;
	
	TRACE(TRACE_ENTER_LEAVE, "Entering targ_digest_recovery\n");

	if (err_rec->pdu_hdr == NULL)
		{
		TRACE_ERROR("iscsi PDU  Header is NULL\n");
		retval= -1;
		goto digest_out;
		}
	else
		{
		opcode = err_rec->pdu_hdr->opcode;
		size = err_rec->pdu_hdr->length;
		if (size > 0xffffff)
			{
			TRACE_ERROR("TotalAHSLength 0x%02x, expected 0x00\n",size >> 24);
			retval = -1;
			goto digest_out;
			}
		}

	if (err_type == HEADER_DIGERR)
		{
		if (size > 0)
			{
				retval = targ_drop_pdu_data(curr_conn, size);
			}
		goto digest_out;
		}

	if (opcode == ISCSI_INIT_SCSI_DATA_OUT)
		{
		if (cmd == NULL)
			{
			/* TRACE_ERROR("Related SCSI Command is NULL\n); */
			retval = -1;
			goto digest_out;
			}

		if (err_type == PAYLOAD_DIGERR)
			{
			/* Roll back received length to original value */
			cmd->data_done -= size;

			data_length = cmd->data_length - cmd->data_done;
			data_offset = cmd->data_length - data_length;

			iscsi_tx_rjt(curr_conn, REASON_DATA_DIGEST_ERR);

			send_recovery_r2t(cmd, data_offset, size, NULL);
			}
		else if (err_type == SEQUENCE_ERR)
			{
			/* error recovery ver ref18_04 - SAI*/
			/* if there is a sequence error and the arrived data seq. indicates
			   lost packets, queue the received packets and request re-transmission 
			   of lost packets */
			queue_data(err_rec);

			data_length = cmd->data_length - cmd->data_done;
			data_offset = cmd->data_length - data_length;
	
			data_out_hdr = (struct iscsi_init_scsi_data_out*)err_rec->pdu_hdr;
			size = (data_out_hdr->offset - data_offset);

			send_recovery_r2t(cmd, data_offset, size, NULL);

			up(&curr_conn->tx_sem);

			}

		}

digest_out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_digest_recovery\n");

	return retval;	
	}

/*******************************************************************
 * Reads the data portion of the dropped PDU due to digest errors. *
 * The data read is also ignored.                                  *
********************************************************************/

int targ_drop_pdu_data(struct iscsi_conn *curr_conn, int size)
	{
	int padding=0, retval=0, data_length=0;
	int total_rx=0, rx_loop=0;
	char		*buffer=NULL;
	struct msghdr	msg;
	struct iovec	iov;
	mm_segment_t oldfs;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_drop_pdu_data\n");

	padding = -(size)&3;
	size += padding;
 
	if (curr_conn->data_crc == 1)
		{
		size += CRC_LEN;
		}

	TRACE(TRACE_DEBUG, "targ_drop_pdu: data_read %d total_data %d\n", data_length, size);

	data_length = size;

	while(data_length > 0)
		{
		if (data_length > MAX_MALLOC_SIZE)
			data_length = MAX_MALLOC_SIZE;

		buffer = kmalloc(data_length*sizeof(char), GFP_KERNEL);
		if (!buffer)
			{
			TRACE_ERROR("Malloc Error : Unable to drop the data PDU\n");
			retval = -ENOMEM;
			goto end_drop_pdu;
			}
	
		iov.iov_base = buffer;	
		iov.iov_len  = data_length;
		msg.msg_iov  = &iov;
		msg.msg_iovlen = 1;

		total_rx=rx_loop=0;
		while(total_rx < data_length)
			{
			oldfs = get_fs();
			set_fs(get_ds());
			rx_loop = sock_recvmsg(curr_conn->conn_socket,&msg, (data_length-total_rx),0);

			set_fs(oldfs);

			if (rx_loop <= 0)
				{
				TRACE_ERROR("trag_drop_pdu: rx_loop %d total_rx %d\n", rx_loop, total_rx);
				break;
				}
			total_rx += rx_loop;

			TRACE(TRACE_DEBUG, "targ_drop_pdu: rx_loop %d total_rx %d\n", rx_loop, total_rx);
			}
	
		size -=data_length;
		data_length = size;

		TRACE(TRACE_DEBUG, "targ_drop_pdu: data_read %d total_data %d\n", data_length, size);

		if (buffer)
			kfree(buffer);
		}

	end_drop_pdu:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_drop_pdu_data\n");

		return retval;	
	}

void deal_with_r2t_timer(unsigned long data);

/*********************************************************************
 * (re)start R2T Timer is used for tracking the R2T sent. This timer
 * helps re-transmit the most recent R2T sent to the initiator.
 ********************************************************************/
/* (re)start the R2T retransmit timer */
void restart_r2t_timer(struct iscsi_conn *conn, unsigned int period)
	{
	TRACE(TRACE_ENTER_LEAVE, "Enter restart_r2t_timer, period %u\n", period);

	if (period > 0 && conn->r2t_timer != NULL && !timer_pending(conn->r2t_timer))
		{
		TRACE(TRACE_DEBUG, "Start r2t timer for %u ticks \n", period);
		conn->r2t_timer->expires = jiffies + period;
		conn->r2t_timer->data = (unsigned long)conn;
		conn->r2t_timer->function = deal_with_r2t_timer;
		add_timer(conn->r2t_timer);
		}
	else
		{
		TRACE_ERROR("R2T timer not started for %u ticks\n", period);
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave restart_r2t_timer\n");
	}

/*********************************************************************
* Called by kernel when R2T periodic session timer expires.
*********************************************************************/
void deal_with_r2t_timer(unsigned long data)
	{
	struct iscsi_conn *this_conn;
	struct iscsi_cmnd *cmnd=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_r2t_timer\n");

	TRACE(TRACE_DEBUG,"R2T retransmit timer expired \n");

	this_conn = (struct iscsi_conn *)data;

	for (cmnd=this_conn->session->cmnd_list; cmnd != NULL; cmnd=cmnd->next)
		{
		if (cmnd->outstanding_r2t > 0 &&
			(jiffies > (cmnd->timestamp + this_conn->dev->r2t_period*HZ)) )
			{
			up(&this_conn->session->retran_sem);
			}
		}

	restart_r2t_timer(this_conn, this_conn->dev->r2t_period*HZ);

	TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_r2t_timer\n");

	}

/***************************************************************************
* create an R2T cookie and store the R2T details in the cookie. This cookie
* is esssential for R2T re-transmssions in error conditions.
***************************************************************************/
struct iscsi_cookie* create_r2t_cookie(struct iscsi_cmnd *cmnd)
{
    struct iscsi_cookie *cookie=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter create_r2t_cookie\n");

    cookie = (struct iscsi_cookie*)kmalloc(sizeof(struct iscsi_cookie), GFP_KERNEL);
    if (cookie == NULL)
    {
        TRACE_ERROR("Malloc problem in creating R2T cookie\n");
        goto end_iscsi_cookie;
    }
    cookie->next = NULL;

    if (cmnd->first_r2t_cookie == NULL)
    {
        cmnd->first_r2t_cookie = cmnd->last_r2t_cookie = cookie;
    }
    else
    {
        cmnd->last_r2t_cookie->next = cookie;
        cmnd->last_r2t_cookie = cookie;
    }

end_iscsi_cookie:

	TRACE(TRACE_ENTER_LEAVE, "Leave create_r2t_cookie\n");

    return cookie;
}

/* free the long pending R2T cookie list after the command is completed - SAI */
void free_r2t_cookie(struct iscsi_cmnd *cmnd)
{
    struct iscsi_cookie *cookie=cmnd->first_r2t_cookie;
    struct iscsi_cookie *tmp_cookie=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter free_r2t_cookie\n");

    while(cookie)
    {
        tmp_cookie = cookie;
        cookie = cookie->next;
        kfree(tmp_cookie);
    }

	TRACE(TRACE_ENTER_LEAVE, "Leave free_r2t_cookie\n");

	return;

}

/**************************************************************************
* Re-Transmits the R2T indicated by the r2tsn. This function is invoked by
* SNACK requests and also by local R2T timeouts.
**************************************************************************/
int retran_r2t (struct iscsi_cmnd *cmnd, int r2tsn)
{
    int retval=-1;  /* pessimistic assumption that the R2T will not be found */
    struct iscsi_cookie *cookie=cmnd->first_r2t_cookie;

	TRACE(TRACE_ENTER_LEAVE, "Enter retran_r2t\n");

    while(cookie)
    {
        if (cookie->seq == r2tsn)
        {
            send_recovery_r2t(cmnd, 0, 0, cookie);
            retval=1;
            break;
        }
        cookie=cookie->next;
    }

    if (retval==-1)
        TRACE_ERROR("Unidentified R2T seq. no %d retransmission request received \n", r2tsn);

	TRACE(TRACE_ENTER_LEAVE, "Leave retran_r2t\n");

    return retval;
}

/*******************************************************************************
* This function uses the offset and data length parameter or cookie information
* to build a recovery R2T to be sent to the intiator for missing Dataouts. This
* function is called by SNACK mechanism and also by other places of the code that
* finds a discrepency in Dataouts using data sequence numbering or offsets.
********************************************************************************/
int send_recovery_r2t (struct iscsi_cmnd *cmnd, int data_offset, int datalen, struct iscsi_cookie *cookie)
{
    __u8            iscsi_hdr[ISCSI_HDR_LEN];
    int         err, niov, hdr_crc_len=0, data_length=0;
    struct iscsi_targ_r2t   *hdr;
    struct iovec        *iov;
    struct msghdr       msg;
    int         max_burst_len=0;

	TRACE(TRACE_ENTER_LEAVE, "Enter send_recovery_r2t\n");

	if (cookie)
		data_length = cookie->xfer_len;
	else
		data_length = datalen;
	
	while (data_length > 0)
	{
        niov = 1;

        if (cmnd->conn->hdr_crc == 1)
        {
            hdr_crc_len = CRC_LEN;
            niov++;
        }

        iov = (struct iovec*) kmalloc(niov*sizeof (struct iovec),GFP_KERNEL);

        /*
         * since the entire space will have been malloced by the STML
         * we need to send just one R2T to the Initiator.
         */
        memset (iscsi_hdr, 0x0, ISCSI_HDR_LEN);
        hdr = (struct iscsi_targ_r2t *) iscsi_hdr;

        hdr->opcode     = ISCSI_TARG_R2T;

        // only for version 8
        if( cmnd->conn->session->draft_number <= DRAFT8 )
            hdr->opcode     |= F_BIT;
        if( cmnd->conn->session->draft_number <= DRAFT9 )
            hdr->opcode |= I_BIT;

        hdr->flags      |= F_BIT;
        hdr->init_task_tag  = cpu_to_be32 (cmnd->init_task_tag);
        hdr->target_xfer_tag= cpu_to_be32 (cmnd->target_xfer_tag);
        hdr->stat_sn    = cpu_to_be32 (cmnd->conn->stat_sn);
        hdr->exp_cmd_sn = cpu_to_be32 (cmnd->session->exp_cmd_sn);
        hdr->max_cmd_sn = cpu_to_be32 (cmnd->session->max_cmd_sn);
		if (!cookie)
		{
        	hdr->r2t_sn     = cpu_to_be32 (cmnd->r2t_sn++);
        	hdr->offset     = cpu_to_be32 (data_offset);

        	if( cmnd->conn->session->draft_number <= DRAFT13 )
	            max_burst_len = cmnd->conn->session->oper_param->MaxBurstSize;
			else if( cmnd->conn->session->draft_number >= DRAFT18 )
				max_burst_len = cmnd->conn->session->oper_param->MaxBurstLength;

			cookie = create_r2t_cookie(cmnd);
			if (cookie)
			{
				cookie->seq = cmnd->r2t_sn-1;
				cookie->offset = data_offset;
				cookie->xfer_len = (data_length<=max_burst_len)?data_length:max_burst_len;
				TRACE(TRACE_DEBUG, "send_recovery_r2t: R2TSN %d offset %d xfer_len %d\n", cookie->seq, cookie->offset, cookie->xfer_len);
			}

	        if( data_length <= max_burst_len )
			{
				hdr->xfer_len = cpu_to_be32 (data_length);
				data_length = 0;
			}
			else
			{
				hdr->xfer_len = cpu_to_be32 (max_burst_len);
				data_offset += be32_to_cpu(hdr->xfer_len);
				data_length -= max_burst_len;
			}
		}
		else
		{
        	hdr->r2t_sn     = cpu_to_be32 (cookie->seq);
        	hdr->offset     = cpu_to_be32 (cookie->offset);
			hdr->xfer_len   = cpu_to_be32 (cookie->xfer_len);
			data_length = 0;
		}

        iov[0].iov_base = iscsi_hdr;
        iov[0].iov_len  = ISCSI_HDR_LEN;

        fill_msg_hdr (&msg);
        msg.msg_iov = iov;
        msg.msg_iovlen  = niov;

        err = tx_data (cmnd->conn, &msg, ISCSI_HDR_LEN + hdr_crc_len);

        if (err != ISCSI_HDR_LEN + hdr_crc_len)
        {
            TRACE_ERROR("iscsi_tx_r2t: tx_data returned < ISCSI_HDR_LEN\n");
            return -1;
        }

        if (TRACE_TEST(TRACE_ISCSI_FULL))
        {
            int i=0;
            for (i=0; i<ISCSI_HDR_LEN; i++)
            {
                printk("%02x ", iscsi_hdr[i]);
                if (i==24) printk("\n");
            }
            printk("\n");
        }

        TRACE(TRACE_ISCSI, "iscsi_tx_%d: r2t command sent\n",
            cmnd->conn->conn_id);

        if (TRACE_TEST(TRACE_ISCSI_FULL))
            print_targ_r2t(hdr);

    }

    cmnd->state = ISCSI_R2T_SENT;

    /* store the r2t time stamp - SAI */
    cmnd->timestamp = jiffies;

	TRACE(TRACE_ENTER_LEAVE, "Leave send_recovery_r2t\n");

    return (0);
}

/***********************************************************************
* This thread remains blocked at a semaphore for most part of it's time
* When a r2t timeout happens thsi thread is awakened for re-transmitting
* the r2t request.
***********************************************************************/

void iscsi_retran_thread (void *param)
{
	struct iscsi_session *current_session = (struct iscsi_session *)param;
	struct iscsi_cmnd *cmnd=NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_retran_thread\n");

	lock_kernel();

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGS);
	daemonize();
	sprintf (current->comm, "iscsi_tran_%d", current_session->devdata->ntsih);
	current_session->retran_thread = current;
	unlock_kernel();

	while(1)
	{
		down_interruptible(&current_session->retran_sem);
	
		if (signal_pending(current))
			break;

		for(cmnd=current_session->cmnd_list; cmnd != NULL; cmnd=cmnd->next)
		{
			if (cmnd->outstanding_r2t > 0 &&
				(jiffies > (cmnd->timestamp + current_session->devdata->r2t_period*HZ)) )
			{
				TRACE(TRACE_ISCSI, "Re-Transmitting most recent R2T\n");
				retran_r2t(cmnd, cmnd->r2t_sn-1);
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_retran_thread\n");

	current_session->retran_thread=NULL;
	up(&current_session->thr_kill_sem);
	
	return;
}

void add_data_to_queue(struct iscsi_cmnd *cmd, struct iscsi_cookie *dataq)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering add_data_to_queue\n");

	if (!cmd->first_data_q)
	{
		cmd->first_data_q = dataq;
		cmd->last_data_q  = dataq;
	}
	else
	{
		cmd->last_data_q->next = dataq;
		cmd->last_data_q = cmd->last_data_q->next;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving add_data_to_queue\n");

	return;
}

/********************************************************************
* This function drives the out-of-order dataouts received into the 
* appropriate scatter gather buffer location maintained by scsi. The
* information about the out-of-order dataout and also the corresponding
* scatter offsets are queued for future use.
**********************************************************************/
int queue_data(struct targ_error_rec *err_rec)
{
	int offset=0, size=0, list_offset=0;
	int count=0;
	int niov, padding, err, retval=0, data_crc_len=0;
	__u32 data_sn, digest, pad_bytes;

	struct iscsi_conn	*conn=NULL;
	struct iscsi_cmnd *cmd;
	struct iscsi_init_scsi_data_out	*hdr;
	struct iscsi_cookie *dataq=NULL;
	struct scatterlist	*st_list;
	struct iovec		*iov=NULL;
	struct msghdr		msg;

	TRACE(TRACE_ENTER_LEAVE, "Entering queue_data\n");

	cmd = err_rec->cmd;	
	hdr = (struct iscsi_init_scsi_data_out *)err_rec->pdu_hdr;

	list_offset = offset = hdr->offset;
	data_sn = hdr->data_sn;
	size = hdr->length;
	conn = err_rec->curr_conn;


	st_list = (struct scatterlist *)cmd->cmnd->req->sr_buffer;	

	while ( st_list[count].length <= list_offset)
	{
		list_offset -= st_list[count].length;
		count ++;
	}

	st_list += count;
	niov = find_iovec_needed(size, st_list, list_offset);

	if (niov < 0)
	{
		TRACE_ERROR("queue data: find_iovec_needed returned an error\n");
		return -1;
	}

	padding = -(size)&3;
	if (padding)
		niov++;

	if (conn->data_crc == 1)
		niov++;

	iov = (struct iovec*)kmalloc(niov*sizeof(struct iovec), GFP_KERNEL);

	if (!iov)
	{
		TRACE_ERROR("queue_data: Could not malloc iov\n");
		return -1;
	}

	TRACE(TRACE_DEBUG, "queue_data: no. of iovec needed %d\n", niov);

	retval = fill_iovec(iov, niov, st_list, &list_offset, size);

	if (padding)
	{
		iov[niov-1 - conn->data_crc].iov_base = &pad_bytes;
		iov[niov-1 - conn->data_crc].iov_len = padding;
	}

	if (conn->data_crc == 1)
	{
		data_crc_len = CRC_LEN;
		iov[niov-1].iov_base = &digest;
		iov[niov-1].iov_len = CRC_LEN;
	}

	fill_msg_hdr(&msg);
	msg.msg_iov = iov;
	msg.msg_iovlen = niov;

	err = rx_data(cmd->conn, &msg, size + padding + data_crc_len);

	if (err != (size + padding + data_crc_len))
	{
		TRACE_ERROR("queue_data: Trouble in rx_data\n");
		if (err == PAYLOAD_DIGERR)
			return err;
			
	}

	dataq = kmalloc(sizeof(struct iscsi_cookie), GFP_KERNEL);

	if (!dataq)
	{
		TRACE_ERROR("add_data_to_queue: Malloc returned an error\n");
		return -1;
	}
	
	dataq->offset = offset;
	dataq->xfer_len = size;
	dataq->list_offset = list_offset;
	dataq->list_count = retval;
	dataq->next = NULL;

	add_data_to_queue(cmd, dataq);

	TRACE(TRACE_ENTER_LEAVE, "Leaving queue_data\n");

	return 0;
}

/***********************************************************************
* This function is called everytime a in-sequence dataout pdu is
* received. The function verifies if there was any previous
* out-of-order dataouts for the command and updates the state variables
* accordingly.
***********************************************************************/
void search_data_q(struct iscsi_cmnd *cmd)
{
	int unhook_flg=0;
	struct iscsi_cookie	*dataq, *prev, *tmp_q;

	TRACE(TRACE_ENTER_LEAVE, "Entering search_data_q\n");

	if (!cmd->first_data_q)
		return;

	dataq = prev = cmd->first_data_q;

	while(dataq)
	{
		if (cmd->data_done > dataq->offset)
			unhook_flg=1;
		else if (cmd->data_done == dataq->offset)
		{
			cmd->scatter_list_count += dataq->list_count;
			cmd->scatter_list_offset = dataq->list_offset;
			cmd->data_done += dataq->xfer_len;
			unhook_flg=1;
		}

		if (unhook_flg)
		{
			/* unhook data node */
			if (prev == cmd->first_data_q)
			{
				cmd->first_data_q = dataq->next;
				prev = dataq->next;
			}
			else if (dataq == cmd->last_data_q)
				cmd->last_data_q = prev;
			else
				prev->next = dataq->next;

			tmp_q = dataq;
			dataq = dataq->next;
			kfree(tmp_q);
			unhook_flg=0;
		}
		else
		{
			prev = dataq;
			dataq = dataq->next;
		}

		if (cmd->data_done >= cmd->data_length)
			break;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving search_data_q\n");
}
