#define	SNACK_TYPE				0x0f
#define DATA_R2T_SNACK			0x00
#define	STATUS_SNACK			0x01
#define	DATACK_SNACK			0x02

#define SESSION_RECOVERY		0
#define	DIGEST_RECOVERY			1
#define	CONNECTION_RECOVERY		2

#define DATA_SNACK_ENABLE		0x01
#define STATUS_SNACK_ENABLE		0x02

#define STATUS_BIT				0x01
#define SNACK_RJT_BIT			0x02
#define STATUS_CONNREC			0x4
#define STATUS_LOGGED_OUT		0x5

struct semaphore	init_sess_recovery_sem;
struct task_struct	*init_rec_thread=NULL;

void init_session_recovery(struct init_error_rec *err_rec);
void init_recovery_thread(void *arg);
int init_digest_recovery(struct init_error_rec *err_rec);
void init_send_snack( struct init_error_rec *err_rec, int begrun, int runlen, __u8 type);
void print_init_snack(struct iscsi_init_snack *cmd);
void print_begrun(unsigned int begrun);
void print_runlen(unsigned int runlen);
void restore_global_data(struct iscsi_hostdata *global_hostdata);

/************************************************************************
*  The Main Error Recovery Handler Called from iscsi_initiator when it  *
*  finds any error.                                                     *
************************************************************************/

void init_do_error_recovery(struct init_error_rec *err_rec)
	{
	
	int retval=1;
	struct connection	*curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_do_error_recovery \n");

	if (err_rec == NULL)
		{
		TRACE_ERROR("Error Recovery structure is NULL\n");
		return;
		}

	if (err_rec->curr_conn == NULL)
		{
		TRACE_ERROR("Connection Pointer is NULL in Error Recovery structure\n");
		return;
		}

	curr_conn = err_rec->curr_conn;

	switch(curr_conn->my_session->oper_param->ErrorRecoveryLevel)
		{
		case SESSION_RECOVERY:
				init_MUTEX_LOCKED(&init_sess_recovery_sem);
				init_session_recovery(err_rec);
				break;
		case DIGEST_RECOVERY:
		case CONNECTION_RECOVERY:
				retval = init_digest_recovery(err_rec);
				break;
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave init_do_error_recovery \n");

	return;
	}

/************************************************************************
* This function handles the session recovery. It creates a recovery     *
* thread that will do session recovery.                                 *
************************************************************************/

void init_session_recovery(struct init_error_rec *err_rec)
{
	struct connection	*current_connection;

	current_connection = err_rec->curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_session_recovery \n");

	/* Release any resources held-up */
	spin_unlock(&my_lock);

	/* Create a recovery thread */
	kernel_thread((int (*)(void *))init_recovery_thread, (void *)err_rec, 0);

	TRACE(TRACE_ISCSI, "Executing Init Session Recovery - cancelling Receive Thread\n");

	/* Wait till the recovery thread completes processing */
	down_interruptible(&init_sess_recovery_sem);

	/* Release the attained recovery lock */
	up(&init_sess_recovery_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave init_session_recovery \n");

	return;
}

/* Create a new recovery session for the failing session before cleaning-up
   the failing session */
struct session*	get_rec_sess(__u32 target_id, __u32 lun)
	{
	int 			retval=0;
	struct connection	*conn;
	struct session	*prev_session, *new_session = global_hostdata->rec_sess_head;

	TRACE(TRACE_ENTER_LEAVE, "Enter get_rec_sess \n");

	prev_session = new_session;

	for (; new_session!= NULL; new_session = new_session->next)
		{
		if (new_session->scsi_target_id == target_id)
			break;
		prev_session = new_session;
		}

	if (new_session == NULL)
		{
		TRACE(TRACE_DEBUG, "Previous Recovery list empty - creating new list\n");
		new_session = (struct session *)my_kmalloc(sizeof(struct session), "rec_session");
		if (new_session == NULL)
			{
			TRACE_ERROR("No Memory during malloc\n");
			}
		else
			{
			memset(new_session, 0, sizeof(struct session));
			/*  initialize most of the new struct session fields here */
			new_session->scsi_target_id = target_id;
			new_session->lun_bits = 1 << lun;
			new_session->session_state = STATUS_NOT_PRESENT;
			new_session->cur_task_tag = 88888;      /* something unique for testing */
			new_session->cur_cmd_sn = 22222;        /* something unique for testing */
			new_session->max_cmd_sn = new_session->exp_cmd_sn =
													new_session->cur_cmd_sn + 1;
			init_MUTEX_LOCKED(&new_session->tx_sem);
			init_MUTEX_LOCKED(&new_session->tx_done_sem);
			init_MUTEX_LOCKED(&new_session->task_mgt_sem);

			/*  use the configured values to initialize following fields */
			new_session->session_flags = global_hostdata->init_sess_flags;
			new_session->init_conn_flags = global_hostdata->init_conn_flags;
			new_session->nop_period = global_hostdata->nop_period;
			new_session->draft_number = global_hostdata->draft_number;

			TRACE(TRACE_DEBUG, "session->draft_number %u\n", global_hostdata->draft_number);

			/* Added data and status snack support - SAI */
			new_session->init_snack_flg = global_hostdata->init_snack_flg;

			/* Added the retranmit check interval and wait period - SAI */
			new_session->xmit_chk_interval = global_hostdata->xmit_chk_interval;
			new_session->retran_period = global_hostdata->retran_period;

			/* Store the connection scheduling scheme - SAI */
			new_session->sched_scheme = global_hostdata->sched_scheme;
			/* Initialize the connection recovery related fields - SAI */
			new_session->n_rec_tasks = 0;


			/* allocate space for configured parameters to use during session login */
			if( (new_session->session_params = my_kmalloc(
								MAX_CONFIG_PARAMS*sizeof(struct parameter_type),
										"rec_session param_tbl")) == NULL )
				{/* no memory for session parameter table */
				retval = -ENOMEM;
				goto out1;
				}

			/* init session parameters with copy of the configured parameter table */
			param_tbl_cpy(*new_session->session_params, *global_hostdata->param_tbl);

			/*  finish allocating remaining structures new struct session points to */
			new_session->oper_param = (struct session_operational_parameters *)
           				     my_kmalloc(sizeof(struct session_operational_parameters),
			                "rec_session oper_param");
			if( new_session->oper_param == NULL )
				{/* no memory for operational parameters structure */
				retval = -ENOMEM;
				goto out2;
				}

			set_session_parameters(new_session->oper_param, *new_session->session_params, new_session->draft_number);

			/* create the first connection on this session */
			new_session->nconnections = 1;
			new_session->connection_head = conn = (struct connection *)
									my_kmalloc(sizeof(struct connection), "rec_connection");
			if( conn == NULL )
				{/* no memory for new connection structure */
				retval = -ENOMEM;
				goto out3;
				}

			/*  reset all fields in new struct connection (although few remain 0) */
			memset(conn, 0, sizeof(struct connection));

			conn->max_send_length = 8192;/* target's default MaxRecvDataSegmentLength */
			conn->max_recv_length = 8192;/* initiator's MaxRecvDataSegmentLength too */
			conn->nop_counter = 1;
			conn->my_session = new_session;
			conn->connection_flags = new_session->init_conn_flags;
			conn->connection_state = STATUS_NOT_PRESENT;
			init_MUTEX_LOCKED(&conn->rx_done_sem);

			if (global_hostdata->rec_sess_head == NULL)
				global_hostdata->rec_sess_head = new_session;
			else
				prev_session->next = new_session;
	
			}
		}

	TRACE(TRACE_DEBUG, "Recovery session %p returned\n", new_session);
	TRACE(TRACE_ENTER_LEAVE, "Leave get_rec_sess \n");

	return new_session;

out3:
	my_kfree((void **)&new_session->oper_param, "rec_session oper_param");
out2:
    /* here to free structure for configured parameters used in login */
    param_tbl_uncpy(*new_session->session_params);
    my_kfree((void **)&new_session->session_params, "rec_session param_tbl");

out1:
    /* here to free new session structure */
    my_kfree((void **)&new_session, "rec_session");

	return NULL;
	}

void clean_rec_sess(struct session *new_session)
	{
	struct session			*sess=global_hostdata->rec_sess_head;
	struct session			*prev=sess;

	TRACE(TRACE_ENTER_LEAVE, "Enter clean_rec_sess \n");

	for(; sess != NULL; sess = sess->next)
		{
		if (sess->scsi_target_id ==	new_session->scsi_target_id)
			{
			if (sess == global_hostdata->rec_sess_head)
				global_hostdata->rec_sess_head = NULL;
			else
				prev->next = sess->next;

			break;
			}

		prev = sess;
		}	

	TRACE(TRACE_ENTER_LEAVE, "Leave clean_rec_sess \n");

	}

/************************************************************************
* This thread handles the session recovery. Session Recovery will       *
* abort all pending tasks and closes all transport connection           *
* Section (6.12.4)                                                      *
************************************************************************/

void init_recovery_thread(void *arg)
	{
	int			ip, port, target, lun, cid;
	int			received=0, retval=0, size=0, opcode=0;
	u_int			checksum=0;
	char			*buffer=NULL;
	char			*over_flow_buffer=NULL;
	unsigned int	redirect_ip=0, redirect_port=0;
	struct response_pdu	*header=NULL;
	struct init_error_rec	*err_rec;
	struct connection	*current_connection, *conn;
	struct session		*current_session, *new_session;
	struct command		*cmd;

	TRACE(TRACE_ENTER_LEAVE, "Enter init_recovery_thread \n");

	/* Lock kernel to daemonize the thread and update init_rec_thread pointer */

	lock_kernel();
	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGS);
	daemonize();
	strcpy(current->comm, "iscsi_init_rec_thread");
	init_rec_thread = current;
	unlock_kernel();

	err_rec = (struct init_error_rec *)arg;
	current_connection = err_rec->curr_conn;
	current_session = current_connection->my_session;


	target = current_session->scsi_target_id;
	ip = current_connection->target_address;
	port = current_connection->target_port;
	cid = current_connection->connection_id;

	/* Abort all queued scsi commands with an appropriate scsi service response Section - 6.12.4 */

	drive_logout(current_session, 0, 0);

	if (current_connection->rx_buf == NULL)
		{
		TRACE_ERROR("scsi Response Header is NULL\n");
		goto rec_thread_out;
		}
	else
		{
		header = (struct response_pdu *)current_connection->rx_buf;
		opcode = header->opcode;
		size = ntohl(header->length);
		}

	/* if (size > 0 && err_rec->err_type != PAYLOAD_DIGERR) */
	if (size > 0)
		{
		if ( (buffer = my_kmalloc(size, "Digest Error Data")) != NULL)
			{
	 		retval = recv_a_message(current_connection, size, buffer);

			retval = recv_locally(size, 0, current_connection, 0, &checksum, &over_flow_buffer);
			my_kfree((void **)&over_flow_buffer, "local buffer");
			my_kfree((void **)&buffer, "Digest Error Data");
			}
		}

        while( 1 )
		{
                size = ISCSI_HDR_LEN;
                if( current_connection->connection_flags & USE_HEADERDIGEST )
                        size += CRC_LEN;

                if( (received = recv_a_message(current_connection, size,
                                                                                        current_connection->rx_buf)) <= 0 )
                        {/* connection died or we got a kill signal, terminate this thread */
                        break;
                        }

                header = (struct response_pdu *)current_connection->rx_buf;

                size = ntohl(header->length);

                if( size > 0xffffff )           /* 2^24 - 1 */
                        {
                        TRACE_ERROR("TotalAHSLength 0x%02x, expected 0x00\n", size>>24);
                        break;
                        }

                /* mark that we got some activity from the target */
                current_connection->connection_flags |= GOT_ACTIVITY;

        	/* Print the received iSCSI header */
                PRINT_BUFF(current_connection->rx_buf, ISCSI_HDR_LEN);

                if( current_session->session_flags & NEED_TX_WAKEUP )
                        {/* turn off the need tx thread wakeup bit */
                        current_session->session_flags &= ~NEED_TX_WAKEUP;
                        /* then get the tx thread started again */
                        up(&current_session->tx_sem);
                        }
	
		opcode = header->opcode;

		if (opcode == ISCSI_TARG_LOGOUT_RSP)
			{
			TRACE(TRACE_DEBUG, "Got LogOut Response\n");
			break;
			}

		if (size > 0)
			{
			if ( (buffer = my_kmalloc(size, "Digest Error Data")) != NULL)
				{
	 			retval = recv_a_message(current_connection, size, buffer);

				retval = recv_locally(size, 0, current_connection, 0, &checksum, &over_flow_buffer);
				my_kfree((void **)&over_flow_buffer, "local buffer");
				my_kfree((void **)&buffer, "Digest Error Data");
				}
			}
		}

	new_session = get_rec_sess(current_session->scsi_target_id, current_session->lun_bits);

	conn = new_session->connection_head;

	if (current_session->connection_head->connection_flags & USE_HEADERDIGEST)
		conn->connection_flags |= USE_HEADERDIGEST;
	if (current_session->connection_head->connection_flags & USE_DATADIGEST)
		conn->connection_flags |= USE_DATADIGEST;

	global_hostdata->init_sessrec_flg = 1;

	for( lun = 0;  lun < global_host->max_lun;  lun++ )
		{
		if( current_session->lun_bits & (1<<lun) )
			{
				if( remove_session_from_hostdata(current_session->scsi_target_id, lun, ALL_ONES, global_hostdata) == 1 )
					{/* just removed the whole session structure */
						break;
					}
			}
		}

	conn->target_address = ip;
	conn->target_port = port;
	conn->connection_id = cid;
	new_session->cur_cmd_sn = 22222;
	new_session->exp_cmd_sn = 22223;
	retval = add_session_to_hostdata(conn, lun, new_session, global_host);
	if (retval == 1)
		{/* = 1: adding complete new session, lun and connection */
		/*  Finish initializing the session and its first connection,
		which includes logging it in */
		if( (retval=init_session(new_session, conn, lun, *global_hostdata->param_tbl,
								global_hostdata->chap_parameter,
								&redirect_ip, &redirect_port)) != 0 )
			{/* initialization or login failed or got redirection */
			goto out4;
			}
		/* Add the connection to round robin connection head - SAI */
		new_session->rr_conn_head = conn;
		new_session->nconnections = 1;
		retval = 1;

		/* start up the single tx_thread for this new session */
		kernel_thread((int (*)(void *))iscsi_initiator_tx_thread,
													(void *)new_session, 0);

		/* wait for the threads to set themselves up */
		down_interruptible(&conn->rx_done_sem);
		down_interruptible(&new_session->tx_done_sem);

		/* a normal session, SCSI Mid-level drives it via queuecommand() */
		TRACE(TRACE_ISCSI, "Enter normal session Full Feature Phase\n");
		clean_rec_sess(new_session);
		}

	for (cmd=conn->pending_commands; cmd != NULL; cmd = cmd->next)
		{
		up(&conn->my_session->tx_sem);
		new_session->cur_cmd_sn ++;
		}

	/* create_session(target, ip, port, lun, cid, global_host); */

rec_thread_out:
	
	global_hostdata->init_sessrec_flg = 0;

	up(&init_sess_recovery_sem);


	TRACE(TRACE_ENTER_LEAVE, "Leave init_recovery_thread \n");

	init_rec_thread = NULL;

	return;

out4:
    /* Here to recover from a late error, after host's list was updated. */
    remove_session_from_hostdata(target, lun, cid, global_hostdata);
	goto out2;

out2:
    /* here to free structure for configured parameters used in login */
    param_tbl_uncpy(*new_session->session_params);
    my_kfree((void **)&new_session->session_params, "session param_tbl");
    /* here to free new session structure */
    my_kfree((void **)&new_session, "session");

    goto rec_thread_out;
	}

/************************************************************************
* This function handles the digest errors and sequence errors that may  *
* have caused because of previous digest errors.                        *
************************************************************************/

int init_digest_recovery(struct init_error_rec *err_rec)
	{
	int			retval=1, over_flow_data_size=0;
	int			size=0, err_type=0;
	int			serial_no=0;
	u_int			checksum=0;
	char			*buffer=NULL;
	char			*over_flow_buffer=NULL;
	unsigned int		opcode=0;
	struct response_pdu	*header;
	struct iscsi_targ_scsi_rsp	*rsp_hdr;
	struct connection	*curr_conn;
	struct command		*related_cmd;
	struct iscsi_targ_scsi_data_in	*data_in_hdr;

	curr_conn = err_rec->curr_conn;
	related_cmd = err_rec->related_cmd;
	err_type = err_rec->err_type;
	
	TRACE(TRACE_ENTER_LEAVE, "Enter init_digest_recovery \n");

	if (curr_conn->rx_buf == NULL)
		{
		TRACE_ERROR("scsi Response Header is NULL\n");
		return -1;
		}
	else
		{
		header = (struct response_pdu *)curr_conn->rx_buf;
		opcode = header->opcode;
		size = ntohl(header->length);
		}

	if (err_type == HEADER_DIGERR)
		{
		TRACE(TRACE_DEBUG, "Handling Header Digest Error \n");

		if (size > 0)
			{
			if ( (buffer = my_kmalloc(size, "Digest Error Data")) != NULL)
				{
		 		retval = recv_a_message(curr_conn, size, buffer);
				}
			else
				init_session_recovery(err_rec);
	
			retval = recv_locally(size, 0, curr_conn, 0, &checksum, &over_flow_buffer);
			my_kfree((void **)&over_flow_buffer, "local buffer");
			my_kfree((void **)&buffer, "Digest Error Data");
			}
		return retval;
		}

	if (opcode == ISCSI_TARG_SCSI_DATA_IN)
		{
		if (err_rec->related_cmd == NULL)
			{
			TRACE_ERROR("Related SCSI Command is NULL - Snack Request not sent\n");
			return -1;
			}

		if (err_type == PAYLOAD_DIGERR)
			{
			TRACE(TRACE_DEBUG, "Handling Payload Digest Error \n");

			/* roll-back to previous DataSN */
			related_cmd->data_in_sn -= 1;
		
			if ( (over_flow_data_size = header->offset + size - related_cmd->SCpnt->request_bufflen) > 0)
				{
				size = related_cmd->SCpnt->request_bufflen - header->offset;
				}

			serial_no = related_cmd->data_in_sn;
			init_send_snack(err_rec, serial_no, 0, DATA_R2T_SNACK);
			}
		else if (err_type == SEQUENCE_ERR)
			{
			TRACE(TRACE_DEBUG, "Handling Data Sequence Error \n");

			if (size > 0)
			{
			if ( (buffer = my_kmalloc(size, "Digest Error Data")) != NULL)
				{
		 		retval = recv_a_message(curr_conn, size, buffer);
				}
			else
				init_session_recovery(err_rec);
			}
	
			retval = recv_locally(size, 0, curr_conn, 0, &checksum, &over_flow_buffer);

			data_in_hdr = (struct iscsi_targ_scsi_data_in *)curr_conn->rx_buf;
			if (ntohl(data_in_hdr->data_sn) > related_cmd->data_in_sn)
				{
				serial_no = related_cmd->data_in_sn;
				init_send_snack(err_rec, serial_no, 0, DATA_R2T_SNACK);
				}
			else
				{
				TRACE(TRACE_ISCSI, "Discarding DataIn PDU repeated for ITT %u \n",
										ntohl(header->init_task_tag));
				}

			my_kfree((void **)&over_flow_buffer, "local buffer");
			my_kfree((void **)&buffer, "Digest Error Data");
			}

		/*
		if (header->flags & S_BIT)
			{
			if (!(header->flags & F_BIT))
				{
				TRACE_ERROR("DataIn has S bit = 1 but F bit = 0, expected 1\n");
				}
			related_cmd->snack_flags |= STATUS_BIT;
			}
		*/
		}
	else if (opcode == ISCSI_TARG_R2T)
		{
		TRACE(TRACE_DEBUG, "Handling R2T Sequence Error \n");
		serial_no = related_cmd->r2t_sn;
		init_send_snack(err_rec, serial_no, 0, DATA_R2T_SNACK);
		}
	else if (opcode == ISCSI_TARG_SCSI_RSP)
		{

		TRACE(TRACE_DEBUG, "Handling Response Sequence Error \n");

		rsp_hdr = (struct iscsi_targ_scsi_rsp *)curr_conn->rx_buf;

		if (related_cmd == NULL)
			{
			TRACE_ERROR("Related SCSI Command is NULL - Snack Request not sent\n");
			return retval;
			}

		if (related_cmd->data_in_sn != rsp_hdr->exp_data_sn)
			{
			serial_no = related_cmd->data_in_sn;
			init_send_snack(err_rec, serial_no, 0, DATA_R2T_SNACK);
			}

		if ( ntohl(rsp_hdr->stat_sn) < curr_conn->exp_stat_sn)
			{
			TRACE(TRACE_ISCSI, "Discarding already received status PDU\n");
			}
		else
			{
			related_cmd->snack_flags |= STATUS_BIT;
			serial_no = curr_conn->exp_stat_sn;
			init_send_snack(err_rec, serial_no, 1, STATUS_SNACK);
			}
		}
			
	TRACE(TRACE_ENTER_LEAVE, "Leave init_digest_recovery %d \n", retval);

	return retval;	
	}

/******************************************************************************************
*       Called to send a SNACK Request to target for retransmission of numbered-responses,
*       data, or R2T PDU.
******************************************************************************************/

void init_send_snack( struct init_error_rec *err_rec, int begrun, int runlen, __u8 type)
	{
	int						need_itt=-1;
	struct connection		*current_connection;
	struct session			*current_session;
	struct command			*new_command=NULL;
	struct iscsi_init_snack	*snack_cmd=NULL;
	struct response_pdu		*header=NULL;


	TRACE(TRACE_ENTER_LEAVE, "Enter init_send_snack\n");

	current_connection = err_rec->curr_conn;
	current_session = current_connection->my_session;
	header = (struct response_pdu *)current_connection->rx_buf;

	if (type == DATA_R2T_SNACK)
		{
		if  ( !(current_session->init_snack_flg & DATA_SNACK_ENABLE))
			{
			TRACE_ERROR("Initiator Data Snack Is Not Enabled - SNACK request Not Sent\n");
			goto out;
			}
		}

	if (type == STATUS_SNACK)
		{
		if ( !(current_session->init_snack_flg & STATUS_SNACK_ENABLE))
			{
			TRACE_ERROR("Initiator Status Snack Is Not Enabled - SNACK request Not Sent\n");
			goto out;
			}
		}

	if( (new_command = setup_command(NULL, current_connection->my_session)) == NULL )
		{/* will not be able to send the snack */
		goto out;
		}

	/* set up the SnackReq header (section 9.16) to be sent to the target */
	snack_cmd = (struct iscsi_init_snack *)&new_command->iscsi_cmd;
	memset(snack_cmd, 0, sizeof(struct iscsi_init_snack));
	snack_cmd->opcode = ISCSI_INIT_SNACK;
	snack_cmd->flags = F_BIT | type;		/* F_BIT always set on SnackReq */
	snack_cmd->length = 0LL;
	snack_cmd->lun = 0LL;

	if (type != STATUS_SNACK)
		{
		snack_cmd->init_task_tag = htonl(err_rec->related_cmd->init_task_tag);
		}
	else
		{
		snack_cmd->init_task_tag = ALL_ONES;
		err_rec->related_cmd->snack_flags |= STATUS_BIT;
		}

	if (type == DATACK_SNACK)
		{
		snack_cmd->target_xfer_tag = htonl(header->target_xfer_tag);
		snack_cmd->lun = header->lun;
		}
	else
		{
		snack_cmd->target_xfer_tag = ALL_ONES;
		}

	snack_cmd->begrun = htonl(begrun);
	snack_cmd->runlen = htonl(runlen);

	snack_cmd->length = htonl(snack_cmd->length);
	snack_cmd->lun    = htonl(snack_cmd->lun);


	/* can now attach the new command to end of pending commands list */
	/* this also sets ExpStatSN */

	new_command->time_stamp = jiffies;
	attach_pending_command(new_command, current_connection, need_itt);

	TRACE(TRACE_ISCSI, "Send SNACK request\n");
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_snack(snack_cmd);

	/*  get the tx thread started, if necessary */
	up(&current_connection->my_session->tx_sem);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_send_snack\n");

	return;
}

/******************************************************************
* Deal with the Connection Recovery after Time2Wait expires
******************************************************************/

void deal_with_connrec(unsigned long data)
	{
    struct session      *current_session;
    struct connection   *fail_connection, *new_connection=NULL;
    struct command      *task_mgt_command=NULL, *cmd, *last_pending_command;
    struct iscsi_init_task_mgt_command *task_mgt_command_hdr=NULL;
	struct init_error_rec	err_rec;

    TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_connrec\n");

    TRACE(TRACE_DEBUG, "Conn Recovery Timer expired\n");

    fail_connection = (struct connection *)data;
    current_session = fail_connection->my_session;

	fail_connection->connection_state = STATUS_CONNREC;

	if (current_session->nconnections <= 1)
		{/* This is the last connection available. Create a new connection */
		create_rec_conn(fail_connection);
		}

	current_session->n_rec_tasks = 0;

    for(cmd=fail_connection->pending_commands; cmd != NULL; cmd = cmd->next)
        {
		if (cmd->iscsi_cmd.opcode != ISCSI_INIT_SCSI_CMND)
			continue;

        new_connection = get_connection(current_session);

		if(new_connection == fail_connection)
			{
			TRACE_ERROR("deal_with_connrec : "
						"get_connection returned the failed connection - Can't
							Re-Assign commands over a failed connection\n");
			goto out;
			}

		if (!new_connection)
			{
			TRACE_ERROR("deal_with_connrec : "
						"get_connection returned a NULL connection\n");
			err_rec.curr_conn = fail_connection;
			init_session_recovery(&err_rec);
			goto out;
			}

		if( (task_mgt_command = setup_command(NULL, current_session)) == NULL )
			{
			goto out;
			}

        TRACE(TRACE_DEBUG, "Populating task mgt command for connection recovery\n");
        task_mgt_command->task_mgt_function = TASK_REASSIGN;

        task_mgt_command->task_mgt_response = FUNCTION_REJECTED;
        task_mgt_command->ref_task_tag = cmd->init_task_tag;

        /* fill up the task_mgt_header info.*/
        task_mgt_command_hdr = (struct iscsi_init_task_mgt_command *)
                                                &task_mgt_command->iscsi_cmd;
		task_mgt_command_hdr->opcode = ISCSI_INIT_TASK_MGMT_CMND;
        task_mgt_command_hdr->opcode |= I_BIT;
        task_mgt_command_hdr->function = TASK_REASSIGN;

        if( current_session->draft_number >= DRAFT11 )
            {/* draft 11 or after, F bit now set to 1 (was 0 in earlier drafts)*/
            task_mgt_command_hdr->function |= F_BIT;
            }
        task_mgt_command_hdr->lun = cmd->iscsi_cmd.lun;
        task_mgt_command_hdr->ref_task_tag = htonl(cmd->init_task_tag);

        task_mgt_command_hdr->ref_cmd_sn = cmd->iscsi_cmd.cmd_sn;

		/* Setup Time Stamp for task mgt. command */
		task_mgt_command->time_stamp = jiffies;

        /* update the command linked list with the task mgt. command */
        attach_pending_command(task_mgt_command, new_connection, 1);

        TRACE(TRACE_ISCSI, "Send TASK MGT Cmnd, ITT %u, %u bytes\n",
                        task_mgt_command->init_task_tag, task_mgt_command->tx_size);
        if( TRACE_TEST(TRACE_ISCSI_FULL) )
            print_init_task_mgt_command(task_mgt_command_hdr);

        /* signal the tx thread for this session that it has work to do */
        up(&current_session->tx_sem);

		current_session->n_rec_tasks ++;
        TRACE(TRACE_DEBUG,"Waiting to get task mgt. rsp from the target\n");
        }

		if (current_session->n_rec_tasks > 0)
			{
			TRACE(TRACE_DEBUG, "No. of commands recovered - %d\n", current_session->n_rec_tasks);

			/* link the old command to the new connection */
			last_pending_command = new_connection->pending_commands;
			if (last_pending_command == NULL)
				{
				new_connection->pending_commands = fail_connection->pending_commands;
				}
			else
				{
				while(last_pending_command->next != NULL)
					last_pending_command = last_pending_command->next;
				last_pending_command->next = fail_connection->pending_commands;
				}
			}
		else
			{/* no tasks to be transferred */
			}

		fail_connection->connection_state = STATUS_NOT_PRESENT;
		fail_connection->connection_id = 0;
		current_session->nconnections--;
		close_connection(fail_connection);

out:
    TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_connrec\n");
	}

void time2wait_function(unsigned long wait_sem)
	{
	struct semaphore *rec_wait_sem = (struct semaphore *)wait_sem;
	up(rec_wait_sem);
	}

/***************************************************************
	Start Time2Wait Timer for Connection Recovery
***************************************************************/

void start_connrec_timer( struct session *current_session, struct semaphore *rec_wait_sem, unsigned int period )
    {

    TRACE(TRACE_ENTER_LEAVE, "Enter start_connrec_timer, period %u\n", period);

    if( period > 0  &&  current_session->connrec_timer != NULL  &&
                                    !timer_pending(current_session->connrec_timer) )
        {/* we can start the timer for a non-zero period of time (in ticks) */
        TRACE(TRACE_DEBUG, "Start timer for %u ticks\n", period);
        current_session->connrec_timer->expires = jiffies + period;
        current_session->connrec_timer->data = (unsigned long)rec_wait_sem;
        current_session->connrec_timer->function = time2wait_function;
        add_timer(current_session->connrec_timer);
        }
    else
        {/* else period is 0 or no timer or timer already going,
            don't start it now*/
        TRACE_ERROR("Conn Recovery Timer not started for %u ticks\n", period);
        }


    TRACE(TRACE_ENTER_LEAVE, "Leave start_connrec_timer\n");
    }

void deal_with_retran_timer(unsigned long data);

/********************************************************************
	Start or restart the Command retry timer here
 ********************************************************************/

void restart_retran_timer( struct session *current_session, unsigned int period )
    {

    TRACE(TRACE_ENTER_LEAVE, "Enter restart_retran_timer, period %u\n", period);
    
    if( period > 0  &&  current_session->retran_timer != NULL  &&
                                    !timer_pending(current_session->retran_timer) )
        {/* we can start the timer for a non-zero period of time (in ticks) */
        TRACE(TRACE_DEBUG, "Start retransmit timer for %u ticks\n", period);
        current_session->retran_timer->expires = jiffies + period;
        current_session->retran_timer->data = (unsigned long)current_session;
        current_session->retran_timer->function = deal_with_retran_timer;
        add_timer(current_session->retran_timer);
        }
    else
        {/* else period is 0 or no timer or timer already going,
            don't start it now*/
        TRACE_ERROR("Retransmit Timer not started for %u ticks\n", period);
        }


    TRACE(TRACE_ENTER_LEAVE, "Leave restart_retran_timer\n");
    }

/**********************************************************************
	Deal with command retransmission after the retransmit wait period
	has expired
 *********************************************************************/

void deal_with_retran_timer( unsigned long data )
    {
    struct session  *this_session;


    TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_retran_timer\n");

    TRACE(TRACE_DEBUG, "Retransmit Timer expired\n");

    this_session = (struct session *)data;

    /*  turn off the need tx thread wakeup bit */
    this_session->session_flags &= ~NEED_TX_WAKEUP;

    /*  get the tx thread started, if necessary */
    up(&this_session->tx_sem);

    /*  and keep the timer going */
    restart_retran_timer(this_session, this_session->xmit_chk_interval*HZ);

    TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_retran_timer\n");

    }

void unhook_connection(struct connection *current_connection)
	{
	struct session	*current_session;
	struct connection	*conn, *prev;

	current_session = current_connection->my_session;
	prev = conn = current_session->connection_head;

	while(conn)
		{
		if (conn == current_connection)
			{
			prev->next = conn->next;
			close_connection(current_connection);
			break;
			}

		prev = conn;
		conn = conn->next;
		}
	return;
	}

void restore_global_data(struct iscsi_hostdata *global_hostdata)
	{
	char *targ_name, *init_name;
	struct parameter_type	*dst, *param;
	/* Uncopy the old values from the parameter table */

	TRACE(TRACE_ENTER_LEAVE, "Entering restore_global_data\n");

	dst = (*global_hostdata->param_tbl);

	param = find_parameter("TargetName", dst);
	targ_name = strdup(param->str_value);

	param = find_parameter("InitiatorName", dst);
	init_name = strdup(param->str_value);

	param_tbl_uncpy(dst);
	param_tbl_cpy(dst, config_params);

	param = find_parameter("TargetName", dst);
	strreplace(&param->str_value, targ_name);
	param->neg_info = KEY_TO_BE_NEGOTIATED;	
	param = find_parameter("InitiatorName", dst);
	strreplace(&param->str_value, init_name);
	param->neg_info = KEY_TO_BE_NEGOTIATED;	
	
	TRACE(TRACE_ENTER_LEAVE, "Leaving restore_global_data\n");
	}
