/*	defined at the end of this file */
int check_range_list_complete( struct order_range *head );

void free_range_list( struct order_range *head );

/*	holds bytes of zeros that are sent as pad bytes when needed */
static int			pad_bytes = 0;

/*	holds list of currently unused cookies */
static struct r2t_cookie	*available_cookies = NULL;

/*	holds maximum number of cookies actually allocated */
static int max_cookies = 0;


/*	called once (from iscsi_initiator_detect) when initiator module is loaded */
void setup_r2t_cookies( void )
	{
	available_cookies = NULL;
	max_cookies = 0;
	}


/*	"prints" statistics on r2t cookie usage into buf,
	returns no. of chars printed.
*/
int print_r2t_cookies( char *buf )
	{
	if( buf == NULL )
		return printk("iscsi: Maximum of %d r2t cookies allocated\n",
																max_cookies);
	else
		return sprintf(buf,"Maximum of %d r2t cookies allocated\n",max_cookies);
	}


/*	called once (from iscsi_initiator_release) when initiator module is unloaded
*/
void endup_r2t_cookies( void )
	{
	struct r2t_cookie	*cookie;

	while( (cookie = available_cookies) != NULL )
		{
		available_cookies = cookie->next;
		my_kfree((void **)&cookie, "r2t cookie");
		}
	print_r2t_cookies(NULL);
	}


/*	allocates a new r2t cookie and adds it to end of command's cookie list */
struct r2t_cookie *create_r2t_cookie( struct session *this_session,
									  struct command *this_command, __u32 tag )
	{
	struct r2t_cookie	*cookie;

	if( (cookie = available_cookies) != NULL )
		{/* reuse an old cookie */
		available_cookies = available_cookies->next;
		}
	else
		{/* allocate a new cookie */
		cookie = (struct r2t_cookie *)my_kmalloc(sizeof(struct r2t_cookie),
																"r2t cookie");
		if( cookie == NULL )
			return NULL;
		max_cookies++;			/* keep a count of number of allocations */
		}

	cookie->target_xfer_tag = tag;
	if( tag != ALL_ONES )
		{/* this is a "real" (not implied) r2t.
			count it and check number outstanding */
		this_command->n_r2t_cookies++;
		if( this_command->n_r2t_cookies >
								this_session->oper_param->MaxOutstandingR2T )
			{/* too many outstanding R2Ts */
			TRACE_ERROR("Target sent %u outstanding R2Ts, "
						"but MaxOutstandingR2T is %u\n",
						this_command->n_r2t_cookies,
						this_session->oper_param->MaxOutstandingR2T);
			}
		}

	/*	Draft 18, Section 9.7.5 DataSN
		"For output (write) data PDUs, the DataSN is the Data-Out PDU number
		within the current output sequence.  The current output sequence is
		either identified by the Initiator Task Tag (for unsolicited data) or
		is a data sequence generated for one R2T (for data solicited through
		R2T)."

		Therefore, each R2T resets data_out_sn to 0.
	*/
	cookie->data_out_sn = 0;

	cookie->next = NULL;	/* this is last in cookie list */
	if( this_command->first_cookie == NULL )
		{/* this is first cookie for this command */
		this_command->first_cookie = cookie;
		}
	else
		{/* already have cookies for this command */
		this_command->last_cookie->next = cookie;
		}
	this_command->last_cookie = cookie;
	return cookie;
	}


/*	removes first cookie from this command's cookie list and returns
	pointer to new first cookie on that list or NULL if list now empty */
struct r2t_cookie *free_r2t_cookie( struct command *this_command )
	{
	struct r2t_cookie	*cookie;

	if( (cookie = this_command->first_cookie) != NULL )
		{/* command actually had a first cookie (not always true!) */

		if( cookie->target_xfer_tag != ALL_ONES )
			{/* this is a "real" (not implied) r2t, reduce number outstanding */
			this_command->n_r2t_cookies--;
			}

		/* snip this cookie off front of list for this command */
		this_command->first_cookie = cookie->next;

		/* then add it to front of list of available cookies */
		cookie->next = available_cookies;
		available_cookies = cookie;
		}

	return this_command->first_cookie;
	}


/*	keeps track of maximum  number of io vector slots actually used */
static int	max_iov_count = 0;

/*	Called to build a DataOut pdu header in the "data" area of the
 *	"related_command" if that area is currently unused (else does nothing).
 *	The amount of data to send is controlled by the first "cookie" for
 *	that command, but the location(s) and layout of the data in memory
 *	are controlled by the SCpnt field in the "related_command" when
 *	setting up the msghdr and io vector in the "related_command".
 *	Returns number of bytes newly set up for transmission (0 if none).
*/
int setup_dataoutpdu( struct connection *related_connection,
						struct command *related_command )
	{
	struct r2t_cookie	*cookie;
	int					len, prev_offset, cur_offset, count, reset_crc=1, size;
	char				*cur_ptr;
	struct	iovec		*iov;
	struct scatterlist	*sg;
	struct session		*related_session;


	TRACE(TRACE_ENTER_LEAVE, "Enter setup_dataoutpdu\n");

	size = 0;
	if( related_command->tx_size != 0 )
		{/* command's single data slot is busy, don't overwrite it */
		TRACE(TRACE_ISCSI, "setup_dataoutpdu: slot busy, tx_size %u\n",
													related_command->tx_size);
		goto out;
		}

	/* always set up data from the command's first cookie (should have one!) */
	if( (cookie = related_command->first_cookie) == NULL )
		{
		TRACE_ERROR("No cookie for command!\n");
		goto out;
		}

	/* sanity check -- we should never try to send no data! */
	while( (int)cookie->r2t_xfer_length <= 0 )
		{
		TRACE_ERROR("setup_dataoutpdu: r2t_xfer_length %d not positive\n",
												(int)cookie->r2t_xfer_length);
		if( (cookie = free_r2t_cookie(related_command)) == NULL )
			goto out;
		}

	/* set up the DataOut PDU in command's single data slot */
	memset(&related_command->data, 0, ISCSI_HDR_LEN);

	related_command->data.opcode = ISCSI_INIT_SCSI_DATA_OUT;
	related_command->data.target_xfer_tag = cookie->target_xfer_tag;
	related_command->data.init_task_tag = htonl(related_command->init_task_tag);
	related_command->data.exp_stat_sn = htonl(related_connection->exp_stat_sn);
	related_command->data.data_sn = htonl(cookie->data_out_sn);

	/* assume this is not the last DataOut PDU to be sent in this sequence */
	related_session = related_connection->my_session;
	if( related_session->draft_number >= DRAFT9 )
		{/* drafts 9 and after */
		len = related_connection->max_send_length;
		}
	else
		{/* drafts 8 and before */
		len = related_session->oper_param->DataPDULength;
		}

	if( cookie->r2t_xfer_length <= len )
		{/* assumption was wrong, this is the last DataOut PDU to be sent */
		len = cookie->r2t_xfer_length;
		related_command->data.flags |= F_BIT;
		}

	related_command->data.length = htonl(len);

	/* get offset into original SCSI CMD buffer of next byte to be sent */
	if( related_session->oper_param->DataPDUInOrder == 0 )
		{/* DataPDUInOrder==No, send Data Out PDUs backwards */
		cookie->offset -= len;					/* next Data Out PDU limit */
		prev_offset = cookie->offset;			/* this Data Out PDU offset */
		}
	else
		{/* DataPDUInOrder==Yes, send Data Out PDUs in order */
		prev_offset = cookie->offset;			/* this Data Out PDU offset */
		cookie->offset += len;					/* next Data Out PDU offset */
		}

	related_command->data.offset = htonl(prev_offset);

	TRACE(TRACE_DEBUG, "Set up DataOut PDU, data_offset %d payload_length %d\n",
															prev_offset, len);

	/* update values in related command cookie to account for this PDU */
	cookie->r2t_xfer_length -= len;				/* total amount left to send */
	cookie->data_out_sn++;						/* Seqno of next Data Out PDU */

	/* set up the TCP io vector, beginning with no. of bytes to xfer */
	size = ISCSI_HDR_LEN + len;

	/* reset the message header, then point it to the io vector */
	memset(&related_command->tx_msghdr, 0, sizeof(struct msghdr));
	iov = related_command->iov;
	related_command->tx_msghdr.msg_iov = iov;

	/* DataOut pdu headers built in separate "data" area of struct command */
	iov->iov_len = ISCSI_HDR_LEN;
	iov->iov_base = &related_command->data;

	related_command->header_opcode = related_command->data.opcode;
	related_command->header_flags = related_command->data.flags;

	/* now compute and add the header digest if we are using them */
	if( related_connection->connection_flags & USE_HEADERDIGEST )
		{
		do_crc((unsigned char *)&related_command->data, ISCSI_HDR_LEN, 3,
										&related_command->data.header_digest);
		iov->iov_len += CRC_LEN;
		size += CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Attach HeaderDigest\n");
		}

	/* now fill in the iovector to point to data to be sent */
	iov++;
	count = 2;			/* always at least 2 items in iovector */
	if( related_command->SCpnt->use_sg > 0 )
		{/* SCSI gave us a scatter-gather list */
		sg = (struct scatterlist *)related_command->SCpnt->request_buffer;
		cur_offset = 0;
		cur_ptr = sg->address;

		/*	skip thru scatterlist past data already sent */
		while( cur_offset < prev_offset )
			{
			if( cur_offset + sg->length > prev_offset )
				{
				cur_ptr = cur_ptr + (prev_offset - cur_offset);
				cur_offset = prev_offset;
				}
			else
				{
				cur_offset += sg->length;
				sg++;
				cur_ptr = sg->address;
				}
			}

		/* sg now points at first list item to xfer */
		iov->iov_len = (sg->address + sg->length) - cur_ptr;
		iov->iov_base = cur_ptr;
		sg++;

		len -= iov->iov_len;		/* no. of bytes left to fill */
		while( len > 0 )
			{/* need to add another element in iovector */
			/* but first compute the partial data digest if we are using them */
			if( related_connection->connection_flags & USE_DATADIGEST )
				{
				do_crc(iov->iov_base, iov->iov_len, reset_crc,
											&related_command->data_checksum);
				reset_crc=0;
				}
			count++;
			iov++;
			iov->iov_base = sg->address;
			iov->iov_len = sg->length;
			len -= sg->length;
			sg++;
			}

		if( len < 0 ) 
			{/* only need to send part of the last list item */
			iov->iov_len += len;
			}
		}
	else 
		{/* all the data is in a single buffer (no scatter gather list) */
		iov->iov_base = related_command->SCpnt->request_buffer + prev_offset;
		iov->iov_len = len;
		}

	/*	now deal with any padding that is needed */

	/*	Draft 18, Section 9.1 iSCSI PDU Length and Padding
		"iSCSI PDUs are padded to the closest integer number of four byte
		words.  The padding bytes SHOULD be 0."
	*/
	if( (len = ((-size) & 3)) != 0 )
		{/* need to send len pad bytes in another element in iovector */
		/* but first compute the partial data digest if we are using them */
		if( related_connection->connection_flags & USE_DATADIGEST )
			{
			do_crc(iov->iov_base, iov->iov_len, reset_crc,
											&related_command->data_checksum);
			reset_crc=0;
			}
		count++;
		iov++;
		iov->iov_base = &pad_bytes;
		iov->iov_len = len;
		size += len;						/* is now a multiple of 4 */
		TRACE(TRACE_ISCSI, "Attach %d pad bytes\n", len);
		}

	/*	now compute and add the final data digest if we are using them */
	if( related_connection->connection_flags & USE_DATADIGEST )
		{
		do_crc(iov->iov_base, iov->iov_len, reset_crc|0x02,
											&related_command->data_checksum);
		count++;
		iov++;
		iov->iov_base = &related_command->data_checksum;
		iov->iov_len = CRC_LEN;
		size += CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
		}

	/* finally, save size and number of iovector elements in message header */
	related_command->tx_size = size;				/* marks this "in use" */
	related_command->tx_msghdr.msg_iovlen = count;

	if( max_iov_count < count )
		max_iov_count = count;		/* keep track of max no. of iovs ever used*/

	TRACE(TRACE_ISCSI, "Send DataOut, %u bytes, %d iovs\n", size, count);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_scsi_data_out(&related_command->data);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave setup_dataoutpdu, size %u\n", size);

	return size;
	}


/*	"prints" statistics on iov usage into buf,
	returns no. of chars printed.
*/
int print_iovs( char *buf )
	{
	if( buf == NULL )
		return printk("iscsi: Maximum of %d iov slots used out of %d\n",
												max_iov_count, MAX_IOV_SLOTS);
	else
		return sprintf(buf, "Maximum of %d iov slots used out of %d\n",
												max_iov_count, MAX_IOV_SLOTS);
	}


void endup_iovs( void )
	{
	print_iovs(NULL);
	}


/*	Reads "orig_size" bytes into the area pointed to by "base" from the socket
 *	attached to "conn".  Uses the rx_msg and rx_iov structures attached to
 *	"conn" to control the read.
 *  "orig_size" MUST be positive!!
 *	my_lock MUST be held by the caller!!
 *	Returns "orig_size" on success, 0 on failure with my_lock held,
 *	<0 on failure with my_lock not held.
*/
int recv_a_message( struct connection *conn, int orig_size, char *base )
	{
	int				retval, size;
	struct msghdr	*msg;
	struct iovec	*iov;
	mm_segment_t	oldfs;


	TRACE(TRACE_ENTER_LEAVE, "Enter recv_a_message\n");

	/* first set up connection's reusable message header and i/o vector */
	/* NOTE: sock_recvmsg() destroys some part of this header, so it MUST be
	   set up EACH TIME through here */

	msg = &conn->rx_msg;	/* reusable structure for this connection's rx */
	iov = &conn->rx_iov;	/* reusable structure for this connection's rx */

	size = orig_size;
	while( size > 0 )
		{
		memset(msg, 0, sizeof(struct msghdr));		/* most fields are 0/NULL */

		msg->msg_iov = iov;
		msg->msg_iovlen = 1;
		msg->msg_flags = MSG_NOSIGNAL;

		iov->iov_len = size;
		iov->iov_base = base;

		/*	allow other threads to run in case we block in the receive */
		spin_unlock(&my_lock);

		oldfs = get_fs(); set_fs(get_ds());
		retval = sock_recvmsg(conn->sock, msg, size, msg->msg_flags);
		set_fs(oldfs);

		TRACE(TRACE_DEBUG, "sock_recvmsg got %d bytes\n", retval);

		if( retval <= 0 )
			{/* sock_recvmsg hit an error (or was interrupted by a signal) */
			if( retval == 0 || retval == -ECONNRESET || retval == -ERESTARTSYS )
				{/* got an EOF indication, which means remote peer is gone */
				printk("iscsi: Disconnected from %u.%u.%u.%u:%u\n",
								(conn->target_address >> 24) & 0xff,
								(conn->target_address >> 16) & 0xff,
								(conn->target_address >> 8) & 0xff,
								conn->target_address & 0xff,
								conn->target_port);
				orig_size = -ECONNRESET;
				goto out;
				}
			TRACE_ERROR("sock_recvmsg error %d\n", retval);
			orig_size = retval;
			goto out;
			}

		/*	get back our exclusive access to the iscsi structures */
		spin_lock(&my_lock);

		base += retval;
		size -= retval;
		}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave recv_a_message, returning %d\n", orig_size);

	return orig_size;
	}


/****************************************************************************
* This function is responsible for making the checks for exp_cmd_sn and
* max_cmd_sn values in the received iSCSI header from the target.
* Parameters: received_exp_cmd_sn in the PDU (in network byte order)
*			  received_max_cmd_sn in the PDU (in network byte order)
*			  cur_session, pointing to the struct session
* Return value: 0, if the received_exp_cmd_sn and received_max_cmd_sn are valid
* 				-1, if either received_exp_cmd_sn or received_max_cmd_sn is
*				invalid.
****************************************************************************/
int check_and_update_cmd_sns( __u32 received_exp_cmd_sn,
            			__u32 received_max_cmd_sn, struct session *cur_session )
    {
	int				retval = 0, delta;
	__u32			exp_cmd_sn, max_cmd_sn;


	TRACE(TRACE_ENTER_LEAVE, "Enter check_and_update_cmd_sns\n");

	exp_cmd_sn = ntohl(received_exp_cmd_sn);
	max_cmd_sn = ntohl(received_max_cmd_sn);

	/*	Draft 18, Section 2.2.2.1 Command Numbering and Acknowledging

		"Comparisons and arithmetic on ExpCmdSN and MaxCmdSN MUST use
		Serial Number Arithmetic as defined in [RFC1982] where
		SERIAL_BITS = 32.

		The target MUST NOT transmit a MaxCmdSN that is less than ExpCmdSN-1.
		...

		MaxCmdSN and ExpCmdSN fields are processed by the initiator as
		follows:

		-If the PDU MaxCmdSN is less than the PDU ExpCmdSN-1 (in Serial,
			Arithmetic Sense), they are both ignored.

		-If the PDU MaxCmdSN is greater than the local MaxCmdSN (in Serial
			Arithmetic Sense), it updates the local MaxCmdSN; otherwise,
			it is ignored.

		-If the PDU ExpCmdSN is greater than the local ExpCmdSN (in Serial
			Arithmetic Sense), it updates the local ExpCmdSN; otherwise,
			it is ignored.

		This sequence is required because updates may arrive out of order
		(e.g., the updates are sent on different TCP connections).

		iSCSI initiators and targets MUST support the command numbering
		scheme."
	*/

	delta = max_cmd_sn - (exp_cmd_sn - 1);
	if( delta < 0 )
		{/* MaxCmdSN < ExpCmdSN - 1, ignore both PDU values */
		TRACE(TRACE_ISCSI_FULL, "Ignoring PDU's ExpCmdSN %u and MaxCmdSN %u\n",
														exp_cmd_sn, max_cmd_sn);
		}
	else
		{/* values in the PDU are ok relative to each other */
		delta = max_cmd_sn - cur_session->max_cmd_sn;
		if( delta < 0 )
			{/* PDU.MaxCmdSN < local.MaxCmdSN, ignore PDU's MaxCmdSN */
			TRACE(TRACE_ISCSI_FULL, "Ignoring PDU's MaxCmdSN %u\n", max_cmd_sn);
			}
		else if( delta > 0 )
			{/* greater than, use it to update the local MaxCmdSN */
			cur_session->max_cmd_sn = max_cmd_sn;
			}

		delta = exp_cmd_sn - cur_session->exp_cmd_sn;
		if( delta < 0 )
			{/* PDU.ExpCmdSN < local.ExpCmdSN, ignore PDU's ExpCmdSN */
			TRACE(TRACE_ISCSI_FULL, "Ignoring PDU's ExpCmdSN %u\n",
															exp_cmd_sn);
			}
		else if( delta > 0 )
			{/* greater than, check it against local CmdSN */

			delta = cur_session->cur_cmd_sn - exp_cmd_sn;
			if( delta < 0 )
				{
				TRACE_ERROR("Got ExpCmdSN %u, exceeds next CmdSN %u\n",
								 exp_cmd_sn, cur_session->cur_cmd_sn);
				retval = -1;
				}
			else
				{/* not less than, update the local ExpCmdSN */
				cur_session->exp_cmd_sn = exp_cmd_sn;
				}
			}
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave check_and_update_cmd_sns, retval %d\n",
																		retval);

    return retval;
	}


struct command *unhook_pending_command( struct command *old_command,
										struct connection *current_connection )
	{
	struct command	*current_command;


	if( current_connection->pending_commands == old_command )
		{/* this command is first in the list, just move over it */
		current_connection->pending_commands=current_command=old_command->next;
		}
	else
		{/* need to find where in the list this command is */

		for( current_command = current_connection->pending_commands;
			 current_command && (current_command->next != old_command);
			 current_command = current_command->next )
			 {}

		if( current_command == NULL )
			{/* should never happen */
			TRACE_ERROR("Command list is corrupt\n");
			}
		else
			{/* snip this command out of the list */
			current_command->next = old_command->next;
			current_command = old_command->next;
			}
		}

	return current_command;
	}


/*	Calls SCSI Mid-level's scsi_done callback with result,
 *	(and with my_lock held).
 *	Then frees up storage allocated to old command,
 *	and returns next command in list (NULL if none).
 *	Must be called with my_lock held by caller.
 */
struct command *free_pending_command( struct command *old_command,
										struct connection *current_connection )
	{
	struct command	*current_command;
	Scsi_Cmnd		*SCpnt;


	TRACE(TRACE_ENTER_LEAVE, "Enter free_pending_command %p\n", old_command);

	current_command = unhook_pending_command(old_command, current_connection);

	/*	release the storage occupied by any range list used by this command */
	free_range_list(&old_command->r2t_range_list);

	/*	Always release the storage used by this command */
	SCpnt = old_command->SCpnt;
	my_kfree((void **)&old_command, "command");

	/*	Call back the Mid-level's "scsi_done" function if there is one */
	if( SCpnt != NULL  &&  SCpnt->scsi_done != NULL )
		(*SCpnt->scsi_done)(SCpnt);

	TRACE(TRACE_ENTER_LEAVE, "Leave free_pending_command\n");

	return current_command;
	}


/*	process the common parts of a SCSIResponse PDU and a DataIn PDU with the
	S-bit set ("phase collapse") */
void do_scsi_response(	struct connection *current_connection,
						struct command *related_command,
						struct iscsi_targ_scsi_rsp *header,
						int received,
						unsigned char *buffer )
	{
	int				missing;
	int				slength;
	unsigned int	residual;


	TRACE(TRACE_ENTER_LEAVE, "Enter do_scsi_response\n");

	if( current_connection->my_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but with field names from draft11) */
		residual = ntohl(header->target_xfer_tag);
		}
	else
		{/* draft 11 or after */
		residual = ntohl(header->resid);
		}

	/*	see if initiator and target agree on amount of data transfered */
	if( related_command->r2t_range_list.next != NULL )
		{/* Command with DataSequenceInOrder==No */
		/*	expected range of any command is offset = 0 limit = bufflen */
		related_command->r2t_range_list.offset = 0;
		related_command->r2t_range_list.limit =
										related_command->SCpnt->request_bufflen;
		missing = check_range_list_complete(&related_command->r2t_range_list);
		}
	else
		{/* Command with DataSequenceInOrder==Yes */
		missing = related_command->SCpnt->request_bufflen -
												related_command->data_offset;
		}

	if( header->flags & U_BIT )
		{/* U bit is set, indicating underflow, check residual count */
		TRACE(TRACE_ISCSI,"Underflow: Residual count %u\n", residual);
		if( missing != residual )
			{
			TRACE_ERROR("Underflow Residual count %u doesn't match "
						"expected %d\n", residual, missing);
			}
		}
	else if( header->flags & O_BIT )
		{/* O bit is set, indicating overflow, check residual count */
		TRACE(TRACE_ISCSI,"Overflow: Residual count %u\n", residual);
		if( missing != 0 )
			{
			TRACE_ERROR("Overflow bit set, but still missing %d bytes\n",
						missing);
			}
		else if( residual == 0 )
			{
			TRACE_ERROR("Overflow bit set, but Residual count is 0\n");
			}
		}
	else if( residual != 0 )
		{
		TRACE_ERROR("Neither Overflow nor Underflow set, but Residual count is"
					" %u\n", residual);
		}

	/*	pass the final status back up to the SCSI Mid-level */
	/*	the "result" field of the command contains 4 independent bytes:
		byte 0	the status byte returned from the device itself in bits 1 thru 5
		byte 1	the status returned from the host adapter (always 0 for us)
		byte 2	the status returned by the low-level driver
		byte 3	"simply the message byte that comes back" -- whatever that is!

		The byte 2 errors are defined in /usr/src/linux/drivers/scsi/scsi.h
		and are symbols of the form DID_xxx.  DID_OK is conveniently 0.
	*/
	related_command->SCpnt->result = header->status;

	if( header->status & STATUS_MASK )
		{
		TRACE(TRACE_ISCSI, "Error detected at target, status 0x%02x\n",
															header->status);

		/* mark this as an internal error of some sort */
		related_command->SCpnt->result |= (DID_ERROR << 16);

		/*	Linux defines CHECK_CONDITION as 0x01,
			and SAM-2 requires it to be 0x02 (see Draft 18, Section 9.4.2).
			This is because Linux does not consider the low order bit
			to be part of the status, since SAM-2 requires that bit to be 0.
			The status_byte() macro in linux/drivers/scsi/scsi.h handles this.
			See also linux/include/scsi/scsi.h. */

		if( status_byte(header->status) == CHECK_CONDITION )
			{/* Draft 18, Section 9.4.7 Data Segment - Sense and Response
				Data Segment
				"iSCSI targets MUST support and enable autosense.  If Status
				is CHECK CONDITION (0x02), then the Data Segment contains
				sense data for the failed command."
			*/
			slength = 0;
			if( received <= 0  ||  buffer == NULL )
				{
				TRACE_ERROR("CHECK CONDITION but no sense data\n");
				}
			else
				{
				if( received < 2 )
					{
					TRACE_ERROR("%d data bytes not enough to contain sense "
														"length\n", received);
					}
				else
					{
					slength = (buffer[0]<<8)+buffer[1]; /* big-endian 16-bits */
					if( slength > received-2 )
						{
						TRACE_ERROR("%d sense data bytes but only %d actually "
											"received\n", slength, received-2);
						slength = received-2;
						}
					if( slength > SCSI_SENSE_BUFFERSIZE )
						{
						TRACE(TRACE_ISCSI,
							"***Warning***: received %d sense data "
							"bytes but mid-level buffer holds %d\n", slength,
							SCSI_SENSE_BUFFERSIZE);
						slength = SCSI_SENSE_BUFFERSIZE;
						}
					if( slength > 0 )
						{/* have some sense data, copy it into buffer */
						memcpy(related_command->SCpnt->sense_buffer, &buffer[2],
																	slength);
						/* Note -- we expect the mid-level to have zeroed the
						   sense buffer before calling queuecommand().
						   Therefore, we do not have to zero the unfilled part
						   of the buf (and do not have to do anything if sense
						   data is missing completely).
						*/
						/* mark that driver sense data is present */
						related_command->SCpnt->result |= (DRIVER_SENSE << 24);
						}
					}
				}
			}
		}

	free_pending_command(related_command, current_connection);

	TRACE(TRACE_ENTER_LEAVE, "Leave do_scsi_response\n");

	}


/*	Called to allocate space for a struct command and then initialize it.
 *	Returns pointer to newly allocated and initialized command if all ok,
 *	else NULL.
 *	When called, the my_lock must be held by the calling thread.
*/
struct command *setup_command(Scsi_Cmnd *Cmnd, struct session *current_session)
	{
	struct command	*new_command;


	new_command=(struct command *)my_kmalloc(sizeof(struct command),"command");
	if( new_command != NULL )
		{
		/*	populate the command struct for bookkeeping (most fields 0/NULL)

			Draft 18, Section 9.7.5 DataSN
			"For input (read) or bidirectional Data-In PDUs, the DataSN is the
			input PDU number within the data transfer for the command identified
			by the Initiator Task Tag."

			Draft 18, Section 9.8.2 R2TSN
			"R2TSN is the R2T PDU input PDU number within the command identified
			by the Initiator Task Tag."
		*/
		memset(new_command, 0, sizeof(struct command));

		new_command->SCpnt = Cmnd;

		/*	now set up the transmit i/o vector for header */
		new_command->tx_size = ISCSI_HDR_LEN;
		new_command->tx_msghdr.msg_iov = new_command->iov;
		new_command->iov->iov_base = &new_command->iscsi_cmd;
		new_command->iov->iov_len = ISCSI_HDR_LEN;
		new_command->tx_msghdr.msg_iovlen = 1;
		}

	return new_command;
	}


/*	Called to receive "extra" data into a locally allocated buffer
 *  and to compute final data digest checksum on it if necessary.
 *	Returns > 0 if ok, 0 on err & still hold my_lock,
 *	<0 on err and lost my_lock.
 *	Returns ptr to buffer if successful (and final_size is set) else NULL
 *	Note: if not NULL, this buffer must be freed by caller!
 *	Also Note: if set, the final_size contains the padding bytes if any and the
 *	data digest size if in use.
 *	On entry can have prev_size >= 0 and new_size >= 0
 *	but MUST have (prev_size + new_size) > 0.
*/
int recv_locally( int prev_size, int new_size,
				struct connection *current_connection,
				int reset_crc, u_int *checksum, char **buffer )
	{
	int	retval = 1, n;


	TRACE(TRACE_ENTER_LEAVE, "Enter recv_locally, prev_size %d, new_size %d\n",
														prev_size, new_size);

	/*	First deal with any padding that is needed */

	/*	Draft 18, Section 9.1 iSCSI PDU Length and Padding
		"iSCSI PDUs are padded to the closest integer number of four byte
		words.  The padding bytes SHOULD be 0."
	*/
	if( (n = ((prev_size + new_size) & 3)) != 0 )
		{/* need to receive pad bytes */
		TRACE(TRACE_ISCSI, "Expecting %d pad bytes\n", 4-n);
		new_size += 4-n;
		}

	/*	Next deal with data digest (if any) */
	if( current_connection->connection_flags & USE_DATADIGEST )
		{
		if( new_size > 0 )
			{/* will read data and/or padding, allow for data digest after it */
			new_size += CRC_LEN;
			}
		else if( !reset_crc )
			{/* previously received some data, now read the data digest */
			new_size = CRC_LEN;
			}
		/*	else no additional or previous data, so no data digest either */

		/*	Draft 18, Section 9.2.3 Header Digest and Data Digest
			"A zero-length Data Segment also implies a zero-length data-digest."
		*/
		}

	if( new_size > 0 )
		{/* receiving additional data and/or padding and/or crc */
		if( (*buffer = (char *)my_kmalloc(new_size, "local buffer")) == NULL )
			{
			new_size = 0;
			}
		else
			{
			TRACE(TRACE_DEBUG, "Receiving %d bytes in local buffer\n",new_size);

			retval = recv_a_message(current_connection, new_size, *buffer);
			if( retval <= 0 )
				{/* hit eof or got an error or a kill signal */
				my_kfree((void **)buffer, "local buffer");
				new_size = 0;
				}
			else
				{/* actually received "retval" bytes of data */
				if( retval != new_size )
					{/* should never happen */
					TRACE_ERROR("received %d bytes, expected %d\n", retval,
																	new_size);
					}
				/* compute the final data digest if we are using them */
				if( current_connection->connection_flags & USE_DATADIGEST )
					{
					if( new_size > CRC_LEN )
						{
						do_crc(*buffer, new_size-CRC_LEN, reset_crc, checksum);
						TRACE(TRACE_DEBUG,
									"do_crc %d bytes, reset %d, crc 0x%08x\n",
									new_size-CRC_LEN, reset_crc, *checksum);
						}
					if( *checksum != ntohl(*(unsigned int *)
											(*buffer + new_size - CRC_LEN)) )
						{/* crc check not == */
						TRACE_ERROR(
							"Data CRC wrong: got 0x%08x, expected 0x%08x\n",
										ntohl(*(unsigned int *)
										(*buffer + new_size - CRC_LEN)),
										*checksum);
						return PAYLOAD_DIGERR;
						}
					}
				}
			}
		}


	TRACE(TRACE_ISCSI_FULL, "Got %d data, pad and crc bytes\n",
														prev_size + new_size);

	TRACE(TRACE_ENTER_LEAVE, "Leave recv_locally, retval %d\n", retval);

	return retval;
	}


/* here to attach cmnds to the list after they have been properly
 * filled in.  The reason we needed to be a bit careful is that other
 * portions of the driver might be attempting to traverse the list
 * at the same time that we are here so we don't want to attach an 
 * invalid entry.  Therefore, the my_lock must be held by
 * the calling process/thread.
 */

void attach_pending_command( struct command *new_command,
							struct connection *current_connection,
							int need_itt )
	{
	struct iscsi_init_scsi_cmnd	*iscsi_cmd;
	struct session				*current_session;
	struct command				*last_pending_command;


	TRACE(TRACE_ENTER_LEAVE, "Enter attach_pending_command, need_itt %d\n",
																	need_itt);

	/*	extract a copy of pdu's opcode and flags for use after sendmsg */
	/*	These bytes are in the "iscsi_cmd" field of the "struct command"
		for all pdu types except DataOut when they are in the "data"
		field of the "struct command" */
	new_command->header_opcode=((__u8 *)(new_command->iov[0].iov_base))[0];
	new_command->header_flags=((__u8 *)(new_command->iov[0].iov_base))[1];

	new_command->next = NULL;		/* this command is always last in list */

	current_session = current_connection->my_session;
	iscsi_cmd = &new_command->iscsi_cmd;

	/* now allocate a session-wide command number and task tag,
	   and fill in other last-minute fields */

	switch( need_itt )
		{
	case 1:
		/* ITT will not contain 0xffffffff */
		/*	Draft 18, Section 2.2.2.1 Command Numbering and Acknowledging
			"Commands meant for immediate delivery are marked with an
			immediate delivery flag; they MUAST also carry the current CmdSN.
			CmdSn does not advance after a command marked for immediate
			delivery is sent.

			Command numbering starts with the first login request on the
			first connection of a session (the leading login on the
			leading connection) and command numbers are incremented by 1
			for every non-immediate command issued afterwards.

			...

			For the numbering mechanism, the initiator and target maintain
			the following three variables for each session:

				- CmdSN - the current command Sequence Number, advanced
				by 1 on each command shipped except for commands marked for
				immediate delivery.  CmdSN always contains the number to be
				assigned to the next Command PDU."

			Draft 18, Section 9.2.1.1 I
			"For request PDUs, the I bit set to 1 is an immediate delivery
			marker."
		*/

		iscsi_cmd->cmd_sn = htonl(current_session->cur_cmd_sn);
		if( !(iscsi_cmd->opcode & I_BIT) )
			{/* not an immediate command, advance CmdSN by 1 */
			current_session->cur_cmd_sn++;
			}
		/* else I bit is set, so this is an immediate command, CmdSN unchanged*/

		new_command->init_task_tag = current_session->cur_task_tag++;
		if( current_session->cur_task_tag == ALL_ONES )
			current_session->cur_task_tag++;
		iscsi_cmd->init_task_tag = htonl(new_command->init_task_tag);
		break;
	case 0:
		/* ITT set to 0xffffffff, so CmdSN not advanced and I bit set to 1 */
		/*	Draft 18, Section 9.18.1 Initiator Task Tag
			"If the Initiator Task Tag contains 0xffffffff, the I bit MUST be
			set to 1 and the CmdSN field is not advanced after this PDU is
			sent."
		*/
		iscsi_cmd->init_task_tag = new_command->init_task_tag = ALL_ONES;
		iscsi_cmd->opcode |= I_BIT;
		iscsi_cmd->cmd_sn = htonl(current_session->cur_cmd_sn);
		break;
	case 2:
		/* reuse old command struct, ITT already set, do not touch it */
		unhook_pending_command(new_command, current_connection);
		break;
		}	/* switch */

	/* Added a DEBUG message for investigating commands sent over different
		connections - SAI */
	TRACE(TRACE_DEBUG, "attach_pending command %p connection %p ###\n", new_command, current_connection);

	iscsi_cmd->exp_stat_sn = htonl(current_connection->exp_stat_sn);

	/* now compute and add the header digest if we are using them */
	if( current_connection->connection_flags & USE_HEADERDIGEST )
		{
		do_crc((unsigned char *)iscsi_cmd, ISCSI_HDR_LEN, 3,
													&iscsi_cmd->header_digest);
		new_command->iov->iov_len += CRC_LEN;
		new_command->tx_size += CRC_LEN;
		TRACE(TRACE_ISCSI_FULL, "Attach HeaderDigest\n");
		}

	/* now actually add this new command to end of pending commands list */
	last_pending_command = current_connection->pending_commands;

	if( last_pending_command == NULL )
		{/* list is totally empty, make this the only command on it */
		TRACE(TRACE_DEBUG, "Attach first pending command, ITT %u\n",
													new_command->init_task_tag);
		current_connection->pending_commands = new_command;
		}
	else
		{/* skip to end of list, then add this command there */
		while( last_pending_command->next != NULL )
			{
			TRACE(TRACE_DEBUG, "Skip pending command, ITT %u\n",
										last_pending_command->init_task_tag);
			last_pending_command = last_pending_command->next;
			}
		TRACE(TRACE_DEBUG, "Skip pending command, ITT %u\n",
										last_pending_command->init_task_tag);
		last_pending_command->next = new_command;
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave attach_pending_command\n");
	}


/*	here to collapse two items into 1 if they are adjacent or overlap */
void collapse( struct order_range *here, struct order_range *next )
	{
	if( next != NULL  &&  here->limit >= next->offset )
		{/* ranges are adjacent or overlap */
		if( here->limit >= next->limit )
			{/* here range completely subsumes next range */
			TRACE(TRACE_ISCSI,"range [%u..%u] subsumes [%u..%u]\n",
						here->offset, here->limit, next->offset, next->limit);
			}
		else
			{/* upper end of here range overlaps lower end of next range */
			if( here->limit > next->offset )
				{/* really have overlap, not just adjacency */
				TRACE(TRACE_ISCSI,"range [%u..%u] overlaps [%u..%u]\n",
						here->offset, here->limit, next->offset, next->limit);
				}
			else
				{/* have adjacency */
				TRACE(TRACE_ISCSI_FULL,"range [%u..%u] precedes [%u..%u]\n",
						here->offset, here->limit, next->offset, next->limit);
				}
			here->limit = next->limit;
			}
		here->next = next->next;
		my_kfree((void **)&next, "order range");
		collapse(here, here->next);
		}
	}


void merge_offset_length( struct order_range *head, unsigned int new_offset,
													unsigned int new_length )
	{
	struct order_range	*ptr, *prev, *newptr;
	unsigned int		new_limit;


	new_limit = new_offset + new_length;

	for( prev=NULL, ptr=head->next;  ptr != NULL;  prev=ptr, ptr=ptr->next )
		{
		if( new_offset < ptr->offset )
			{/* the new range starts below the ptr range,
				insert new item here in list, then try collapsing it */
			break;
			}
		else if( new_offset <= ptr->limit )
			{/* new range starts at or before end of ptr range */
			if( new_limit <= ptr->limit )
				{/* new range completely within ptr range */
				TRACE(TRACE_ISCSI,"range [%u..%u] subsumes [%u..%u]\n",
								ptr->offset, ptr->limit, new_offset, new_limit);
				}
			else
				{/* lower end of new range overlaps upper end of ptr range */
				if( new_offset == ptr->limit )
					{/* actually have adjacency */
					TRACE(TRACE_ISCSI_FULL,"range [%u..%u] precedes [%u..%u]\n",
								ptr->offset, ptr->limit, new_offset, new_limit);
					}
				else
					{/* true overlap */
					TRACE(TRACE_ISCSI,"range [%u..%u] overlaps [%u..%u]\n",
								ptr->offset, ptr->limit, new_offset, new_limit);
					}
				ptr->limit = new_limit;
				collapse(ptr, ptr->next);
				}
			return;
			}
		}
	/* if loop finishes, need to add a new range after prev, before ptr */
	if( (newptr = (struct order_range *)my_kmalloc(sizeof(
					struct order_range), "order range")) == NULL )
		{/* no memory for new range structure */
		return;
		}

	TRACE(TRACE_ISCSI_FULL,"new range [%u..%u]\n", new_offset, new_limit);

	newptr->offset = new_offset;
	newptr->limit = new_limit;
	newptr->next = ptr;
	if( prev == NULL )
		head->next = newptr;			/* first element on list */
	else
		prev->next = newptr;
	collapse(newptr, ptr);
	return;
	}


/*	Checks that range list covers the complete range it was supposed to cover.
	On entry, both head and head->next must NOT be NULL!
	Returns total number of bytes not covered by items in the range list.
*/
int check_range_list_complete( struct order_range *head )
	{
	struct order_range	*ptr, *prev;
	unsigned int		gap;
	int					missing;


	TRACE((TRACE_ENTER_LEAVE|TRACE_ISCSI_FULL),
		"Enter check_range_list_complete [%u..%u]\n",head->offset,head->limit);

	missing = 0;
	if( head->offset < head->next->offset )
		{/* first range starts later than it was supposed to start */
		gap = head->next->offset - head->offset;
		TRACE(TRACE_ISCSI,"gap of %u before first range [%u..%u]\n", gap,
										head->next->offset, head->next->limit);
		missing += gap;
		}
	for( ptr=head->next, prev=NULL;  ptr != NULL;  prev = ptr, ptr = ptr->next )
		{
		if( prev != NULL  &&  prev->limit < ptr->offset )
			{/* have a gap between the ranges of two list items */
			gap = ptr->offset - prev->limit;
			TRACE(TRACE_ISCSI,"gap of %u between range [%u..%u] and [%u..%u]\n",
					gap, prev->offset, prev->limit, ptr->offset, ptr->limit);
			missing += gap;
			}
		}
	if( prev->limit < head->limit )
		{/* last range ends before it was supposed to end */
		gap = head->limit - prev->limit;
		TRACE(TRACE_ISCSI,"gap of %u after last range [%u..%u]\n", gap,
													prev->offset, prev->limit);
		missing += gap;
		}

	TRACE((TRACE_ENTER_LEAVE|TRACE_ISCSI_FULL),
					"Leave check_range_list_complete, missing %d\n", missing);

	return missing;
	}


/*	frees all elements in a range list and then sets head->next to NULL */
void free_range_list( struct order_range *head )
	{
	struct order_range	*ptr, *next;


	for( ptr=head->next;  ptr != NULL;  ptr=next )
		{
		next = ptr->next;
		TRACE(TRACE_ISCSI_FULL,"free range [%u..%u]\n",ptr->offset,ptr->limit);
		my_kfree((void **)&ptr, "order range");
		}
	head->next = NULL;
	}
