/******************************************************************************
*
*	ref18_01/initiator/iscsi_initiator.c 
*
*	This file contains the core functions for iscsi initiator code that are 
*	responsible for connecting and maintaining sessions between initiator 
*	and target. The object file generated is loadable kernel module.
*
*   IMPORTANT: READ BEFORE DOWNLOADING, COPYING, INSTALLING OR USING.
*   By downloading, copying, installing or using the software you agree to this
*   license. If you do not agree to this license, do not download, install,
*   copy or use the software.
*
*   License Agreement
*
*   Copyright (C) 2001  Inter-Operability Lab.(IOL),Univ. of New Hampshire(UNH),
*                       Durham, NH 03824
*
*   Redistribution and use in source and binary forms, with or without
*   modification, are permitted provided that the following conditions are met:
*
*   -Redistributions of source code must retain the above copyright notice,
*   this list of conditions and the following disclaimer.
*
*   -Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
*
*   The name of IOL/UNH may not be used to endorse or promote products derived
*   from this software without specific prior written permission.
*
*   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
*   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
*   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL IOL,UNH OR CONTRIBUTORS
*   BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF
*   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
*   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*   POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************/

#include "iscsi_initiator.h"

/* chap support - CHONG */

/*	max number of redirects to follow during login on any connection */
#define	MAX_REDIRECTION_RETRIES	2

/*	Used for all access to the shared structures, and during the operation
	of any single thread to prevent things from disappearing from under
	the thread.
*/
spinlock_t	my_lock = SPIN_LOCK_UNLOCKED;

#include "initiator_utilities.h"

/*	bits in session_flags */
#define NEED_TX_WAKEUP		0x01

#include "initiator_error_rec.h"

/*	routine to pack an ordinary lun number into an 8-byte LUN structure
	(see SAM-2, Section 4.12.3 page 39)
	Thanks to Bill Conway of EMC for information about this.
*/
unsigned long long pack_lun( unsigned int lun, unsigned int flags )
	{
	unsigned long long	result;


	/* Intel CPU chips are little-endian */
	result = ((lun & 0xff) << 8);		/* LSB of lun into byte 1 big-endian */

	if( flags & USE_FLAT_SPACE_LUN )
		{/* use flat space addressing method, SAM-2 Section 4.12.4
			-	high-order 2 bits of byte 0 are 01
			-	low-order 6 bits of byte 0 are MSB of the lun
			-	all 8 bits of byte 1 are LSB of the lun
			-	all other bytes (2 thru 7) are 0
		*/
		result |= 0x40 | ((lun >> 8) & 0x3f);
		}
	/* else use peripheral device addressing method, Sam-2 Section 4.12.5
			-	high-order 2 bits of byte 0 are 00
			-	low-order 6 bits of byte 0 are all 0
			-	all 8 bits of byte 1 are the lun
			-	all other bytes (2 thru 7) are 0
	*/

	return result;
	}


unsigned int unpack_lun( unsigned char *lun_ptr )
	{
	unsigned int	result, temp;


	/* Intel CPU chips are little-endian */
	result = *(lun_ptr+1);			/* LSB of lun from byte 1 big-endian */

	switch( temp=((*lun_ptr)>>6) )	/* high 2 bits of byte 0 big-endian */
		{
	case 0:	/* peripheral device addressing method, Sam-2 Section 4.12.5
			-	high-order 2 bits of byte 0 are 00
			-	low-order 6 bits of byte 0 are all 0
			-	all 8 bits of byte 1 are the lun
			-	all other bytes (2 thru 7) are 0
			*/
		if( *lun_ptr != 0 )
			{
			TRACE_ERROR("Illegal Byte 0 in LUN peripheral device addressing "
						"method %u, expected 0\n", *lun_ptr);
			}
		break;
	case 1: /* flat space addressing method, SAM-2 Section 4.12.4
			-	high-order 2 bits of byte 0 are 01
			-	low-order 6 bits of byte 0 are MSB of the lun
			-	all 8 bits of byte 1 are LSB of the lun
			-	all other bytes (2 thru 7) are 0
			*/
		result += ((*lun_ptr) & 0x3f) << 8;
		break;
	default:/* (extended) logical unit addressing */
		TRACE_ERROR("Unimplemented LUN addressing method %u, "
					"PDA method used instead\n", temp);
		break;
		}	/* switch */

	return result;
	}


/*	attributes for each opcode that can be received from target */
#define	ITT_OPTIONAL	0x0000
#define	ITT_REQUIRED	0x0100
#define	ITT_RESERVED	0x0200
#define ITT_MASK		0x0300
#define	F_MUST_BE_1		0x0400
#define ILLEGAL_REPLY	0x0800
#define DSL_MUST_BE_0	0x1000
#define LUN_MUST_BE_0	0x2000

/*	table of attributes (in byte 1) and related opcode (in byte 0)
	for each opcode that can be received from target
*/
static unsigned short	opcode_attributes[32] = {
/* 0x20	NopIn ==> NopOut */
						F_MUST_BE_1 | ITT_OPTIONAL | ISCSI_INIT_NOP_OUT,
/* 0x21	SCSI Response ==> SCSI Command */
			LUN_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x22	Task Management Response ==> Task Management Request */
		LUN_MUST_BE_0 | DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED |
		ISCSI_INIT_TASK_MGMT_CMND,
/* 0x23	Login Response ==>Login Request */
						ITT_REQUIRED | ISCSI_INIT_LOGIN_CMND,
/* 0x24	Text Response ==>Text Request */
						ITT_REQUIRED | ISCSI_INIT_TEXT_CMND,
/* 0x25	DataIn ==> SCSI Command */
					 	ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x26	Logout Response ==> Logout Request */
		LUN_MUST_BE_0 | DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED |
		ISCSI_INIT_LOGOUT_CMND,
/* 0x27 */							ILLEGAL_REPLY,
/* 0x28 */							ILLEGAL_REPLY,
/* 0x29 */							ILLEGAL_REPLY,
/* 0x2a */							ILLEGAL_REPLY,
/* 0x2b */							ILLEGAL_REPLY,
/* 0x2c */							ILLEGAL_REPLY,
/* 0x2d */							ILLEGAL_REPLY,
/* 0x2e */							ILLEGAL_REPLY,
/* 0x2f */							ILLEGAL_REPLY,
/* 0x30 */							ILLEGAL_REPLY,
/* 0x31	R2T ==> SCSI Command */
			DSL_MUST_BE_0 | F_MUST_BE_1 | ITT_REQUIRED | ISCSI_INIT_SCSI_CMND,
/* 0x32	Asynch ==> -none- */
						F_MUST_BE_1 | ITT_RESERVED,
/* 0x33 */							ILLEGAL_REPLY,
/* 0x34 */							ILLEGAL_REPLY,
/* 0x35 */							ILLEGAL_REPLY,
/* 0x36 */							ILLEGAL_REPLY,
/* 0x37 */							ILLEGAL_REPLY,
/* 0x38 */							ILLEGAL_REPLY,
/* 0x39 */							ILLEGAL_REPLY,
/* 0x3a */							ILLEGAL_REPLY,
/* 0x3b */							ILLEGAL_REPLY,
/* 0x3c */							ILLEGAL_REPLY,
/* 0x3d */							ILLEGAL_REPLY,
/* 0x3e */							ILLEGAL_REPLY,
/* 0x3f	Reject ==> -none- */
									LUN_MUST_BE_0 | F_MUST_BE_1 | ITT_RESERVED
											};


void close_connection( struct connection *conn );


/*	add padding bytes, if necessary, to buffer which already has size bytes */
/*	update the command's iov and msghdr to point at this buffer */
/*	then compute and add data digest, if necessary */

/*	Draft 18, Section 9.1 iSCSI PDU Length and Padding
	"iSCSI PDUs are padded to the closest integer number of
	four byte words.  The padding bytes SHOULD be sent as 0."
*/
void add_padding( struct connection *current_connection,
				  struct command *new_command, char *buffer, int size )
	{
	int				n;
	struct iovec	*iov;


	if( (n = (size & 3)) != 0 )
		{
		TRACE(TRACE_ISCSI, "Attach %u pad bytes\n", 4-n);
		while( n++ < 4 )
			{/* the pad bytes should be 0 */
			*(buffer+size) = '\0';
			size++;
			}
		}

	/* get ptr to i/o vector which is already set up for header */
	iov = new_command->iov;
	iov++;
	iov->iov_base = buffer;
	iov->iov_len = size;
	new_command->tx_size += size;
	new_command->tx_msghdr.msg_iovlen++;

	/* compute the final data digest if we are using them */
	if( current_connection->connection_flags & USE_DATADIGEST )
		{
		do_crc(iov->iov_base, iov->iov_len, 3, &new_command->data_checksum);
		iov++;
		iov->iov_base = &new_command->data_checksum;
		iov->iov_len = CRC_LEN;
		new_command->tx_size += CRC_LEN;
		new_command->tx_msghdr.msg_iovlen++;
		TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
		}
	}


/***************************************************************************
*	Called only from iscsi_initiator_detect().
*	Initializes the hostdata structure.
*	Creates the parameter table for this hostdata.
*	Returns non-NULL pointer to parameter table if ok, NULL if error.
***************************************************************************/
static void *init_initiator( struct iscsi_hostdata *hostdata )
	{
	TRACE(TRACE_ENTER_LEAVE, "Enter init_initiator, hostdata %p\n", hostdata);

	/*	Initialize all fields to 0 or NULL */
	memset(hostdata, 0, sizeof(struct iscsi_hostdata));

	/*	Then allocate space for a copy of the parameter table */
	if( (hostdata->param_tbl = my_kmalloc(MAX_CONFIG_PARAMS*
						sizeof(struct parameter_type), "param_tbl")) == NULL )
		{
		goto out;
		}

	/* Initialize the session recovery related parameters - SAI */
	hostdata->init_sessrec_flg = 0;

    /*	Initialize parameter list to a copy of the default set of parameters */
	param_tbl_cpy(*hostdata->param_tbl, config_params);

	/* chap support - CHONG */
	hostdata->chap_parameter.local_ctx = CHAP_InitializeContext();
	hostdata->chap_parameter.peer_ctx = CHAP_InitializeContext();
		
	/*	Initialize the draft number to the default draft value */
	hostdata->draft_number = DEFAULT_DRAFT;
	TRACE(TRACE_DEBUG, "default hostdata->draft_number %u\n", DEFAULT_DRAFT);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave init_initiator, param_tbl %p\n",
														hostdata->param_tbl);

	return hostdata->param_tbl;
	}


/***************************************************************************
*	Called by the scsi subsystem when the iscsi_initiator module is loaded.
*	This allows the iscsi_initiator driver code to register itself with the
*	SCSI Mid-level as being the driver for this type of host adapter.
*	It also registers space for iscsi_hostdata struct with the SCSI Mid-level.
*	It returns:	 n > 0 if ok, where n is number of host adapters of this type
*							on this system (all will be driven by this code);
*				-1 on error.
***************************************************************************/

int iscsi_initiator_detect(Scsi_Host_Template * tmpt)
	{
	int		retval = 1;
	struct Scsi_Host *host = NULL;


	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_detect, tmpt %p\n", tmpt);

	/*	scsi_register() is defined in linux/drivers/scsi/hosts.c */
	host = scsi_register(tmpt, sizeof(struct iscsi_hostdata));

	TRACE(TRACE_DEBUG, "Back from scsi_register, host %p\n", host);

	if( host == NULL )
		{
		retval = -1;
		goto out;
		}

	TRACE(TRACE_ISCSI, "registered iSCSI host %p, host number %d\n",
														host, host->host_no);

	/*	setup the r2t_cookie management list */
	setup_r2t_cookies();

	/*	three standard default values which can be overridden:
		host->max_channel = 0;
		host->max_id = 8;
		host->max_lun = 8;
	*/

	/*	set the number of allowed "targets" this adaptor "card" can handle. */
	/*	the "target" numbers will be [0..max_id-1]. */
	/*	each "target" will have LUN numbers in [0..max_lun-1]. */
	host->max_id = MAX_TARGET_IDS;
	host->max_lun = MAX_LUNS;

	/*	scsi_register() allocated space for our private structure at the end */
	/*	of the host structure. */
	/*	This global variable is cheating! */
	/*	Potentially each (physical) adaptor card can have its own hostdata. */
	/*	We get away with having a single global variable here because */
	/*	we know we will never have more than 1 adaptor of this type ever. */
	global_hostdata = (struct iscsi_hostdata *)host->hostdata;
	global_host = host;

	/* Initialize the adapter's private data structure */
	if( init_initiator(global_hostdata) == NULL )
		{
		TRACE(TRACE_ISCSI, "unregister iSCSI host %p, host number %d\n",
														host, host->host_no);
		scsi_unregister(host);
		global_hostdata = NULL;
		global_host = NULL;
		retval = -1;
		goto out;
		}

	TRACE(TRACE_DEBUG, "Register iSCSI host adapter \"%s\"\n", tmpt->proc_name);

out:
	TRACE(TRACE_ENTER_LEAVE,"Leave iscsi_initiator_detect, retval %d\n",retval);

	return retval;
	}


/*	Called to send a NopOut to target as:
		a new nop (nopin_cmd == NULL, size == 0 )
	or
		a new ping (nopin_cmd == NULL, size != 0 )
	or
		a response to a NopIn (nopin_cmd != NULL).
	Buffer points to data to send, and size is number of bytes of data to send.
	If size > 0 then better have buffer != NULL.
	Returns 1 if all ok, else 0.
*/
int tx_nopout( struct connection *current_connection, __u32 i_bit,
				struct iscsi_targ_nopin *nopin_cmd, char **buffer, int size )
	{
	int							need_itt, retval = 1;
	struct command				*new_command;
	struct iscsi_init_nopout	*nopout_cmd;


	TRACE(TRACE_ENTER_LEAVE, "Enter tx_nopout, size = %d\n", size);

	if( (new_command = setup_command(NULL, current_connection->my_session))
																	== NULL )
		{/* will not be able to send the nop */
		retval = 0;
		goto out;
		}

	/*	set up the NopOut header (Draft 18, Section 9.18) to be sent to the
		target */
	nopout_cmd = (struct iscsi_init_nopout *)&new_command->iscsi_cmd;
	nopout_cmd->opcode = ISCSI_INIT_NOP_OUT | i_bit;
	nopout_cmd->flags = F_BIT;					/* always set on NopOut */

	if( size > 0 )
		{/* we have data to send to target */
		nopout_cmd->length = htonl(size);

		/* now add padding bytes and data digest, if necessary */
		add_padding(current_connection, new_command, *buffer, size);

		/* mark that buffer will be freed in tx_thread after NopOut is sent */
		new_command->buffer_to_free = *buffer;
		*buffer = NULL;
		}
	/* else no data to send to target (DSL field initialized to 0) */

	/*	Draft 18, Section 9.18.2 Target Transfer Tag
		"The NOP-Out MUST only have the Target Transfer Tag set if it is
		issued in response to a NOP-In with a valid Target Transfer Tag.
		In this case, it copies the Target Transfer Tag from the NOP-In PDU.
		Otherwise, the Target Transfer Tag MUST be set to 0xffffffff.

		When the Target Transfer Tag is set to a value other than 0xffffffff,
		the LUN field MUST also be copied from the NOP-In."
	*/
	if( nopin_cmd != NULL )
		{/* this is a response to a NopIn from the target */
		nopout_cmd->target_xfer_tag = nopin_cmd->target_xfer_tag;
		nopout_cmd->lun = nopin_cmd->lun;
		need_itt = 0;		/* need to set ITT to 0xffffffff */
		}
	else
		{/* this is a new nop or nop ping command to the target */
		nopout_cmd->target_xfer_tag = ALL_ONES;	/* LUN is 0 by default */

		/*	Draft 18, Section 9.18.1 Initiator Task Tag
			"The NOP-Out MUST have the Initiator Task Tag set to a valid value
			only if a response in the form of NOP-In is requested (i.e., the
			NOP-Out is used as a ping request).  Otherwise, the Initiator Task
			Tag MUST be set to 0xffffffff."
		*/
		if( size == 0 )
			need_itt = 0;
		else
			need_itt = 1;		/* set ITT to next in sequence */
		}

	/* can now attach the new command to end of pending commands list */
	/* this also sets the ITT, the CmdSN and the ExpStatSN */
	attach_pending_command(new_command, current_connection, need_itt);

	if( new_command->init_task_tag == ALL_ONES )
		{
		TRACE(TRACE_ISCSI, "Send NopOut, ITT 0x%08x, %u bytes\n",
							new_command->init_task_tag, new_command->tx_size);
		}
	else
		{
		TRACE(TRACE_ISCSI, "Send NopOut, ITT %u, %u bytes\n",
							new_command->init_task_tag, new_command->tx_size);
		}
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_nopout(nopout_cmd);

	/*	get the tx thread started, if necessary */
	up(&current_connection->my_session->tx_sem);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave tx_nopout\n");

	return retval;
	}


void deal_with_tx_timer( unsigned long data );

/*	(re)start the tx_timer -- period must already be in ticks */
void restart_tx_timer( struct session *current_session, unsigned int period )
	{

	TRACE(TRACE_ENTER_LEAVE, "Enter restart_tx_timer, period %u\n", period);

	if( period > 0  &&  current_session->tx_timer != NULL  &&
									!timer_pending(current_session->tx_timer) )
		{/* we can start the timer for a non-zero period of time (in ticks) */
		TRACE(TRACE_DEBUG, "Start timer for %u ticks\n", period);
		current_session->tx_timer->expires = jiffies + period;
		current_session->tx_timer->data = (unsigned long)current_session;
		current_session->tx_timer->function = deal_with_tx_timer;
		add_timer(current_session->tx_timer);
		}
	else
		{/* else period is 0 or no timer or timer already going,
			don't start it now*/
		TRACE(TRACE_DEBUG, "Timer not started for %u ticks\n", period);
		}


	TRACE(TRACE_ENTER_LEAVE, "Leave restart_tx_timer\n");
	}


/*	Called by kernel when periodic session timer expires */
void deal_with_tx_timer( unsigned long data )
	{
	struct session	*this_session;


	TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_tx_timer\n");

	TRACE(TRACE_DEBUG, "Timer expired\n");

	this_session = (struct session *)data;

	/*	turn off the need tx thread wakeup bit */
	this_session->session_flags &= ~NEED_TX_WAKEUP;

	/*	get the tx thread started, if necessary */
	up(&this_session->tx_sem);

	/*	and keep the timer going */
	restart_tx_timer(this_session, this_session->nop_period*HZ);

	TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_tx_timer\n");

	}


/*	drives sending of a nop out to the target */
int drive_nopout( struct connection *conn, struct session *current_session,
					unsigned int i_bit, unsigned int want_data )
	{
	char			*buffer;
	int				size;


	TRACE(TRACE_ENTER_LEAVE, "Enter drive_nopout\n");

	if( !want_data )
		{/* don't want any data attached to this nop out */
		buffer = NULL;
		if( i_bit )
			size = 0;		/* send it as immediate */
		else
			size = -1;		/* send it as normal, which forces empty ping */
		}

	/* else want data,  allocate a multiple of 4 bytes in case of padding */
	else if( (buffer = my_kmalloc(32, "NopOut data")) == NULL )
		{/* couldn't get any buffer space (should never happen) */
		size = 0;
		}

	else
		{/* send the '\0' at the end of the string too */
		size = 1 + sprintf(buffer, "NopOut_Ping_Number=%u",conn->nop_counter++);

		TRACE(TRACE_DEBUG, "Send \"%s\", size %d\n", buffer, size);
		}

	tx_nopout(conn, i_bit, NULL, &buffer, size);
	
	TRACE(TRACE_ENTER_LEAVE, "Leave drive_nopout\n");

	return 0;
	}

/*	drives the discovery session from within the tx_thread */
/*	returns 1 on success, 0 on failure */
int drive_text_negotiate( struct connection *conn,
								struct session *current_session, int discovery )
	{
	struct command				*new_command;
	struct iscsi_init_text_cmnd	*text_cmd;
	int							retval, size;
	char						*buffer;


	TRACE(TRACE_ENTER_LEAVE, "Enter drive_text_netotiate\n");

	if( (new_command = setup_command(NULL, current_session)) == NULL )
		{/* will not be able to send the text request */
		retval = 0;
		goto out;
		}

	/*	set up the Text Request header (Draft 18, Section 9.10) to be sent to
		the target */
	text_cmd = (struct iscsi_init_text_cmnd *)&new_command->iscsi_cmd;
	text_cmd->opcode = ISCSI_INIT_TEXT_CMND;
	text_cmd->flags = F_BIT;					/* we have only 1 command */
	text_cmd->target_xfer_tag = ALL_ONES;		/* reset the target */
												/* LUN is 0 by default */

	if( discovery )
		{
		/* must allocate a multiple of 4 bytes in case of padding */
		if( (buffer = my_kmalloc(32, "Text data")) == NULL )
			{/* no space for the text request, free up the command structure */
			my_kfree((void **)&new_command, "command");
			retval = 0;
			goto out;
			}

		/* save ptr to this buffer so we can free it in rx_text_rsp when done */
		new_command->buffer_to_free = buffer;

		/* send the '\0' at the end of the string too */
		size = 1 + sprintf(buffer, "SendTargets=%s",
					current_session->draft_number <= DRAFT9 ? "all" : "All");

		text_cmd->length = htonl(size);

		/* now add padding bytes and data digest, if necessary */
		add_padding(conn, new_command, buffer, size);
		}

	/* can now attach the new command to end of pending commands list */
	/* this also sets the ITT, the CmdSN and the ExpStatSN */
	attach_pending_command(new_command, conn, 1);

	TRACE(TRACE_ISCSI, "Send Text, ITT %u, %u bytes\n",
							new_command->init_task_tag, new_command->tx_size);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_text_cmnd(text_cmd);

	/*	get the tx thread started, if necessary */
	up(&current_session->tx_sem);

	/*	return success */
	retval = 1;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave drive_text_netotiate, retval = %d\n", retval);

	return retval;
	}


/*	drives the logout of a session from within the tx_thread */
/*	returns 1 on success, 0 on failure */
/* Modified to include connection logout - SAI */

int drive_logout( struct session *current_session, int cid, int reason )
	{
	struct command					*new_command;
	struct iscsi_init_logout_cmnd	*logout_cmd;
	int								retval;
	struct session					*sess;
	struct connection				*conn;
	int	result;


	TRACE(TRACE_ENTER_LEAVE, "Enter drive_logout\n");

	if( (new_command = setup_command(NULL, current_session)) == NULL )
		{/* will not be able to send the logout request */
		retval = 0;
		goto out;
		}

	/* For connection logout use the cid to find the appropriate connection
	   to logout - SAI */
	if (reason == 2)
		result = find_connection(current_session->scsi_target_id, cid, 
								global_hostdata, &conn, &sess);
	else
		conn = current_session->connection_head;

	/*	set up the Logout Request header (Draft 18, Section 9.14) to send to
		the target */
	logout_cmd = (struct iscsi_init_logout_cmnd *)&new_command->iscsi_cmd;
	logout_cmd->opcode = ISCSI_INIT_LOGOUT_CMND;
	logout_cmd->flags = F_BIT;					/* must be set to 1 on logout */
	if( current_session->draft_number < DRAFT13 )
    	logout_cmd->rsvd3 = reason;            /* use the reason code - SAI*/
	else
    	logout_cmd->flags |= (reason & LOGOUT_REASON); /* use the reason code - SAI*/

	conn->connection_state = STATUS_LOGGED_OUT; /* Make sure that no one uses
								a logged-out connection - SAI */
									
	/* Set Time Stamp for logout Retries - SAI */
	new_command->time_stamp = jiffies;

	/*	leave reasoncode field set to 0 to mean "close session"
		see Draft 18, Section 9.14.1 Reason Code */

	/* can now attach the new command to end of pending commands list */
	/* this also sets the ITT, the CmdSN and the ExpStatSN */
	attach_pending_command(new_command, conn, 1);

	TRACE(TRACE_ISCSI, "Send Logout, ITT %u, %u bytes\n",
							new_command->init_task_tag, new_command->tx_size);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_logout_cmnd(logout_cmd);

	/*	get the tx thread started, if necessary */
	up(&current_session->tx_sem);

	/*	return success */
	retval = 1;

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave drive_logout, retval = %d\n", retval);

	return retval;
	}


/****************************************************************************
* The tx thread is woken up anytime someone has something 
* that needs to be sent.  each time it is woken up it searches through the
* command table for stuff that it needs to send.  one problem with the current
* implementation is that a send blocking will block up everything.
* There is single thread for each session.
*******************************************************************************/

void iscsi_initiator_tx_thread(void * arg)
	{
	mm_segment_t		oldfs;
	struct session		*current_session = (struct session *) arg;
	struct connection	*current_connection;
	int					size, sent;
	int					hold_lock = -1;		/* 0 if lock held; -1 if not */
	struct iovec		*iov;
	struct command		*current_command;
	unsigned int		lun;
	int					serial_no=0;
	struct init_error_rec	err_rec;
	

	TRACE(TRACE_ISCSI, "tx_thread starting, session %p\n", current_session);
	
	lock_kernel();
	
	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGS);
	
	daemonize();
	
	/*	NOTE: in struct task_struct have declaration char comm[16] */
	strcpy(current->comm, "iscsi_tx_thread");
	
	current_session->tx_thread = current;
	
	unlock_kernel();
	
	/* if NopOut pings to target are to be sent periodically, start timer now */
	restart_tx_timer(current_session, current_session->nop_period*HZ);

    if (current_session->oper_param->ErrorRecoveryLevel > 0
        && current_session->retran_period > 0
        && current_session->xmit_chk_interval > 0)
        {
        /* start timer for command retransmissions - SAI */
        restart_retran_timer(current_session, current_session->xmit_chk_interval*HZ);
        }

	/*	signal anyone waiting that the tx thread is up and running */
	up(&current_session->tx_done_sem);

	TRACE(TRACE_ISCSI, "tx_thread started\n");
	
	while( 1 )
		{
		TRACE(TRACE_SEM, "tx_thread blocked on tx_sem %p\n", current_session);
		if( down_interruptible(&current_session->tx_sem) < 0 )
			break;					/* did not get the sem but got signalled */
		TRACE(TRACE_SEM, "tx_thread unblocked on tx_sem %p\n", current_session);

		/* now get exclusive access to the iscsi structures */
		spin_lock(&my_lock);
		hold_lock = 0;

		/* Free Unused Connections - SAI */
		clear_free_conn(current_session);

		for( current_connection = current_session->connection_head;
			 current_connection != NULL;
			 current_connection = current_connection->next )
			{
			TRACE(TRACE_DEBUG, "tx_thread for connection %p\n", 
															current_connection);

			/* if timer went off and this connection has had no activity since
			   the last timer then send a NopOut ping to the target */
			if( current_session->nop_period > 0  &&
								current_connection->pending_commands == NULL )
				{/* timer is going and nothing pending on this connection */
				if( !(current_connection->connection_flags & GOT_ACTIVITY) )
					{/* no activity since last timer timed out, send nop ping */
					drive_nopout(current_connection, current_session, 0, 1);
					}
				else
					{/* had activity, turn off the flag to start next period */
					current_connection->connection_flags &= ~GOT_ACTIVITY;
					}
				}

			for( current_command = current_connection->pending_commands;
				 current_command != NULL;
				 current_command = current_command->next )
				{
				TRACE(TRACE_DEBUG, "tx_thread for command %p\n",
															current_command);

				/* Only Logout command allowed on a logged-ut connection - SAI */
				if (current_connection->connection_state != STATUS_LOGGED_IN)
					{
					if (current_command->iscsi_cmd.opcode != ISCSI_INIT_LOGOUT_CMND)
						continue;
					}
				if(current_connection->sock == NULL)
					continue;
					
			/* Should tx_thread be dependant on rx_thread ?? - SAI */
			/*
				if( current_connection->rx_thread == NULL )
					goto out;
			*/

			/* If there is no transport connections, quit the tx_thread too - SAI */
				if (current_session->nconnections < 1)
					goto out;

                /* Check if command re-transmission required - SAI */

                if (current_session->oper_param->ErrorRecoveryLevel > 0
                    && current_session->retran_period > 0
                    && current_session->xmit_chk_interval > 0)
                    {
					/* If timer expired and no activity, assume command was lost and re-transmit command - SAI */
                    if (jiffies > (current_command->time_stamp + ((current_session->retran_period)*HZ)) )
						{
						if (!current_command->activity_flg)
							{
							if (current_command->tx_size == 0)
								{
								TRACE(TRACE_DEBUG,  "Re-Transmitting lost Command PDU\n");
								current_command->tx_size = current_command->retransmit_tx_size;
								}
							}
						else
							{/* for read commands, the data_in or the scsi response could be lost due to packet drops
								because of bad connection or during connection recovery. This function pro-actively
								fills the gap in data-in pdus and scsi responses. This uses the command re-transmit
								timer for checking and filling gaps - SAI */
							if (current_command->SCpnt)
								{
								if(current_command->SCpnt->sc_data_direction == SCSI_DATA_READ )
									{
									err_rec.curr_conn = current_connection;
									err_rec.related_cmd = current_command;

                                    /* take care of out-of-order PDUs too - new 18_04 SAI */
                                    if (current_session->oper_param->DataSequenceInOrder &&
                                        current_session->oper_param->DataPDUInOrder)
										{
										if (current_command->data_offset != current_command->SCpnt->request_bufflen)
											{
											serial_no = current_command->data_in_sn;
											init_send_snack(&err_rec, serial_no, 0, DATA_R2T_SNACK);
											}
										}
                                    else
                                        {
                                        if (check_range_list_complete(&current_command->r2t_range_list) > 0)
                                            {
                                            serial_no = current_command->data_in_sn;
                                            init_send_snack(&err_rec, serial_no, 0, DATA_R2T_SNACK);
                                            }
                                        }
									}
								serial_no = current_connection->exp_stat_sn;
								init_send_snack(&err_rec, serial_no, 1, STATUS_SNACK);
   		                     	/* Timestamp the command transmission time - SAI */
								current_command->time_stamp = jiffies;
								}
							}
						}
					}

                if (current_command->tx_size == 0)
                    continue;

				/* have a command that needs to be sent on this connection */
				do
					{
					size = current_command->tx_size;
					iov = current_command->iov;

					TRACE(TRACE_DEBUG, "--send %d\n", size);
	
					/* allow other threads to run while we send */
					spin_unlock(&my_lock);
					hold_lock = -1;

					oldfs = get_fs(); set_fs(get_ds());
					sent = sock_sendmsg(current_connection->sock, 
										&current_command->tx_msghdr, size);
					set_fs(oldfs);

					TRACE(TRACE_DEBUG,"--sent %d\n", sent);

					/* now get back exclusive access to the iscsi structures */
					spin_lock(&my_lock);
					hold_lock = 0;

					if( sent < size )
						{/* did not send what we expected to send, why? */

						if( current_session->tx_thread==NULL || sent==-EINTR )
							{/* thread was killed as part of session shutdown */
							goto out;
							}

						TRACE(TRACE_ISCSI, "sock_sendmsg returned %d, "
												"expected %d\n", sent, size);

						if( sent < 0  &&  sent != -EAGAIN )
							{/* hard error -- report it, then ignore it! */
							TRACE_ERROR("sock_sendmsg error %d\n", sent);
							}

						else
							{/* sent could now be -EAGAIN or value >= 0.
								in either case need to try again later */

							if( sent > 0 )
								{/* compute amount left to send later */
								current_command->tx_size -= sent;
								}

							/* update the iov to send rest later */
							while( sent > 0 )
								{
								if( iov->iov_len <= sent )
									{/* all of this iov already sent */
									sent -= iov->iov_len;
									iov->iov_len = 0;
									iov++;
									}
								else
									{/* part of this iov not sent yet */
									iov->iov_base += sent;
									iov->iov_len -= sent;
									sent = 0;
									}
								}
							
							/* iov adjusted (as needed), try again later */
							current_session->session_flags |= NEED_TX_WAKEUP;
							restart_tx_timer(current_session,TX_RESTART_PERIOD);

							/* do nothing more on this command or connection */
							current_command = NULL;
							break;
							}
						}
		
                    TRACE(TRACE_DEBUG, "Header opcode 0x%.2x flags 0x%.2x\n",
												current_command->header_opcode,
												current_command->header_flags);

					if( (current_command->header_opcode & ISCSI_OPCODE) ==
													ISCSI_INIT_SCSI_CMND )
						{/* this was a SCSI Request */
						TRACE(TRACE_DEBUG,
							"Sent CDBopcode 0x%.2x cmd_sn %d "
							"init_task_tag %d exp_stat_sn %d length %d\n",
							current_command->iscsi_cmd.cdb[0],
							ntohl(current_command->iscsi_cmd.cmd_sn),
							ntohl(current_command->iscsi_cmd.init_task_tag),
							ntohl(current_command->iscsi_cmd.exp_stat_sn),
							current_command->tx_size);

                        /* Timestamp the command transmission time - SAI */
                        current_command->time_stamp = jiffies;
						}

                    /* Store the tx_size for command re-transmissions - SAI */
                    current_command->retransmit_tx_size = current_command->tx_size;
					
					/*	Mark that current command has been sent.  This allows
						new transfers to be set up in this command slot. */
					current_command->tx_size = 0;
				
                    if( !(current_command->header_flags & F_BIT) )
                        {/* more DataOut PDUs to be sent */
                        setup_dataoutpdu(current_connection, current_command);
                        }
					else if( (current_command->header_opcode & ISCSI_OPCODE) ==
													ISCSI_INIT_SCSI_DATA_OUT )
						{/* this was the last DataOut PDU in an R2T sequence */
						if( free_r2t_cookie(current_command) != NULL )
							{/* send first DataOut PDU of next R2T */
							setup_dataoutpdu(current_connection,
															current_command);
							}
						}
					else if( (current_command->header_opcode & ISCSI_OPCODE) ==
															ISCSI_INIT_NOP_OUT )
						{/* this was the last (only) NopOut PDU sent */
	    				if( current_command->init_task_tag == ALL_ONES )
							{/* this was a ping response to target's NopIn */

							/* Free up any locally allocated data buffer */
							my_kfree(&current_command->buffer_to_free,
																"NopOut data");

							/*	Now free up the command itself since we will
								get no further response from target. */
							if( (current_command = free_pending_command(
								current_command, current_connection)) == NULL )
								break;
							}
						}
						/* Freeing SNACK Request from pending commands list - SAI */
					else if ( (current_command->header_opcode & ISCSI_OPCODE) == ISCSI_INIT_SNACK)
						{
						if ( (current_command = free_pending_command(current_command, current_connection)) == NULL)
							break;
						}
	
                    } while( current_command->tx_size != 0 );

				if( current_command == NULL )
					break;
                }   /* end of for command loop */

			} /* end of for connection loop */

		/* release our exclusive access to the iscsi structures */
		spin_unlock(&my_lock);
		hold_lock = -1;
		}	/* end of while loop */

out:

	TRACE(TRACE_ISCSI, "tx_thread exiting\n");

	/* get exclusive access to iscsi structures if we don't already have it */
	if( hold_lock < 0 )
		spin_lock(&my_lock);

	if( current_session->tx_thread != NULL )
		{/* no other thread is already shutting this session down */

		/* first indicate that this thread is shutting itself down */
		current_session->tx_thread = NULL;

		/* then release our exclusive access to the iscsi structures */
		spin_unlock(&my_lock);

		/* remove all valid luns in session (last lun will take out session) */
		for( lun = 0;  lun < global_host->max_lun;  lun++ )
			{
			if( current_session->lun_bits & (1<<lun) )
				{
				if(remove_session_from_hostdata(current_session->scsi_target_id,
										lun, ALL_ONES, global_hostdata) == 1 )
					{/* just removed the whole session structure */
					break;
					}
				}
			}

		/* get back exclusive access to the iscsi structures */
		spin_lock(&my_lock);
		}
	/* else some other thread is already shutting this session down */

	/*	signal anyone waiting that the tx thread is finished */
	up(&current_session->tx_done_sem);

	/* finally, release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ISCSI, "tx_thread exited\n");
	}
	

/****************************************************************************
* The rx thread exists for each connection.  the thread sits
* blocked on recvmsg until it can successfully read an iscsi header.  at
* this point it passes off everything to the appropriate function based
* on the type of header it received.  this function processes the rest of
* the frame.  this way data can be read right into the correct buffers.
****************************************************************************/
void iscsi_initiator_rx_thread(void * arg)
	{
	struct connection	*current_connection;
	struct session		*current_session;
	unsigned int		size, i=0;
	int					received;
	unsigned int		checksum, local_itt;
	unsigned int		opcode, attributes;
    struct command		*related_command;
	struct init_error_rec	err_rec;
	char				*buffer;
	struct response_pdu	*header;

	
	current_connection = (struct connection *) arg;
	
	TRACE(TRACE_ISCSI,"rx_thread starting, connection %p\n",current_connection);
	
	current_session = current_connection->my_session;

	/* Lock kernel to daemonize the thread and update the rx_thread pointer */
	lock_kernel();
	
	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGS);
	
	daemonize();
	
	/*	NOTE: in struct task_struct have declaration char comm[16] */
	strcpy(current->comm, "iscsi_rx_thread");
	
	current_connection->rx_thread = current;
	
	unlock_kernel();
	
	/*	signal anyone waiting that the rx thread is up and running */
	up(&current_connection->rx_done_sem);

	TRACE(TRACE_ISCSI, "rx_thread started\n");
	
	/*	the rx_thread normally holds exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	received = 1;
	while( received > 0 )
		{
		size = ISCSI_HDR_LEN;
		if( current_connection->connection_flags & USE_HEADERDIGEST )
			size += CRC_LEN;

		/*	a hole in our exclusive access opens up while waiting in here */
		if( (received = recv_a_message(current_connection, size,
											current_connection->rx_buf)) <= 0 )
			{/* connection died or we got a kill signal, terminate this thread*/
			/* Check if connection recovery is supported - SAI */
			if (current_connection->my_session->oper_param->ErrorRecoveryLevel > 1)
				{/* Perform connection recovery */
				deal_with_connrec((unsigned long)current_connection);
				}
			break;
			}
			
		header = (struct response_pdu *)current_connection->rx_buf;

		/*	pick up number of bytes of data attached to this PDU (if any) */
		/*	plus the AHS in the high-order byte */
		size = ntohl(header->length);

		/* we do not yet handle AHS, so appearance of one is fatal */
		if( size > 0xffffff )		/* 2^24 - 1 */
			{
			TRACE_ERROR("TotalAHSLength 0x%02x, expected 0x00\n", size>>24);
			break;
			}

		/* mark that we got some activity from the target */
		current_connection->connection_flags |= GOT_ACTIVITY;

        /* Print the received iSCSI header */ 
		TRACE(TRACE_BUFF, "rx_thread got message\n");
		PRINT_BUFF(current_connection->rx_buf, ISCSI_HDR_LEN);
		
		if( current_session->session_flags & NEED_TX_WAKEUP )
			{/* turn off the need tx thread wakeup bit */
			current_session->session_flags &= ~NEED_TX_WAKEUP;
			/* then get the tx thread started again */
			up(&current_session->tx_sem);
			}

		if( current_connection->connection_flags & USE_HEADERDIGEST )
			{/* header digests are in use, compute the crc and check it */
			do_crc((unsigned char *)current_connection->rx_buf, ISCSI_HDR_LEN,
																3, &checksum);
			if( checksum != header->header_digest )
				{/* crc check should have been == and was not */
				TRACE_ERROR("Header CRC wrong: got 0x%08x, expected 0x%08x\n",
							ntohl(header->header_digest), ntohl(checksum));
				/* Header Digest Error Recovery - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = NULL;
				err_rec.err_type = HEADER_DIGERR;
				init_do_error_recovery(&err_rec);
				continue;
				}
			}

		/*	pick up the opcode byte, with bits that are not in opcode itself */
        opcode = header->opcode;

		/*	Check that the old X bit (bit 7 in byte at offset 0) is 0 */
		if( opcode & OLD_X_BIT )
			{/* bit is set, this is bad in later drafts */
			if( current_session->draft_number > DRAFT8 )
				{
				TRACE(TRACE_ISCSI, "***Warning***: Response has Bit 0 of "
									"Byte 0 = 1, expected 0\n");
				}
			opcode &= ~OLD_X_BIT;	/* turn it off, so it has expected value */
			}
		else
			{/* bit is not set, this is bad in earlier drafts */
			if( current_session->draft_number <= DRAFT8 )
				{
				TRACE_ERROR("Response has X Bit = 0, expected 1\n");
				}
			}

		/*	Check the I bit (bit 6 in byte at offset 0):
			-- should NOT be set in later drafts > draft9
			-- must be set in earlier drafts <= draft9
		*/
		if( opcode & I_BIT )
			{/* bit is set, this is bad in later drafts */
			if( current_session->draft_number > DRAFT9 )
				{
				TRACE(TRACE_ISCSI,
						"***Warning***: Response has I Bit = 1, expected 0\n");
				}
			opcode &= ~I_BIT;		/* turn it off, leaving opcode only */
			}
		else
			{/* bit is not set, this is bad in earlier drafts */
			if( current_session->draft_number <= DRAFT9 )
				{
				TRACE_ERROR("Response has I Bit = 0, expected 1\n");
				}
			}

		/*	get attributes for this opcode from table */
		attributes = opcode_attributes[opcode & ~REPLY_BIT];

		/*	Check that this response opcode is legal, fatal if not */
		if( !(opcode & REPLY_BIT)  ||  (attributes & ILLEGAL_REPLY) )
			{
			TRACE_ERROR("Illegal response opcode 0x%x, %d bytes\n", 
															opcode, received);
			for( i = 0;  i < received;  i++ )
				{
				if( (i&7) == 0 )
					{
					if( i )
						PRINT("\n");
					PRINT("%4u:", i);
					}
				PRINT(" %02x",
						(unsigned char)(current_connection->rx_buf[i]));
				}
			PRINT("\n");
			received = 0;
			break;
			}

		/*	Check F bit (bit 7 in byte 1), most responses require it to be 1 */
        if( !(header->flags & F_BIT) )
			{/* F bit not set, this may still be ok but needs further check */
			if( attributes & F_MUST_BE_1 )
				{/* a mistake in draft 12 requires F=0 in SCSI RESPONSE PDUs */
				if( current_session->draft_number != DRAFT12  ||
												opcode != ISCSI_TARG_SCSI_RSP )
					{
					TRACE(TRACE_ISCSI,
						"***Warning***: Response has F Bit = 0, expected 1\n");
					}
				}
			}

		/*	ITT is in same position in all pdus, pick it up once here */
		local_itt =  ntohl(header->init_task_tag);

		/* Find the matching command to which this is a response */
		related_command = NULL;
		switch( attributes & ITT_MASK )
			{
		case ITT_RESERVED:
			if( current_session->draft_number <= DRAFT9 )
				{/* draft 9 and earlier, reserved value should be 0 */
				if( local_itt != 0 )
					{
					TRACE(TRACE_ISCSI,
					"***Warning***: Response has ITT = 0x%08x, expected 0\n",
					local_itt);
					}
				}
			else
				{/* drafts after draft 9, reserved value should be 0xffffffff */
				if( local_itt != ALL_ONES )
					{
					TRACE(TRACE_ISCSI,
					"***Warning***: Response has ITT = 0x%08x, expected 0x%x\n",
					local_itt, ALL_ONES);
					}
				}
			break;
		case ITT_OPTIONAL:
			if( local_itt == ALL_ONES )
				{/* ITT has this reserved value in all drafts */
				break;
				}
			/* fall thru to required */
		case ITT_REQUIRED:
			if( local_itt == ALL_ONES )
				{/* ITT has reserved value when it should not */
				TRACE(TRACE_ISCSI,
					"***Warning***: Response has ITT reserved value 0x%08x, "
					"should not\n", local_itt);
				break;
				}
			/* this ITT is not reserved, we should find a matching command */
			for( related_command = current_connection->pending_commands;
				 related_command != NULL; 
				 related_command = related_command->next )
				{
				TRACE(TRACE_DEBUG, "Check pending command ITT %u\n",
												related_command->init_task_tag);
				if( related_command->init_task_tag == local_itt )
					{
					related_command->activity_flg = GOT_ACTIVITY; /* Command Activity flag - SAI*/
				 	break;
					}
				}
			if( related_command == NULL )
				{
				TRACE_ERROR("No command has matching ITT %u\n", local_itt);
				}
			else if( (attributes & ISCSI_OPCODE) !=
							(related_command->iscsi_cmd.opcode & ISCSI_OPCODE) )
				{
				TRACE_ERROR("Command matching ITT %u has opcode "
						"0x%02x, expected 0x%02x\n", local_itt,
						related_command->iscsi_cmd.opcode & ISCSI_OPCODE,
						attributes & ISCSI_OPCODE);
				}
			break;
			}	/* switch */

		if( header->lun != 0LL )
			{/* the LUN is not 0, which may be ok, just need to check */
			if( attributes & LUN_MUST_BE_0 )
				{/* the LUN in this response should be reserved */
				TRACE(TRACE_ISCSI,"***Warning***: Response LUN is %s"
						", expected 0\n", string_llx(header->lun));
				}
			}

		/*	no buffer allocated yet for data attached to this pdu */
		buffer = NULL;

		TRACE(TRACE_DEBUG, "Response DSL %u\n", size);
			
		if( size > 0 )
			{/* a data segment is attached, is this ok? */
			if( attributes & DSL_MUST_BE_0 )
				{/* this response cannot have any data attached -- fatal */
				TRACE_ERROR("Response has DSL = %u, expected 0\n", size);
				break;
				}

			/*	If payload is greater than initiator's MaxRecvDataSegmentLength
				(or MaxRecvPDULength for drafts 9 through 12,
				or DataPDULength for drafts up through 8) it is an error */
			if( current_session->draft_number >= DRAFT13 )
				{/* drafts 13 and after */
				if( size>current_connection->max_recv_length )
					{
					TRACE_ERROR("DSL %u greater than initiator's "
						"MaxRecvDataSegmentLength %u\n", size,
						current_connection->max_recv_length);
					}
				}
			else if( current_session->draft_number >= DRAFT9 )
				{/* drafts 9 through 12 */
				if( size > current_connection->max_recv_length )
					{
					TRACE_ERROR("DSL %u greater than initiator's "
							"MaxRecvPDULength %u\n", size,
							current_connection->max_recv_length);
					}
				}
			else
				{/* drafts 8 and before */
				if( size > current_session->oper_param->DataPDULength )
					{
					TRACE_ERROR("DSL %u greater than negotiated "
							"DataPDULength %u\n", size,
							current_session->oper_param->DataPDULength);
					}
				}

			/*	for all commands except DataIn, read all data attached to
				the pdu into a local buffer.  This includes, if necessary,
				reading any pad bytes and the data digest, and checking
				the data data digest. */

			/* Better place to read the rejected PDU is rx_rjt - SAI*/

			if( opcode != ISCSI_TARG_SCSI_DATA_IN && opcode != ISCSI_TARG_RJT)
				{
				if( (received = recv_locally(0, size, current_connection, 1,
													&checksum, &buffer)) <= 0 )
					{
					/* If There is a Data Digest Error Handle appropriately - SAI */

					if (received == PAYLOAD_DIGERR)
						{
						err_rec.curr_conn = current_connection;
						err_rec.related_cmd = related_command;
						err_rec.err_type = PAYLOAD_DIGERR;
						init_do_error_recovery(&err_rec);
						}
					else
						{/* got error or killed by a signal */
						break;
						}
					}
				}
			}

		/*	Check StatSN field in all responses except DataIn with S=0 */
		if( opcode != ISCSI_TARG_SCSI_DATA_IN  ||  (header->flags & S_BIT) )
			{/*	check exp_stat_sn with stat_sn value received in the header */
			i =  ntohl(header->stat_sn);
			if( i != current_connection->exp_stat_sn )
				{
				TRACE_ERROR("Got StatSN %u, expected %u\n", i,
											current_connection->exp_stat_sn);

                /* for iscsi responses, trigger error recovery if the statsn does not match
                                                                    - new 18_04 - SAI */
                if (opcode == ISCSI_TARG_SCSI_RSP)
                    {
                    /* DataSN Sequence Error Recovery - SAI */
                    err_rec.curr_conn = current_connection;
                    err_rec.related_cmd = related_command;
                    err_rec.err_type = SEQUENCE_ERR;
                    init_do_error_recovery(&err_rec);
                    }
                }

			/*	the StatSN field advances for all responses in which it is
				checked, except R2T and NopIn with TTT not reserved */

			/*	Draft 18, Section 9.8.3 StatSN (under R2T)
				"The StatSN field will contain the next StatSN.  The StatSN for
				this connection is not advanced after this PDU is sent." */

			/*	Draft 18, Section 9.19.2 StatSN (under NOP-In)
				"The StatSN field will contain always the next StatSN.
				However, when the Initiator Task Tag is set to 0xfffffff StatSN
				for the connection is not advanced after this PDU is sent." */

			if( opcode != ISCSI_TARG_R2T  &&  (opcode != ISCSI_TARG_NOP_IN  ||
											header->init_task_tag != ALL_ONES) )
				{
				current_connection->exp_stat_sn++;
				}
			}

		/*	Check ExpCmdSN and MaxCmdSN fields in all responses */
		check_and_update_cmd_sns(header->exp_cmd_sn, header->max_cmd_sn,
															current_session);

        switch( opcode )
			{
			case ISCSI_TARG_NOP_IN:			/* NopIn PDU */
				received = rx_nopin(current_connection, local_itt,
												related_command, size, &buffer);
				break;
			case ISCSI_TARG_SCSI_RSP:		/* SCSIResponse PDU */
				received = rx_rsp(current_connection, local_itt,
												related_command, size, buffer);
				break;
            case ISCSI_TARG_TASK_MGMT_RSP:	/* TaskManagementResponse PDU */
                received=rx_task_mgt_rsp(current_connection, local_itt,
															related_command);
                break;
			case ISCSI_TARG_LOGIN_RSP:		/* LoginResponse PDU */
				TRACE_ERROR("Unexpected Login Response, status 0x%02x, "
								"%d bytes\n", header->status_class, received);
				received = 0;				/* fatal error */
				break;
			case ISCSI_TARG_TEXT_RSP:		/* TextResponse PDU */
				received = rx_text_rsp(current_connection, local_itt,
												related_command, size, buffer);
				break;
			case ISCSI_TARG_SCSI_DATA_IN:	/* DataIn PDU */
				received = rx_data(current_connection, local_itt,
														related_command, size);
				break;
			case ISCSI_TARG_LOGOUT_RSP:		/* LogoutResponse PDU */
				received = rx_logout_rsp(current_connection, local_itt,
															related_command);
				break;
			case ISCSI_TARG_R2T:			/* R2T PDU */
				received = rx_r2t(current_connection,local_itt,related_command);
				break;
			case ISCSI_TARG_ASYNC_MSG:		/* Asynchronous Message PDU */
				received = rx_async_msg(current_connection, size, buffer);
				break;
            case ISCSI_TARG_RJT:			/* Reject PDU */
                received = rx_rjt(current_connection);
                break;
			}	/* switch */

		/*	We are now finished with any locally allocated data buffer
			that is not still in use (i.e., being sent in NopOut). */
		my_kfree((void **)&buffer, "local buffer");
		}	/* while */

	/*	the rx_thread is finished, tell the world */
	TRACE(TRACE_ISCSI, "rx_thread exiting, received %d\n", received);

	/* get exclusive access to iscsi structures if we don't already have it */
	if( received < 0 )
		spin_lock(&my_lock);

	/* If connection recovery active, do not free pending commands - SAI */
	if (current_session->n_rec_tasks <= 0)
		{
		TRACE(TRACE_DEBUG, "Freeing pending commands\n");

		i = 0;
		while( (related_command = current_connection->pending_commands) != NULL )
			{
			i++;
			while( free_r2t_cookie(related_command) != NULL )
				{}
			if( related_command->SCpnt != NULL )
				{/* pass the final status back up to SCSI Mid-level */
				related_command->SCpnt->result = DID_ERROR << 16;
				/* should we call scsi_done() for this command? */
				}
			free_pending_command(related_command, current_connection);
			}
		}

	TRACE(TRACE_DEBUG, "%d pending commands freed\n", i);

	current_connection->connection_state = STATUS_DISCONNECTED;

	/* save the lun number of this connection in case we need it below */
	for( i = 0;  i < 32  && !(current_session->lun_bits & (1<<i));  i++ )
		{}

    /* NULL the rx_thread before exit new 18_04 - SAI */
    current_connection->rx_thread = NULL;

	/*	signal anyone waiting that the rx thread is finished */
	up(&current_connection->rx_done_sem);

	/*	release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ISCSI, "rx_thread exited\n");

	if( current_session->oper_param->SessionType )
		{/* this is a discovery session, shut it down */
		remove_session_from_hostdata( current_session->scsi_target_id, i,
							current_connection->connection_id, global_hostdata);
		}
	}


/*	utility function called by all rx___() functions when they determine that
*	their reserved fields are not all 0
*/
void reserved_not_0( char *pdu_name )
	{
	TRACE(TRACE_ISCSI, "***Warning***: Reserved fields in %s"
						"header not all 0\n", pdu_name);
	}


/****************************************************************************
* This function processes the DataIn payload of DataIn PDU. It fills the 	
* scatter-gather buffers provided by the SCSI Mid-level. The data digest
* is also checked (if such checking has been enabled).
* Parameters: -	current_connection which is a ptr to the struct where the SCSI 
*				DataIn Header was received.
*			  -	related_command which is a ptr to the struct for the SCSI READ
*				command to which this DataIn is related.
* Return Value: >0 success
*				=0 failure, my_lock is locked
*				<0 failure, my_lock is NOT locked
* While reading, the following variables in the related_command maintain state:
*	r2t_range_list.limit	sum of the size of all recvd data-in pdus in command
*							incremented by size of every recvd data-in pdu
*	r2t_range_list.offset	offset to first pdu in current sequence.
*							set at start of an out-of-order seq if in-order pdus
*							set at end of sequence of out-of-order pdus
*	data_offset				expected offset in next in-order data-in pdu
*							incremented by size of every recvd data-in pdu
*							set at start of an out-of-order sequence
*	recvd_length			sum of size of recvd data-in pdus in this sequence
*							incremented by size of every recvd data-in pdu
*							reset to 0 at end of each sequence
*							used to test for start of a new sequence when 0
*	pdu_range_list.limit	largest limit of any out-of-order pdu in current seq
*	pdu_range_list.offset	smallest offset of any out-of-order pdu in cur seq
*							used to set r2t_range_list.offset at end of seq
****************************************************************************/

int rx_data( struct connection *current_connection, unsigned int local_itt,
							struct command *related_command, unsigned int size )
	{
	struct iscsi_targ_scsi_data_in	*data_in_header;
	struct scatterlist				*sg;
	u_char							*cur_ptr;
	u_int							cur_offset, checksum;
	int								retval, received=1, reset_crc=1;
	int								over_flow_data_size = 0;
	struct session					*related_session;
	struct init_error_rec			err_rec;
	char							*over_flow_buffer = NULL;
	unsigned int					lun;


	
	TRACE(TRACE_ENTER_LEAVE, "Enter rx_data, current_connection %p\n", 
														current_connection);
	
	data_in_header = (struct iscsi_targ_scsi_data_in *) 
													current_connection->rx_buf;
	
	if( data_in_header->flags & S_BIT )
		TRACE(TRACE_ISCSI, "Got DataIn Hdr, ITT %u, DSL %u, S bit 1\n",
															local_itt, size);
	else
		TRACE(TRACE_ISCSI, "Got DataIn Hdr, ITT %u, DSL %u\n", local_itt, size);

	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_scsi_data_in(data_in_header);

	related_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( related_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but using field names as in draft 11) */
		if( !((data_in_header->rsvd1 == 0) &&
			((data_in_header->flags & (W_BIT|BRO_BIT|BRU_BIT))==0)  &&
			(data_in_header->lun==0LL) &&
			(data_in_header->resid==0)) ) 
			{
			reserved_not_0("SCSI DataIn");
			}
		}
	else
		{/* draft 11 or after */
		if( !((data_in_header->rsvd1 == 0) &&
			((data_in_header->flags & (W_BIT|BRO_BIT|BRU_BIT))==0)) )
			{
			reserved_not_0("SCSI DataIn");
			}
		if( data_in_header->flags & A_BIT )
			{/* A bit is 1 -- TTT and LUN must be valid */
			/*	Draft 18, Section 9.7.4 Target Transfer Tag and LUN
				"On incoming data, the Target Transfer Tag and LUN MUST be
				provided by the target if the A bit is set to 1; otherwise
				they are reserved. ...

				The Target Transfer Tag values are not specified by this
				protocol except that the value 0xffffffff is reserved and
				means that the Target Transfer Tag is not supplied.  If the
				Target Transfer Tag is provided, then the LUN field MUST hold
				a valid value and be consistent with whatever was specified
				with the command; otherwise, the LUN field is reserved."
			*/
			if( data_in_header->target_xfer_tag == ALL_ONES )
				{/* TTT is reserved, this is an error */
				TRACE_ERROR("DataIn with A bit = 1 has reserved TTT 0x%08x, "
							"expected valid value\n", ALL_ONES);
				}
			else if( related_command != NULL && related_command->SCpnt != NULL)
				{/* TTT has valid value, LUN must be same as in related Cmnd */
				lun = unpack_lun((unsigned char *)&data_in_header->lun);
				if( lun != related_command->SCpnt->lun )
					{
					TRACE_ERROR("Invalid DataIn LUN %u, expected %u\n",
											lun, related_command->SCpnt->lun);
					}
				}
			}
		else
			{/* A bit is 0 -- TTT can be reserved, doesn't have to be */
			if( data_in_header->target_xfer_tag == ALL_ONES )
				{/* TTT is reserved */
				if( data_in_header->lun != 0LL )
					{/* have reserved TTT, non-reserved LUN, this is an error */
					TRACE_ERROR("DataIn with reserved TTT 0x%08x, "
							"non-reserved LUN %s\n",
							 ALL_ONES, string_llx(data_in_header->lun));
					}
				/* else LUN is also reserved, all ok */
				}
			else if( related_command != NULL && related_command->SCpnt != NULL)
				{/* TTT has valid value, LUN must be same as in related Cmnd */
				lun = unpack_lun((unsigned char *)&data_in_header->lun);
				if( lun != related_command->SCpnt->lun )
					{
					TRACE_ERROR("Invalid DataIn LUN %u, expected %u\n",
											lun, related_command->SCpnt->lun);
					}
				}
			}
		}
			
	if( data_in_header->flags & S_BIT )
		{/* this command is a "phase collapse" with "good" status */
		if( !(data_in_header->flags & F_BIT) )
			{/* F bit not set, but it should be when S bit is set */

			/*	Draft 18, Section 9.7.3 Flags (byte 1)
				"bit 7 S (status) - set to indicate that the Command
				Status field contains status.  If this bit is set
				to 1, the F bit MUST also be set to 1."
			*/
			TRACE_ERROR("DataIn has S bit = 1 but F Bit = 0, expected 1\n");
			}
		}

	TRACE(TRACE_DEBUG,"Data from 0x%08x\n",current_connection->target_address);	
	
	if( related_command != NULL )
		{
		if (related_command->SCpnt == NULL)
			{/*If we lost ptr to the SCSI Midlevel struct,we can't do anything*/
			TRACE_ERROR("SCpnt in command struct is NULL\n");		
			}
		else
			{/* have matched the DataIn to a legitimate SCSI READ request */
			data_in_header->offset = ntohl(data_in_header->offset);
			
			/* Check if DataSN field in header is what we expected */
			if( ntohl(data_in_header->data_sn) != related_command->data_in_sn )
				{
				TRACE_ERROR("Expected DataSN %d, got %d\n",
										related_command->data_in_sn,
										ntohl(data_in_header->data_sn));
				/* DataIn PDU Sequence Error - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = related_command;
				err_rec.err_type = SEQUENCE_ERR;
				init_do_error_recovery(&err_rec);
				received = 1;
				goto out;
				}

			/*	Draft 18, Section 9.7.1 F (Final) Bit
				"Splitting the data stream into sequences does not affect
				DataSN counting on Data-In PDUs."

				Draft 18, Section 9.7.5 DataSN
				"For input (read) or bidirectional Data-IN PDUs, the DataSN is
				the input PDU number within the data transfer for the command
				identified by the Initiator Task Tag."

				Therefore, we do not reset data_in_sn at the end of a data-in
				sequence -- it continues incrementing over all data-in sequence
				boundaries.
			*/
			related_command->data_in_sn++;

			/* Check if we are overflowing the SCSI Mid-level buffers */
			if( (over_flow_data_size = data_in_header->offset + size
								- related_command->SCpnt->request_bufflen) > 0 )
				{
				TRACE_ERROR("Received buffer size %u bytes will"
												" overflow SCSI buffer by %d\n",
												size, over_flow_data_size);
				if( related_command->SCpnt->request_bufflen >
											data_in_header->offset )
					size = related_command->SCpnt->request_bufflen -
											data_in_header->offset;
				else
					size = 0;
				}
			else
				{/* be sure this is not negative for use later on! */
				over_flow_data_size = 0;
				}
			
			TRACE(TRACE_DEBUG,"Local data offset %u, request length %d\n",
									related_command->data_offset, 
									related_command->SCpnt->request_bufflen);
			TRACE(TRACE_DEBUG,"DataIn hdr data offset %u, payload len %u\n",
											data_in_header->offset, size);

			if( related_command->recvd_length == 0 )
				{/* this is first PDU in a sequence */
				related_command->r2t_range_list.offset = data_in_header->offset;
				/*	can sequences be out-of-order? */
				if( !related_session->oper_param->DataSequenceInOrder )
					{/* DataSequenceInOrder==No, so seqs can be out-of-order */
					related_command->data_offset = data_in_header->offset;
					}
				/* can the PDU's within a sequence be out-of-order? */
				if( !related_session->oper_param->DataPDUInOrder )
					{/* DataPDUInOrder==No, so PDUs can be out-of-order */
					related_command->pdu_range_list.limit = 0;
					related_command->pdu_range_list.offset = 0xffffffff;
					}
				}

			/* can the PDU's within a sequence be out-of-order? */
			if( !related_session->oper_param->DataPDUInOrder )
				{/* DataPDUInOrder==No, so PDU can be out-of-order */
				u_int	limit;

				if( related_command->pdu_range_list.offset >
														data_in_header->offset )
					{/* this is smallest offset seen so far */
					related_command->pdu_range_list.offset =
														data_in_header->offset;
					}
				limit = data_in_header->offset + size;
				if( related_command->pdu_range_list.limit < limit )
					{/* this is largest limit seen so far */
					related_command->pdu_range_list.limit = limit;
					}
				/*	each DataIn PDU is its own range */
				merge_offset_length(&related_command->pdu_range_list,
								data_in_header->offset, size);
				}
			else
				{/* DataPDUInOrder==Yes, so PDU must be in-order within seq */
				/* does buffer offset in DataIn header have expected value? */
				if( related_command->data_offset != data_in_header->offset )
					{
					TRACE_ERROR("DataIn has Buffer Offset %u, expected %u\n",
								data_in_header->offset,
								related_command->data_offset);
					}
				}

			if( size == 0 )
				{
				TRACE(TRACE_ISCSI,"***Warning***: DataIn PDU with no data\n");
				}
			else
				{/* Check if we have scatter-gather list to be filled */
				if( related_command->SCpnt->use_sg == 0 )
					{/* No scatter-gather list, just one big buffer */

					cur_ptr = related_command->SCpnt->request_buffer 
												+ data_in_header->offset;
			
					received = recv_a_message(current_connection,size,cur_ptr);
					if( received <= 0 )
						{/* connection died, terminate this thread */
						goto out;
						}

					/* compute the partial data digest if we are using them */
					if( current_connection->connection_flags & USE_DATADIGEST )
						{
						do_crc(cur_ptr, received, reset_crc, &checksum);
						TRACE(TRACE_DEBUG,
									"do_crc %d bytes, reset %d, crc 0x%08x\n",
									received, reset_crc, checksum);
						reset_crc=0;
						}
					}
				else
					{/* We have to fill scatter-gather list */

					sg = (struct scatterlist *) 
									related_command->SCpnt->request_buffer;
					cur_offset = 0;
					cur_ptr = sg->address;

					/* first find where to put data in the scatterlist */ 
					while( cur_offset < data_in_header->offset )
						{
						TRACE(TRACE_DEBUG, "cur_offset while %d %d %d %p %p\n",
										cur_offset, data_in_header->offset,
											sg->length, cur_ptr, sg->address);
						if( cur_offset+sg->length>data_in_header->offset )
							{
							cur_ptr = sg->address + data_in_header->offset
																- cur_offset;
							cur_offset = data_in_header->offset;
							}
						else
							{
							cur_offset += sg->length;
							sg++;
							cur_ptr = sg->address;
							}
						}

					/* fill in the buffers till we receive amount of data 
					   corresponding to the datasegmentlength in the DataIn
					   header */
					for( received = 0;  received < size;  received += retval )
						{
						int		cur_size;

						cur_size = (u_int)sg->address+sg->length-(u_int)cur_ptr;
						if( cur_size > size - received )
							cur_size = size - received;

						retval = recv_a_message(current_connection, cur_size,
																	cur_ptr);
						if( retval <= 0 )
							{/* connection died, terminate this thread */
							received = retval;
							goto out;
							}
		
						/* Print the data buffer */				
						PRINT_BUFF(cur_ptr, retval);

						/* compute the partial data digest if we're using them*/
						if(current_connection->connection_flags&USE_DATADIGEST)
							{
							do_crc(cur_ptr, retval, reset_crc, &checksum);
							TRACE(TRACE_DEBUG,
									"do_crc %d bytes, reset %d, crc 0x%08x\n",
									retval, reset_crc, checksum);
							reset_crc=0;
							}

						if( retval == cur_size )
							{/* filled current sg list item, move to next */
							sg++;
							cur_ptr = sg->address;
							}
						else
							{/* only partially filled current sg list item */
							cur_ptr += retval;
							}
						}
					}
				}

			if( size+over_flow_data_size > 0 )
				{
				if( (received = recv_locally(size, over_flow_data_size,
								current_connection, reset_crc,
								&checksum, &over_flow_buffer)) <= 0 )
					{
					/* If There is a Data Digest Error Handle appropriately - SAI */

					if (received == PAYLOAD_DIGERR)
						{
						err_rec.curr_conn = current_connection;
						err_rec.related_cmd = related_command;
						err_rec.err_type = PAYLOAD_DIGERR;
						init_do_error_recovery(&err_rec);
						received = 1;
						}

					/* got an error of some sort */
					goto out;
					}
				}

			/*	update amount of data transfered so far in this sequence */
			related_command->data_offset += size;
			related_command->recvd_length += size;

			/* Check that we have not received more data than allowed in
			   one input sequence */
			if( related_command->recvd_length >
							(related_session->draft_number >= DRAFT18 ?
								related_session->oper_param->MaxBurstLength :
								related_session->oper_param->MaxBurstSize) )
				{
				TRACE_ERROR("%u bytes received in sequence more than allowed by"
							" %s %u\n", 
							related_command->recvd_length,
							related_session->draft_number >= DRAFT18 ?
							"MaxBurstLength" :
							"MaxBurstSize",
							related_session->draft_number >= DRAFT18 ?
							related_session->oper_param->MaxBurstLength :
							related_session->oper_param->MaxBurstSize);
				}

			/*	update total amount of data transfered so far in this command */
			related_command->r2t_range_list.limit += size;

			/*	if this is the last DataIn PDU expected for the corresponding
				SCSI Command, then check that the F_BIT is set */
			if( related_command->r2t_range_list.limit >= 
									related_command->SCpnt->request_bufflen )
				{
				if( !(data_in_header->flags & F_BIT) )
					{
					TRACE_ERROR("F_BIT not set on last expected DataIn PDU\n"); 
					data_in_header->flags &= F_BIT;		/* turn it on now */
					}
				}
			
			if( (data_in_header->flags & F_BIT) )
				{/* this is last DataIn PDU in a sequence,
					set up for any new sequence that may follow this one */
				if( !related_session->oper_param->DataPDUInOrder )
					{/* DataPDUInOrder==No, pdus can be out-of-order */
					check_range_list_complete(&related_command->pdu_range_list);
					free_range_list(&related_command->pdu_range_list);
					/*	set up offset to first pdu in this sequence */
					related_command->r2t_range_list.offset =
										related_command->pdu_range_list.offset;
					}
				if( !related_session->oper_param->DataSequenceInOrder )
					{/* DataSequenceInOrder==No, sequences can be out-of-order*/
					merge_offset_length(&related_command->r2t_range_list,
								related_command->r2t_range_list.offset,
								related_command->recvd_length);
					}
				else if( related_command->r2t_range_list.offset +
										related_command->recvd_length !=
										related_command->r2t_range_list.limit )
					{/* in-order sequence not in-order */
					TRACE_ERROR("In-order sequence started at offset %u, "
							"expected to start at offset %u\n",
							related_command->r2t_range_list.offset,
							related_command->r2t_range_list.limit -
												related_command->recvd_length);
					}

				/*	mark that next DataIn PDU (if any) starts a new sequence */
				related_command->recvd_length = 0;
				}

			if( data_in_header->flags & S_BIT )
				{/* this command is a "phase collapse" with "good" status */
				do_scsi_response( current_connection, related_command,
						(struct iscsi_targ_scsi_rsp *)data_in_header, 0, NULL);
				}
			/* Update the command timestamp and activity flags - SAI */
			related_command->time_stamp = jiffies;
			}
		}

out:
	my_kfree((void **)&over_flow_buffer, "local buffer");

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_data, retval %d\n", received);

	return received;
	}
	
/****************************************************************************
* This function processes the R2T PDU and sets up the TCP buffers for the 
* first iSCSI DataOut PDU to be sent out to the target. It wakes up the 
* tx_thread so that it can send DataOut PDU to the target.
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the R2T header was received.
* Return Value: >0 success (always)
****************************************************************************/

int rx_r2t( struct connection *current_connection, unsigned int local_itt,
											struct command *related_command )
	{
	struct iscsi_targ_r2t	*r2t_header;
	struct session			*related_session;
	struct init_error_rec	err_rec;
	struct r2t_cookie		*cookie;
	int						over_flow_data_size;
	int						retransmit_r2t_flg=0;

	
	TRACE(TRACE_ENTER_LEAVE, "Enter rx_r2t, current_connection %p\n", 
														current_connection);
	
	r2t_header = (struct iscsi_targ_r2t *)current_connection->rx_buf;
	
	TRACE(TRACE_ISCSI, "Got R2T Hdr, ITT %u, %u bytes, offset %u\n", local_itt,
						ntohl(r2t_header->xfer_len), ntohl(r2t_header->offset));
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_r2t(r2t_header);

	TRACE(TRACE_DEBUG,"R2T from 0x%08x\n",current_connection->target_address);	

	related_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( related_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but using field names as in draft 11) */
		if( !(((r2t_header->flags & ~F_BIT)==0) && (r2t_header->rsvd2==0) &&
													(r2t_header->lun==0LL)) )
			{
			reserved_not_0("R2T");
			}
		}
	else
		{/* draft 11 or after */
		if( !(((r2t_header->flags & ~F_BIT)==0) && (r2t_header->rsvd2==0)) )
			{
			reserved_not_0("R2T");
			}
		}
			
	if( related_command != NULL )
		{	
		/*	is this R2T in the correct sequence of R2Ts for this command? */

		/*	Draft 18, Section 9.8.2 R2TSN
			"R2TSN is the R2T PDU input PDU number within the command
			identified by the Initiator Task Tag."
		*/

        /* new 18_04 - SAI */
		if( related_command->r2t_sn != ntohl(r2t_header->r2t_sn) )
			{
        	if (related_session->oper_param->ErrorRecoveryLevel > 0)
				{
                if (ntohl(r2t_header->r2t_sn) == (related_command->r2t_sn - 1))
                    {/* only one outstanding R2T supported for error recovery levels > 0 */
                    TRACE(TRACE_ISCSI, "Re-transmitted R2T against a dropped DataOut PDU\n");
                    /* if an R2T seq. repeats due to inherent timeout from Target either due to
                    lost or delayed DataOuts remove the old R2T cookie and recreate a new one.
                    Remember to re-transmit all DataOuts requested by the R2T - new 18_04 - SAI */
                    if ( (cookie = related_command->first_cookie) == NULL)
                        TRACE(TRACE_ISCSI, "No cookies available - nothing to do !!\n");
                    else
                        free_r2t_cookie(related_command);
                    related_command->r2t_sn--;
                    related_command->n_r2t_cookies--;
                    retransmit_r2t_flg=1;
                    }
            	}

			if (!retransmit_r2t_flg)
				{
				TRACE_ERROR("R2T contains R2TSN %u, expected %u\n",
							ntohl(r2t_header->r2t_sn), related_command->r2t_sn);
				/* R2TSN Sequence Error Recovery - SAI */
				err_rec.curr_conn = current_connection;
				err_rec.related_cmd = related_command;
				err_rec.err_type = SEQUENCE_ERR;
				init_do_error_recovery(&err_rec);
				goto rx_r2t_out;
				}
			}

		related_command->r2t_sn++;

		/*	Draft 18, Section 9.7.4 Target Transfer Tag and LUN
			"The Target Transfer Tag values are not specified by this protocol
			except that the value 0xffffffff is reserved and means that the
			Target Transfer Tag is not supplied."
		*/
		if( r2t_header->target_xfer_tag == ALL_ONES )
			{
			TRACE_ERROR("R2T has reserved TTT value 0x%08x\n",
												r2t_header->target_xfer_tag);
			}

		if( related_command->SCpnt == NULL )
			{
			TRACE_ERROR("SCpnt in command struct is NULL\n");
			}
		else
			{/* extract useful data from the R2T PDU and store it in a cookie */
			if( (cookie = create_r2t_cookie(related_session, related_command,
										r2t_header->target_xfer_tag)) != NULL )
				{/* fill in remaining fields in the cookie with R2T info */
				cookie->offset = ntohl(r2t_header->offset);
				cookie->r2t_xfer_length = ntohl(r2t_header->xfer_len);

				if( related_session->oper_param->DataSequenceInOrder == 0 )
					{/* DataSequenceInOrder==No */
					merge_offset_length(&related_command->r2t_range_list,
										cookie->offset,cookie->r2t_xfer_length);
					}
				else
					{/* DataSequenceInOrder=Yes */

					/*	Draft 18, Section 9.8 Ready to Transfer (R2T)
						"DataSequenceInOrder governs the buffer offset ordering
						in consecutive R2Ts. If DataSequenceInOrder is Yes,
						then consecutive R2Ts MUST refer to continuous
						non-overlapping ranges except for Recovery-R2Ts."
					*/

					/* Consider out of seq R2T to be recovery R2T - SAI */
					if( cookie->offset != related_command->data_offset )
						{
						TRACE(TRACE_ISCSI,"R2T buffer offset %u, expected %u\n",
							cookie->offset, related_command->data_offset);

						/* Handle Recovery R2T without flagging errors - SAI */
						if( related_session->oper_param->ErrorRecoveryLevel>0 )
							{
							/* error recovery change ver. ref18_04 */
							if( cookie->offset < related_command->data_offset )
								{/* yes, recovery R2T, backup command offset*/
								TRACE(TRACE_ISCSI, "Recovery R2T - Data offset already covered\n");
								related_command->data_offset = cookie->offset;
								}
							}
						/* force cookie offset to be same as command offset */
						cookie->offset = related_command->data_offset;
						}
					}

				/* Check that R2T is not asking to xfer more than MaxBurstSize*/
				if( cookie->r2t_xfer_length > 
							(related_session->draft_number >= DRAFT18 ?
								related_session->oper_param->MaxBurstLength :
								related_session->oper_param->MaxBurstSize) )
					{
					TRACE_ERROR("R2T DDTL of %u bytes exceeds allowed"
								" %s %u\n", 
								cookie->r2t_xfer_length,
								related_session->draft_number >= DRAFT18 ?
								"MaxBurstLength" :
								"MaxBurstSize",
								related_session->draft_number >= DRAFT18 ?
								related_session->oper_param->MaxBurstLength :
								related_session->oper_param->MaxBurstSize);
					/* for now, just ignore this error -- we can handle data
					   beyond MaxBurstLength without problem */
					}

				/* Check if this R2T would overflow the SCSI Mid-level buffers*/
				over_flow_data_size = cookie->offset + cookie->r2t_xfer_length -
										related_command->SCpnt->request_bufflen;
				if( over_flow_data_size > 0 )
						{
						TRACE_ERROR("R2T DDTL of %u bytes"
									" overflows SCSI buffer by %d\n",
										cookie->r2t_xfer_length,
										over_flow_data_size);

						/*	accept only the remaining amount of buffer space */
						cookie->r2t_xfer_length -= over_flow_data_size;
						}

				TRACE(TRACE_DEBUG,
						"recvd r2t command offset %u xfer length %u\n",
						related_command->data_offset, cookie->r2t_xfer_length);

				/* Update command offset to that expected in next in-order R2T*/
				related_command->data_offset += cookie->r2t_xfer_length;
					
				if( related_session->oper_param->DataPDUInOrder == 0 )
					{/* DataPDUInOrder==No, send Data Out PDUs backwards */
					cookie->offset += cookie->r2t_xfer_length;	/* set limit */
					}

				if( setup_dataoutpdu(current_connection, related_command) > 0 )
					{/* we actually set up something to xmit, tell xmit thread*/
					up(&current_connection->my_session->tx_sem);
					}
				}
			}
		}

/* r2t exit added for Error Recovery - SAI */
rx_r2t_out:
	
	TRACE(TRACE_ENTER_LEAVE, "Leave rx_r2t, current_connection %p\n", 
														current_connection);
	return 1;		/* nothing but success here */
	}
	
/****************************************************************************
* This function processes the Asynchronous Message PDU
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the PDU header was received.
* Return Value: >0 success, =0 failure.
****************************************************************************/

int rx_async_msg( struct connection *current_connection, unsigned int size,
																char *buffer )
	{
	struct iscsi_targ_async_msg	*am_header;
	struct session				*related_session;
	struct connection			*conn;
	unsigned int				lun;
	int							retval = 0;		/* assume this fails */

	
	TRACE(TRACE_ENTER_LEAVE, "Enter rx_async_msg, current_connection %p\n", 
														current_connection);
	
	am_header = (struct iscsi_targ_async_msg *)current_connection->rx_buf;
	
	TRACE(TRACE_ISCSI, "Got Async Msg Hdr, AsyncEvent 0x%02x, DSL %u\n",
												am_header->async_event, size);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_async_msg(am_header);

	TRACE(TRACE_DEBUG,"Async Msg from 0x%08x\n",current_connection->target_address);	

	related_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( !(((am_header->flags & ~F_BIT)==0) && (am_header->rsvd2==0) &&
							(am_header->rsvd3==0) && (am_header->rsvd5==0)) )
		{
		reserved_not_0("Async Msg");
		}

	/*	Draft 18, section 9.9.2 AsyncVCode
		"AsyncVcode is a vendor specific detail code that is only valid if the
		AsyncEvent field indicates a vendor specific event.  Otherwise, it is
		reserved."
	*/
	if( am_header->async_vcode != 0  &&  am_header->async_event != 255 )
		{
		TRACE_ERROR("AsyncVCode 0x%02x should be 0 when AsyncEvent is 0x%02x\n",
							am_header->async_vcode, am_header->async_event);
		}

	/*	Draft 18, section 9.9.3 LUN
		"The LUN field MUST be valid if AsyncEvent is 0.  Otherwise, this
		field is reserved."
	*/
	if( am_header->lun != 0  &&  am_header->async_event != 0 )
		{
		TRACE_ERROR("LUN 0x%016llx should be 0 when AsyncEvent is 0x%02x\n",
									am_header->lun, am_header->async_event);
		}

	switch( am_header->async_event )
		{

	case 0:		/* SCSI Asynchronous Event */
		/*	check for valid LUN, per section 9.9.3 quoted above */
		lun = unpack_lun((unsigned char *)&am_header->lun);
		if( !(related_session->lun_bits & (1 <<lun)) )
			{
			TRACE_ERROR("Invalid SCSI Asynch Event LUN %u\n", lun);
			}

		TRACE_ERROR("Unimplemented SCSI Asynchronous Event\n");
		break;

	case 1:		/* target requests Logout */
		if( am_header->parameter1 != 0  ||  am_header->parameter2 != 0 )
			{
			TRACE_ERROR("Parameter1 0x%04x and Parameter2 0x%04x both should "
								"be 0 when AsyncEvent is 0x%02x\n",
								am_header->parameter1, am_header->parameter2,
								am_header->async_event);
			}
		if(current_connection->connection_flags & SEND_NO_REPLY_TO_ASYNC_LOGOUT)
			{
			TRACE(TRACE_ISCSI,
						"Not replying to Async Logout request from target\n");
			retval = 1;		/* success */
			}
		else if( !drive_logout(related_session, current_connection->connection_id, 0) )
			{
			TRACE_ERROR("Unable to send Logout PDU requested by target\n");
			}
		else
			retval = 1;		/* success */
		break;

	case 2:		/* target dropping a specified connection */
		/* NOTE	we cannot just kill the connection, since that might
				require this thread to kill itself and then wait for
				itself to die, which will not work.
		*/
		/* check that this CID actually matches an existing connection id */
		for( conn = related_session->connection_head;  conn != NULL;
															conn = conn->next )
			{
			if( conn->connection_id == am_header->parameter1 )
				break;
			}
		if( conn == NULL )
			{
			TRACE_ERROR("Target dropping connection with nonexistant CID %u\n",
														am_header->parameter1);
			}
		else
			{
			TRACE(TRACE_ISCSI, "Target dropping connection with CID %u\n",
														am_header->parameter1);
			}
		retval = 1;		/* success */
		break;

	case 3:		/* target dropping all connections in a session */
		/* NOTE	we cannot just kill the session, since that might
				require this thread to kill itself and then wait for
				itself to die, which will not work.
		*/
		if( am_header->parameter1 != 0 )
			{
			TRACE_ERROR("Parameter1 0x%04x should be 0 "
								"when AsyncEvent is 0x%02x\n",
								am_header->parameter1, am_header->async_event);
			}
		TRACE(TRACE_ISCSI, "Target dropping all connections in session\n");
		retval = 1;		/* success */
		break;

	case 4:		/* target requests Text negotiation */
		if( !drive_text_negotiate(current_connection, related_session, 0) )
			{
			TRACE_ERROR("Unable to send Text PDU requested by target\n");
			}
		else
			retval = 1;		/* success */
		break;

	case 255:	/* vendor specific iSCSI Event */
		TRACE(TRACE_ISCSI, "Ignoring vendor specific iSCSI Event, "
								"AsyncVcode 0x%02x\n", am_header->async_vcode);
		retval = 1;		/* success */
		break;

	default:
		TRACE_ERROR("Illegal AsyncEvent 0x%02x\n", am_header->async_event);
		retval = 1;		/* success */
		break;
		}	/* switch */

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_async_msg, retval %d\n", retval);
	return retval;
	}
	
/****************************************************************************
* This function processes the task mgt response header, sets up the 
* task_mgt_response field in the command struct and wakes up the 
* initiator_abort function (called by SCSI Midlevel) waiting on the  
* task_mgt_sem semaphore.
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the Task Mgt. Header was received.
* Return Value: >=0, on success (always)
****************************************************************************/
int rx_task_mgt_rsp( struct connection *current_connection,
				    unsigned int local_itt, struct command *related_command )
	{
	struct session						*current_session;
    struct iscsi_targ_task_mgt_response	*header;
	struct command						*cmd;
	int									retval=-1;
 

	TRACE(TRACE_ENTER_LEAVE, "Enter rx_task_mgt_rsp, current_connection"
												" %p\n", current_connection);

	header = (struct iscsi_targ_task_mgt_response *)
												current_connection->rx_buf;

	TRACE(TRACE_ISCSI, "Got TASK MGT Rsp Hdr, ITT %u\n", local_itt);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_task_mgt_response(header);

    TRACE(TRACE_DEBUG, "Got StatSN %u\n", ntohl(header->stat_sn));

	/* Store the session pointer for future use - SAI */
	current_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( current_connection->my_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but using field names as in draft 11) */
		if( !(((header->flags & ~F_BIT)==0)  &&  (header->rsvd4==0)  &&
														(header->rsvd5==0)) )
			{
			reserved_not_0("TM Response");
			}
		}
	else if( current_connection->my_session->draft_number <= DRAFT11 )
		{/* draft 11 */
		if( !(((header->flags & ~F_BIT)==0)  &&  (header->qualifier==0) &&
									(header->rsvd4==0) && (header->rsvd5==0)) )
			{
			reserved_not_0("TM Response");
			}
		}
	else
		{/* draft 12 or after */
		if( !(((header->flags & ~F_BIT)==0)  &&  (header->qualifier==0) &&
									(header->ref_task_tag==0) &&
									(header->rsvd4==0) && (header->rsvd5==0)) )
			{
			reserved_not_0("TM Response");
			}
		}
			
	/* leave if command struct corresponding to the task mgt rsp was not found*/
    if( related_command == NULL )
        {
		goto out;
        }

	/* If the Function is something other than Abort Task, then the ref_task_tag
		should be FFFFFFFF */
	/* Need ref tasktag for TASK_REASSIGN - SAI */
	if( (related_command->task_mgt_function != ABORT_TASK) && 
		(related_command->task_mgt_function != TASK_REASSIGN) &&
											(header->ref_task_tag != ALL_ONES) )
		{
		TRACE_ERROR("Ref_task_tag should be 0xFFFFFFFF when the task mgt "
											"function is not ABORT_TASK\n");
		}		
    
	if( (header->response == FUNCTION_COMPLETE) ||
		(header->response == TASK_WAS_NOT_IN_TASK_SET) ||
        (header->response == LUN_DOES_NOT_EXIST) ||
        (header->response == FUNCTION_REJECTED) )
        related_command->task_mgt_response = header->response;
	else
		{
		TRACE_ERROR("Invalid task mgt response from target 0x%02x,"
		" returning reponse FUNCTION_REJECTED to SCSI mid-level\n",
															header->response);
    	related_command->task_mgt_response = FUNCTION_REJECTED;
		}
	
	/* Added the TASK REASSIGN handling for connection recovery - SAI */
	if (related_command->task_mgt_function == TASK_REASSIGN)
		{
		for(cmd=current_connection->pending_commands; cmd!=NULL; cmd=cmd->next)
			{
			if (cmd->init_task_tag == related_command->ref_task_tag)
				{
				retval = 1;
				break;
				}
			}

		if (retval)
			{/* Free the corresponding task mgt command */
			if (related_command->task_mgt_response == FUNCTION_REJECTED)
				{/* TASK_REASSIGN probably not supported. Free pending command */
				if( cmd->SCpnt != NULL )
					{ /* pass the final status back up to SCSI Mid-level */
					cmd->SCpnt->result = DID_ERROR << 16;
					}

				TRACE_ERROR("Task Reassignment for task ITT %u, returned a response %d\n",
				related_command->ref_task_tag, header->response);
				TRACE(TRACE_DEBUG, "Freeing related command, ITT %u\n", related_command->ref_task_tag);
				free_pending_command(cmd, current_connection);
				}
			}
		else
			{/* could not find the corresponding command - reject task mgt. command */

			TRACE_ERROR("No task with ITT %u found\n", related_command->ref_task_tag);
			}

			TRACE(TRACE_DEBUG, "Freeing Task Mgt. request, ITT %u\n", related_command->init_task_tag);
			free_pending_command(related_command, current_connection);

			current_session->n_rec_tasks --;
		}

	/* wake up the process waiting on task_mgt_sem semaphore */ 
	up(&current_connection->my_session->task_mgt_sem);

out:
	TRACE(TRACE_ENTER_LEAVE,"Leave rx_task_mgt_rsp ,current_connection"
												" %p\n", current_connection);
    return 1;		/* nothing but success here */
	}


/***************************************************************************
* This function processes the Reject header. 
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the Reject header was received.
* Return Value: always 0 to terminate the rx_thread gracefully
***************************************************************************/
int rx_rjt( struct connection * current_connection )
	{
	int retval=1, received=0;
	unsigned int size=0, local_itt=0, checksum=0;
	u_char		*cur_ptr;
	char		*over_flow_buffer=NULL;
	struct iscsi_targ_rjt *rjt_header;
	struct iscsi_init_snack	*header;
	struct command			*related_command;
	struct scsi_sense_data	senseinf;


	TRACE(TRACE_ENTER_LEAVE, "Enter rx_rjt, current_connection %p\n", 
														current_connection);
    
	rjt_header = (struct iscsi_targ_rjt *)current_connection->rx_buf;

	/* Reject does NOT have an ITT value */
	TRACE(TRACE_ISCSI, "Got Reject Hdr\n");
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_rjt(rjt_header);    

	/* Data SNACK rjt to be handled for Error Recovery - SAI */

	/*      allocate buffer into which the rejected PDU header could be read */
	cur_ptr = (u_char *)my_kmalloc(ISCSI_HDR_LEN, "cur_ptr");

	if( cur_ptr == NULL)
		{/* no memory for cur_ptr */
		retval = -ENOMEM;
		goto rjt_out;
		}

	/* Check that reserved fields are zero */
	if( !((rjt_header->flags & ~F_BIT)==0 && rjt_header->rsvd2==0 &&
								rjt_header->rsvd4==0 && rjt_header->rsvd5==0) ) 
		{
		reserved_not_0("Reject");
		}
			
	size = ISCSI_HDR_LEN;
    /* for all type of rejects, abort the command - new 18_04 SAI */
	/*** if (rjt_header->reason == REASON_DATA_SNACK) ***/
		{
		size = htonl(rjt_header->length);
		received = recv_a_message(current_connection,size,cur_ptr);
		if( received <= 0 )
			{/* connection died, terminate this thread */
			retval = received;
			goto rjt_out;
			}

		/* Read data digests if any */
		received = recv_locally(size, 0, current_connection, 0, &checksum, &over_flow_buffer);

        /* ignore target digest errors and wait for target to send-out any
            recovery R2Ts new 18_04 - SAI */
        if (rjt_header->reason == REASON_DATA_DIGEST_ERR)
            goto rjt_out;

		header = (struct iscsi_init_snack *)cur_ptr;
		local_itt = htonl(header->init_task_tag);

		/* If the rejected SNACK is anyother than Data/R2T or Status, it's an
			error - SAI */
		if ( ((header->flags & SNACK_TYPE) != DATA_R2T_SNACK) &&
		   ((header->flags & SNACK_TYPE) != STATUS_SNACK))
			{
			TRACE_ERROR("Expected a Data/R2T or Status header in Reject PDU, found type 0x%02x",
					(header->flags & SNACK_TYPE) );
			retval = 0;
			goto rjt_out;
			}

		/* this ITT is not reserved, we should find a matching command */
		for( related_command = current_connection->pending_commands;
				related_command != NULL;
				related_command = related_command->next )
			{
            /* Status SNACK always have the ITT as ALL_ONES. So for status snack,
                do not check for ITT - new 18_04 SAI */
            if ((header->flags & SNACK_TYPE) == STATUS_SNACK)
                {
                if (related_command->snack_flags & STATUS_BIT)
                    {
                    TRACE(TRACE_DEBUG, "Found the corresponding Status SNACK request - ITT %u\n",
                                            related_command->init_task_tag);
                    break;
                    }
                }
			else
				{
				if( related_command->init_task_tag == local_itt )
					break;
				}
			}
		if( related_command == NULL )
			{
			TRACE_ERROR("No command has matching ITT %u\n", local_itt);
			/* silently discard any snack rejects that appear delayed - SAI */
			goto rjt_out;
			}

		/* all rejects should terminate the related command - SAI */
		/* if (related_command->snack_flags & STATUS_BIT) */
			{
			if( related_command->SCpnt == NULL )
				{/* If we don't have the pointer to the SCSI Mid-level struct,
					bail out */
				TRACE_ERROR("SCpnt in command struct is NULL\n");
				retval=0;
				goto rjt_out;
				}
			else
				{
				/* We need to increment the exp_stat_sn as we will not receive the missing status anymore */

				current_connection->exp_stat_sn ++;

				related_command->SCpnt->result |= (DID_ERROR << 16);
				memset((struct scsi_sense_data *)&senseinf, 0, sizeof(struct scsi_sense_data));

				if (current_connection->my_session->draft_number >= DRAFT13)
					{
					senseinf.response = 0xf0;  /* Valid Bit set, Response = 0x70 */
					senseinf.sense_key = 0x0b; /* Protocol CRC Error */
					senseinf.asl = 0x07;       /* n-7 = 14-7 = 7 */
					senseinf.asc = 0x47;       /* Protocol CRC Error Additional Sense Code */
					senseinf.ascq = 0x05;      /* Protocol CRC Error Additional Sense code qualifier */
					}
				else
					{
					senseinf.response = 0xf0;  /* Valid Bit set, Response = 0x70 */
					senseinf.sense_key = 0x0b; /* SNACK rejected sense key */
					senseinf.asl = 0x07;       /* n-7 = 14-7 = 7 */
					senseinf.asc = 0x11;       /* SNACK rejected Additional Sense Code */
					senseinf.ascq = 0x13;      /* SNACK rejected Additional Sense code qualifier */
					}

				memcpy(related_command->SCpnt->sense_buffer, (char *)&senseinf, sizeof(senseinf));
				related_command->SCpnt->result |= (DRIVER_SENSE << 24);

				free_pending_command(related_command, current_connection);
				}
			}
			my_kfree((void **)&over_flow_buffer, "local buffer");
			retval = 1;
		}
		/* ignore target digest errors and wait for target to send-out any
			recovery R2Ts - SAI */
		if (rjt_header->reason == REASON_DATA_DIGEST_ERR)
			{
			size = htonl(rjt_header->length);
			received = recv_a_message(current_connection,size,cur_ptr);
			if( received <= 0 )
				{/* connection died, terminate this thread */
				retval = received;
				goto rjt_out;
				}

			/* Read data digests if any */
			received = recv_locally(size, 0, current_connection, 0, &checksum, &over_flow_buffer);
			}

rjt_out:

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_rjt\n");

	return retval;      /* Do not terminate on SNACK reject - SAI */
	}


/***************************************************************************
* This function processes the NopIn header. 
* 4 combinations of ITT and TTT to look for (resvd value is 0xffffffff):
*		TTT			ITT			StatSN	Respond
*		resvd		resvd		?		no		simple update to command window
*		resvd		!=resvd		+1		no		response to NopOut sent by us
*		!=resvd		resvd		+0		yes		provoke "ping" response from us
*		!=resvd		!=resvd		NA		NA		illegal
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the NopIn header was received.
* Return Value: >0 success (always)
***************************************************************************/
int rx_nopin( struct connection * current_connection, unsigned int local_itt,
			struct command *related_command, unsigned int size, char **buffer )
	{
	struct iscsi_targ_nopin		*header;
	struct session				*related_session;
	struct iscsi_init_nopout	*nopout_cmd;
	unsigned int				lun;


	TRACE(TRACE_ENTER_LEAVE, "Enter rx_nopin, current_connection %p\n", 
														current_connection);
    
	header = (struct iscsi_targ_nopin *)current_connection->rx_buf;

	if( TRACE_TEST(TRACE_ISCSI) )
		{
		if( local_itt == ALL_ONES )
			TRACE(TRACE_ISCSI, "Got NopIn Hdr, ITT 0x%08x\n", local_itt);
		else
			TRACE(TRACE_ISCSI, "Got NopIn Hdr, ITT %u\n", local_itt);
		}
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_nopin(header);    

	/* Check that reserved fields are zero */
	if( !((header->flags & ~F_BIT)==0 && header->rsvd1==0 && header->rsvd2==0 &&
															header->rsvd3==0) ) 
		{
		reserved_not_0("NopIn");
		}
			
	related_session = current_connection->my_session;

	if( size > current_connection->max_send_length  &&
									related_session->draft_number >= DRAFT9 )
		{/* don't send back more than target wants to get (draft 9 and after)*/
		size = current_connection->max_send_length;
		}
	else if( size > related_session->oper_param->DataPDULength )
		{/* don't send back more than target wants to get (draft 8 and before)*/
		size = related_session->oper_param->DataPDULength;
		}

	if( header->target_xfer_tag == ALL_ONES )
		{/* sent by target either as a response to a NopOut sent earlier by us
			or as a means to update the command window.  In both cases, no
			response is needed from the initiator. */

		if( local_itt == ALL_ONES )
			{/* this is not a response to a NopOut sent earlier by us,
				but is just a simple update of the command window
				and this has already been done, so we are finished. */
			}
		else
			{/* this NopIn is a response to a NopOut sent earlier by us */
    		if( related_command != NULL )
				{/* found the corresponding NopOut command, check it */
				nopout_cmd = (struct iscsi_init_nopout *)
													&related_command->iscsi_cmd;
				if( (nopout_cmd->opcode & ISCSI_OPCODE) != ISCSI_INIT_NOP_OUT )
					{
					TRACE_ERROR("Command matching NopIn has opcode "
											"0x%02x, expected 0x%02x\n",
											nopout_cmd->opcode & ISCSI_OPCODE,
											ISCSI_INIT_NOP_OUT);
					}

				if( size != ntohl(nopout_cmd->length) )
					{
					TRACE_ERROR("NopIn DSL is %u, expected %u\n",
											size, ntohl(nopout_cmd->length));
					}
				else if( size > 0  &&  memcmp(*buffer,
								related_command->iov[1].iov_base, size) != 0 )
					{
					TRACE_ERROR("NopIn Data \"%s\", expected \"%s\"\n",
							*buffer, (char *)related_command->iov[1].iov_base);
					}
				else
					{
					TRACE(TRACE_ISCSI_FULL, "NopIn matches NopOut\n");
					}

				/* in any case, free up the related command */
				/* (but first free up any buffer it sent) */
				my_kfree(&related_command->buffer_to_free, "NopOut data");
				free_pending_command(related_command, current_connection);
				}
			}
		}
	else
		{/* sent by target on its own to provoke a "ping" response from us */

		if( local_itt != ALL_ONES )
			{
			TRACE_ERROR("Invalid NopIn ITT %u when TTT is %u\n",
						local_itt, ntohl(header->target_xfer_tag));
			}

		/*	Draft 18, Section 9.19.3 Lun
			"A LUN MUST be set to a correct value when the Target Transfer
			Tag is valid (not the reserved value 0xffffffff)."
		*/
		lun = unpack_lun((unsigned char *)&header->lun);
		if( !(related_session->lun_bits & (1 <<lun)) )
			{
			TRACE_ERROR("Invalid NopIn LUN %u\n", lun);
			}

		/*	Now send a NopOut PDU in response to this NopIn */
		if(current_connection->connection_flags & SEND_NO_REPLY_TO_NOP_PING)
			{
			TRACE(TRACE_ISCSI, "Not replying to NopIn ping from target\n");
			}
		else
			{
			tx_nopout(current_connection, 0, header, buffer, size);
			}
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_nopin\n");

	return 1;		/* only success here */
	}

/*	print out all keys sent by target in a text response */
/*	returns 1 on success, 0 on error (unterminated key) */
int text_key_dump( char *buffer, unsigned int size, char *message )
	{
	int		retval = 1;
	char	*ptr, *ptr2;

	if( buffer != NULL  &&  size > 0 )
		{/* print out the info sent by target */
		for( ptr = buffer;  size > 0;  ptr = ptr2 )
			{
			for( ptr2 = ptr;  *ptr2 != '\0'  &&  size > 0;  ptr2++, size-- )
				{/* skip over key itself */
				}
			if( *ptr2 != '\0' )
				{
				*ptr2 = '\0';
				TRACE_ERROR("Unterminated key %s\n", ptr);
				retval = 0;
				}
			printk("iscsi %s: %s\n", message, ptr);
			for( ;  *ptr2 == '\0'  &&  size > 0;  ptr2++, size-- )
				{/* skip over nulls between keys */
				}
			}
		}
	return retval;
	}

/***************************************************************************
* This function processes the TextResponse header. 
*
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the TextResponse header was received.
* Return Value: >0 success; ==0 on error
***************************************************************************/
int rx_text_rsp( struct connection * current_connection, unsigned int local_itt,
			struct command *related_command, unsigned int size, char *buffer )
	{
	struct iscsi_targ_text_rsp	*header;
	struct session				*related_session;
	struct iscsi_init_text_cmnd	*text_cmd;
	int							retval = 1;


	TRACE(TRACE_ENTER_LEAVE, "Enter rx_text_rsp, current_connection %p\n", 
														current_connection);
    
	header = (struct iscsi_targ_text_rsp *)current_connection->rx_buf;

	TRACE(TRACE_ISCSI, "Got Text Rsp Hdr, ITT %u\n", local_itt);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_text_rsp(header);    

	related_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( related_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but using field names as in draft 11) */
		if( !((header->flags & ~F_BIT)==0  &&  header->rsvd2==0  &&
				header->lun==0LL  &&  header->rsvd4==0  &&  header->rsvd5==0) ) 
			{
			reserved_not_0("Text Response");
			}
		}
	else if( related_session->draft_number < DRAFT13 )
		{/* draft 11 or 12 */
		if( !((header->flags & ~F_BIT)==0  &&  header->rsvd2==0  &&
									header->rsvd4==0  &&  header->rsvd5==0) ) 
			{
			reserved_not_0("Text Response");
			}
		}
	else
		{/* draft 13 or after */
		if( !((header->flags & ~(F_BIT|C_BIT))==0  &&  header->rsvd2==0  &&
									header->rsvd4==0  &&  header->rsvd5==0) ) 
			{
			reserved_not_0("Text Response");
			}
		else if( header->flags & C_BIT )
			{/* C bit is set in Text Response */
			/*	Draft 18, Section 9.11.2 C (Continue) Bit
				"A Text Response with the C bit set to 1 MUST have the
				F bit set to 0."
			*/
			if( header->flags & F_BIT )
				{
				TRACE_ERROR("C=1 in Text Response requires F=0, not F=1\n");
				retval = 0;
				}
			else
				{/* for now */
				TRACE_ERROR("C=1 in Text Response not implemented yet\n");
				retval = 0;
				}
			}
		}
			
	if( related_session->oper_param->SessionType )
		{/* this is a discovery session, print out all keys sent by target */
		text_key_dump(buffer, size, "discovery");
		}
	else
		{/* not a discovery session, must be text negotiation */
		TRACE_ERROR("Unimplemented text negotiation\n");
		text_key_dump(buffer, size, "text negotiation");
		}

	if( header->flags & F_BIT )
		{/* F Bit is 1, Target is finished sending everything */
		if( header->target_xfer_tag != ALL_ONES )
			{/*	Draft 18, Section 9.11.1 F (Final) Bit
				"A Text Response with the F bit set to 1 MUST have a Target
				Transfer Tag field set to the reserved value of 0xffffffff."
			*/
			TRACE_ERROR("F=1 in Text Response requires TTT=0xffffffff, not "
						"0x%8x\n", header->target_xfer_tag);
			retval = 0;
			}

		/* in any case, free up the related command (if any) */
		/* (but first free up any buffer it sent) */
		if( related_command != NULL )
			{
			my_kfree(&related_command->buffer_to_free, "Text data");
			free_pending_command(related_command, current_connection);
			}

		if( related_session->oper_param->SessionType )
			{/* this is a discovery session, send a logout to shut it down */
			drive_logout(related_session, 0, 0);
			}
		}
	else
		{/* F bit is 0, target has more to send following this TextResponse */
		if( header->target_xfer_tag == ALL_ONES )
			{/*	Draft 18, Section 9.11.1 F (Final) Bit
				"A text response with the F bit set to 0 MUST have a Target
				Transfer Tag field set to a value other than the reserved
				0xffffffff."
			*/
			TRACE_ERROR("F=0 in Text Response requires TTT != "
						"0x%8x\n", header->target_xfer_tag);
			retval = 0;
			}
		else if( related_command != NULL )
			{/*	must now queue up an empty TextRequest to keep Target going */
			text_cmd=(struct iscsi_init_text_cmnd *)&related_command->iscsi_cmd;

			/*	Draft 18, Section 9.11.4 Target Transfer Tag
				"The initiator MUST copy the Target Transfer Tag and LUN in its
				next request to indicate that it wants the rest of the data."
			*/
			text_cmd->target_xfer_tag = header->target_xfer_tag;
			text_cmd->lun = header->lun;

			/* this follow-on TextRequest pdu has no data attached to it */
			/* so iov and msghdr are reset so only header is sent */
			text_cmd->length = 0;
			related_command->tx_size = ISCSI_HDR_LEN;
			related_command->iov->iov_len = ISCSI_HDR_LEN;
			related_command->tx_msghdr.msg_iovlen = 1;

			/* can now attach the new command to end of pending commands list */
			/* this also sets the ITT, the CmdSN and the ExpStatSN */
			attach_pending_command(related_command,
											related_session->connection_head,2);

			TRACE(TRACE_ISCSI,"Send Text, ITT %u, %u bytes\n",
					related_command->init_task_tag, related_command->tx_size);
			if( TRACE_TEST(TRACE_ISCSI_FULL) )
				print_init_text_cmnd(text_cmd);

			/*	get the tx thread started, if necessary */
			up(&related_session->tx_sem);

			}
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_text_rsp, retval %d\n", retval);

	return retval;
	}


/***************************************************************************
* This function processes the LogoutResponse header. 
*
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the LogoutResponse header was received.
* Return Value: =0 terminate session on return, (my_lock still locked)
***************************************************************************/
int rx_logout_rsp( struct connection * current_connection,
					unsigned int local_itt, struct command *related_command )
	{
	int reason=0;
	struct iscsi_init_logout_cmnd	*logout_cmd;
	struct iscsi_targ_logout_rsp	*header;
	struct session					*current_session=NULL;
	struct semaphore				rec_wait_sem;


	TRACE(TRACE_ENTER_LEAVE, "Enter rx_logout_rsp, current_connection %p\n", 
														current_connection);
    
	header = (struct iscsi_targ_logout_rsp *)current_connection->rx_buf;

	TRACE(TRACE_ISCSI,"Got Logout Rsp Hdr, ITT %u\n", local_itt);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_logout_rsp(header);    

	/* store the session pointer - SAI */
	current_session = current_connection->my_session;

	/* Check if reserved fields are zero or not */
	if( !((header->flags & ~F_BIT)==0 && header->rsvd1==0 &&
					header->rsvd3==0 && header->rsvd4==0 && header->rsvd5==0) ) 
		{
		reserved_not_0("Logout Response");
		}
			
	/* this PDU should be a response to a LogoutRequest sent earlier by us.  */
	if( related_command != NULL )
		{/* found the corresponding LogoutRequest command */

		/* If connection recovery, start a Wait timer - SAI */
		logout_cmd = (struct iscsi_init_logout_cmnd *)&(related_command->iscsi_cmd);

		if( current_session->draft_number < DRAFT13)
			reason = logout_cmd->rsvd3 & LOGOUT_REASON;
		else
			reason = logout_cmd->flags & LOGOUT_REASON;
			
		if (reason == 2)
			{
			/* If the response is zero and ErrorRecoveryLevel > 1 then wait
				until Time2Wait - SAI */
			if (header->response == 0 &&
				current_session->oper_param->ErrorRecoveryLevel > 1)
				{
				free_pending_command(related_command, current_connection);
				if (be16_to_cpu(header->time2wait) > 0)
					{
					init_MUTEX_LOCKED(&rec_wait_sem);
					/* Start connection recovery wait timer now - SAI */

					start_connrec_timer(current_session, &rec_wait_sem, be16_to_cpu(header->time2wait)*HZ);
					down_interruptible(&rec_wait_sem);
					}
				deal_with_connrec((unsigned long)current_connection);
				down_interruptible(&rec_wait_sem);
				return 1;
				}
			}
		else
			{

			free_pending_command(related_command, current_connection);
			}
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_logout_rsp\n");

	return 0;		/* always shut down the session on return */
	}


/***************************************************************************
* This function processes the SCSI response header. The status is passed onto 
* the SCSI Mid-level. The sense data, if present, is received but not passed to
* the SCSI Midlevel. 
* Parameters: current_connection which is a ptr to the connection struct 
*			  where the SCSI Response header was received.
* Return Value: >0 success (always)
***************************************************************************/
int rx_rsp( struct connection * current_connection, unsigned int local_itt,
		    struct command *related_command, unsigned int size, char *buffer )
	{
	struct iscsi_targ_scsi_rsp  *scsi_rsp_header;
	struct init_error_rec		err_rec;


	TRACE(TRACE_ENTER_LEAVE, "Enter rx_rsp, current_connection %p\n", 
														current_connection);
    
	scsi_rsp_header = (struct iscsi_targ_scsi_rsp *)current_connection->rx_buf;

	TRACE(TRACE_ISCSI, "Got SCSI Rsp Hdr, ITT %u\n", local_itt);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_targ_scsi_rsp(scsi_rsp_header);    

	/* Check if reserved fields are zero or not */
	if( current_connection->my_session->draft_number <= DRAFT9 )
		{/* draft 9 or before (but using field names as in draft 11) */
		if( !(((scsi_rsp_header->flags & (R_BIT|W_BIT|S_BIT))==0)  &&
			(scsi_rsp_header->bidi_resid==0)) )
			{
			reserved_not_0("SCSI Response");
			}
		}
	else
		{/* draft 11 or after */
		if( !(((scsi_rsp_header->flags & (R_BIT|W_BIT|S_BIT))==0)  &&
			(scsi_rsp_header->target_xfer_tag==0)) )
			{
			reserved_not_0("SCSI Response");
			}
		else if( (scsi_rsp_header->flags & F_BIT)  &&
					current_connection->my_session->draft_number == DRAFT12 )
			{/* draft 12 has a mistake:  the F bit in a SCSI Response
				is reserved (it is set to 1 for all other drafts) */
			TRACE(TRACE_ISCSI,
					"***Warning***: SCSI Response has F Bit = 1, expected 0\n");
			}
		}
			
	/*	Draft 18, Section 9.4.2 Status
		"The Status field is used to report the SCSI status of the command
		(as specified in [SAM2]) and is valid only if the Response Code is
		Command Completed at target."

		Draft 18, Section 9.4.3 Response
		"0x00 - Command Completed at Target"
	*/
	if( scsi_rsp_header->response != 0 && scsi_rsp_header->status != 0 )
		{
		TRACE_ERROR("Status %u should be 0 when Response is %u\n",
					scsi_rsp_header->status, scsi_rsp_header->response);
		}

    TRACE(TRACE_DEBUG, "Got StatSN %u\n", ntohl(scsi_rsp_header->stat_sn));

    if( related_command != NULL )
        {/* Check if ExpDataSN field in header is what we expected */

		/*	Draft 18, Section 9.4.8 ExpDataSN
			"The number of Data-In (read) PDUs the target has sent for the
			command.

			This field is reserved if the response code is not Command
			Completed at Target or the command is a write command."

			Draft 18, Section 9.4.3 Response
			"0x00 - Command Completed at Target"
		*/
		if( scsi_rsp_header->response == 0  &&
			ntohl(scsi_rsp_header->exp_data_sn) != related_command->data_in_sn )
			{
			TRACE_ERROR("Expected ExpDataSN %d, got %d\n",
									related_command->data_in_sn,
									ntohl(scsi_rsp_header->exp_data_sn));
			/* DataSN Sequence Error Recovery - SAI */
            /* decrement the exp_stat_sn as it will be requested for re-transmission
                using STATUS_SNACK - new 18_04 SAI */
            current_connection->exp_stat_sn--;
			err_rec.curr_conn = current_connection;
			err_rec.related_cmd = related_command;
			err_rec.err_type = SEQUENCE_ERR;
			init_do_error_recovery(&err_rec);
			goto rx_rsp_out;
			}

        if( related_command->SCpnt == NULL )
			{/* If we don't have the pointer to the SCSI Mid-level struct, 
																bail out */		
            TRACE_ERROR("SCpnt in command struct is NULL\n");
       		}
		else
			{
			/*	Do all the rest of the work in a common function shared by
				rx_data() when it detects a "phase collapse" */
			do_scsi_response(current_connection, related_command,
											scsi_rsp_header, size, buffer);
			}
        }

/* rsp exit added for Error Recovery - SAI */
rx_rsp_out :

	TRACE(TRACE_ENTER_LEAVE, "Leave rx_rsp\n");
    
	return 1;		/* nothing but success here */
	}


/****************************************************************************
*	Called by the SCSI Mid-level when the iscsi_initiator module is removed. 
*	Parameter: Ptr. to Scsi_Host
*	Return Value:	0, on success
*					-1, on error
****************************************************************************/
int iscsi_initiator_release(struct Scsi_Host *host)
	{
	int						retval = 0;
	struct iscsi_hostdata	*hostdata;
	struct session			*sess;
	unsigned int			lun;

	
	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_release, host %p\n", host);
	
	TRACE(TRACE_ISCSI, "Releasing iSCSI module\n");

	/* If the host struct is NULL, it is an error and we want to return */
	if( host == NULL  ||  global_host == NULL  ||  global_hostdata == NULL )
		{
		retval = -1;
		goto out;
		}
	
	/* If the hostdata struct is not what we expect, we want to return */
	if( (hostdata=(struct iscsi_hostdata *)host->hostdata) != global_hostdata )
		{
		retval = -1;
		goto out;
		}

	/* Clear the session struct linked list */
	while( (sess = hostdata->session_head) != NULL )
		{
		if( sess->lun_bits == 0 )
			{/* all luns already removed, just release this structure */
			hostdata->session_head = sess->next;
			my_kfree((void **)&sess, "session");
			}
		else
			{/* remove all valid luns in session
				(last lun will take out session) */
			for( lun = 0;  lun < host->max_lun;  lun++ )
				{
				if( sess->lun_bits & (1<<lun) )
					{
					if( remove_session_from_hostdata(sess->scsi_target_id, lun,
													ALL_ONES, hostdata) == 1 )
						{/* just removed the whole session structure */
						break;
						}
					}
				}
			}
		}

	/*	endup the r2t_cookie management list */
	endup_r2t_cookies();

	/* global pointer is no longer valid after unregister */
	if( global_hostdata != NULL )
		{
		/* chap support - CHONG */
		CHAP_FinalizeContext(global_hostdata->chap_parameter.peer_ctx);
		CHAP_FinalizeContext(global_hostdata->chap_parameter.local_ctx);

		if( global_hostdata->param_tbl != NULL )
			{
			param_tbl_uncpy(*global_hostdata->param_tbl);
			my_kfree((void **)&global_hostdata->param_tbl, "param_tbl");
			}
		}

	global_hostdata = NULL;
	global_host = NULL;

	/* check for any unfreed memory */
	my_kempty();

	/* print statistics about io vector usage */
	endup_iovs();

	TRACE(TRACE_ISCSI, "unregister iSCSI host %p, host number %d\n",
														host, host->host_no);
	scsi_unregister(host);

	TRACE(TRACE_ISCSI, "Released iSCSI module\n");

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_release, retval %d\n",
																	retval);

	return retval;
	}
	
/* Create a new recovery connection to keep the iSCSI session active - Remember,
iscsi requires atleast one connection in a session to be active for the session
to be alive - SAI */
void create_rec_conn(struct connection *conn)
	{
	int connid=0;
	struct session *sess = conn->my_session;

	TRACE(TRACE_ENTER_LEAVE, "Entering create_rec_con\n");

	TRACE(TRACE_DEBUG, "This is the only connection available - "
				"Need to create a new connection for recovery\n");
	connid = conn->connection_id;
	restore_global_data(global_hostdata);
	global_hostdata->draft_number = DEFAULT_DRAFT;
	global_hostdata->nop_period = 0;
	global_hostdata->init_sess_flags = 0;
	global_hostdata->init_conn_flags = 0;
	global_hostdata->init_sessrec_flg = 0;

	create_session(sess->scsi_target_id, conn->target_address, 
					conn->target_port, ((sess->lun_bits)>>1), 
					connid+1, global_host);

	TRACE(TRACE_ENTER_LEAVE, "Leaving create_rec_con\n");

	}

/* clear disconnected and unused connections - SAI */
void clear_free_conn(struct session *sess)
{
	struct connection					*prev, *conn;
	
	TRACE(TRACE_ENTER_LEAVE, "Entering clear_free_conn\n");

	if (!sess->connection_head)
		{
		TRACE_ERROR("No Valid connections available\n");
		return;
		}

	if (sess->nconnections == 1 && sess->connection_head->connection_state == STATUS_CONNREC)
		{
		/* No connections probably available - Need to create a new one - SAI */
		create_rec_conn(sess->connection_head);
		}

	for( prev = conn = sess->connection_head;  conn != NULL;  conn = conn->next )
		{
		if ( (conn->connection_state == STATUS_NOT_PRESENT) ||
			(conn->connection_state == STATUS_DISCONNECTED) )
			{
			if (conn == sess->connection_head)
				sess->connection_head = conn->next;
			else
				prev->next = conn->next;

			/* Adjust the round robin connection head - SAI */
			if (conn == sess->rr_conn_head)
				{
				sess->rr_conn_head = sess->rr_conn_head->next;
				if (sess->rr_conn_head == NULL)
					sess->rr_conn_head = sess->connection_head;
				}
			
			if( conn->sock != NULL )
				{
				TRACE(TRACE_NET, "Releasing socket %p\n", conn->sock);
				sock_release(conn->sock);
				TRACE(TRACE_NET, "Released socket %p\n", conn->sock);
				conn->sock = NULL;
				}
			my_kfree((void **)&conn, "connection");	/* now release this! */
			break;
			}
		prev = conn;
		}

	TRACE(TRACE_ENTER_LEAVE, "Leaving clear_free_conn\n");

	return;
}

/* Multiple Connection Scheduling Algorithm - SAI */
struct connection* get_connection(struct session *current_session)
{
    int nconn = 0;
    struct connection   *current_connection=current_session->connection_head;

	TRACE(TRACE_ENTER_LEAVE, "Entering get_connection\n");

    switch(current_session->sched_scheme)
        {
        case CONN_SCHED_RR:

            /* get the next connection in the round-robin */

			/* clear_free_conn(current_session); */

			if (current_session->nconnections == 1)
				current_connection = current_session->connection_head;
			else	
				{
				current_session->rr_conn_head = current_session->rr_conn_head->next;
				if (current_session->rr_conn_head == NULL)
					current_session->rr_conn_head = current_session->connection_head;
				while(nconn < current_session->nconnections)
					{
					if (current_session->rr_conn_head->connection_state ==
							STATUS_LOGGED_IN)
						break;

					current_session->rr_conn_head = current_session->rr_conn_head->next;
					if (current_session->rr_conn_head == NULL)
						current_session->rr_conn_head = current_session->connection_head;
					nconn++;
					}
				current_connection = current_session->rr_conn_head;
				}
            break;

        case CONN_SCHED_LUN:

        default:
            /* use only one connection */
            current_connection = current_session->connection_head;

        }

	TRACE(TRACE_ENTER_LEAVE, "Leaving get_connection\n");

    return current_connection;
}


/*****************************************************************************
* This function is called by the SCSI Mid-level whenever a Read/Write Request 
* has to be satisfied. This function will set up TCP buffers for sending SCSI
* Command and, if the negotiated ImmediateData is 1, ImmediateData. After
* setting up the TCP buffers, the tx_thread is woken up which is waiting on 
* tx_sem. 
* Parameters: Scsi_Cmnd struct, 
*			  done function which is called when any SCSI Command is satisfied.
* Return Value: 0, on success
* 				-1, on failure (any non-zero value seems to work)
*
* Note that the done() function cannot be called until queuecommand returns!
* Also note that on entry, the io_request_lock is held by this process/thread!
*
*****************************************************************************/
int iscsi_initiator_queuecommand(Scsi_Cmnd * Cmnd, void (*done) (Scsi_Cmnd *))
	{
	int							retval, reset_crc, count, len;
	struct iscsi_hostdata		*hostdata;
	struct iovec				*iov;
	struct iscsi_init_scsi_cmnd	*iscsi_cmd;
	struct command				*new_command;
	struct session				*current_session;
	struct connection			*current_connection; 
	struct r2t_cookie			*cookie;
	int							negotiated_immediate_data_length;
	int							actual_immediate_data_length=0;
	int							still_to_send, space_left;
	

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_queuecommand\n");
	
	retval = -1;		/* pessimistic assumption this function will fail */

	/* get exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	if( global_host == NULL  ||  global_hostdata == NULL )
		{
		TRACE_ERROR("queuecommand called after shutdown\n");
		goto out;
		}

	/* get a pointer to the hostdata struct registered to the SCSI Mid-level*/
	if( (hostdata = (struct iscsi_hostdata *)Cmnd->host->hostdata) !=
															global_hostdata )
		{
		TRACE_ERROR("Hostdata struct %p not same as global_hostdata %p\n",
													hostdata, global_hostdata);
		goto out;
		}

	/* tell Mid-level when I am done receiving the SCSI Cmnd info. */
	Cmnd->scsi_done = done;

	/* Find out the session corresponding to the target_id */
	for( current_session = hostdata->session_head;
			current_session != NULL;
			current_session = current_session->next )
		{ 
		if( current_session->scsi_target_id == Cmnd->target )
			break;
		}

	/*	Check that we have finished the Login Phase with the concerned target
		and are now in the full feature phase of a normal session */
	if( current_session == NULL  || 
		current_session->session_state != STATUS_LOGGED_IN  ||
		current_session->tx_thread == NULL  ||
		current_session->connection_head == NULL  ||
		current_session->connection_head->rx_thread == NULL  ||
		current_session->oper_param->SessionType )
		{

		/* Check to see if session recovery is active - SAI */
		if (global_hostdata && global_hostdata->init_sessrec_flg)
			{/* changes for session recovery new 18_04 - SAI */
			TRACE(TRACE_ISCSI, "iscsi_initiator_queuecommand : "
				"Recovery in Progress for target %u\n", Cmnd->target);

            /* create a new session recovery structure or get the already
                existing session recovery structure depending on whether one or
                multiple session participate in session recovery - new 18_04 SAI */
            current_session = get_rec_sess(Cmnd->target, Cmnd->lun);

            if (current_session)
                {
                current_connection = current_session->connection_head;
                TRACE(TRACE_DEBUG, "Recovery session %p, connection %p received\n",
                                    current_session, current_connection);
                goto rec_in_progress;
                }
            }

		printk("iscsi: Target %u not logged in\n", Cmnd->target);
		Cmnd->result = DID_NO_CONNECT << 16;
		if( done != NULL )
			done(Cmnd);		/* should not do this before queuecommand returns!*/
		goto out_ok;
		}
	
    /* multiple connections usage - SAI */
    current_connection = get_connection(current_session);
	if (!current_connection)
		{
		TRACE_ERROR("iscsi_initiator_queuecommand : "
			"get_connection returned a NULL connection\n");
		goto out;
		}

	/* make sure connection is still in FFP */
	if( current_connection->connection_state != STATUS_LOGGED_IN )
		{/* rx_thread gone -- peer probably reset the connection */
		Cmnd->result = DID_NO_CONNECT << 16;
		if( done != NULL )
			done(Cmnd);		/* should not do this before queuecommand returns!*/
		goto out_ok;
		}

rec_in_progress:

	/* get a new struct command to send SCSI Request PDU to target */
	if( (new_command = setup_command(Cmnd, current_session)) == NULL )
		{
		goto out;
		}

	count = 1;			/* no. of iov slots set up so far (i.e., the header) */
	reset_crc = 1;		/* start a new data digest crc calculation */

	/* set up the SCSI_Cmnd header that is to be sent to the target */
	iscsi_cmd = &new_command->iscsi_cmd;
	iscsi_cmd->opcode = ISCSI_INIT_SCSI_CMND;
    iscsi_cmd->lun = pack_lun(Cmnd->lun, current_connection->connection_flags);
	iscsi_cmd->xfer_len = htonl(Cmnd->request_bufflen);

	TRACE(TRACE_DEBUG, "Xfer len for queuecommand %d\n", Cmnd->request_bufflen);

	/*	copy the SCSI CDB into the PDU */
	memcpy(iscsi_cmd->cdb, Cmnd->cmnd, Cmnd->cmd_len);
	if( Cmnd->cmd_len < 16 )
		memset(&iscsi_cmd->cdb[Cmnd->cmd_len], 0, 16-Cmnd->cmd_len);
	
	if( Cmnd->sc_data_direction == SCSI_DATA_WRITE )
		{/* SCSI Command has write data to be sent */

		iscsi_cmd->flags = W_BIT;
		if(Cmnd->request_bufflen && current_session->oper_param->ImmediateData)
			{/* we are allowed to send immediate data */
	
			/* get the smaller of target's MaxRecvDataSegmentLength or
				FirstBurstSize */
			if( current_session->draft_number >= DRAFT9 )
				{/* drafts 9 and after */
				negotiated_immediate_data_length =
									current_connection->max_send_length;
				}
			/* get the smaller of target's DataPDULength or FirstBurstSize */
			else
				{/* drafts 8 and before */
				negotiated_immediate_data_length =
									current_session->oper_param->DataPDULength;
				}
			if( negotiated_immediate_data_length >
							(current_session->draft_number >= DRAFT18 ?
								current_session->oper_param->FirstBurstLength :
								current_session->oper_param->FirstBurstSize) )
				{
				negotiated_immediate_data_length =
								(current_session->draft_number >= DRAFT18 ?
								current_session->oper_param->FirstBurstLength :
								current_session->oper_param->FirstBurstSize);
				}
			
			/*	actual datalength to be sent as immediate data is minimum of
				what command wants to send and the negotiated mimimum allowed
				for immediate data */
			actual_immediate_data_length = Cmnd->request_bufflen;
			if( actual_immediate_data_length>negotiated_immediate_data_length )
				actual_immediate_data_length = negotiated_immediate_data_length;

            TRACE(TRACE_DEBUG, "Sending immediate data %d\n",
                                                actual_immediate_data_length);

			/* get ptr to i/o vector which is already set up for header */
			iov = new_command->iov;
	
			if( Cmnd->use_sg > 0 )
				{/* we have a scatter gather list to be sent */

				int immediate_data_to_be_filled = actual_immediate_data_length;
				struct scatterlist *sg =
									(struct scatterlist *)Cmnd->request_buffer;
		
				while( immediate_data_to_be_filled >0 )
					{/* set up socket iovector for sending immediate_data */
					TRACE(TRACE_DEBUG, 
						"sg_address %p sg_length %d immediate_data_filled %d\n",
						sg->address, sg->length, immediate_data_to_be_filled);
					count++;
					iov++;
					iov->iov_base = sg->address;
					iov->iov_len = sg->length;	
					immediate_data_to_be_filled -= sg->length;
					if( immediate_data_to_be_filled < 0 )
						{/* went too far, last element needs to be shortened */
						iov->iov_len += immediate_data_to_be_filled;
						}
					/* compute the partial data digest if we are using them */
					if( current_connection->connection_flags & USE_DATADIGEST )
						{
						do_crc(iov->iov_base, iov->iov_len, reset_crc,
												&new_command->data_checksum);
						reset_crc=0;
						}
					sg++;
					}
				}
			else
				{/* we have single buffer ptr. passed from the SCSI Mid-level */
				count++;
				iov++;
				iov->iov_len = actual_immediate_data_length;
				iov->iov_base = Cmnd->request_buffer;
				/* compute the data digest if we are using them */
				if( current_connection->connection_flags & USE_DATADIGEST )
					{
					do_crc(iov->iov_base, iov->iov_len, reset_crc,
												&new_command->data_checksum);
					reset_crc=0;
					}
				}

			iscsi_cmd->length = htonl(actual_immediate_data_length);
			new_command->tx_size += actual_immediate_data_length;
			new_command->data_offset += actual_immediate_data_length;

			if( current_session->oper_param->DataSequenceInOrder == 0 )
				{/* DataSequenceInOrder==No */
				merge_offset_length(&new_command->r2t_range_list,
											0, actual_immediate_data_length);
				}

			/*	now deal with any padding that is needed */

			/*	Draft 18, Section 9.1 iSCSI PDU Length and Padding
				"iSCSI PDUs are padded to the closest integer number
				of four byte words.  The padding bytes SHOULD be sent as 0."
			*/
			if( (len = ((-actual_immediate_data_length) & 3)) != 0 )
				{/* need to send len pad bytes in another element in iovector */
				count++;
				iov++;
				iov->iov_base = &pad_bytes;
				iov->iov_len = len;
				new_command->tx_size += len;		/* is now a multiple of 4 */
				TRACE(TRACE_ISCSI, "Attach %d pad bytes\n", len);
				/* compute the partial data digest if we are using them */
				if( current_connection->connection_flags & USE_DATADIGEST )
					{
					do_crc(iov->iov_base, iov->iov_len, reset_crc,
												&new_command->data_checksum);
					reset_crc=0;
					}
				}

			/* if data digests are in use, be sure to send it */
			if( current_connection->connection_flags & USE_DATADIGEST )
				{
				new_command->data_checksum = htonl(new_command->data_checksum);
				count++;
				iov++;
				iov->iov_base = &new_command->data_checksum;
				iov->iov_len = CRC_LEN;
				new_command->tx_size += CRC_LEN;
				TRACE(TRACE_ISCSI_FULL, "Attach DataDigest\n");
				}
			}
	
		still_to_send = Cmnd->request_bufflen - actual_immediate_data_length;
		space_left =
					(current_session->draft_number >= DRAFT18 ?
					current_session->oper_param->FirstBurstLength :
					current_session->oper_param->FirstBurstSize)
												- actual_immediate_data_length;

		if( current_session->oper_param->InitialR2T  || still_to_send <= 0  ||
															space_left <= 0 )
			{/* set the F bit as no separate unsolicted data PDUs can follow */
			new_command->iscsi_cmd.flags |= F_BIT;
			}
		else if( !(current_session->oper_param->InitialR2T) )
			{/* we are allowed to send unsolicited data PDUs after this command,
				but do we need to and is there any burst space left to do so? */
			if( still_to_send > 0  &&  space_left > 0 )
				{/* still data left to send, and there is room in first burst */
				/*  set up to send unsolicited data PDUs next */
				if( (cookie = create_r2t_cookie(current_session, new_command,
															ALL_ONES)) != NULL )
					{
					cookie->offset = new_command->data_offset;
					cookie->r2t_xfer_length = still_to_send;
					if( still_to_send > space_left )
						cookie->r2t_xfer_length = space_left;
					new_command->data_offset += cookie->r2t_xfer_length;
					if( current_session->oper_param->DataSequenceInOrder == 0 )
						{/* DataSequenceInOrder==No */
						merge_offset_length(&new_command->r2t_range_list,
										cookie->offset,cookie->r2t_xfer_length);
						}
					if( current_session->oper_param->DataPDUInOrder == 0 )
						{/* DataPDUInOrder==No, send Data Out PDUs backwards */
						cookie->offset+=cookie->r2t_xfer_length;/* set limit */
						}
					}
				}
			}
		}
	else
		{/* not a write -- probably a read, but check out of curiousity */
		/*  must always set the F_BIT if it is not a WRITE SCSI Commamnd */ 
		if( Cmnd->sc_data_direction == SCSI_DATA_READ )
			{/* definitely a read */
			iscsi_cmd->flags = F_BIT|R_BIT;			/* F and R bits set */
			/***** start testing *****
			 * on INQUIRY, set the EDTL too small to test overflow response *
			if( iscsi_cmd->cdb[0] == 0x12 )
				{
				iscsi_cmd->xfer_len = htonl(36);
				}
			 * on TEST UNIT READY, set W bit instead of R bit *
			if( iscsi_cmd->cdb[0] == 0x00 )
				{
				iscsi_cmd->flags = F_BIT|W_BIT;
				}
			 * on TEST UNIT READY, set neither W bit nor R bit *
			if( iscsi_cmd->cdb[0] == 0x00 )
				{
				iscsi_cmd->flags = F_BIT;
				}
			 * on TEST UNIT READY, set both W bit and R bit *
			if( iscsi_cmd->cdb[0] == 0x00 )
				{
				iscsi_cmd->flags = F_BIT|R_BIT|W_BIT;
				}
			***** end testing *****/
			}
		else
			{/* direction given as neither Read nor Write */
			TRACE(TRACE_DEBUG, "sc_data_direction %u not READ or WRITE\n",
													Cmnd->sc_data_direction);
			if( Cmnd->request_bufflen != 0 )
				{
				/*	Draft 18, Section 9.3.1 Flags and Task Attributes (byte 1)
					"Either or both of R and W MAY be 1 when either the Expected
					Data Transfer Length and/or Bidirectional Read Expected Data
					Transfer Length are 0, but they MUST NOT both be 0 when the
					Expected Data Transfer Length and/or Bidirectional Read
					Expected Data Transfer Length are not 0 (i.e., when some
					data transfer is expected the transfer direction is
					indicated by the R and/or W bit).
				*/
        		TRACE_ERROR("request_bufflen %u not zero when "
							"sc_data_direction %u not READ or WRITE\n",
							Cmnd->request_bufflen, Cmnd->sc_data_direction);
				iscsi_cmd->flags = F_BIT|R_BIT;		/* F and R bits set */
				}
			else
				{/* no data expected to be transfered, not READ or WRITE */
				iscsi_cmd->flags = F_BIT;			/* only the F bit set */
				}
			}
		}
			 
	/* set the number of iov slots actually used */
	new_command->tx_msghdr.msg_iovlen = count;

	if( max_iov_count < count )
		max_iov_count = count;		/* keep track of max no. of iovs ever used*/

	/* Set Time Stamp for Command Retries - SAI */
	new_command->time_stamp = jiffies;

	/* can now attach the new command to end of pending commands list */
	attach_pending_command(new_command, current_connection, 1);

	TRACE(TRACE_ISCSI, "Send SCSI %s, ITT %u, %u bytes %d iovs\n",
						printable_scsi_op(iscsi_cmd->cdb[0]),
						new_command->init_task_tag,
						new_command->tx_size, count);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_scsi_cmnd(iscsi_cmd);

	/* Check if a Connection Recovery Test needs to be initiated - SAI */
	if (current_connection->rec_tests == 8)
		{
		if (current_session->nconnections <= 1)
			create_rec_conn(current_session->connection_head);
		drive_logout(current_session, current_connection->connection_id, current_session->oper_param->ErrorRecoveryLevel);
		current_connection->rec_tests = 0;
		}

    /* do not awake the semaphore during session recovery - new 18_04 SAI */
    if (!global_hostdata->init_sessrec_flg)
		/* signal the tx thread for this session that it has work to do */
		up(&current_connection->my_session->tx_sem);
	
out_ok:

	retval = 0;			/* pessimistic assumption about this function wrong! */

out:

	/* release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_queuecommand, retval %d\n",
																		retval);
	return retval;
	}
	
/****************************************************************************
* This function is called by the system when the abort_timer set in
* the iscsi_initiator_abort function expires. This function wakes up 
* the thread/process which is waiting on the task_mgt_sem semaphore.
****************************************************************************/

static void abort_timer_function(unsigned long data)
	{
	TRACE(TRACE_ENTER_LEAVE, "Enter abort_timer_function\n");
	up((struct semaphore *)data);
	TRACE(TRACE_ENTER_LEAVE, "Leave abort_timer_function\n");
	}

/***************************************************************************
* This function is called by the SCSI Mid-level when it has to abort any 
* SCSI Command. This function will check if the command has been sent to the 
* target for satisfying the request. If it has been sent, then this function
* will set up TCP buffers to send out a Task Mgt. Command. 
* When called by the SCSI Mid-level (from linux/drivers/scsi/scsi_error.c)
* the io_request_lock is held by the caller.
****************************************************************************/
int iscsi_initiator_abort(Scsi_Cmnd * Cmnd)
	{
										/* pessimistic assumption */
	int									retval=SCSI_ABORT_NOT_RUNNING;
	__u8								response;
	struct timer_list					abort_timer;
	struct iscsi_init_task_mgt_command	*task_mgt_command_hdr;
    struct session						*related_session;
    struct connection					*related_connection;
    struct command						*related_scsi_command;
    struct command						*task_mgt_command;
    struct command						*tm_command;
    struct iscsi_hostdata				*hostdata;
	unsigned int						related_itt;


	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_abort\n");

	/* get exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	if( (hostdata = (struct iscsi_hostdata *)Cmnd->host->hostdata) !=
															global_hostdata )
		{
		TRACE_ERROR("Hostdata struct %p not same as global_hostdata %p\n",
													hostdata, global_hostdata);
		goto out;
		}


	/* Find out the session that corresponds to the target ID */
	for( related_session = hostdata->session_head;  related_session != NULL;
									related_session = related_session->next )
		{
		if( related_session->scsi_target_id == Cmnd->target )
			break;
		}
 
	/* check if we have session info. for that target */
	if( related_session == NULL )
        {
        TRACE_ERROR("No session for target %u\n", Cmnd->target);
        goto out;
        }

	/* check if the session is in login phase or not */
    if( related_session->session_state != STATUS_LOGGED_IN  ||
		related_session->tx_thread == NULL  ||
		(related_connection = related_session->connection_head) == NULL  ||
		related_connection->rx_thread == NULL  ||
		related_connection->connection_state != STATUS_LOGGED_IN )
        {
        TRACE_ERROR("No connection to target %u\n", Cmnd->target);
        goto out;
        }
	
	/* search for the PDU containing the SCSI command which has to be aborted */
    for( related_scsi_command = related_connection->pending_commands;
		                    related_scsi_command != NULL;
                            related_scsi_command = related_scsi_command->next )
        {
        if( related_scsi_command->SCpnt == Cmnd  &&
			(related_scsi_command->iscsi_cmd.opcode & ISCSI_OPCODE) ==
														ISCSI_INIT_SCSI_CMND )
            break;
        }

    if( related_scsi_command == NULL )
        {
        TRACE_ERROR("Command struct related to SCSI Cmnd is NULL\n");
        goto out;
        }

	related_itt = related_scsi_command->init_task_tag;

	TRACE_ERROR("Aborting SCSI Cmnd with ITT %u\n", related_itt);

    if( (task_mgt_command = setup_command(NULL, related_session)) == NULL )
        {
        goto out;
        }

    TRACE(TRACE_DEBUG, "Populating task mgt command\n");
	task_mgt_command->task_mgt_function = ABORT_TASK;

    task_mgt_command->task_mgt_response = response = FUNCTION_REJECTED;

	/* fill up the task_mgt_header info.*/
	task_mgt_command_hdr = (struct iscsi_init_task_mgt_command *)
												&task_mgt_command->iscsi_cmd;
	task_mgt_command_hdr->opcode = ISCSI_INIT_TASK_MGMT_CMND;
    task_mgt_command_hdr->function = ABORT_TASK;
	if( related_session->draft_number >= DRAFT11 )
		{/*	draft 11 or after, F bit now set to 1 (was 0 in earlier drafts)*/
    	task_mgt_command_hdr->function |= F_BIT;
		}
    task_mgt_command_hdr->lun = pack_lun(Cmnd->lun,
										related_connection->connection_flags);
    task_mgt_command_hdr->ref_task_tag = htonl(related_itt);
 
	task_mgt_command_hdr->ref_cmd_sn = related_scsi_command->iscsi_cmd.cmd_sn;

	/* update the command linked list with the task mgt. command */
	attach_pending_command(task_mgt_command, related_connection, 1);

	TRACE(TRACE_ISCSI, "Send TASK MGT Cmnd, ITT %u, %u bytes\n",
					task_mgt_command->init_task_tag, task_mgt_command->tx_size);
	if( TRACE_TEST(TRACE_ISCSI_FULL) )
		print_init_task_mgt_command(task_mgt_command_hdr);

	/* signal the tx thread for this session that it has work to do */
    up(&related_session->tx_sem);

	TRACE(TRACE_DEBUG,"Waiting to get task mgt. rsp from the target\n");

	/* set up timer and wait on task_mgt_sem until the timer expires */
	/* or we receive the task management response from the target */
	init_timer(&abort_timer);
	abort_timer.expires = jiffies + 3*HZ;
	abort_timer.data = (unsigned long)&related_session->task_mgt_sem;
	abort_timer.function = abort_timer_function;
	add_timer(&abort_timer);

	/* release our exclusive access to the iscsi structures while waiting */
	spin_unlock(&my_lock);

	/* wait for our timer to expire */
	down(&related_session->task_mgt_sem);

	/* get back exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	del_timer(&abort_timer);

	/*	we have to do all this all over again because we may have been
		"away" for a long time and our operating environment could have
		changed during that interval -- the session and/or connection
		could have been closed (unlikely) or the command could have
		completed  and been removed from the pending commands list
		(more likely).
	*/
	if( hostdata != global_hostdata )
		{
		goto do_nothing;
		}

	related_session = hostdata->session_head;

	/* Find out the session that corresponds to the target ID */
	while( related_session != NULL )
		{
		if( related_session->scsi_target_id == Cmnd->target )
			break;
		related_session = related_session->next;
		}
 
	/* check if we have session info. for that target */
	if( related_session == NULL )
        {/* the session was closed */
        goto do_nothing;
        }

	/* check if the session is in login phase or not */
    if( related_session->session_state != STATUS_LOGGED_IN  ||
		related_session->tx_thread == NULL  ||
		related_session->connection_head == NULL  ||
		related_session->connection_head->rx_thread == NULL )
        {
        goto do_nothing;
        }
	
	/* get ptr. to the head of connection linked list */
	if( (related_connection = related_session->connection_head) == NULL )
        {/* the connection was closed */
		goto do_nothing;
		}

	/* search for the PDU containing the SCSI command which was aborted */
    for( related_scsi_command = related_connection->pending_commands;
		                    related_scsi_command != NULL;
                            related_scsi_command = related_scsi_command->next )
        {
		if( related_scsi_command->init_task_tag == related_itt )
			break;
        }

    if( related_scsi_command != NULL )
        {/* Free command struct for the SCSI command that had to be aborted */
		related_scsi_command->SCpnt->result = DID_ERROR << 16; /* error recovery - SAI */
		free_pending_command(related_scsi_command, related_connection);
    	TRACE(TRACE_DEBUG, "Freed related command, ITT %u\n", related_itt);
        }
	/*	else the command to be aborted disappeared while we were "away" */

	/* search for the PDU containing the Task Management command */
    for( tm_command = related_connection->pending_commands;
		                    tm_command != NULL;
                            tm_command = tm_command->next )
        {
		if( tm_command == task_mgt_command )
			{/* Save the response byte, then free the task_mgt_command struct */
			response = task_mgt_command->task_mgt_response;
			free_pending_command(task_mgt_command, related_connection);
			break;
			}
        }

do_nothing:

    TRACE(TRACE_DEBUG, "Task mgt response 0x%02x\n", response);

    /* if( response == FUNCTION_COMPLETE ) */ /* error recovery - SAI */
		{/* task management command itself completed successfully */
        retval = SCSI_ABORT_SUCCESS;

		/* the command was aborted successfully */
		Cmnd->result = DID_ABORT << 16;
		if( Cmnd->scsi_done != NULL )
			Cmnd->scsi_done(Cmnd);
		}
	/* else task management command itself did not complete successfully */
	/* not clear what to do about the Cmnd itself in this case */

out:

	/* release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_abort, 0x%04x\n", retval);
	return retval;
	}
	
/*****************************************************************************
* This function is called by the SCSI Mid-level.
* need to implement this function
*****************************************************************************/
int iscsi_initiator_reset(Scsi_Cmnd * Cmnd )
	{
	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_reset\n");
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_reset, 0x%04x\n",
															SCSI_RESET_SUCCESS);
	return SCSI_RESET_SUCCESS;
	}

	
/****************************************************************************
*	Initializes the fields of the first connection in a new session,
*	connects to a remote target, logs in to it,
*	and creates the rx_thread for that connection
*	if full feature phase has been attained successfully.
*	Called by create_session() when adding new connection to existing session
*	and by init_session() when this is first connection in a new session.
*	Return value:	=1 for redirection,
*					=0 for any other success,
*					<0 for error
****************************************************************************/

int init_connection( struct connection *conn, unsigned int when_called,
					 struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
					 struct chap_parameter_type chap_parameter,
					 unsigned int *redirect_ip, unsigned int *redirect_port )
/* chap support - CHONG */
	{
	mm_segment_t		oldfs;
	int					retval, optlen;
	unsigned int		ip = conn->target_address;
	struct sockaddr_in	serv_addr;
	struct session		*sess = conn->my_session;
	/* chap support - CHONG */
	struct chap_parameter_type new_chapparam;


	TRACE(TRACE_ENTER_LEAVE,
				"Enter init_connection, conn %p, sess %p, ip 0x%08x, port %u\n",
				conn, sess, ip, conn->target_port);

	TRACE(TRACE_ISCSI, "Create connection %p, target %u CID %u\n",
							conn, sess->scsi_target_id, conn->connection_id);

	/* chap support- CHONG */
	new_chapparam.chap_flags = 0;
	new_chapparam.local_ctx = NULL;
	new_chapparam.peer_ctx = NULL;

	/*	finish allocating fields of new struct connection */

	retval = sock_create(AF_INET, SOCK_STREAM, 0, &conn->sock);

	TRACE(TRACE_NET, "sock_create returned %u\n", retval);

	if( retval < 0 )
		{
		TRACE_ERROR("Could not create socket\n");
		goto out;
		}

	TRACE(TRACE_NET, "Create socket %p\n", conn->sock);

	memset(&serv_addr, 0, sizeof(struct sockaddr_in));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(conn->target_port);
	serv_addr.sin_addr.s_addr = htonl(ip);

	retval = conn->sock->ops->connect(conn->sock, (struct sockaddr *)&serv_addr,
													sizeof(struct sockaddr), 0);

	printk("iscsi: %s to %u.%u.%u.%u:%u\n",
					retval<0? "***ERROR*** Could not connect" : "Connected",
					(ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff,
					ip & 0xff, conn->target_port);

	if( retval < 0 )
		goto out1;

	/*	Need to turn off the Nagle Algorithm on this connection */
	/*	4th parameter is any non-zero value to mean "Turn NODELAY on" */
	oldfs = get_fs(); set_fs(get_ds());
	retval = conn->sock->ops->setsockopt(conn->sock, SOL_TCP, TCP_NODELAY,
													(char *)&sess, sizeof(int));
	set_fs(oldfs);
	if( retval < 0 )
		{
		TRACE_ERROR("Can't turn off Nagle Algorithm, retval %d\n", retval);
		}
	else
		{/* looks like we were able to turn it off, verify it */
		optlen = sizeof(int);
		oldfs = get_fs(); set_fs(get_ds());
		conn->sock->ops->getsockopt(conn->sock, SOL_TCP, TCP_NODELAY,
													(char *)&retval, &optlen);
		set_fs(oldfs);

		printk("iscsi: Nagle Algorithm is %s\n", retval ? "off" : "on");
		}


	/*	allocate buffer into which rx_thread will read PDU headers from target*/
	/*	for now, we ignore AHS, so space is needed only for BHS and digest */
	conn->rx_buf = (u_char *)my_kmalloc(ISCSI_HDR_LEN+CRC_LEN, "rx_buf");
		
	if( conn->rx_buf == NULL)
		{/* no memory for rx_buf */
		retval = -ENOMEM;
		goto out1;
		}

	/*	we are connected to a target */
	conn->connection_state = STATUS_CONNECTED;

	/* chap support - CHONG */
	new_chapparam.chap_flags = chap_parameter.chap_flags;
	new_chapparam.local_ctx = CHAP_CloneContext(chap_parameter.local_ctx);
	new_chapparam.peer_ctx = CHAP_CloneContext(chap_parameter.peer_ctx);

	/*	now go do the login phase for this new connection */
	if( (retval = iscsi_initiator_login(conn, when_called, p_param_tbl, new_chapparam,
		redirect_ip, redirect_port)) != 0 )
		{
		goto out2;
		}
		
	/* Create connection's rx_thread only if get into Full Feature Phase */
	kernel_thread((int (*)(void *))iscsi_initiator_rx_thread, (void *)conn, 0);

out:
	CHAP_FinalizeContext(new_chapparam.peer_ctx);
	CHAP_FinalizeContext(new_chapparam.local_ctx);
	TRACE(TRACE_ENTER_LEAVE, "Leave init_connection, retval %d\n", retval);

	return retval;

out2:
	conn->connection_state = STATUS_NOT_PRESENT;
	my_kfree((void **)&conn->rx_buf, "rx_buf");

out1:
	sock_release(conn->sock);
	conn->sock = NULL;
	goto out;
	}
	

/****************************************************************************
*	Called only by create_session() to
*	1.	initialize a new session structure
*	2.	create a new first connection for the session
*	3.	connect to the target
*	4.	conduct the complete login phase with the target
*	
*	Return value:	=1 on redirection
*					=0 on all other successes,
*					<0 on failure
****************************************************************************/
int init_session( struct session *sess, struct connection *conn,
		unsigned int lun, struct parameter_type param_tbl[MAX_CONFIG_PARAMS],
		struct chap_parameter_type chap_param,
		unsigned int *redirect_ip, unsigned int *redirect_port )
/* chap support - CHONG */
	{
	int		retval = 0;


	TRACE(TRACE_ENTER_LEAVE,"Enter init_session, sess %p\n", sess);

   /* if not already allocated, allocate now - new 18_04 SAI */
    if (!sess->oper_param)
        {
		/*	finish allocating remaining structures new struct session points to */
		sess->oper_param = (struct session_operational_parameters *) 
					my_kmalloc(sizeof(struct session_operational_parameters),
					"session oper_param");
		}

	if( sess->oper_param == NULL )
		{/* no memory for operational parameters structure */
		retval = -ENOMEM;
		goto out;
		}

	/* always set up timer to restart tx_thread whenever necessary */
	if( (sess->tx_timer = (struct timer_list *) 
					my_kmalloc(sizeof(struct timer_list), "tx_timer")) == NULL )
		{/* no memory for tx_timer */
		retval = -ENOMEM;
		goto out1;
		}

	/*	from timer.h: "Timers may be dynamically created and destroyed, and
		should be initialized by a call to init_timer() upon creation."
	*/
	init_timer(sess->tx_timer);		/* we MUST do this right after creation */

	TRACE(TRACE_ISCSI, "Create session %p, target %u lun %u CID %u\n",
						sess, sess->scsi_target_id, lun, conn->connection_id);

	/* Connect to the target and go through the complete login phase */
	/* chap support - CHONG */
	if( (retval = init_connection(conn, LEADING_ONLY|INITIAL_ONLY|ALL,
					*sess->session_params, chap_param, redirect_ip, redirect_port)) != 0 )
		goto out2;

	sess->session_state = STATUS_LOGGED_IN;

	/* successfully created a new connection for a new session */
	/* Set up the session-wide negotiated operational parameters for this
	   new session */
	set_session_parameters(sess->oper_param, *sess->session_params,
															sess->draft_number);

	TRACE(TRACE_DEBUG, "ErrorRecoveryLevel %u\n",
										sess->oper_param->ErrorRecoveryLevel);

	/* setup the retransmit timer - SAI */
	if (sess->oper_param->ErrorRecoveryLevel > 0
			&& sess->retran_period > 0
			&& sess->xmit_chk_interval > 0)
		{
		if ((sess->retran_timer = (struct timer_list *)
			my_kmalloc(sizeof(struct timer_list), "retran_timer")) == NULL)
			{
			retval = -ENOMEM;
			goto out1;
			}

		init_timer(sess->retran_timer);
		}
	else
		{
		sess->retran_timer = NULL;
		}

	/* setup the conn recovery timer - SAI */
	if (sess->oper_param->ErrorRecoveryLevel > 1)
		{
		if ((sess->connrec_timer = (struct timer_list *)
			my_kmalloc(sizeof(struct timer_list), "connrec_timer")) == NULL)
			{
			retval = -ENOMEM;
			goto out1;
			}
			init_timer(sess->connrec_timer);
		}
	else
		{
		sess->connrec_timer = NULL;
		}

out:

	TRACE(TRACE_ENTER_LEAVE, "Leave init_session, retval %d\n", retval);
	
	return retval;	


out2:
	my_kfree((void **)&sess->tx_timer, "tx_timer");

out1:
	my_kfree((void **)&sess->oper_param, "session oper_param");
	goto out;
	}
	
/****************************************************************************
*	Creates and initializes a new session and its first connection struct by 
*	calling corresponding functions. It then tries to connect to the target.
*	Called only from iscsi_initiator_proc_info() when a line such as
*		"iscsi_initiator ip 1234abcd port 4000 target 1 lun 2 cid 3"
*	is written to /proc/scsi/iscsi_initiator/0
*	Return value:	=1 on success -- new session added, logged in, now in FFP
*					=2 on success -- new lun added to existing session
*					=0 on success -- new connection added, logged in, now in FFP
*									 or new discovery session added, now in FFP
*					<0 on failure -- value is errno negated
****************************************************************************/
int create_session( __u32 target, __u32 ip, __u32 port, __u32 lun, __u32 cid,
														struct Scsi_Host *host )
	{
	int						i, retval, redirect_count = 0;
	unsigned int			redirect_ip, redirect_port;
	struct session			*new_session;
	struct connection		*conn;
	struct iscsi_hostdata	*hostdata = (struct iscsi_hostdata *)host->hostdata;


	TRACE(TRACE_ENTER_LEAVE, "Enter create_session, ip 0x%08x, port %u, "
					"target %u, lun %u, cid %u\n", ip, port, target, lun, cid);

retry:	/* here to try again on a redirection */

	redirect_ip = redirect_port = retval = 0;

	new_session = (struct session *)my_kmalloc(sizeof(struct session),
																"session");
	if( new_session == NULL )
		{/* no memory for new session structure */
		retval = -ENOMEM;
		goto out;
		}

	/*	reset all fields in new struct session to 0 (although few remain 0) */
	memset(new_session, 0, sizeof(struct session));

	/*	initialize most of the new struct session fields here */
	new_session->scsi_target_id = target;
	new_session->lun_bits = 1 << lun;
	new_session->session_state = STATUS_NOT_PRESENT;
	new_session->cur_task_tag = 88888;		/* something unique for testing */
	new_session->cur_cmd_sn = 22222;		/* something unique for testing */
	new_session->max_cmd_sn = new_session->exp_cmd_sn =
													new_session->cur_cmd_sn + 1;
	init_MUTEX_LOCKED(&new_session->tx_sem);
	init_MUTEX_LOCKED(&new_session->tx_done_sem);
    init_MUTEX_LOCKED(&new_session->task_mgt_sem);

	/*	use the configured values to initialize following fields */
	new_session->session_flags = hostdata->init_sess_flags;
	new_session->init_conn_flags = hostdata->init_conn_flags;
	new_session->nop_period = hostdata->nop_period;
	new_session->draft_number = hostdata->draft_number;

	TRACE(TRACE_DEBUG, "session->draft_number %u\n", hostdata->draft_number);

	/* Added data and status snack support - SAI */
	new_session->init_snack_flg = global_hostdata->init_snack_flg;

	/* Added the retranmit check interval and wait period - SAI */
	new_session->xmit_chk_interval = global_hostdata->xmit_chk_interval;
	new_session->retran_period = global_hostdata->retran_period;

	/* Store the connection scheduling scheme - SAI */
	new_session->sched_scheme = global_hostdata->sched_scheme;

	/* Initialize the connection recovery related fields - SAI */
	new_session->n_rec_tasks = 0;


	/* allocate space for configured parameters to use during session login */
	if( (new_session->session_params = my_kmalloc(
								MAX_CONFIG_PARAMS*sizeof(struct parameter_type),
								"session param_tbl")) == NULL )
		{/* no memory for session parameter table */
		retval = -ENOMEM;
		goto out1;
		}

	/* init session parameters with copy of the configured parameter table */
	param_tbl_cpy(*new_session->session_params, *hostdata->param_tbl);

	/* create the first connection on this session */		
	new_session->nconnections = 1;
	new_session->connection_head = conn = (struct connection *)
							my_kmalloc(sizeof(struct connection), "connection");
	if( conn == NULL )
		{/* no memory for new connection structure */
		retval = -ENOMEM;
		goto out2;
		}

	/*	reset all fields in new struct connection (although few remain 0) */
	memset(conn, 0, sizeof(struct connection));

	conn->target_address = ip;
	conn->target_port = port;
	conn->connection_id = cid;
	conn->max_send_length = 8192;/* target's default MaxRecvDataSegmentLength */
	conn->max_recv_length = 8192;/* initiator's MaxRecvDataSegmentLength too */
	conn->nop_counter = 1;
	conn->my_session = new_session; 
	conn->connection_flags = new_session->init_conn_flags;
	conn->connection_state = STATUS_NOT_PRESENT;
	init_MUTEX_LOCKED(&conn->rx_done_sem);	/* unlocked when rx_thread done */

	/* add the session and/or connection to the global hostdata structure */
	retval = add_session_to_hostdata(conn, lun, new_session, host);

	/*	retval < 0: error, everything already exists or too many sessions
		retval = 0: adding a new connection (only)  to an existing session
		retval = 1: adding complete new session, lun and connection
		retval = 2: adding a new lun (only)  to an existing session
	*/
	switch( retval )
		{
	case 2:		/* = 2: adding a new lun (only) to an existing session */
		TRACE(TRACE_ISCSI, "Add LUN to session %p, target %u lun %u\n",
													new_session, target, lun);

	default:	/* < 0: error, everything already exits or too many sessions */
		/* free up newly created connection and session structures */
		goto out3;

	case 0:		/* = 0: adding a new connection (only) to an existing session */
		/* connection must be negotiated using same draft as existing session */
		if( conn->my_session->draft_number != new_session->draft_number )
			{
			TRACE_ERROR("Can't negotiate connection with draft %u belonging to "
						"session with draft %u\n",
						conn->my_session->draft_number,
						new_session->draft_number);
			goto out3;
			}
		/* finish initializing new connection and then log it in */
		if( (retval = init_connection(conn, INITIAL_ONLY|ALL,
					*new_session->session_params,
					hostdata->chap_parameter, 
					&redirect_ip, &redirect_port)) /* chap support - CHONG */
																		!= 0 )
			{/* could not bring up the new connection, remove it from session */
			goto out4;
			}

		/* wait for the new connection's rx_thread to set itself up */
		down_interruptible(&conn->rx_done_sem);

		/* don't need the session structure or the session parameter table */
		goto out2;

	case 1:		/* = 1: adding complete new session, lun and connection */
		/*	Finish initializing the session and its first connection,
			which includes logging it in */
		if( (retval=init_session(new_session, conn, lun, *hostdata->param_tbl,
										hostdata->chap_parameter,
										&redirect_ip, &redirect_port)) != 0 )
										/* chap support - CHONG */
			{/* initialization or login failed or got redirection */
			goto out4;
			}
		/* Add the connection to round robin connection head - SAI */
		new_session->rr_conn_head = conn;
		new_session->nconnections = 1;
		retval = 1;

		/* start up the single tx_thread for this new session */
		kernel_thread((int (*)(void *))iscsi_initiator_tx_thread,
														(void *)new_session, 0);

		/* wait for the threads to set themselves up */
		down_interruptible(&conn->rx_done_sem);
		down_interruptible(&new_session->tx_done_sem);

		/* everything is now up and running in the new session */
		if( new_session->oper_param->SessionType )
			{/* a discovery session, we have to drive it */
			TRACE(TRACE_ISCSI, "Enter discovery session Full Feature Phase\n");
			if( drive_text_negotiate(conn, new_session, 1) )
				retval = 0;
			}
		else
			{/* a normal session, SCSI Mid-level drives it via queuecommand() */
			TRACE(TRACE_ISCSI, "Enter normal session Full Feature Phase\n");
			}
		}	/* switch */

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave create_session, retval %d\n", retval);

	return retval;

out4:
	/* Here to recover from a late error, after host's list was updated. */
	if( (i = remove_session_from_hostdata(target, lun, cid, hostdata)) == 1 )
		{/* memory for entire session already freed, just leave */
		goto out0;
		}
	else if( i == 0 )
		{/* only memory for new connection freed, get rid of new session too */
		goto out2;
		}

	/* should never happen */
	TRACE_ERROR("non-existent target %u LUN %u CID %u\n", target, lun, cid);

out3:
	/* here to free all new session structs and the connection struct */
	my_kfree((void **)&new_session->connection_head, "connection");

out2:
	/* here to free structure for configured parameters used in login */
	param_tbl_uncpy(*new_session->session_params);
	my_kfree((void **)&new_session->session_params, "session param_tbl");

out1:
	/* here to free new session structure */
	my_kfree((void **)&new_session, "session");

out0:
	/* now check to see if failure was caused by request for redirection */
	if( retval == STAT_CLASS_REDIRECTION )
		{
		if( redirect_count++ < MAX_REDIRECTION_RETRIES )
			{
			TRACE(TRACE_ISCSI,"Attempting Redirection number %u\n",
															redirect_count);
			ip = redirect_ip;
			port = redirect_port;
			goto retry;
			}
		TRACE_ERROR("Too many redirection attempts, max of %d\n",
													MAX_REDIRECTION_RETRIES );
		retval = -ECONNABORTED;		/* software caused connection abort */
		}
	goto out;
	}


/****************************************************************************
*	This function adds a new session struct to the end of the hostdata's
*	session_head linked list of sessions
*	Return value:	=1 on success -- new session struct added to list ok
*					=2 on success -- adding new lun to existing session
*					=0 on success -- adding new connection to existing session
*					<0 on failure -- value is errno negated
****************************************************************************/
int add_session_to_hostdata(struct connection *new_connection, unsigned int lun,
						struct session *new_session, struct Scsi_Host *host)
	{
	struct session			*old_session;
	struct connection		*conn, *prev;
	int						retval = 1;
	struct iscsi_hostdata	*hostdata = (struct iscsi_hostdata *)host->hostdata;
	

	TRACE(TRACE_ENTER_LEAVE, "Enter add_session_to_hostdata, new_session %p\n",
																new_session);

	/*	When changing the session list, be sure nobody else tries to use it */
	spin_lock(&my_lock);

	if( (old_session = hostdata->session_head) == NULL )
		{/* this is the first session on the list for this host */
		hostdata->session_head = new_session;
		hostdata->nsessions = 1;
		}
	else
		{/* other sessions already on the list for this host,  add this
			at end of list unless session with same target id already there */
		do	{
			if( old_session->scsi_target_id == new_session->scsi_target_id )
				{/* already have a session with this target id */

				/*	is this adding a different LUN to same target? */
				if( !(old_session->lun_bits & new_session->lun_bits) )
					{/* yes, multiple luns can share same session */
					old_session->lun_bits |= new_session->lun_bits;
					retval = 2;
					}
				else
					{/* found identical luns */
					/*	is this adding a new connection to existing session? */
					retval = 0;		/* assume it is adding a new connection */
					prev = NULL;
					for( conn = old_session->connection_head;  conn != NULL;
															conn = conn->next )
						{
						if( conn->connection_id==new_connection->connection_id )
							{/* already have a connection with this cid */
							TRACE_ERROR("Already have target %u LUN %u "
										"CID %u\n", new_session->scsi_target_id,
										lun, new_connection->connection_id);
							retval = -EBUSY;
							break;
							}
						prev = conn;
						}
					if( retval == 0 )
						{/* add new connection to end of existing list */
						/*  but first be sure session allows more connections */
						if( old_session->nconnections >=
									old_session->oper_param->MaxConnections )
							{/* session already has max number of connections */
							TRACE_ERROR("Already have maximum of %u connections"
									" for target %u\n",
									old_session->nconnections,
									old_session->scsi_target_id);
							retval = -EBUSY;
							break;
							}

						old_session->nconnections++;
						if( prev == NULL )
							old_session->connection_head = new_connection;
						else
							prev->next = new_connection;
						/* override initializations that depend on session */
						new_connection->my_session = old_session; 
						new_connection->connection_flags =
												old_session->init_conn_flags;
						}
					}
				break;
				}

			/* no target match so far, are we at end of host's session list? */
			if( old_session->next == NULL )
				{/* at end of host's list, try to add new session here */
				if( hostdata->nsessions < host->max_id )
					{/* ok to add another session */
					old_session->next = new_session;
					hostdata->nsessions++;

					/* Strore the connection in connection list - SAI */
					new_session->rr_conn_head = new_connection;
					new_session->nconnections = 1;
					}
				else
					{/* this host already has maximum number of sessions */
					TRACE_ERROR("Max of %u sessions, can't add session for "
									"target %u LUN %u CID %u\n", host->max_id,
									new_session->scsi_target_id,
									lun, new_connection->connection_id);
					retval = -EBUSY;
					}
				break;
				}

			/* more sessions on host's list, continue checking */
			old_session = old_session->next; 
			}
		while( 1 );
		}
	
	spin_unlock(&my_lock);

	TRACE(TRACE_ENTER_LEAVE, "Leave add_session_to_hostdata, retval %d\n",
																		retval);

	return retval;
	}


/****************************************************************************
*	Removes the session struct for this target from the session struct list 
*	Return value:	=1 on success -- complete session removed
*					=2 on success -- old lun removed from old session
*					=0 on success -- old connection removed from old session
*					<0 on failure -- value is errno negated
****************************************************************************/
int remove_session_from_hostdata( __u32 target, __u32 lun, __u32 cid,
											struct iscsi_hostdata *hostdata )
	{
	int					retval = -ENODEV;	/* assume failure */
	struct session		*current_session;
	struct session		*prev;
	struct connection	*conn, *prev_conn;
	

	TRACE(TRACE_ENTER_LEAVE, "Enter remove_session_from_hostdata, target %u "
													"lun %u\n", target, lun);

	/*	When changing the session list, be sure nobody else tries to use it */
	spin_lock(&my_lock);

	for( prev = NULL, current_session = hostdata->session_head;
			current_session != NULL; current_session = current_session->next )
		{
		if( current_session->scsi_target_id == target )
			{/* this is the matching session structure, see if lun matches */
			if( current_session->lun_bits & (1<<lun) )
				{/* lun matches, see if cid matches */
				prev_conn = NULL;
				for( conn = current_session->connection_head;  conn != NULL;
															conn = conn->next )
					{
					if( conn->connection_id == cid )
						{/* found a matching connection */
						if( current_session->nconnections > 1 )
							{/* at least 1 more connection will remain */
							current_session->nconnections--;
							if( prev_conn == NULL )
								current_session->connection_head = conn->next;
							else
								prev_conn->next = conn->next;
							retval = 0;
							}
						else
							{/* this is last connection, remove entire session*/
							current_session->lun_bits = 0;
							conn = NULL;
							}
						break;
						}
					prev_conn = conn;
					}
				if( conn == NULL )
					{/*	did not find matching connection, or matching connection
					was only connection left, turn off bit for lun */
					current_session->lun_bits &= ~(1<<lun);
					if( current_session->lun_bits == 0 )
						{/* no luns remain, remove entire session */
						if( prev == NULL )
							hostdata->session_head = current_session->next;
						else
							prev->next = current_session->next;
						hostdata->nsessions--;
						retval = 1;
						}
					else
						{/* other luns remain, don't do anything else here */
						retval = 2;
						}
					}
				}
			else
				{/* lun does not match, remove entire session */
				current_session->lun_bits = 0;
				if( prev == NULL )
					hostdata->session_head = current_session->next;
				else
					prev->next = current_session->next;
				hostdata->nsessions--;
				retval = 1;
				}
			break;
			}
		prev = current_session;
		}
	
	spin_unlock(&my_lock);

	if( retval == 0 )
		{/* close only this connection, which does not release the socket yet */
		close_connection(conn);
		/* go back and release the socket and free the structure */
		if( conn->sock != NULL )
			{
			TRACE(TRACE_NET, "Releasing socket %p\n", conn->sock);
			sock_release(conn->sock);
			TRACE(TRACE_NET, "Released socket %p\n", conn->sock);
			conn->sock = NULL;
			}
		my_kfree((void **)&conn, "connection");		/* now release this! */
		}

	else if( retval == 1 )
		{/* close this entire session, including all remaining connections */
		TRACE(TRACE_ISCSI, "Remove entire session %p, target %u\n",
													current_session, target);

		/*	free up any storage connected to this old session */
		if( current_session->session_state == STATUS_NOT_PRESENT )
			{/* session's connection structure is empty, just free it */
			my_kfree((void **)&current_session->connection_head, "connection");
			}
		else
			close_session(current_session);

		/*	finally free up the old struct session itself */
		param_tbl_uncpy(*current_session->session_params);
		my_kfree((void **)&current_session->session_params,"session param_tbl");
		my_kfree((void **)&current_session, "session");
		}

	else if( retval < 0 )
		{
		TRACE(TRACE_ISCSI, "non-existent target %u LUN %u CID %u\n",
															target, lun, cid);
		}

	else
		{/* retval = 2, just removing another lun, do nothing further here */
		TRACE(TRACE_ISCSI, "Remove LUN from session %p, target %u lun %u\n",
												current_session, target, lun);
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave remove_session_from_hostdata, retval %d\n",
																	retval);
	return retval;
	}


/*	When called, spin_lock MUST be held by caller */
/*	Returns 0 on success (and sets old_connection, old_session), <0 on failure*/
int find_connection( unsigned int target, unsigned int cid,
						struct iscsi_hostdata *hostdata,
						struct connection **old_connection,
						struct session **old_session )
	{
	struct session			*session;
	struct connection		*conn;
	int						retval = -ENODEV;	/* assume failure */
	

	TRACE(TRACE_ENTER_LEAVE, "Enter find_connection, target %u cid %u\n",
																target, cid);

	for( session = hostdata->session_head; session != NULL;
									session = session->next )
		{
		if( session->scsi_target_id == target )
			{/* this is the matching session structure, find matching cid */
			for( conn = session->connection_head;  conn != NULL;
															conn = conn->next )
				{
				if( conn->connection_id == cid )
					{/* found a matching connection, pass back answers */
					*old_connection = conn;
					*old_session = session;
					retval = 0;
					break;
					}
				}
			break;
			}
		}

	TRACE(TRACE_ENTER_LEAVE, "Leave find_connection, retval %d\n", retval);

	return retval;
	}


int do_test( unsigned int target, unsigned int cid, unsigned int test_no )
	{
	int					result;
	struct connection	*conn;
	struct session		*sess;


	/*	When using session list, be sure nobody else accesses it */
	spin_lock(&my_lock);

	/* use target and cid to find session and connection */
	result = find_connection(target, cid, global_hostdata, &conn, &sess);

	if( result >= 0 )
		{/* perform the requested test on this connection */

		TRACE(TRACE_ISCSI, "Do test number %u\n", test_no );

		switch( test_no )
			{
		case 2:	/* send nop (not immediate, no data) */
			result = drive_nopout(conn, sess, 0, 0);
			break;
		case 3:	/* send nop (immediate, no data) */
			result = drive_nopout(conn, sess, I_BIT, 0);
			break;
		case 4:	/* send nop ping (not immediate, data) */
			result = drive_nopout(conn, sess, 0, 1);
			break;
		case 5:	/* send nop ping (immediate, data) */
			result = drive_nopout(conn, sess, I_BIT, 1);
			break;
		case 6:	/* stop/start responding to nop ping from target */
			conn->connection_flags ^= SEND_NO_REPLY_TO_NOP_PING;
			break;
		case 7:	/* stop/start responding to async logout request from target */
			conn->connection_flags ^= SEND_NO_REPLY_TO_ASYNC_LOGOUT;
			break;
		case 8: /* perform connection re-assignment - SAI */
			conn->rec_tests = test_no;
			break;
		default:
			result = -EINVAL;
			break;
			}	/* switch */
		}

	/* release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	return result;
	}


/***************************************************************************
*	Cleans up a connection struct whenever a connection is ended.
*	This does NOT close the socket or free the connection structure
*	Called from close_session() and remove_session_from_hostdata()
*	When called, my_lock is NOT held
***************************************************************************/
void close_connection( struct connection *conn )
	{
	void			*temp;


	TRACE(TRACE_ENTER_LEAVE, "Enter close_connection %p\n", conn);

	if( conn == NULL  ||  conn->connection_state == STATUS_NOT_PRESENT )
		{/* nothing to do in this case, really shouldn't be here, but ... */
		goto out;
		}

	TRACE(TRACE_ISCSI, "Closing connection %p, target %u CID %u\n", conn,
						conn->my_session->scsi_target_id, conn->connection_id);

	/* get exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	if( conn->rx_thread != NULL )
		{/* thread not previously shut down, do it now */
        /* changes new 18_04 SAI */
		/*** if( conn->connection_state == STATUS_LOGGED_IN ) ***/
			{/* rx_thread still up and running, kill it */
			TRACE(TRACE_ISCSI, "Killing rx_thread\n");
			temp = conn->rx_thread;
			conn->rx_thread = NULL;	/* marks that we are shutting thread down */
			send_sig(SIGKILL, temp, 1);
			}
		
			{/* rx_thread already shut itself down */
			conn->rx_thread = NULL;
			}

		/* then release our exclusive access to the iscsi structures */
		spin_unlock(&my_lock);

		TRACE(TRACE_SEM, "close_connection waiting on rx_done_sem\n");
		down_interruptible(&conn->rx_done_sem);
		TRACE(TRACE_SEM, "close_connection done waiting on rx_done_sem\n");

		/* now get back exclusive access to the iscsi structures */
		spin_lock(&my_lock);
		}

	/* Decrement the number of connections - SAI */
	conn->my_session->nconnections--;

	conn->connection_state = STATUS_NOT_PRESENT;
	my_kfree((void **)&conn->rx_buf, "rx_buf");

	/* then release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ISCSI, "Closed connection %p, target %u CID %u\n", conn,
						conn->my_session->scsi_target_id, conn->connection_id);

out:
	TRACE(TRACE_ENTER_LEAVE, "Leave close_connection\n");
	}


/***************************************************************************
*	Cleans up session struct whenever a session has to be ended.
*	When called, this session struct is no longer linked into the global list.
*	On exit, this session struct is NOT freed.
*	Called only from remove_session_from_hostdata()
*	When called, my_lock is NOT held
***************************************************************************/
void close_session(struct session *sess)
	{
	struct connection	*conn;
	struct connection	*next_conn;
	void				*temp;


	TRACE(TRACE_ENTER_LEAVE, "Enter close_session\n");

	TRACE(TRACE_ISCSI, "Closing session %p, target %u\n",
												sess, sess->scsi_target_id);

	/* get exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	/* do not allow tx_timer to restart itself */
	sess->nop_period = 0;

	if( sess->tx_thread != NULL )
		{/* no other thread is already shutting this session down */

		/* indicate that this thread is going to do the shut down */
		TRACE(TRACE_ISCSI, "killing tx_thread\n");
		temp = sess->tx_thread;
		sess->tx_thread = NULL;		/* marks that we are shutting thread down */
		send_sig(SIGKILL, temp, 1);

		/* then release our exclusive access to the iscsi structures */
		spin_unlock(&my_lock);

		/* then wait for tx thread to actually die */
		TRACE(TRACE_SEM, "Waiting on tx_done_sem \n");
		down_interruptible(&sess->tx_done_sem);
		TRACE(TRACE_SEM, "Done waiting on tx_done_sem \n");

		}
	else
		{/* some other thread is already shutting this session down */

		/* just release our exclusive access to the iscsi structures */
		spin_unlock(&my_lock);
		}

	/* close all connections, but do not release the socket yet */
	for( conn = sess->connection_head;  conn != NULL;  conn = next_conn )
		{
		next_conn = conn->next;	
		TRACE(TRACE_DEBUG, "Going to close_connection(%p), next_conn %p\n",
															conn, next_conn);
		close_connection(conn);
		TRACE(TRACE_DEBUG, "Back from close_connection(%p), next_conn %p\n",
															conn, next_conn);
		}

	TRACE(TRACE_ISCSI, "All connections closed for session %p, target %u\n",
												sess, sess->scsi_target_id);

	/* get exclusive access to the iscsi structures */
	spin_lock(&my_lock);

	/* do not allow tx_timer to restart itself */
	sess->nop_period = 0;

	/* now delete the tx_timer (del_timer checks for timer active or not) */
	if( sess->tx_timer != NULL )
		{
		TRACE(TRACE_DEBUG, "Deleting tx_timer %p\n", sess->tx_timer);
		del_timer(sess->tx_timer);
		TRACE(TRACE_DEBUG, "Deleted tx_timer\n");
		my_kfree((void **)&sess->tx_timer, "tx_timer");
		}

	/* delete command retransmission timers - SAI */
	if (sess->retran_timer != NULL)
		{
		TRACE(TRACE_DEBUG, "Deleting retran_timer %p\n", sess->retran_timer);
		del_timer(sess->retran_timer);
		TRACE(TRACE_DEBUG, "Deleted retran_timer\n");
		my_kfree((void **)&sess->retran_timer, "retran_timer");
		}

	/* delete connection recovery timer - SAI */
	if (sess->connrec_timer != NULL)
		{
		TRACE(TRACE_DEBUG, "Deleting connrec_timer %p\n", sess->connrec_timer);
		del_timer(sess->connrec_timer);
		TRACE(TRACE_DEBUG, "Deleted connrec_timer\n");
		my_kfree((void **)&sess->connrec_timer, "connrec_timer");
		}

	/* go back and release the connections' sockets and free the structures */
	for( conn = sess->connection_head;  conn != NULL;  conn = next_conn )
		{
		next_conn = conn->next;	
		if( conn->sock != NULL )
			{
			TRACE(TRACE_NET, "Releasing socket %p\n", conn->sock);
			sock_release(conn->sock);
			TRACE(TRACE_NET, "Released socket %p\n", conn->sock);
			conn->sock = NULL;
			}
		my_kfree((void **)&conn, "connection");			/* now release this! */
		}

	/* the connection list for this session is now empty */
	sess->connection_head = NULL;

	my_kfree((void **)&sess->oper_param, "session oper_param");
	
	/* now release our exclusive access to the iscsi structures */
	spin_unlock(&my_lock);

	TRACE(TRACE_ISCSI, "Closed session %p, target %u\n",
												sess, sess->scsi_target_id);

	TRACE(TRACE_ENTER_LEAVE, "Leave close_session\n");
	}


/******************************************************************************
*	Called when the host file "0" or "1" or ...
*	in the /proc/scsi/iscsi_initiator/ directory is read or written.
*	The parameter "inout" is 0 for a read, 1 for a write operation.
*	A write expects the buffer to contain something in the format:
*		"iscsi_initiator ip 1234abcd port 4000 target 6 lun 0"
*	or
*		"iscsi_initiator manage 5 key=value"
*	or
		"iscsi_iniitator manage 32 o"
*	or
*		"iscsi_initiator test 7"
*	Return value:	0, on success
*					-Exxxx, on failure 
******************************************************************************/
int iscsi_initiator_proc_info(char *buffer, char **start, off_t offset,
											int length, int inode, int inout)
	{
	int					pos, begin, result, i;
	unsigned int		ip, target, lun, port, cid, c_len;
	char				*ptr;
	int					param_neg_info;
	int					test_no;
	char				*ptr_to_keytext;
	struct session		*sess;
	struct connection	*conn;
	

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_initiator_proc_info\n");
	TRACE(TRACE_DEBUG, "buffer %p, offset %ld, length %d, inout %d\n",
	buffer, offset, length, inout);
	
	if( inout )                                     
		{/* write to this file, buffer should be:
				"iscsi_initiator ip 1234abcd port 4000 target 1 lun 2 cid 3"
			or
				"iscsi_initiator manage
					<neg_info number for "set/setr/unset/setp/restore/force">
					<parameter for "set/setr/unset/setp/restore/force"> "
		*/
		if( global_host == NULL )
			{/* this should never happen */
			TRACE_ERROR("global_host is NULL\n");
			result = -ENODEV;
			}
		else if( global_hostdata == NULL )
			{/* this should never happen */
			TRACE_ERROR("global_hostdata is NULL\n");
			result = -ENODEV;
			}
		else if( buffer == NULL  ||  length > PAGE_SIZE )
			result = -EINVAL;
		else
			{
			result = length;
			if( length == PAGE_SIZE )
				length--;
			buffer[length] = '\0';  /* make the buffer into a string */
			
			TRACE(TRACE_ISCSI, "iSCSI write to proc_info (%d): %s\n",
																length, buffer);

			if( length >= 19 && strncmp(buffer,"iscsi_initiator ip ",19) == 0 )
				{/* this buffer has expected preface, scan ip address */
				
				ip = simple_strtoul(buffer+19, &ptr, 0);
				
				if( strncmp(ptr, " port ", 6) != 0)
					{
					result = -EINVAL;
					goto out;
					}
				
				/* this buffer has expected preface, scan port number */
				port = simple_strtoul(ptr+6, &ptr, 0);

				if( port == 0  ||  port > 0xffff  ||
									strncmp(ptr, " target ", 8) != 0)
					{
					result = -EINVAL;
					goto out;
					}

				/* this buffer has expected preface, scan target number */
				target = simple_strtoul(ptr+8, &ptr, 0);

				if( target >= global_host->max_id  ||
										strncmp(ptr, " lun ", 5) != 0 )
					{
					result = -EINVAL;
					goto out;
					}

				/* this buffer has expected preface, scan lun number */
				lun = simple_strtoul(ptr+5, &ptr, 0);
				
				if( lun >= global_host->max_lun  ||
										strncmp(ptr, " cid ", 5) != 0 )
					{
					result = -EINVAL;
					goto out;
					}
				
				/* this buffer has expected preface, scan cid number */
				cid = simple_strtoul(ptr+5, &ptr, 0);
				
				if( cid > 0xff  ||  strspn(ptr, WHITE_SPACE) != strlen(ptr) )
					{
					result = -EINVAL;
					goto out;
					}
				
				if( ip == 0 )
					{/* "down target lun cid" on this host */
					if( (i = remove_session_from_hostdata(target, lun, cid,
														global_hostdata)) <= 0 )
						{/* error (<0) or old connection (only) (==0)
							pass this back as the result */
						result = i;
						}
					/*	else >0, removed old session or lun, result==length */
					}
				else
					{/* "up target lun cid" to ip:port on this host.  Returns:
						< 0: error
						= 0: added a new connection (only) to existing session
						= 1: added complete new session, lun and connection
						= 2: added a new lun (only) to existing session
					*/
					if( (i = create_session(target, ip, port, lun, cid,
															global_host)) <= 0 )
						{/* error (<0) or new connection (only) (==0)
							pass this back as the result */
						result = i;
						}
					/*	else >0, added new session or lun, result==length */
					}
				}
            else if( length >= 23  &&
						strncmp(buffer, "iscsi_initiator manage ", 23) == 0 )
                {/* this is for configuring the login parameters */
 
                /*	this buffer has expected interface,
					scan number for "set/setr/unset/setp/restore/force" */
                param_neg_info = simple_strtoul( buffer+23, &ptr_to_keytext, 0);
 
                /* Move a position to ignore space */
                ptr_to_keytext++;
 
                /* Now configure the parameters */
				if( param_neg_info == FORCE )
					{/* special case to force security or operational subphase
						during login */
					if( strcmp(ptr_to_keytext, "s") == 0 )
						global_hostdata->init_conn_flags |= USE_SECURITY_PHASE;
					else if( strcmp(ptr_to_keytext, "o") == 0 )
						global_hostdata->init_conn_flags|=USE_OPERATIONAL_PHASE;
					else if( strcmp(ptr_to_keytext, "f") == 0 )
						global_hostdata->init_conn_flags |= USE_FLAT_SPACE_LUN;
					else if( strcmp(ptr_to_keytext, "r") == 0 )
						global_hostdata->init_conn_flags |= USE_FULL_REPLIES;
					else if( strncmp(ptr_to_keytext, "n=", 2) == 0 )
						{/* set periodic timer for sending nop-out to target */
						global_hostdata->nop_period =
									simple_strtoul(ptr_to_keytext+2, NULL, 0);
						}
					else if( strncmp(ptr_to_keytext, "v=", 2) == 0 )
						{/* set draft number to enforce during operation */
						global_hostdata->draft_number =
									simple_strtoul(ptr_to_keytext+2, NULL, 0);
						TRACE(TRACE_ISCSI,"Set Initiator to enforce Draft %u\n",
											global_hostdata->draft_number/100);


						/* now adjust Burst parameter units to match draft */
						adjust_version_diff(*global_hostdata->param_tbl,
												global_hostdata->draft_number);
						}
					/* Store the xmit_chk_interval and retran_interval - SAI */
					else if(strncmp(ptr_to_keytext, "i=", 2) == 0)
						{
						global_hostdata->xmit_chk_interval = simple_strtoul(ptr_to_keytext+2, NULL, 0);
						TRACE(TRACE_DEBUG, "Set Initiator to enforce retransmit check interval as %u\n", global_hostdata->xmit_chk_interval);
						}
					else if(strncmp(ptr_to_keytext, "p=", 2) == 0)
						{
						global_hostdata->retran_period = simple_strtoul(ptr_to_keytext+2, NULL, 0);
						TRACE(TRACE_DEBUG, "Set Initiator to enforce retransmit wait period as %u\n", global_hostdata->retran_period);
						}
					/* Store the connection scheduling scheme - SAI */
					else if(strncmp(ptr_to_keytext, "sch=", 4) == 0)
						{
						global_hostdata->sched_scheme = simple_strtoul(ptr_to_keytext+4, NULL, 0);
						TRACE(TRACE_DEBUG, "Set Initiator to enforce Connection Scheduling Scheme as %u\n", global_hostdata->sched_scheme);
						}
					/* chap support - CHONG */
					else if (strcmp(ptr_to_keytext, "t") == 0)
					{
						TRACE(TRACE_ISCSI, "target confirmation enabled\n");
						global_hostdata->chap_parameter.chap_flags |= USE_TARGET_CONFIRMATION;
					}
					else if (strcmp(ptr_to_keytext, "b") == 0)
					{
						TRACE(TRACE_ISCSI, "base64 format enabled\n");
						global_hostdata->chap_parameter.chap_flags |= USE_BASE64;
						global_hostdata->chap_parameter.local_ctx->number_format = BASE64_FORMAT;
						global_hostdata->chap_parameter.peer_ctx->number_format = BASE64_FORMAT;
					}
				else if (strncmp(ptr_to_keytext, "px=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "peer secret set to \"%s\"\n", ptr_to_keytext+3);
						CHAP_SetSecret(ptr_to_keytext+3, global_hostdata->chap_parameter.peer_ctx);
					}
				else if (strncmp(ptr_to_keytext, "pn=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "peer name set to \"%s\"\n", ptr_to_keytext+3);
						CHAP_SetName(ptr_to_keytext+3, global_hostdata->chap_parameter.peer_ctx);
					}
				else if (strncmp(ptr_to_keytext, "lx=", 3) == 0)
					{
						TRACE(TRACE_ISCSI,"local secret set to \"%s\"\n", ptr_to_keytext+3);
						CHAP_SetSecret(ptr_to_keytext+3, global_hostdata->chap_parameter.local_ctx);
					}
				else if (strncmp(ptr_to_keytext, "ln=", 3) == 0)
					{
						TRACE(TRACE_ISCSI, "local name set to \"%s\"\n", ptr_to_keytext+3);
						CHAP_SetName(ptr_to_keytext+3, global_hostdata->chap_parameter.local_ctx);
					}
				else if (strncmp(ptr_to_keytext, "cl=", 3) == 0)
					{
						c_len = simple_strtoul(ptr_to_keytext+3, NULL, 0);
						if ((c_len <= 0) || (c_len > MAX_CHALLENGE_LENGTH))
						{
							TRACE_ERROR("invalid challenge length: \"%s\"\n", ptr_to_keytext+3);
						}
						else
						{
						TRACE(TRACE_ISCSI, "challenge length set to %d\n", c_len);
						CHAP_SetChallengeLength(c_len, global_hostdata->chap_parameter.local_ctx);
						}
					}
				else
					{
					TRACE_ERROR("Unknown force \"%s\"\n", ptr_to_keytext);
					}
				}
				else if (param_neg_info == SNACK)
					{
					/* Error Recovery Options to enable or disable snack - SAI */
					if( strncmp(ptr_to_keytext, "d=", 2) == 0 )
						{/* enable or disable data snack */
						if ( strcmp(ptr_to_keytext+2, "y") == 0)
							global_hostdata->init_snack_flg |= DATA_SNACK_ENABLE;
						else if ( strcmp(ptr_to_keytext+2, "Y") == 0)
							global_hostdata->init_snack_flg |= DATA_SNACK_ENABLE;
						}
					else if( strncmp(ptr_to_keytext, "s=", 2) == 0 )
						{/* enable or disable status snack */
						if ( strcmp(ptr_to_keytext+2, "y") == 0)
							global_hostdata->init_snack_flg |= STATUS_SNACK_ENABLE;
						else if (strcmp(ptr_to_keytext+2, "Y") == 0)
							global_hostdata->init_snack_flg |= STATUS_SNACK_ENABLE;
					}
					TRACE(TRACE_DEBUG, "SNACK Flags : datasnack %d statsnack %d\n",
							(global_hostdata->init_snack_flg & DATA_SNACK_ENABLE),
							( (global_hostdata->init_snack_flg & STATUS_SNACK_ENABLE) >> 1) );
					}
				else if( param_neg_info == RESTORE )
					{/* restore all settings to default values */

    				/*	reset parameter list to copy of default param set */
					param_tbl_uncpy(*global_hostdata->param_tbl);
					param_tbl_cpy(*global_hostdata->param_tbl, config_params);

					/* chap support - CHONG */
					CHAP_FinalizeContext(global_hostdata->chap_parameter.peer_ctx);
					CHAP_FinalizeContext(global_hostdata->chap_parameter.local_ctx);
					global_hostdata->chap_parameter.chap_flags = 0;
					global_hostdata->chap_parameter.local_ctx = CHAP_InitializeContext();
					global_hostdata->chap_parameter.peer_ctx = CHAP_InitializeContext();
					/*	reset the draft number to the default draft value */
					global_hostdata->draft_number = DEFAULT_DRAFT;

					/*	reset all other fields to 0 */
					global_hostdata->nop_period = 0;
					global_hostdata->init_sess_flags = 0;
					global_hostdata->init_conn_flags = 0;
					global_hostdata->init_sessrec_flg = 0;
					}
				else
					{/* a "normal" key=value type of thing */
					configure_parameter(param_neg_info, ptr_to_keytext,
												*global_hostdata->param_tbl);
					}
				}
			else if(length >=  21  &&
							strncmp(buffer, "iscsi_initiator test ", 21) == 0 )
				{
				/* This is for triggering test actions in the initiator */
				TRACE(TRACE_DEBUG, "trigger test action\n");

				ptr = buffer + 20;

				if( strncmp(ptr, " target ", 8) != 0)
					{
					result = -EINVAL;
					goto out;
					}

				/* this buffer has expected preface, scan target number */
				target = simple_strtoul(ptr+8, &ptr, 0);

				TRACE(TRACE_DEBUG, "got target %u\n", target);

				if( target >= global_host->max_id  ||
										strncmp(ptr, " cid ", 5) != 0 )
					{
					result = -EINVAL;
					goto out;
					}

				/* this buffer has expected preface, scan cid number */
				cid = simple_strtoul(ptr+5, &ptr, 0);
				
				TRACE(TRACE_DEBUG, "got cid %u\n", cid);

				if( cid > 0xff  ||  strncmp(ptr, " test ", 6) != 0 )
					{
					result = -EINVAL;
					goto out;
					}
				
				/* this buffer has expected preface, scan test number */
				test_no = simple_strtoul(ptr+6, &ptr, 0);

				TRACE(TRACE_DEBUG, "got test_no %u\n", test_no);

				if( strspn(ptr, WHITE_SPACE) != strlen(ptr) )
					{
					result = -EINVAL;
					goto out;
					}

				if( (i = do_test(target, cid, test_no)) < 0 )
					result = i;
				}
			else
				{
				TRACE(TRACE_ISCSI, "Illegal write to proc_info\n");
				result = -EINVAL;
				}
			}
		}
	else
		{/* read from this file */
		/*	this is what gets printed by "cat /proc/scsi/iscsi_initiator/0" */
		pos = sprintf(buffer, "%s", OUR_NAME);
		if( global_host != NULL )
			{
			pos += sprintf(buffer+pos, " Draft %u\n",
											global_hostdata->draft_number/100);
			pos += sprintf(buffer+pos, "host is %p, host number is %d\n",
											global_host, global_host->host_no);
			pos += sprintf(buffer+pos, "host has %u sessions\n",
											global_hostdata->nsessions);
			for( sess = global_hostdata->session_head;
								sess != NULL;
								sess = sess->next )
				{
				pos += sprintf(buffer+pos,
									"    session for target %u is %p\n",
									sess->scsi_target_id, sess);
				pos += sprintf(buffer+pos, "    session has luns:");
				for( i = 0;  i < global_host->max_lun;  i++ )
					{
					if( sess->lun_bits & (1<<i) )
						{
						pos += sprintf(buffer+pos, " %d", i);
						}
					}
				pos += sprintf(buffer+pos, "\n    session has %u connections\n",
															sess->nconnections);
				for( conn = sess->connection_head; conn != NULL;
															conn = conn->next )
					{
					i = conn->target_address;
					pos += sprintf(buffer+pos,
										"        connection with CID %u is %p\n"
										"        connection has socket %p ",
										conn->connection_id, conn, conn->sock);
					if( conn->connection_state == STATUS_LOGGED_IN )
						pos += sprintf(buffer+pos,
										"connected to %d.%d.%d.%d:%u\n",
										(i >> 24) & 0xff, (i >> 16) & 0xff,
										(i >> 8) & 0xff, i & 0xff,
										conn->target_port);
					else
						pos += sprintf(buffer+pos, "disconnected\n");
					}
				pos += sprintf(buffer+pos, "\n");
				}
			}
		else
			{
			pos += sprintf(buffer+pos, "\n");
			}
		pos += print_r2t_cookies(buffer+pos);
		pos += print_my_kmemory(buffer+pos);
		pos += print_iovs(buffer+pos);
		if( pos < offset )
			{/* reader already read everything we have to say, send back EOF */
			result = 0;
			begin = pos;
			}
		else
			{/* reader had not read everything yet, send back something */
			result = pos;
			begin = 0;
			}
		*start = buffer + (offset - begin); /* Start of wanted data */
		result -= (offset - begin);
		if (result > length)
			result = length;
		}
	
out:
	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_initiator_proc_info, result %d\n",
																	result);

	return result;
	}

/*****************************************************************************
*	Sets the session operation parameters and changes the session
*	state to LOGGED_IN, meaning the initiator is in full feature phase.
*	Called only from init_connection()
*	Return value:	=1 on redirection
*					=0 on any other success
*					<0 on failure
*****************************************************************************/

int iscsi_initiator_login( struct connection *current_connection,
					 unsigned int when_called,
					 struct parameter_type p_param_tbl[MAX_CONFIG_PARAMS],
					 struct chap_parameter_type chap_param,
					 unsigned int *redirect_ip, unsigned int *redirect_port )
/* chap support - CHONG */
	{
	int					retval = 0, cid;
	int					login_task_tag;
	unsigned char		isid[6];
	struct session		*current_session;
	unsigned int		pdu_version_number = 0;		/* draft 11 and after */
	unsigned int		draft_number;
	unsigned int		port, ip;
	struct parameter_type	*p;
	char				*colon, *ptr;


	TRACE(TRACE_ENTER_LEAVE,
						"Enter iscsi_initiator_login, current connection %p\n",
															current_connection);

	current_session = current_connection->my_session;
	draft_number = current_session->draft_number;

	if( draft_number <= DRAFT8 )
		{/* set up the isid and cid fields for draft 8 and before,
			where it was really <cid, reserved, isid> */
		isid[0] = 0;									/* draft 8 cid */
		isid[1] = current_connection->connection_id;	/* draft 8 cid */
		isid[2] = 0;									/* draft 8 reserved */
		isid[3] = 0;									/* draft 8 reserved */
		isid[4] = 0x78;									/* draft 8 isid */
		isid[5] = current_session->scsi_target_id;		/* draft 8 isid */
		cid = 0;										/* offset 20 reserved */
		pdu_version_number = 2;							/* draft 8 version no.*/
		}
	else
		{/* set up the isid for draft 9 and after */
		if( draft_number == DRAFT9 )
			{
			isid[0] = 0x02;						/* draft 9 "random" option */
			pdu_version_number = 3;				/* draft 9 version number */
			}
		else
			isid[0] = 0x80;							/* Drft 11 "random" option*/
		isid[1] = 0x12;								/* Random number */
		isid[2] = 0x34;								/* Random number */
		isid[3] = 0x56;								/* Random number */
		isid[4] = 0x78;								/* Qualifier */
		isid[5] = current_session->scsi_target_id;	/* Qualifier */
		cid=current_connection->connection_id;		/* for offset 20 */
		}

	/* initialize task tag for the whole login process */
	/* then update the session's task tag counter */
	login_task_tag = current_session->cur_task_tag++;

	TRACE(TRACE_DEBUG, "iscsi_initiator_login draft_number %u, "
				"pdu_version_number %u\n", draft_number, pdu_version_number);

    /* In login phase always disable Header and Data Digests if they are enabled new 18_04 - SAI */
    current_connection->connection_flags &= ~USE_HEADERDIGEST;
    current_connection->connection_flags &= ~USE_DATADIGEST;

	if( (retval = initiator_parameter_negotiate(current_connection->sock, 
			p_param_tbl,							/* key table */
			chap_param,								/* chap parameter - CHONG */
			&current_session->cur_cmd_sn,
 			login_task_tag,
			isid,									/* we set this here */
			&current_session->tsih,					/* target returns this */
			cid,									/* cid, we set this here */
			&current_connection->connection_flags,	/* controls login phases */
			draft_number,							/* draft no. to enforce */
			pdu_version_number,						/* version no. in login */
			when_called,							/* new session/conn flags */
			&current_connection->exp_stat_sn,		/* set on return */
			&current_connection->max_send_length,	/* set on return */
			&current_connection->max_recv_length))	/* set on return */
			< 0 )
		{
		TRACE(TRACE_ISCSI, "parameter negotiation failed\n");
		retval = -ECONNABORTED;		/* software caused connection abort */
		goto out;
		}
	else if( retval == 0 )
		{/* new connection negotiation finished successfully */
		current_connection->connection_state = STATUS_LOGGED_IN;
		}
	else
		{/* have a redirection attempt */
		if( (p = find_parameter(TARGETADDRESS, p_param_tbl)) == NULL )
			{/* could not find TargetAddress key (should never happen) */
			TRACE_ERROR("Cannot find %s key\n", TARGETADDRESS);
			retval = -ECONNABORTED;		/* software caused connection abort */
			}
		else if( !IS_KEY_GOT_FROM_OTHER_SIDE(p->neg_info) )
			{/* target never sent us this key */
			TRACE_ERROR("Target never sent TargetAddress key on Redirection\n");
			retval = -ECONNABORTED;		/* software caused connection abort */
			}
		else if( strchr(p->str_value, ',') != NULL )
			{/* Draft 18, section 11.8 TargetAddress
				"If the TargetAddress is returned as the result of a redirect
				status in a login response, the comma and portal group tag
				MUST be omitted." */
			retval = -ECONNABORTED;		/* software caused connection abort */
			}
		else
			{
			if( (colon = strchr(p->str_value, ':')) == NULL )
				{/* no port number specified, use IANA's WKP for iSCSI */
				port = ISCSI_PORT;
				}
			else
				{/* user specified an explicit port number */
				*colon++ = '\0';	/* terminate the IP address string */
				if( (port = simple_strtoul(colon, &ptr, 10)) == 0  ||
												port > 0xffff  ||  *ptr != '\0')
					{
					TRACE_ERROR("Illegal Port Number %s\n", colon);
					retval = -ECONNABORTED;
					goto out;
					}
				}

			/*	should do more checking here for valid ip address */
			ip = ntohl(in_aton(p->str_value));
			TRACE(TRACE_ISCSI,"Redirecting to %u.%u.%u.%u:%u\n",
				(ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff,
				ip & 0xff, port);

			/* finally, set up for a redirection */
			*redirect_port = port;
			*redirect_ip = ip;
			}
		}

out:
	TRACE(TRACE_ENTER_LEAVE,"Leave iscsi_initiator_login, retval %d\n",retval);
	
	return retval;
	}


/*	Note the name "driver_template" is required for "scsi_module.c" to work! */
/*	Used by: "/usr/src/linux/drivers/scsi/scsi_module.c" */
/*	Wrap this in an ifdef because "scsi_module.c" says to! */
#ifdef MODULE

MODULE_AUTHOR("Bob Russell");
MODULE_DESCRIPTION(OUR_NAME);
MODULE_SUPPORTED_DEVICE("sd,sr,sg,st");
#ifdef MODULE_LICENSE
	MODULE_LICENSE("GPL");
#endif

/*	the single static instance of this structure for this adaptor type */
/*	ISCSI_INITIATOR is defined in "iscsi_initiator.h" */
static Scsi_Host_Template driver_template = ISCSI_INITIATOR;

#include "scsi_module.c"

EXPORT_NO_SYMBOLS;
#endif
